/****
 *
 * Arquivo: ArvoreRN.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2016
 * ltima modificao: 22/07/2016
 *
 * Descrio: Implementao do mdulo ArvoreRN
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h>   /* Alocao dinmica           */
#include <string.h>   /* Funes strXXX() e memXXX() */
#include "ArvoreRN.h" /* Interface deste mdulo      */
#include "ASSEGURA.h" /* Macro ASSEGURA              */

/****************** Definies de Tipos *******************/

typedef enum { VIOLACAO_NENHUMA = 1, VIOLACAO_ARVORE_DE_BUSCA,
               VIOLACAO_VERMELHA, VIOLACAO_PRETA } tViolacao;

/***************** Definies de Funes ******************/

/********************* Funes Locais *********************/

/****
 *
 * EhVermelho(): Verifica se um n  vermelho
 *
 * Parmetros:
 *       pNO (entrada) - o n a ser verificado
 *
 * Retorno: 1, se o n for vermelho; 0, em caso contrrio
 *
 ****/
static int EhVermelho(tNoRN *pNo)
{
   return pNo && pNo->cor == VERMELHO;
}

/****
 *
 * RotacaoSimples(): Efetua uma rotao simples sobre
 *                   um n de uma rvore rubro-negra
 *                   na direo especificada
 *
 * Parmetros:
 *      *x (entrada/sada) - o n em torno do qual ser
 *                           efetuada a rotao
 *       dir (entrada) - sentido da rotao (0 = esquerda;
 *                       1 = direita)
 *
 * Retorno: A nova raiz aps a rotao
 *
 * Observao: Assume-se que todos os ns so vlidos
 *
 ****/
static tNoRN *RotacaoSimples(tNoRN *x, tDirecao dir)
{
   tNoRN *y; /* Nova raiz aps a rotao */

      /* Faz y apontar para o filho da raiz */
      /* que est do lado oposto da rotao */
   y = x->filho[!dir];

      /* O novo filho de x que estar do lado oposto da */
      /* rotao ser o filho de y do lado da rotao   */
   x->filho[!dir] = y->filho[dir];

      /* O filho da nova raiz do lado  */
      /* da rotao ser a antiga raiz */
   y->filho[dir] = x;

   y->cor = PRETO; /* A nova raiz ser preta */

      /* A antiga raiz ser vermelha para no */
      /* causar violao preta da nova raiz   */
   x->cor = VERMELHO;

   return y;
}

/****
 *
 * RotacaoDupla(): Efetua uma rotao dupla sobre
 *                 um n de uma rvore rubro-negra
 *                 na direo especificada
 *
 * Parmetros:
 *      *x (entrada/sada) - o pai do n em torno do qual
 *                           ser efetuada a primeira rotao
 *       dir (entrada) - sentido da rotao (0 = esquerda;
 *                       1 = direita)
 *
 * Retorno: A nova raiz aps a rotao
 *
 * Observao: Assume-se que todos os ns so vlidos
 *
 ****/
static tNoRN *RotacaoDupla(tNoRN *x, tDirecao dir)
{
      /* Efetua uma rotao em torno do filho oposto de x */
      /* na direo contrria especificada por 'dir'      */
   x->filho[!dir] = RotacaoSimples(x->filho[!dir], !dir);

      /* Agora gira o filho */
   return RotacaoSimples(x, dir);
}

/****
 *
 * NovoNoRN(): Cria e inicia um novo n de uma rvore rubro-negra
 *
 * Parmetros:
 *       conteudo (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 * Observao: Esta funo no insere o novo n na rvore
 *
 ****/
static tNoRN *NovoNoRN(const tConteudo *conteudo)
{
   tNoRN *pNovo = malloc(sizeof *pNovo);

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(pNovo, "Erro: Nao foi possivel alocar no'");

   pNovo->cor = VERMELHO;
   pNovo->conteudo = *conteudo;
   pNovo->filho[ESQUERDA] = pNovo->filho[DIREITA] = NULL;

   return pNovo;
}

/********************* Funes Globais ********************/

/****
 *
 * CriaArvoreRN(): Cria uma rvore rubro-negra vazia
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo da rvore criada
 *
 ****/
tArvoreRN *CriaArvoreRN(void)
{
   tArvoreRN *arvore = malloc(sizeof *arvore);

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(arvore, "Erro: Nao foi possivel alocar arvore");

   arvore->raiz = NULL;
   arvore->nNos = 0;

   return arvore;
}

/****
 *
 * EstaVaziaArvoreRN(): Verifica se uma rvore rubro-negra
 *                      est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreRN(tArvoreRN *arvore)
{
   return arvore->nNos == 0;
}

/****
 *
 * DestroiArvoreRN(): Destri uma rvore rubro-negra
 *
 * Parmetros:
 *     arvore (entrada/sada) - ponteiro para a raiz da rvore
 *                              que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreRN(tArvoreRN *arvore)
{
   tNoRN *noCorrente = arvore->raiz;
   tNoRN *q;

      /* Gira os filhos esquerdos de modo que  */
      /* a operao possa ser tratada como uma */
      /* destruio de lista encadeada         */
   while (noCorrente) {
      if (!noCorrente->filho[ESQUERDA]) {
            /* Como no h filho esquerdo, remove-se */
            /* o n corrente e passa-se adiante      */
         q = noCorrente->filho[DIREITA];
         free(noCorrente);
      } else {
            /* Gira o filho esquerdo e checa novamente */
         q = noCorrente->filho[ESQUERDA];
         noCorrente->filho[ESQUERDA] = q->filho[DIREITA];
         q->filho[DIREITA] = noCorrente;
      }

      noCorrente = q;
   }

   free(arvore);
}

/****
 *
 * BuscaRN(): Retorna o endereo do contedo do n que possui uma
 *            determinada chave numa rvore rubro-negra
 *
 * Parmetros:
 *      *arvore (entrada) - rvore que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tConteudo *BuscaRN(tArvoreRN *arvore, tChave chave)
{
   tNoRN *noCorrente = arvore->raiz;
   int    compara;

   while (noCorrente) {
         /* Compara a chave do n corrente */
         /* com a chave de busca           */
      compara = strcmp(noCorrente->conteudo.chave, chave);

         /* Verifica se a chave foi encontrada */
      if (!compara)
         break; /* Chave encontrada */

         /* Desce na direo indicada pelo */
         /* valor da varivel 'compara'    */
      noCorrente = noCorrente->filho[compara < 0];
   }

   return noCorrente ? &noCorrente->conteudo : NULL;
}

/****
 *
 * InsereRNRec(): Insere um novo n numa rvore rubro-negra
 *
 * Parmetros:
 *    *raiz (entrada/sada) - raiz da rvore rubro-negra na
 *                            qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Endereo da raiz, se a insero for bem sucedida
 *          NULL, em caso contrrio
 *
 ****/
static tNoRN *InsereRNRec( tNoRN *raiz,
                           const tConteudo *conteudo )
{
   tDirecao dir; /* Direo (esquerda/direita) de descida */
   int      compara; /* Resultado de comparao de chaves */

      /* Se a rvore estiver vazia, cria uma raiz e retorna-a */
   if (!raiz) {
      return NovoNoRN(conteudo);
   }

      /* Compara a chave a ser inserida com */
      /* a chave que se encontra na raiz    */
   compara = strcmp(raiz->conteudo.chave, conteudo->chave);

      /* Se a chave j existir, ela no ser inserida */
      /* novamente, pois ela  primria               */
   if (!compara) {
      return NULL; /* Chave j existe */
   }

      /*                                 */
      /*  preciso descer na rvore para */
      /* encontrar o local da insero   */
      /*                                 */

      /* 'dir' indicar se a descida ser pela esquerda */
      /* (dir = 0) ou pela direita (dir = 1)            */
   dir = (tDirecao) (compara < 0);

      /* Chamada recursiva para insero  */
      /* direita ou esquerda de uma folha  */
   raiz->filho[dir] = InsereRNRec(raiz->filho[dir], conteudo);

      /*                                              */
      /* Se for necessrio corrigir alguma violao,  */
      /* a correo ocorrer na instruo if a seguir */
      /*                                              */

      /* Verifica se o n inserido ainda  vermelho */
   if (EhVermelho(raiz->filho[dir])) {
         /* Verifica se o irmo do n inserido  vermelho */
      if (EhVermelho(raiz->filho[!dir])) {
            /* Caso 1 de violao. Esse caso  corrigido por */
            /* meio de troca de cores entre pai e filhos     */
         raiz->cor = VERMELHO; /* Pai se torna comunista */

            /* Filhos se tornam pretos */
         raiz->filho[ESQUERDA]->cor = PRETO;
         raiz->filho[DIREITA]->cor = PRETO;
      } else {
            /* Casos 2 e 3 de violao. Aqui, */
            /* apenas rotaes resolvem.      */
         if (EhVermelho(raiz->filho[dir]->filho[dir])) {
               /* O filho esquerdo do filho esquerdo ou o    */
               /* filho direito do filho direito  vermelho. */
               /* Uma simples rotao na direo contrria   */
               /* resolve essa violao.                     */
            raiz = RotacaoSimples(raiz, !dir);
         } else if (EhVermelho(raiz->filho[dir]->filho[!dir])) {
               /* O filho direito do filho esquerdo ou o      */
               /* filho esquerdo do filho direito  vermelho. */
               /* Uma rotao dupla na direo contrria      */
               /* resolve essa violao.                      */
            raiz = RotacaoDupla(raiz, !dir);
         }
      }
   }

   return raiz;
}

/****
 *
 * InsereRN(): Insere um novo n numa rvore rubro-negra
 *
 * Parmetros:
 *    *arvore (entrada) - rvore rubro-negra na qual ser
 *                        feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: 1, se a insero for bem sucedida
 *          0, em caso contrrio
 *
 ****/
int InsereRN(tArvoreRN *arvore, const tConteudo *conteudo)
{
   tNoRN *p;

      /* De fato, quem faz a insero  a funo InsereRNRec()*/
   p = InsereRNRec(arvore->raiz, conteudo);

      /* Altera a raiz da rvore apenas se ocorreu insero */
   if (p) {
         /* Altera a raiz da rvore para refletir a insero */
      arvore->raiz = p;

         /* Assegura que a raiz  preta */
      arvore->raiz->cor = PRETO;

      ++arvore->nNos; /* Mais um n foi inserido */
   }

   return p ? 1 : 0;
}

/****
 *
 * CorrigeRemocao(): Corrige eventuais violaes ocorridas numa
 *                   remoo de n de uma rvore rubro-negra
 *
 * Parmetros:
 *     raiz (entrada) - ponteiro para a raiz da rvore na qual
 *                      ser feita a correo
 *     dir (entrada) - direo (esquerda/direita) da correo
 *     concluido (entrada/sada) - indica quando a correo
 *                                 est concluda
 *
 * Retorno: Endereo da raiz da rvore corrigida
 *
 ****/
static tNoRN *CorrigeRemocao( tNoRN *raiz, tDirecao dir,
                              int *concluido )
{
   tNoRN *pai = raiz,
         *filho = raiz->filho[!dir],
         *r;
   tCor   guardaCor; /* Guarda a cor de um n */

      /* Determina qual  o caso de violao e o corrige */
   if (filho && !EhVermelho(filho)) {
         /* Casos de irmo preto */
      if ( !EhVermelho(filho->filho[ESQUERDA]) &&
           !EhVermelho(filho->filho[DIREITA]) ) {
         if (EhVermelho(pai)) {
            *concluido = 1;
         }

         pai->cor = PRETO;
         filho->cor = VERMELHO;
      } else {
         guardaCor = raiz->cor;

         if (EhVermelho(filho->filho[!dir])) {
            pai = RotacaoSimples(pai, dir);
         } else {
            pai = RotacaoDupla(pai, dir);
         }

         pai->cor = guardaCor;
         pai->filho[ESQUERDA]->cor = PRETO;
         pai->filho[DIREITA]->cor = PRETO;

         *concluido = 1;
      }
   } else if (filho->filho[dir] != NULL) {
         /* Casos de irmo vermelho */
      r = filho->filho[dir];

      if ( !EhVermelho(r->filho[ESQUERDA]) &&
           !EhVermelho(r->filho[DIREITA]) ) {
         pai = RotacaoSimples(pai, dir);
         pai->filho[dir]->filho[!dir]->cor = VERMELHO;
      } else {
         if (EhVermelho(r->filho[dir])) {
            filho->filho[dir] = RotacaoSimples(r, !dir);
         }

         pai = RotacaoDupla(pai, dir);
         filho->filho[dir]->cor = PRETO;
         pai->filho[!dir]->cor = VERMELHO;
      }

      pai->cor = PRETO;
      pai->filho[dir]->cor = PRETO;

      *concluido = 1;
   }

   return pai;
}

/****
 *
 * RemoveRNRec(): Recursivamente, remove de uma rvore
 *                rubro-negra um n que contm a chave
 *                especificada como parmetro
 *
 * Parmetros:
 *     raiz (entrada/sada) - ponteiro para a raiz da rvore
 *                            na qual ser feita a remoo
 *     chave (entrada/sada) - chave do n a ser removido
 *     concluido (entrada/sada) - indica quando se ser
 *                                 necessria correo
 *     removido (entrada/sada) - indica quando ocorre remoo
 *
 * Retorno: Endereo da raiz da rvore na qual ocorreu a remoo
 *
 * OBSERVAO IMPORTANTE: O parmetro chave pode ser alterado!
 *
 ****/
static tNoRN *RemoveRNRec( tNoRN *raiz, tChave chave,
                           int *concluido, int *removido )
{
   tNoRN   *sucessor, /* Apontar para o sucessor de um n */
           *novaRaiz; /* Um ponteiro auxiliar */
   tDirecao dir;  /* Direo (esquerda/direita) de descida */
   int      compara; /* Resultado de comparao de chaves */

   if (!raiz) { /* rvore vazia */
      *concluido = 1; /* No h mais o que fazer */
   } else {
         /* Compara a chave do n corrente   */
         /* com a chave do n a ser removido */
      compara = strcmp(raiz->conteudo.chave, chave);

         /* Verifica se a chave foi encontrada */
      if (!compara) { /* Chave encontrada */
         if (!raiz->filho[ESQUERDA] || !raiz->filho[DIREITA]) {
            /*                              */
            /* Caso trivial: O n no tem   */
            /* filho ou tem apenas um filho */
            /*                              */

               /* Faz novaRaiz apontar para o filho direito  */
               /* da raiz se o filho esquerdo for NULL. Se   */
               /* o filho esquerdo da raiz no for NULL,     */
               /* novaRaiz apontar para ele. Se os 2 filhos */
               /* forem NULL, novaRaiz tambm ser NULL.     */
            novaRaiz = raiz->filho[!raiz->filho[ESQUERDA]];

               /* Se a nova raiz for vermelha, torna-a preta */
            if (EhVermelho(novaRaiz)) {
                  /* Lembre-se que ns nulos so pretos. */
                  /* Ento, no h possibilidade de      */
                  /* corrupo de memria aqui.          */
               novaRaiz->cor = PRETO;
            }

               /* Libera a antiga raiz */
            free(raiz);

               /* No h mais nenhuma correo a ser feita */
            *concluido = 1;

               /* Retorna a nova raiz */
            return novaRaiz;
         } else { /* O n tem dois filhos */
            /*                           */
            /* Encontra o sucessor do n */
            /*                           */

               /* Primeiro, desce-se at o filho direito */
            sucessor = raiz->filho[DIREITA];

               /* Agora, desce-se sempre  esquerda */
               /* at encontrar uma folha           */
            while (sucessor->filho[ESQUERDA]) {
               sucessor = sucessor->filho[ESQUERDA];
            }

            /* Sucessor encontrado */

               /* Copia o contedo do sucessor */
               /* para o n que seria removido */
            raiz->conteudo = sucessor->conteudo;

               /* Copia a chave do sucessor para */
               /* a chave do n a ser removido   */
            strcpy(chave, sucessor->conteudo.chave);
         }

         *removido = 1; /* Houve remoo */
      }

      /*                                  */
      /* A chave ainda no foi encontrada */
      /* e a busca continua               */
      /*                                  */

         /* Obtm a direo da descida  */
         /* (0 = esquerda; 1 = direita) */
      dir = compara < 0;

         /* Efetua a remoo na subrvore indicada por 'dir' */
      raiz->filho[dir] = RemoveRNRec( raiz->filho[dir],
                                      chave, concluido,
                                      removido );

         /* Efetua eventuais correes de violaes  */
         /* se o parmetro 'concluido' assim indicar */
      if (!*concluido) {
         raiz = CorrigeRemocao(raiz, dir, concluido);
      }
   }

   return raiz;
}

/****
 *
 * RemoveRN(): Remove de uma rvore rubro-negra um n que
 *             contm a chave especificada como parmetro
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro que representa a rvore
 *                        na qual ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: 1, se a remoo for bem sucedida
 *          0, em caso contrrio (i.e., se a chave no
 *          for encontrada)
 *
 ****/
int RemoveRN(tArvoreRN *arvore, tChave chave)
{
   int concluido = 0, /* Indicar se ser necessrio checar    */
                      /* possveis violaes (concluido = 0)   */
                      /* ou no (concluido = 1). Em princpio, */
                      /* supe-se que sim. RemoveRNRec()       */
                      /* confirmar ou no essa suposio      */
       houveRemocao = 0; /* Indicar se houve remoo */

      /* Chama a funo que realmente efetua a remoo */
   arvore->raiz = RemoveRNRec( arvore->raiz, chave, &concluido,
                               &houveRemocao );

      /* A raiz da rvore deve ser preta */
   if (arvore->raiz) {
      arvore->raiz->cor = PRETO;
   }

   return houveRemocao;
}

/****
 *
 * AlteraConteudoArvoreRN(): Altera o contedo efetivo de um
 *                           n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreRN( tArvoreRN *arvore, tChave chave,
                            tConteudo *conteudo )
{
   tConteudo *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaRN(arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;

         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /*                                            */
      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
      /*                                            */

      /* Verifica se a chave antiga se encontra na rvore */
   if (!BuscaRN(arvore, chave)) {
      return 1; /* Chave no foi encontrada */
   }

      /* Verifica se a chave nova j se encontra na rvore */
   if (BuscaRN(arvore, conteudo->chave)) {
      return 1; /* Chave nova j existe na rvore */
   }

      /* Remove o n contendo a chave antiga */
   ASSEGURA( RemoveRN(arvore, chave),
             "Erro na alteracao de um no" );

   ASSEGURA( InsereRN(arvore, conteudo),
             "Erro na alteracao de um no");

   return 0;
}

/****
 *
 * NumeroNosRN(): Obtm o nmero de ns de uma rvore rubro-negra
 *
 * Parmetros:
 *     *arvore (entrada) - a rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosRN(tArvoreRN *arvore)
{
   return arvore->nNos;
}

/****
 *
 * CaminhamentoInfixo(): Efetua um caminhamento em ordem
 *                       infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixo( tNoRN *raiz, tOperacao op,
                         FILE *stream )
{
   if (raiz) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixo(raiz->filho[ESQUERDA], op, stream);

      op(raiz->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixo(raiz->filho[DIREITA], op, stream);
   }
}

/****
 *
 * ProfundidadeRN(): Calcula a profundidade de uma
 *                   rvore rubro-negra de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A profundidade da referida rvore
 *
 ****/
int ProfundidadeRN(tNoRN *raiz)
{
   int profEsq, profDir;

   if (!raiz) {
      return 0; /* rvreo vazia tem profundidade zero */
   } else {
        /* Calcula a profundidade da subrvore esquerda */
      profEsq = 1 + ProfundidadeRN(raiz->filho[ESQUERDA]);

        /* Calcula a profundidade da subrvore direita */
      profDir = 1 + ProfundidadeRN(raiz->filho[DIREITA]);

        /* A profundidade da rvore  a maior */
      return profEsq > profDir ? profEsq : profDir;
   }
}

/****
 *
 * MenorChave(): Obtm o contedo do n que possui a menor
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MenorChave(tArvoreRN *arvore)
{
   tNoRN *p = arvore->raiz,
         *q; /* Apontar para o n que */
             /* armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p) {
      return NULL;
   }
      /* A menor chave encontra-se no n mais */
      /*  esquerda da rvore de busca        */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  esquerda */
      p = p->filho[ESQUERDA];
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a menor chave      */
   return &q->conteudo;
}

/****
 *
 * MaiorChave(): Obtm o contedo do n que possui a maior
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MaiorChave(tArvoreRN *arvore)
{
   tNoRN *p = arvore->raiz,
         *q; /* Apontar para o n que */
             /* armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p) {
      return NULL;
   }
      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  direita */
      p = p->filho[DIREITA];
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a maior chave      */
   return &q->conteudo;
}

/****
 *
 * EhRubroNegra(): Verifica se uma rvore supostamente
 *                 rubro-negra  realmente rubro-negra
 *
 * Parmetros:
 *      raiz (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: Um valor indicando se ocorreu alguma violao
 *          das regras de rvores rubro-negras
 *
 ****/
static tViolacao EhRubroNegra(tNoRN *raiz)
{
   int    alturaEsq,
          alturaDir,
          comparaEsq, comparaDir;
   tNoRN *noEsquerda,
         *noDireita;

   if (raiz) {
      return 1;
   } else {
      noEsquerda = raiz->filho[0];
      noDireita = raiz->filho[1];

         /* Checa violao vermelha */
      if (EhVermelho(raiz)) {
         if (EhVermelho(noEsquerda) || EhVermelho(noDireita)) {
            return VIOLACAO_VERMELHA;
         }
      }

      alturaEsq = EhRubroNegra(noEsquerda);
      alturaDir = EhRubroNegra(noDireita);

         /* Checa se  rvore binria de busca */

      if (noEsquerda) {
         comparaEsq = strcmp( noEsquerda->conteudo.chave,
                              raiz->conteudo.chave );
      }

      if (noDireita) {
         comparaDir = strcmp( noDireita->conteudo.chave,
                              raiz->conteudo.chave );
      }

      if ( (noEsquerda && comparaEsq >= 0) ||
           (noDireita && comparaDir <= 0) ) {
         return VIOLACAO_ARVORE_DE_BUSCA;
      }

         /* Checa alturas pretas */
      if (alturaEsq && alturaDir && alturaEsq != alturaDir) {
         return VIOLACAO_PRETA;
      }

         /* Conta apenas ns pretos */
      if (alturaEsq && alturaDir) {
         return EhVermelho(raiz) ? alturaEsq : alturaEsq + 1;
      } else {
         return VIOLACAO_PRETA;
      }
   }
}

/****
 *
 * EhRubroNegra(): Verifica se uma rvore supostamente
 *                 rubro-negra  realmente rubro-negra
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro que representa a rvore
 *
 * Retorno: 1, se a rvore for rubro-negra. Caso contrrio, 0
 *
 ****/
void AvaliaArvoreRubroNegra(tArvoreRN *arvore)
{
   tViolacao resultado;

   resultado = EhRubroNegra(arvore->raiz);

   switch (resultado) {
      case VIOLACAO_VERMELHA:
         puts("\n\t>>> Ocorreu uma violacao vermelha <<<");
         return;

      case VIOLACAO_PRETA:
         puts("\n\t>>> Ocorreu uma violacao preta <<<");
         return;

      case VIOLACAO_ARVORE_DE_BUSCA:
         puts("Essa nao e' uma arvore binaria de busca <<<<");
         return;

      case VIOLACAO_NENHUMA:
         puts( "\n\t>>> Essa e' uma legitima arvore "
               "rubro-negra <<<" );
         return;

      default:
         puts("\n\t>>> Este programa comtem um erro");
         return;
   }
}
