/****
 *
 * Ttulo: Copiando Arquivos
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2012
 * ltima modificao: 16/06/2012
 *
 * Descrio: Copia um arquivo para outro byte a byte
 *
 * Entrada: Nome do arquivo a ser copiado e nome do arquivo que
 *          recebe a cpia
 *
 * Sada: Um novo arquivo contendo a cpia
 *
 ****/

#include <stdio.h>  /* Entrada e sada */
#include <string.h> /* Funo strcmp() */

/****
 *
 * CopiaArquivo(): Copia contedo de um arquivo byte a byte
 *
 * Parmetros:
 *      streamEntrada (entrada) - stream associado ao arquivo que ser copiado
 *      streamSaida (entrada) - stream associado ao arquivo que
 *                              receber a cpia
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 * Observao: O stream que ser lido deve estar aberto num modo
 *             que permita leitura e o stream que ser escrito
 *             deve estar aberto num modo que permite escrita.
 *
 ****/
int CopiaArquivo(FILE *streamEntrada, FILE *streamSaida)
{
   int c; /* Armazenar cada byte lido e escrito  */

      /* Garante que a leitura comea no incio do arquivo */
   rewind(streamEntrada);

      /* O lao encerra quando houver tentativa de leitura   */
      /* alm do final do arquivo de entrada ou ocorrer erro */
      /* de leitura ou escrita em qualquer dos arquivos      */
   while (1) {
         /* L um byte no arquivo de entrada */
      c = fgetc(streamEntrada);

         /* Testa se final do arquivo de entrada foi */
         /* atingido ou ocorreu erro de leitura      */
      if (feof(streamEntrada) || ferror(streamEntrada)) {
         break; /* Processamento encerrado */
      }

         /* Escreve o byte lido no arquivo de sada */
      fputc(c, streamSaida);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamSaida)) {
         break;
      }
   }

   /************************************************/
   /* O processamento est terminado, mas a funo */
   /* no deve fechar os arquivos, j que ela no  */
   /* foi responsvel pelas aberturas.             */
   /************************************************/

      /* Se ocorreu erro de escrita ou leitura, o   */
      /* retorno ser 1. Caso contrrio, ser zero. */
   return ferror(streamEntrada) || ferror(streamSaida);
}

/****
 *
 * main(): Copia o contedo de um arquivo para outro byte a byte
 *
 * Parmetros:
 *     argc (entrada) - Nmero de argumentos de linha de comando
 *     argv (entrada) - Array de strings presentes na linha de
 *                      comando quando o programa  executado
 *
 * Retorno: Zero, se no ocorrer nenhum erro.
 *          Um valor diferente de zero em caso contrrio.
 *
 * Observaes:
 *     1. Os nomes dos arquivos devem acompanhar o nome na
 *        linha de comando
 *     2. O nome do arquivo de entrada  o primeiro e o nome do
 *        arquivo de sada  o segundo argumento de linha de
 *        comando
 *
 ****/
int main(int argc, char *argv[])
{
   FILE *streamEntrada, /* Stream de entrada */
        *streamSaida;   /* Streams de sada  */
   int   resultado;

      /* Verifica se o usurio informou quais  */
      /* sero os arquivos envolvidos na cpia */
   if (argc != 3) {
      printf( "\n\t>>> Este programa deve ser usado assim:"
              "\n\t> %s arquivo-a-ser-copiado "
              "arquivo-que-recebe-a-copia\n", argv[0] );
      return 1;
   }

      /* Verifica se os arquivos de entrada e de sada so os mesmos */
   if (!strcmp(argv[1], argv[2])) {
      printf("\nOs nomes dos arquivos nao podem ser iguais\n");
      return 1;
   }

      /* Tenta abrir o arquivo de entrada */
   streamEntrada = fopen(argv[1], "rb");

      /* Verifica se o arquivo de entrada foi aberto */
   if (!streamEntrada) {
      printf("\nO arquivo \"%s\" nao pode ser aberto\n", argv[1]);
      return 1;
   }

      /*                                                    */
      /* Aqui, o arquivo de entrada foi aberto com sucesso. */
      /* Se o arquivo de sada no for aberto, deve-se      */
      /* fechar o arquivo de entrada antes de retornar.     */
      /*                                                    */

      /* Tenta abrir arquivo de sada */
   streamSaida = fopen(argv[2], "wb");

      /* Verifica se o arquivo de sada foi aberto */
   if (!streamSaida) {
      printf("\nO arquivo \"%s\" nao pode ser aberto\n", argv[2]);
      fclose(streamEntrada);
      return 1;
   }

      /* Efetua a cpia */
   resultado = CopiaArquivo(streamEntrada, streamSaida);

      /* Os arquivos no precisam mais estar abertos */
   fclose(streamEntrada);
   fclose(streamSaida);

      /* Comunica ao usurio o resultado da operao */
   if (!resultado) {
      printf( "\n\t>>> O arquivo %s foi copiado em %s\n",
              argv[1], argv[2] );
   } else {
      printf("\n\t>>> Impossivel copiar o arquivo %s\n", argv[1]);
      return 1; /* Operao falhou */
   }

   return 0;
}

/******************* Exemplos de Execuo ******************

[NB: O nome do programa executvel  "copia"]

Exemplo 1:
=========

C:\Programas>copia Tudor.txt Tudor.txt

Os nomes dos arquivos nao podem ser iguais

Exemplo 2:
=========

C:\Programas>copia Tudor.txt TudorBK.txt

        >>> O arquivo Tudor.txt foi copiado em TudorBK.txt

***********************************************************/
