/****
 *
 * Exemplo: Busca binria/por interpolao com registros
 *          armazenados em memria secundria (arquivo)
 *
 * Seo: 3.5
 *
 * Descrio:
 *    -> Uma tabela de busca  implementada como lista indexada
 *       ordenada contendo em que cada elemento contm uma
 *       a chave e a posio do respetivo registro no arquivo
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> A tabela  criada sem ordenao e depois  ordenada
 *       com qsort(). Criar a tabela j ordenada leva uma
 *       eternidade.
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Entrada e sada           */
#include <string.h>   /* Funes strXXX()          */
#include <stdlib.h>   /* malloc()                  */
#include "ASSEGURA.h" /* Macro ASSEGURA()          */
#include "Tabela.h"   /* Interface do tipo tTabelaIdx */

/****************** Complemento de Tipo *******************/

struct rotTabelaIdx {
   tCEP_Ind *elementos; /* Ponteiro para o array que */
                         /* contm os elementos       */
   int        nElementos; /* Nmero de elementos */
   int        tamanhoArray;
};

/***************** Definies de Funes ******************/

/****
 *
 * CriaTabelaIdx(): Cria uma tabela
 *
 * Parmetros: No h
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
tTabelaIdx CriaTabelaIdx(void)
{
   tTabelaIdx tabela;

      /* Aloca espao para a estrutura que armazena a tabela */
   tabela = malloc(sizeof(struct rotTabelaIdx));

      /* Garante que a alocao realmente ocorreu */
   ASSEGURA(tabela, "Nao foi possivel alocar tabela");

      /* Aloca previamente um array com um elemento */
   tabela->elementos = calloc(1, sizeof(tCEP_Ind));

      /* Garante que o array foi realmente alocado */
   ASSEGURA( tabela->elementos, "Nao foi possivel alocar "
             "array para conter elementos da tabela" );

      /* Atualiza a varivel que armazena o tamanho do array */
   tabela->tamanhoArray = 1;

      /* Inicialmente a tabela est vazia */
   tabela->nElementos = 0;

   return tabela;
}

/****
 *
 * DestroiTabelaIdx(): Libera o espao ocupado pela tabela
 *
 * Parmetros: tabela (entrada) - a tabela
 *
 * Retorno: Nada
 *
 ****/
void DestroiTabelaIdx(tTabelaIdx tabela)
{
      /* Libera o array que contm os elementos */
   free(tabela->elementos);

      /* Libera a prpria tabela */
   free(tabela);
}

/****
 *
 * ComprimentoIdx(): Calcula o comprimento de uma tabela
 *
 * Parmetros: *tabela (entrada) - a tabela cujo comprimento
 *                                 ser calculado
 *
 * Retorno: o nmero de elementos na tabela
 *
 ****/
int ComprimentoIdx(tTabelaIdx tabela)
{
   return tabela->nElementos; /* No h o que calcular */
}

/****
 *
 * ObtemElementoIdx(): Obtm o elemento de uma tabela que
 *                  se encontra numa dada posio
 *
 * Parmetros:
 *     *tabela (entrada) - a tabela
 *      indice (entrada) - o elemento
 *
 * Retorno: o nmero de elementos na tabela
 *
 ****/
tCEP_Ind ObtemElementoIdx(tTabelaIdx tabela, int indice)
{
   ASSEGURA( indice >= 0 && indice < tabela->nElementos,
             "Elemento inexistente" );

   return tabela->elementos[indice];
}

/****
 *
 * RemoveElementoIdx(): Remove um elemento numa dada posio
 *                   de uma tabela sem ordenao
 *
 * Parmetros:
 *      *tabela (entrada/sada) - tabela onde ser feita a
 *                               remoo
 *       indice (entrada) - posio do elemento que ser removido
 *
 * Retorno: O item removido
 *
 ****/
tCEP_Ind RemoveElementoIdx(tTabelaIdx tabela, int indice)
{
   tCEP_Ind itemRemovido;

         /* Verifica se o ndice  vlido */
   ASSEGURA( indice >= 0 && indice < tabela->nElementos,
             "Posicao de remocao inexistente" );

   itemRemovido = tabela->elementos[indice];

      /* Remover um elemento significa mover cada   */
      /* elemento uma posio para trs a partir    */
      /* do sucessor do elemento que ser  removido */
   for (int i = indice; i < tabela->nElementos - 1; i++)
      tabela->elementos[i] = tabela->elementos[i + 1];

   --tabela->nElementos; /* O tamanho da tabela diminuiu */

   return itemRemovido;
}

/****
 *
 * EstaVaziaIdx(): Verifica se uma tabela sem ordenao est vazia
 *
 * Parmetros: *tabela (entrada) - a tabela que ser testada
 *
 * Retorno: 1, se a tabela estiver vazia; 0, em caso contrario
 *
 ****/
int EstaVaziaIdx(tTabelaIdx tabela)
{
   return tabela->nElementos == 0;
}

/****
 *
 * AcrescentaElementoIdx(): Acrescenta um novo elemento ao final
 *                       de uma tabela sem ordenao
 *
 * Parmetros:
 *      tabela (sada) - tabela na qual ser feita o acrscimo
 *      elemento (entrada) - valor do elemento que ser
 *                           acrescentado
 *
 * Retorno: Nada
 *
 ****/
void AcrescentaElementoIdx(tTabelaIdx tabela, const tCEP_Ind *elemento)
{
   tCEP_Ind *novoArray; /* Ponteiro para o array */
                         /* redimensionado        */

      /* Se o array estiver repleto, tenta redimension-lo */
   if (tabela->nElementos >= tabela->tamanhoArray) {
         /* Dobra o tamanho do array */
      tabela->tamanhoArray *= 2;

         /* Tenta redimensionar o array */
      novoArray = realloc(tabela->elementos,
                          tabela->tamanhoArray*sizeof(tCEP_Ind));

         /* Checa se houve realocao de memria */
      ASSEGURA(novoArray, "Nao houve redimensionamento");

         /* O ponteiro que apontava para o incio */
         /* do array pode no ser mais vlido.    */
         /* Portanto,  preciso atualiz-lo.      */
      tabela->elementos = novoArray;
   }

      /* Acrescenta o novo elemento ao final da tabela */
   tabela->elementos[tabela->nElementos] = *elemento;

   ++tabela->nElementos; /* O tamanho da tabela aumentou */
}

/****
 *
 * InsereEmOrdemIdx(): Insere um novo elemento numa tabela de
 *                  modo a preservar a ordenao da tabela
 *
 * Parmetros:
 *      tabela (entrada e sada) - tabela na qual ser feita a insero
 *      elemento (entrada) - elemento que ser inserido
 *
 * Retorno: Nada
 *
 ****/
void InsereEmOrdemIdx(tTabelaIdx tabela, const tCEP_Ind *elemento)
{
   int        posicao; /* Posio de insero do elemento */
   tCEP_Ind *novoArray; /* Ponteiro para o array redimensionado */

      /* Se o array estiver repleto, tenta redimension-lo */
   if (tabela->nElementos >= tabela->tamanhoArray) {
      tabela->tamanhoArray *= 2; /* Dobrar o tamanho do array */

         /* Tenta redimensionar o array */
      novoArray = realloc(tabela->elementos, tabela->tamanhoArray*sizeof(tCEP_Ind));

         /* Checa se houve realocao de memria */
      ASSEGURA(novoArray, "Nao houve redimensionamento");

         /* O ponteiro que apontava para o incio do array pode no */
         /* ser mais vlido.  Portanto,  preciso atualiz-lo.      */
      tabela->elementos = novoArray;
   }

      /* Se a tabela estiver vazia, o elemento  */
      /* inserido na primeira posio do array  */
   if (EstaVaziaIdx(tabela)) {
         /* O elemento ser o primeiro da tabela */
      tabela->elementos[0] = *elemento;

      ++tabela->nElementos; /* A tabela cresceu */

      return;
   }

      /* Encontra a posio no array onde o elemento ser inserido */
   for (posicao = 0; posicao < tabela->nElementos; ++posicao)
      if ( strcmp(tabela->elementos[posicao].chave, elemento->chave) > 0 )
         break; /* Posio de insero encontrada */

      /* Abre espao para o novo elemento */
   for (int i = tabela->nElementos - 1; i >= posicao; --i)
         /* Move cada elemento uma posio adiante a partir do  */
         /* elemento que ora se encontra na posio de insero */
      tabela->elementos[i + 1] = tabela->elementos[i];

   tabela->elementos[posicao] = *elemento; /* Insere o novo elemento */

   ++tabela->nElementos; /* A tabela cresceu */
}

/****
 *
 * BuscaBinariaIdx(): Efetua uma busca binria numa tabela de busca
 *                    implementada como lista indexada ordenada
 *
 * Parmetros:
 *     *tabela (entrada) - tabela na qual ser efetuada a busca
 *      chave (entrada) - a chave de busca
 *
 * Retorno: ndice do primeiro elemento encontrado na tabela
 *          que casa com o valor recebido como parmetro.
 *          Um valor negativo, se o valor no for encontrado.
 *
 * Observao: Na busca binria, quando uma chave  encontrada
 *             e existe mais de uma ocorrncia dessa chave na
 *             tabela, o ndice retornado no corresponde
 *             necessariamente ao menor ndice de elemento
 *             que contm a chave procurada.
 *
 ****/
int BuscaBinariaIdx(tTabelaIdx tabela, tCEP chave)
{
   int inf, /* Limite inferior da busca */
       sup, /* Limite superior da busca */
       meio, /* Meio do intervalo */
       teste; /* Resultado da comparao de duas chaves */

   inf = 0; /* Limite inferior inicial */
   sup = tabela->nElementos - 1; /* Limite superior inicial */

      /* Efetua a busca binria */
   while (inf <= sup) {
         /* Calcula o meio do intervalo */
      meio = inf + (sup - inf)/2;

      teste = strcmp(chave, tabela->elementos[meio].chave);

         /* Verifica se a chave se encontra */
         /* no meio do intervalo            */
      if (!teste)
         return tabela->elementos[meio].valor; /* Sucesso */

         /* Ajusta o intervalo de busca */
      if (teste < 0)
         sup = meio - 1;
      else
         inf = meio + 1;
   }

   return -1; /* Elemento no foi encontrado */
}

/****
 *
 * BuscaInterpolacaoIdx(): Efetua uma busca por interpolao numa
 *                      tabela de busca implementada como lista
 *                      indexada ordenada
 *
 * Parmetros:
 *      tabela (entrada) - tabela na qual ser efetuada a busca
 *      chave (entrada) - a chave de busca
 *
 * Retorno: ndice do primeiro elemento encontrado na lista
 *          que casa com a chave de busca. Um valor negativo,
 *          se a chave no for encontrada.
 *
 ****/
int BuscaInterpolacaoIdx(tTabelaIdx tabela, tCEP chave)
{
   int inf, /* Limite inferior da busca */
       sup, /* Limite superior da busca */
       meio, /* Meio do intervalo */
       chaveInf, chaveSup, /* Chaves dos limites inferior e */
                           /* superior convertidas em int   */
       teste; /* Resultado da comparao de duas chaves */

   inf = 0; /* Limite inferior inicial */
   sup = tabela->nElementos - 1; /* Limite superior inicial */

      /* Efetua a busca por interpolao */
   while (inf <= sup) {
         /* Converte as chaves nos extremos do intervalo */
      chaveInf = atoi(tabela->elementos[inf].chave);
      chaveSup = atoi(tabela->elementos[sup].chave);

         /* Calcula o meio do intervalo */
      meio = inf + (sup - inf)*((atoi(chave) - chaveInf)/(chaveSup - chaveInf));

         /* Compara a chave de busca com aquela  */
         /* que se encontra no meio do intervalo */
      teste = strcmp(chave, tabela->elementos[meio].chave);

         /* Verifica se o elemento encontra-se no meio do intervalo */
      if (!teste)
         return tabela->elementos[meio].valor; /* Sucesso */

         /* Ajusta o intervalo de busca */
      if (teste < 0)
         sup = meio - 1;
      else
         inf = meio + 1;
   }

   return -1; /* Elemento no foi encontrado */
}

/****
 *
 * BuscaSequencialIdx(): Efetua uma busca sequencial numa tabela de
 *                    busca implementada como lista indexada
 *
 * Parmetros:
 *      *tabela (entrada) - tabela na qual ser efetuada a busca
 *       chave (entrada) - chave de busca
 *
 * Retorno: ndice do primeiro elemento no array que casa com
 *          o valor recebido como parmetro.
 *          Um valor negativo, se o valor no for encontrado.
 *
 ****/
int BuscaSequencialIdx(tTabelaIdx tabela, tCEP chave)
{
   for (int i = 0; i < tabela->nElementos; ++i)
      if (!strcmp(tabela->elementos[i].chave, chave))
         return tabela->elementos[i].valor; /* Sucesso */

   return -1; /* Elemento no foi encontrado */
}

/****
 *
 * BuscaBinariaIdx2(): Efetua uma busca binria numa tabela de
 *                  busca implementada como lista indexada
 *                  ordenada usando qsort()
 *
 * Parmetros:
 *      *tabela (entrada) - tabela na qual ser efetuada a busca
 *       chave (entrada) - chave de busca
 *       FComp (entrada) - ponteiro para a funo que
 *                         compara elementos da tabela
 *
 * Retorno: Valor associado  chave que casa com a chave de
 *          busca, se tal casamento ocorrer. Se a chave de
 *          busca no for encontrada, o retorno  -1.
 *
 ****/
int BuscaBinariaIdx2( tTabelaIdx tabela, tCEP chave,
                   int (*FComp) (const void *, const void *) )
{
   tCEP_Ind *pElemento; /* Ponteiro para o elemento encontrado */

   pElemento = bsearch( chave, tabela->elementos,
                        tabela->nElementos,
                        sizeof(tabela->elementos[0]), FComp );

      /* Se a chave foi encontrada, retorna o valor associado */
      /* a ela, que  o ndice do registro no arquivo         */
   return pElemento ? pElemento->valor : -1;
}

/****
 *
 * OrdenaTabelaIdx(): Ordena a tabela usando qsort()
 *
 * Parmetros:
 *       tabela (entrada e sada) - tabela que ser ordenada
 *       FComp (entrada) - ponteiro para a funo que compara elementos da tabela
 *
 * Retorno: Nada
 *
 ****/
void OrdenaTabelaIdx( tTabelaIdx tabela,
                      int (*FComp) (const void *, const void *) )
{
      /* A funo qsort() faz o servio */
   qsort( tabela->elementos, tabela->nElementos,
          sizeof(tabela->elementos[0]), FComp );
}
