/****
 *
 * Exemplo: Busca com Chaves Secundrias
 *
 * Seo: 3.7.2
 *
 * Descrio:
 *    -> Uma tabela de busca  implementada como lista ENCADEADA
 *       em que cada elemento contm uma chave e a posio
 *       do respetivo registro no arquivo
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Tipo de busca: sequencial com chaves secundrias
 *
 ****/

/************************ Includes ************************/

#include <stdio.h>    /* Entrada e sada         */
#include <stdlib.h>   /* exit()                  */
#include <string.h>   /* Funes strXXX()        */
#include "ASSEGURA.h" /* Macro ASSEGURA          */
#include "Lista.h"    /* Interface deste arquivo */

/****************** Definies de Funes *****************/

/****
 *
 * IniciaListaSE(): Inicia uma lista simplesmente encadeada
 *
 * Parmetros: *lista (sada) - a lista que ser iniciada
 *
 * Retorno: Nada
 *
 * Observao: Esta funo independe do tipo de contedo
 *             efetivo armazenado em cada n da lista
 *
 ****/
void IniciaListaSE(tListaSE *lista)
{
   *lista = NULL;
}

/****
 *
 * ComprimentoListaSE(): Calcula o nmero de ns de uma lista
 *                       simplesmente encadeada
 *
 * Parmetros:
 *     lista (entrada) - ponteiro para o primeiro n da lista
 *                       cujo comprimento ser calculado
 *
 * Retorno: O nmero de ns da referida lista
 *
 * Observao: Esta funo independe do tipo de contedo
 *             efetivo armazenado em cada n da lista
 *
 ****/
int ComprimentoListaSE(tListaSE lista)
{
   int tamanho = 0; /* Armazena o nmero de ns da lista */

      /* Acessa cada n da lista e conta */
      /* quantos ns so acessados       */
   while (lista) {
      ++tamanho; /* Mais um n foi encontrado */
      lista = lista->proximo; /* Passa para o prximo n */
   }

   return tamanho;
}

/****
 *
 * InsereNoLSE(): Insere um novo n no incio de uma
 *                lista simplesmente encadeada
 *
 * Parmetros:
 *    *lista (entrada/sada) - ponteiro para a lista na qual
 *                             ser feita a insero
 *     conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Nada
 *
 * Observao: Como a lista no obedece nenhuma ordem, pode-se
 *             inserir um n em qualquer posio da lista.
 *             Escolheu-se insero no incio da lista porque
 *              mais simples e eficiente.
 *
 ****/
void InsereNoLSE(tListaSE *lista, const tCEP_Ind *conteudo)
{
   tNoSE *ptrNovoNo; /* Apontar para o novo n alocado */

      /* Tenta alocar um novo n */
   ASSEGURA( ptrNovoNo = malloc(sizeof(tNoSE)),
             "Nao foi possivel alocar no'" );

      /* Armazena no novo n os dados recebidos como parmetro */
   ptrNovoNo->conteudo = *conteudo;

      /* O novo n apontar para o incio corrente da lista */
   ptrNovoNo->proximo = *lista;

      /* O incio da lista passa a apontar para o novo n */
   *lista = ptrNovoNo;
}

/****
 *
 * RemoveNoLSE(): Remove de uma lista simplesmente encadeada
 *                um n que contm um determinado valor
 *                especificado como parmetro
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro que representa
 *                             a lista na qual ser feita a remoo
 *     conteudo (entrada) - contedo do n a ser removido
 *
 * Retorno: 0, se a remoo for bem sucedida, ou 1, se o n a
 *          ser removido no for encontrado.
 *
 ****/
int RemoveNoLSE(tListaSE *lista, int indice)
{
   tListaSE p = *lista, /* p aponta para o n corrente     */
            q = NULL;   /* q aponta para o n anterior a p */

      /*                                                     */
      /* O lao while procura um n que cujo contedo case   */
      /* com o parmetro 'conteudo' e tem duas condies de  */
      /* parada: (1) todos os ns so acessados sem que o    */
      /* referido n tenha sido encontrado (neste  caso, p   */
      /* assume NULL); (2) o n que contm o valor procurado */
      /*  encontrado.                                       */
      /*                                                     */
   while (p && p->conteudo.indiceCEP != indice) {
      q = p; /* q passa a apontar para o n corrente */
      p = p->proximo; /* p passa a apontar para o prximo n */
   }

      /* Se p for NULL, a lista foi totalmente examinada   */
      /* sem que o n a ser removido tenha sido encontrado */
   if (!p) {
      return 1; /* Tentativa de remoo de n inexistente */
   }

      /****************************************************/
      /* Neste ponto, p aponta para o n a ser removido e */
      /* q aponta para o n imediatamente anterior a ele  */
      /****************************************************/

      /* Verifica se p aponta para o primeiro */
      /* n da lista porque remoo no incio */
      /* deve ser tratada separadamente       */
   if (p == *lista) { /* Remoo ser no incio da lista */
      (*lista) = p->proximo; /* Altera incio da lista */
   } else {/* N a ser removido NO  o primeiro da lista */
      q->proximo = p->proximo; /* Desvia o n anterior */
                               /* do n a ser removido */
   }

      /* Neste ponto, o n desejado no faz mais parte da */
      /* lista, mas ele ainda ocupa espao em memria.    */
      /* Portanto,  necessrio liberar este espao.      */
   free(p); /* Libera o n removido */

   return 0;
}

/****
 *
 * EstaVaziaLSE(): Verifica se uma lista simplesmente
 *                 encadeada est vazia
 *
 * Parmetros:
 *     lista (entrada) - a lista que ser verificada
 *
 * Retorno: 1, quando a lista est vazia, ou 0, em caso contrrio
 *
 ****/
int EstaVaziaLSE(tListaSE lista)
{
   return lista == NULL;
}

/****
 *
 * BuscaSecundariaLSE(): Retorna o endereo de uma lista
 *                       encadeada contendo chaves que
 *                       casam com uma chave de busca
 *
 * Parmetros:
 *      lista (entrada) - a tabela de busca
 *      chave (entrada) - chave de busca
 *
 * Retorno: Endereo de uma lista encadeada contendo chaves
 *          que casam com uma chave de busca ou NULL se
 *          nenhuma chave for encontrada
 *
 ****/
tListaSE BuscaSecundariaLSE(tListaSE lista, tCEP chave)
{
   tListaSE encontrados = NULL; /* Lista contendo as chaves que */
                                /* casam com a chave de busca   */

      /* Enquanto 'lista' no assume NULL, a busca prossegue */
   while (lista) {
         /* Se ocorrer casamento entre a chave de busca e */
         /* a chave do elemento corrente, acrescenta-se   */
         /* esse elemento  lista de chaves encontradas   */
      if (!strcmp(lista->conteudo.chave, chave)) {
         InsereNoLSE(&encontrados, &lista->conteudo);
      }

      lista = lista->proximo;
   }

   return encontrados;
}

/****
 *
 * BuscaSequencialLSE(): Retorna o endereo do contedo efetivo
 *                       do n que possui uma determinado chave
 *                       numa lista simplesmente encadeada
 *
 * Parmetros:
 *      lista (entrada) - lista que ser pesquisada
 *      chave (entrada) - chave de busca
 *
 * Retorno: Endereo dos dados que contm o referido contedo,
 *          se ele for encontrado. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *BuscaSequencialLSE(tListaSE lista, int chave)
{
      /* Enquanto p no assume NULL ou a chave de um n no */
      /* casa com o parmetro 'chave', a busca prossegue    */
   while (lista) {
      if (lista->conteudo.indiceCEP == chave) {
         return &lista->conteudo;
      }

      lista = lista->proximo;
   }

   return NULL;
}

/****
 *
 * AlteraConteudoLSE(): Altera o contedo efetivo de um n
 *                      de uma lista simplesmente encadeada
 *
 * Parmetros:
 *     lista (entrada) - lista que contm o n que
 *                       ser alterado
 *     antigo (entrada) - antigo valor
 *     novo (entrada) - novo valor
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoLSE( tListaSE lista, int indice,
                       const tCEP_Ind *conteudo )
{
   tCEP_Ind *pConteudo = NULL; /* Ponteiro para o contedo */
                                /* que ser alterado        */

      /* Obtm o endereo do contedo efetivo do */
      /* n que casa com o parmetro 'chave'     */
   while (lista) {
       if (lista->conteudo.indiceCEP == indice) {
          pConteudo = &lista->conteudo;
          break;
       } else {
          lista = lista->proximo;
       }
   }

      /* Se 'pConteudo' for NULL o n no foi encontrado */
   if (!pConteudo) {
      return 1; /* N no foi encontrado */
   }

         /*                                 */
         /* O n foi encontrado. Altera-se  */
         /* seu contedo e retorna-se zero. */
         /*                                 */

   *pConteudo = *conteudo; /* Altera o contedo do n */

   return 0; /* Alterao completa */
}

/****
 *
 * DestroiListaSE(): Libera o espao ocupado pelos ns de
 *                   uma lista encadeada, tornando-a vazia
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro para a
 *                             lista simplesmente encadeada que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiListaSE(tListaSE *lista)
{
   tListaSE p; /* Aponta para o prximo n a ser liberado */

   if (!*lista) { /* Verifica se a lista est vazia */
      return; /* Lista vazia no precisa ser destruda */
   }

   p = *lista; /* Faz p apontar para o incio da lista */

      /* Visita cada n da lista liberando-o */
   do {
         /* Passa para o prximo n antes que o n corrente */
         /* seja destrudo. No importa que o ponteiro para */
         /*o incio da lista seja alterado. Afinal, este    */
         /* ponteiro ser mesmo anulado.                    */
      *lista = (*lista)->proximo;

      free(p); /* Libera o espao do n corrente */

         /* Faz p apontar para o prximo n, cujo */
         /* endereo est armazenado em *lista    */
      p = *lista;
   } while (p);

      /* Neste ponto, a lista ficou vazia, pois o  */
      /* ltimo valor assumido por *lista foi NULL */
}

/****
 *
 * ProximoListaSE(): Encontra o prximo n de uma lista
 *                   simplesmente encadeada
 *
 * Parmetros:
 *     lista (entrada) - ponteiro para a lista
 *                       simplesmente encadeada
 *
 * Retorno: Endereo do contedo do prximo n
 *
 ****/
tCEP_Ind *ProximoListaSE(tListaSE lista)
{
   static tNoSE *proximoNo = NULL;

   if (lista) {
      if (!proximoNo) {
         proximoNo = lista;
      } else {
         proximoNo = proximoNo->proximo;
      }
      return &proximoNo->conteudo;
   } else {
      return NULL;
   }
}
