/****
 *
 * Exemplo: Piso de Chave
 *
 * Seo: 3.7.3
 *
 * Descrio:
 *    -> Uma tabela de busca  implementada como tabela indexada
 *       sem ordenao contendo as chaves e os ndices dos registros
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Arquivo de dados: CEPs.bin
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Entrada e sada         */
#include <string.h>   /* Funes strXXX()        */
#include <stdlib.h>   /* malloc()                */
#include "ASSEGURA.h" /* Macro ASSEGURA()        */
#include "Tabela.h"    /* Interface do tipo tabela */

/****************** Complemento de Tipo *******************/

struct rotTabelaIdx {
   tCEP_Ind *elementos; /* Ponteiro para o array que contm os elementos */
   int        nElementos; /* Nmero de elementos */
   int        tamanhoArray;
};

/***************** Definies de Funes ******************/

/****
 *
 * CriaTabelaIdx(): Cria uma tabela
 *
 * Parmetros: No h
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
tTabelaIdx CriaTabelaIdx(void)
{
   tTabelaIdx tabela;

      /* Aloca espao para a estrutura que armazena a tabela */
   tabela = malloc(sizeof(struct rotTabelaIdx));

      /* Garante que a alocao realmente ocorreu */
   ASSEGURA(tabela, "Nao foi possivel alocar tabela");

      /* Aloca previamente um array para um elemento */
   tabela->elementos = calloc(1, sizeof(tCEP_Ind));

      /* Garante que o array foi realmente alocado */
   ASSEGURA( tabela->elementos, "Nao foi possivel alocar "
             "array para conter elementos da tabela" );

      /* Atualiza a varivel que armazena o tamanho do array */
   tabela->tamanhoArray = 1;

   tabela->nElementos = 0; /* Inicialmente a tabela est vazia */

   return tabela;
}

/****
 *
 * DestroiTabelaIdx(): Libera o espao ocupado pela tabela
 *
 * Parmetros: tabela (entrada) - a tabela
 *
 * Retorno: Nada
 *
 ****/
void DestroiTabelaIdx(tTabelaIdx tabela)
{
   free(tabela->elementos); /* Libera o array que contm os elementos */
   free(tabela); /* Libera a prpria tabela */
}

/****
 *
 * ComprimentoIdx(): Calcula o comprimento de uma tabela
 *
 * Parmetros: *tabela (entrada) - a tabela cujo comprimento ser calculado
 *
 * Retorno: o nmero de elementos na tabela
 *
 ****/
int ComprimentoIdx(tTabelaIdx tabela)
{
   return tabela->nElementos; /* No h o que calcular */
}

/****
 *
 * ObtemElementoIdx(): Obtm o elemento de uma tabela que se
 *                     encontra numa dada posio
 *
 * Parmetros:
 *     *tabela (entrada) - a tabela
 *      indice (entrada) - o elemento
 *
 * Retorno: o nmero de elementos na tabela
 *
 ****/
tCEP_Ind ObtemElementoIdx(tTabelaIdx tabela, int indice)
{
   ASSEGURA(indice >= 0 && indice < tabela->nElementos, "Elemento inexistente");

   return tabela->elementos[indice];
}

/****
 *
 * RemoveElementoIdx(): Remove um elemento numa dada posio
 *                      de uma tabela sem ordenao
 *
 * Parmetros:
 *      *tabela (entrada/sada) - tabela onde ser feita a remoo
 *       indice (entrada) - posio do elemento que ser removido
 *
 * Retorno: O item removido
 *
 ****/
tCEP_Ind RemoveElementoIdx(tTabelaIdx tabela, int indice)
{
   tCEP_Ind itemRemovido;

         /* Verifica se o ndice  vlido */
   ASSEGURA( indice >= 0 && indice < tabela->nElementos,
             "Posicao de remocao inexistente" );

   itemRemovido = tabela->elementos[indice];

      /* Remover um elemento significa mover cada elemento uma posio  */
      /* para trs a partir  do sucessor do elemento que ser  removido */
   for (int i = indice; i < tabela->nElementos - 1; i++)
      tabela->elementos[i] = tabela->elementos[i + 1];

   --tabela->nElementos; /* O tamanho da tabela diminuiu */

   return itemRemovido;
}

/****
 *
 * EstaVaziaIdx(): Verifica se uma tabela sem ordenao est vazia
 *
 * Parmetros: *tabela (entrada) - a tabela que ser testada
 *
 * Retorno: 1, se a tabela estiver vazia; 0, em caso contrario
 *
 ****/
int EstaVaziaIdx(tTabelaIdx tabela)
{
   return tabela->nElementos == 0;
}

/****
 *
 * BuscaSequencialIdx(): Efetua uma busca sequencial numa tabela indexada
 *
 * Parmetros:
 *       tabela (entrada) - tabela na qual ser efetuada a busca
 *       chaveProcurada (entrada) - valor da chave do elemento
 *                                  que ser procurado
 *
 * Retorno: ndice do primeiro elemento no array que casa com
 *          o valor recebido como parmetro.
 *          Um valor negativo, se o valor no for encontrado.
 *
 ****/
int BuscaSequencialIdx(tTabelaIdx tabela, tCEP chaveProcurada)
{
   for (int i = 0; i < tabela->nElementos; ++i)
         /* Verifica se a chave foi encontrada */
      if (!strcmp(tabela->elementos[i].chave, chaveProcurada))
         return i; /* Elemento foi encontrado */

   return -1; /* Elemento no foi encontrado */
}

/****
 *
 * AcrescentaElementoIdx(): Acrescenta um novo elemento ao final
 *                          de uma tabela sem ordenao
 *
 * Parmetros:
 *      tabela (sada) - tabela na qual ser feita o acrscimo
 *      elemento (entrada) - valor do elemento que ser acrescentado
 *
 * Retorno: Nada
 *
 ****/
void AcrescentaElementoIdx(tTabelaIdx tabela, const tCEP_Ind *elemento)
{
   tCEP_Ind *novoArray; /* Ponteiro para array redimensionado */

      /* Se o array estiver repleto, tenta redimension-lo */
   if (tabela->nElementos >= tabela->tamanhoArray) {
      tabela->tamanhoArray *= 2; /* O tamanho do array ir dobrar */

         /* Tenta redimensionar o array */
      novoArray = realloc(tabela->elementos,
                          tabela->tamanhoArray*sizeof(tCEP_Ind));

         /* Checa se houve realocao de memria */
      ASSEGURA(novoArray, "Nao houve redimensionamento");

         /* O ponteiro que apontava para o incio do array   */
         /* pode no ser mais vlido e  preciso atualiz-lo */
      tabela->elementos = novoArray;
   }

      /* Acrescenta o novo elemento ao final da tabela */
   tabela->elementos[tabela->nElementos] = *elemento;

   ++tabela->nElementos; /* O tamanho da tabela aumentou */
}

/****
 *
 * AlteraElementoIdx(): Altera o valor de um elemento que se
 *                   encontra num determinada posio de
 *                   uma tabela sem ordenao
 *
 * Parmetros:
 *      *tabela (entrada e sada) - tabela na qual o elemento
 *                                  se encontra
 *       elemento (entrada) - valor do elemento que ser
 *                            acrescentado
 *
 * Retorno: Nada
 *
 ****/
void AlteraElementoIdx(tTabelaIdx tabela, int indice,
                    const tCEP_Ind *novoValor)
{
         /* Garante que o ndice  vlido */
   ASSEGURA(indice >= 0 && indice < tabela->nElementos,
            "Posicao de aleracao inexistente");

      /* Altera o valor do elemento */
   tabela->elementos[indice] = *novoValor;
}

/****
 *
 * BuscaPisoIdx(): Efetua uma busca de piso sequencial numa tabela indexada
 *
 * Parmetros:
 *      tabela (entrada) - tabela na qual ser efetuada a busca
 *      chave (entrada) - chave cujo piso ser procurado
 *
 * Retorno: Endereo do elemento que contm o piso da chave
 *          recebida como parmetro ou NULL se tal piso no
 *          for encontrado.
 *
 ****/
tCEP_Ind *BuscaPisoIdx( tTabelaIdx tabela, tCEP chave )
{
   tCEP piso;
   int  teste, indicePiso;

   strcpy(piso, "");

   for (int i = 0; i < tabela->nElementos; ++i) {
      teste = strcmp(tabela->elementos[i].chave, chave);

      if (!teste)
         return &tabela->elementos[i]; /* Chave foi encontrada e ser o piso */
      else if (teste < 0)
            /* A chave corrente  menor do que a chave de busca */
         if (strcmp(tabela->elementos[i].chave, piso) > 0) {
               /* A chave corrente  maior do que o piso */
               /* corrente e passar a ser o novo piso   */
            strcpy(piso, tabela->elementos[i].chave);
            indicePiso = i;
         }
   }

   return *piso ? &tabela->elementos[indicePiso] : NULL;
}
