/****
 *
 * Arquivo: ArvoreFunil.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2016
 * ltima modificao: 22/07/2016
 *
 * Descrio: Implementao do mdulo ArvoreFunil
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada    */
#include <stdlib.h>   /* Funo exit()               */
#include <string.h>   /* Funes strXXX() e memXXX() */
#include "ArvoreFunil.h" /* Interface deste mdulo   */
#include "ASSEGURA.h"    /* Macro ASSEGURA           */

/***************** Definies de Funes ******************/

/********************* Funes Locais *********************/

/****
 *
 * ConstroiNoArvoreBB(): Constri um n de uma rvore binria
 *
 * Parmetros:
 *       item (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
static tArvoreBB ConstroiNoArvoreBB(tCEP_Ind item)
{
   tArvoreBB no;

      /* Tenta alocar o novo n */
   no = malloc(sizeof(tNoArvoreBB));

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

      /* Preenche o contedo do n */
   no->conteudo = item;
   no->esquerda = NULL; /* Este n ainda no tem filho */
   no->direita = NULL;

   return no;
}

/****
 *
 * RotacaoDireitaArvoreBB(): executa uma rotao direita sobre uma rvore binria
 *
 * Parmetros:
 *      y (entrada) - raiz da rvore que sofrer rotao
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e a varivel local so rotulados
 *             de acordo com a figura no texto, mas,
 *             diferentemente do que o ocorre no texto, x e y
 *             so ponteiros para ns e no ns.
 *
 ****/
static tArvoreBB RotacaoDireitaArvoreBB(tArvoreBB x)
{
   tArvoreBB y = x->esquerda; /* Guarda o endereo da nova */
                              /* raiz aps a rotao       */

      /*                                             */
      /* A ordem das instrues a seguir  essencial */
      /*                                             */

      /* O filho direito de y torna-se filho esquerdo de x */
   x->esquerda = y->direita;

      /* x torna-se filho direito de y */
   y->direita = x;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * RotacaoEsquerdaArvoreBB(): executa uma rotao esquerda sobre uma
 *                    rvore afunilada
 *
 * Parmetros:
 *      x (entrada) - raiz da rvore que sofrer rotao
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e a varivel local so rotulados
 *             de acordo com a figura no texto, mas,
 *             diferentemente do que o ocorre no texto, x e y
 *             so ponteiros para ns e no ns.
 *
 ****/
static tArvoreBB RotacaoEsquerdaArvoreBB(tArvoreBB x)
{
   tArvoreBB y = x->direita; /* Guarda o endereo da nova */
                                /* raiz aps a rotao       */

      /*                                             */
      /* A ordem das instrues a seguir  essencial */
      /*                                             */

      /* O filho esquerdo de y torna-se filho direito de x */
   x->direita = y->esquerda;

      /* x torna-se filho esquerdo de y */
   y->esquerda = x;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * MenorNoArvoreBB(): Encontra o n que contm a menor chave de uma vore
 *                    binria de busca
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da vore
 *
 * Retorno: Endereo do menor n da rvore
 *
 ****/
static tArvoreBB MenorNoArvoreBB(tArvoreBB arvore)
{
   ASSEGURA(arvore, "Arvore vazia nao tem menor no");

      /* Encontra o n mais  esquerda da rvore */
   while (arvore->esquerda)
      arvore = arvore->esquerda;

   return arvore;
}

/****
 *
 * Afunila(): Afunila de modo descendente um n de uma rvore
 *            binria de busca que contm uma chave especificada
 *
 * Parmetros:
 *      arvore (entrada/sada) - raiz da rvore que ser afunilada
 *      chave (entrada) - chave do n que ser afunilado
 *
 * Retorno: Raiz da rvore aps o afunilamento
 *
 * Observao: Se a chave no for encontrada, o ltimo n
 *             acessado se tornar a raiz da rvore
 *
 ****/
tArvoreBB Afunila(tArvoreBB arvore, tCEP chave)
{
   int         compara;/* Resultado de comparao de duas chaves */
   tNoArvoreBB noAux; /* Um n auxiliar */
   tArvoreBB   esq,   /* Ponteiro auxiliar */
               dir;   /* Outro ponteiro auxiliar */

      /* Passo 1: Se a rvore estiver vazia, retorne */
   if (!arvore)
      return arvore;

      /* Passo 2: Faa 'esq' e 'dir' apontarem para 'noAux' */
   esq = dir = &noAux;
   noAux.esquerda = noAux.direita = NULL;

      /* Passo 3: Enquanto o n contendo a chave a ser   */
      /* afunilada no for a raiz ou filho da raiz, faa */
   while(1) {
         /* Compara a chave do n a ser afunilado com a chave da raiz */
      compara = strcmp(chave, arvore->conteudo.chave);

         /* Passo 3.1: Verifique se a chave encontra-se na subrvore esquerda */
      if (compara < 0) {
            /* Passo 3.1.1: Se a chave no se encontra */
            /* na subrvore esquerda, encerre o lao   */
         if (!arvore->esquerda)
            break;

            /* Passo 3.1.2: Verifique se a chave  menor do */
            /* que a chave da raiz da subrvore esquerda    */

         compara = strcmp(chave, arvore->esquerda->conteudo.chave);

         if (compara < 0) {
               /* Passo 3.1.2.1: Faa uma rotao zig na rvore */
            arvore = RotacaoDireitaArvoreBB(arvore);

               /* Passo 3.1.2.2: Encerre o lao se */
               /* a subrvore esquerda ficou vazia */
            if (!arvore->esquerda)
               break;
         }

            /* Passo 3.1.3: Faa a subrvore esquerda  */
            /* do ponteiro 'dir' apontar para a rvore */
         dir->esquerda = arvore;

            /* Passo 3.1.4: Faa o ponteiro 'dir' apontar para a rvore */
         dir = arvore;

            /* Passo 3.1.5: Faa o ponteiro que representa */
            /* a rvore apontar para seu filho esquerdo    */
         arvore = arvore->esquerda;

         /* Passo 3.2: Verifique se a chave se */
         /* encontra na subrvore direita      */
      } else if (compara > 0) {
            /* Passo 3.2.1: Se a chave no se encontra */
            /* na subrvore direita, encerre o lao    */
         if (!arvore->direita)
            break;

            /* Passo 3.2.2: Verifique se a chave  maior do */
            /* que a chave da raiz da subrvore direita     */

         compara = strcmp(chave, arvore->direita->conteudo.chave);

         if (compara > 0) {
               /* Passo 3.2.2.1: Faa uma rotao zag na rvore */
            arvore = RotacaoEsquerdaArvoreBB(arvore);

               /* Passo 3.2.2.2: Encerra o lao se */
               /* a subrvore direita ficou vazia  */
            if (!arvore->direita)
               break;
         }

            /*  Passo 3.2.3: Faa a subrvore direita do */
            /* ponteiro 'esq' apontar para a rvore      */
         esq->direita = arvore;

            /* Passo 3.2.4: Faa o ponteiro 'esq' apontar para a rvore */
         esq = arvore;

            /*  Passo 3.2.5: Faa o ponteiro que representa */
            /* a rvore apontar para seu filho direito      */
         arvore = arvore->direita;
      } else /* A chave j est na raiz */
         break;
   }

      /* Passo 4: Faa a subrvore direita do ponteiro */
      /* 'esq' apontar para o filho esquerdo da rvore */
   esq->direita = arvore->esquerda;

      /* Passo 5: Faa a subrvore esquerda do ponteiro */
      /* 'dir' apontar para o filho direito da rvore   */
   dir->esquerda = arvore->direita;

      /* Passo 6: Faa o filho esquerdo da rvore */
      /* apontar para o filho direito de 'noAux'  */
   arvore->esquerda = noAux.direita;

      /* Passo 7: Faa o filho direito da rvore  */
      /* apontar para o filho esquerdo de 'noAux' */
   arvore->direita = noAux.esquerda;

   return arvore;
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaArvoreBB(): Inicia uma rvore binria afunilada
 *
 * Parmetros: *arvore (sada) - a arvore que ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void IniciaArvoreBB(tArvoreBB *arvore)
{
   *arvore = NULL;
}

/****
 *
 * NumeroNosArvoreBB(): Calcula o nmero de ns de uma rvore
 *                      binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosArvoreBB(tArvoreBB arvore)
{
      /* Se a rvore for vazia seu nmero de ns  0 */
   if (!arvore)
      return 0;

      /* Leva em considerao a raiz da rvore e os */
      /* ns das subrvores esquerda e direita      */
   return 1 + NumeroNosArvoreBB(arvore->esquerda) +
              NumeroNosArvoreBB(arvore->direita);
}

/****
 *
 * AlturaArvoreBB(): Calcula a altura de uma
 *                         rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A altura da referida rvore
 *
 ****/
int AlturaArvoreBB(tArvoreBB arvore)
{
   int altEsq, altDir;

   if (!arvore)
      return 0; /* rvore vazia tem altura zero */

     /* Calcula a altura da subrvore esquerda */
   altEsq = 1 + AlturaArvoreBB(arvore->esquerda);

     /* Calcula a altura da subrvore direita */
   altDir = 1 + AlturaArvoreBB(arvore->direita);

     /* A altura da rvore  a maior */
   return altEsq > altDir ? altEsq : altDir;
}

/****
 *
 * BuscaArvoreFunil(): Retorna o endereo do contedo do n que
 *                     possui uma determinada chave numa rvore afunilada
 *
 * Parmetros:
 *       arvore (entrada) - rvore que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *BuscaArvoreFunil(tArvoreBB *arvore, tCEP chave)
{
      /* Verifica se a rvore est vazia */
   if (!*arvore)
      return NULL; /* A rvore est vazia */

      /* Afunila o n que contm a chave de busca */
   *arvore = Afunila(*arvore, chave);

      /* Se a chave foi encontrada, ela agora est na raiz */
   if (!strcmp(chave, (*arvore)->conteudo.chave))
      return &(*arvore)->conteudo;

   return NULL; /* A chave no foi encontrada */
}

/****
 *
 * InsereArvoreFunil(): Insere um novo n numa rvore afunilada
 *
 * Parmetros:
 *    *arvore (entrada/sada) - ponteiro para a raiz da rvore
 *                              na qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: 1, se no houver insero porque a chave j existe
 *          0, se a insero ocorrer
 *
 ****/
int InsereArvoreFunil(tArvoreBB *arvore, tCEP_Ind *conteudo)
{
   tArvoreBB pNovoNo; /* Ponteiro para n que ser inserido */
   int          compara; /* Resultado da comparao de duas chaves */

      /* Se a rvore estiver vazia torna o novo n sua raiz e retorna */
   if (!*arvore) {
      *arvore = ConstroiNoArvoreBB(*conteudo);
      return 0;
   }

      /* Afunila o n que contm a chave do n a ser inserido */
   *arvore = Afunila(*arvore, conteudo->chave);

      /* Compara a chave do n a ser inserido com a chave que */
      /* se encontra na raiz da rvore aps o afunilamento    */
   compara = strcmp(conteudo->chave, (*arvore)->conteudo.chave);

      /* Verifica se a chave j existe */
   if (!compara)
      return 1; /* A chave j existe e  primria */

   pNovoNo = ConstroiNoArvoreBB(*conteudo); /* Constri o novo n */

      /* Faz com que o novo n se torne a raiz da rvore */
   if (compara < 0) {
      /*                                                         */
      /* A chave do novo n  menor do que a chave da raiz atual */

         /* Faz o filho direito do novo n apontar para a raiz */
      pNovoNo->direita = *arvore;

         /* Faz o filho esquerdo do novo n apontar */
         /* para o filho esquerdo da raiz           */
      pNovoNo->esquerda = (*arvore)->esquerda;

      (*arvore)->esquerda = NULL; /* Torna nulo o filho esquerdo da raiz */
   } else {
      /*                                                         */
      /* A chave do novo n  maior do que a chave da raiz atual */

         /* Faz o filho esquerdo do novo n apontar para a raiz */
      pNovoNo->esquerda = *arvore;

         /* Faz o filho direito do novo n apontar para o filho direito da raiz */
      pNovoNo->direita = (*arvore)->direita;
      (*arvore)->direita = NULL;
   }

   *arvore = pNovoNo; /* O novo n passa a ser a nova raiz */

   return 0; /* Insero foi OK */
}

/****
 *
 * RemoveArvoreFunil(): Remove de uma rvore afunilada um n que
 *                        contm a chave especificada como parmetro
 *
 * Parmetros:
 *     arvore (entrada/sada) - endereo do ponteiro que representa
 *                              a rvore na qual ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: 0, se a remoo for bem sucedida, ou 1, se o n a
 *          ser removido no for encontrado.
 *
 ****/
int RemoveArvoreFunil(tArvoreBB *arvore, tCEP chave)
{
   tArvoreBB pE, /* Apontar para o filho esquerdo da raiz */
                pD, /* Apontar para o filho direito da raiz  */
                pS, /* Apontar para o sucessor da raiz */
                p;  /* Um ponteiro auxiliar */

      /* Verifica se a rvore est vazia */
   if (!*arvore)
      return 1; /* No h o que remover */

      /* Afunila o n que ser removido */
   *arvore = Afunila(*arvore, chave);

      /* Se a chave foi encontrada, ela agora est na raiz */
   if (strcmp(chave, (*arvore)->conteudo.chave))
      return 1; /* Chave no foi encontrada */

   /*                                                      */
   /* A chave encontra-se na raiz e ser removida a seguir */
   /*                                                      */

      /* Verifica quantos filhos a raiz possui */
   if (!(*arvore)->esquerda && !(*arvore)->direita) {
         /* A raiz no tem filhos logo a rvore ficar vazia aps a remoo */
      free(*arvore);
      *arvore = NULL;
   } else if ((*arvore)->esquerda && (*arvore)->direita) {
      /*                        */
      /* A raiz tem dois filhos */

      pE = (*arvore)->esquerda; /* Faz pE apontar para o filho esquerdo da raiz */

      pD = (*arvore)->direita; /* Faz pD apontar para o filho direito da raiz */

         /* A raiz j pode ser liberada */
      free(*arvore);

         /* Obtm o sucessor imediato da antiga raiz */
      pS = MenorNoArvoreBB(pD);

         /* Afunila o sucessor imediato da antiga raiz */
      pD = Afunila(pD, pS->conteudo.chave);

         /* Apesar de o sucessor ter sido movido para a raiz, */
         /* ele ainda deve ocupar o mesmo endereo em memria */
      ASSEGURA(pS == pD, "Sucessor nao esta' na raiz");

         /* O n apontado por pD no pode ter filho esquerdo */
      ASSEGURA(!pD->esquerda, "Sucessor tem filho esquerdo");

         /* Faz com que o filho esquerdo do sucessor */
         /* seja a rvore apontada por pE            */
      pD->esquerda = pE;

         /* Faz o ponteiro que representa a rvore apontar para a nova raiz */
      *arvore = pD;
   } else {
      /*                                           */
      /* A raiz s tem um filho. Esse filho ser a */
      /* nova raiz e a raiz antiga ser removida.  */
      /*                                           */

      p = *arvore; /* Guarda o endereo da raiz para remov-la depois */

         /* Faz a raiz apontar para seu nico filho */
      *arvore = (*arvore)->esquerda ? (*arvore)->esquerda : (*arvore)->direita;

      free(p); /* Remove a raiz antiga */
   }

   return 0; /* A remoo foi bem sucedida */
}

/****
 *
 * EstaVaziaArvoreFunil(): Verifica se uma rvore binria
 *                      est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreFunil(tArvoreBB arvore)
{
   return arvore == NULL;
}

/****
 *
 * AlteraConteudoArvoreFunil(): Altera o contedo efetivo de um
 *                           n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser
 *                        alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreFunil( tArvoreBB *arvore, tCEP chave,
                               tCEP_Ind *conteudo )
{
   tCEP_Ind *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaArvoreFunil(arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;
         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
   if (!RemoveArvoreFunil(arvore, chave)) {
      if (!InsereArvoreFunil( arvore, conteudo )) {
         return 0; /* Alterao OK */
      } else {
         return 1; /* Insero no foi possvel */
      }
   } else {
      return 1; /* Remoo no foi possvel */
   }
}

/****
 *
 * DestroiArvoreBB(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz da rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBB(tArvoreBB arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBB(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreBB(arvore->direita);

      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * CaminhamentoInfixoBB(): Efetua um caminhamento em ordem
 *                        infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixoBB( tArvoreBB arvore, tOperacao op, FILE *stream )
{
   if (arvore) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixoBB(arvore->esquerda, op, stream);

      op(arvore->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixoBB(arvore->direita, op, stream);
   }
}

/****
 *
 * MenorChaveArvoreBB(): Obtm o contedo do n que possui a menor
 *                       chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *MenorChaveArvoreBB(tArvoreBB p)
{
   tArvoreBB q; /* Apontar para o n que armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p)
      return NULL;

      /* A menor chave encontra-se no n mais  esquerda da rvore de busca */
   while (p) {
      q = p; /* Faz q apontar para o n para o qual p aponta */

      p = p->esquerda; /* Faz p apontar para o n  esquerda */
   }
      /* Neste ponto, q aponta para o n que armazena a menor chave */
   return &q->conteudo;
}

/****
 *
 * MaiorChaveArvoreBB(): Obtm o contedo do n que possui a maior
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *MaiorChaveArvoreBB(tArvoreBB p)
{
   tArvoreBB q; /* Apontar para o n que armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p)
      return NULL;

      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
      q = p; /* Faz q apontar para o n para o qual p aponta */

      p = p->direita; /* Faz p apontar para o n  direita */
   }
      /* Neste ponto, q aponta para o n que armazena a maior chave */
   return &q->conteudo;
}
