/****
 *
 * Exemplo: rvores Binrias Afuniladas
 *
 * Seo: 4.6
 *
 * Descrio:
 *    -> A tabela de busca  implementada como rvore afunilada
 *       em que cada n contm uma a chave e a posio do
 *       respetivo registro no arquivo
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Tipo de busca: com rvore binria afunilada
 *
 *    -> Arquivo de dados: CEPs.bin
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* Funo strtod()          */
#include <time.h>     /* Medio de tempo         */
#include "ArvoreFunil.h" /* Interface do tipo tArvoreFunil */
#include "Leitura.h"   /* Leitura via teclado      */
#include "Registros.h" /* Interface do mdulo Registros */
#include "ASSEGURA.h"  /* Macro ASSEGURA()        */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo de dados */
#define NOME_ARQUIVO_BIN "I:\\Dados\\CEPs\\CEPs.bin"

   /* Nome do arquivo contendo as chaves da tabela de busca */
#define NOME_ARQUIVO_CHAVES "ListaDeChaves.txt"

/**************** Compilao Condidicional ****************/

#define DEPURACAO 0

/***************** Definies de Funes ******************/

/****
 *
 * ExibeConteudo(): Escreve num stream o campo 'conteudo'
 *                  de um n de uma rvore binria
 *
 * Parmetros:
 *      conteudo (entrada) - contedo que ser impresso
 *      stream (entrada) - stream no qual os dados sero
 *                         escritos
 *
 * Retorno: Nenhum
 *
 ****/
void ExibeConteudo(tConteudo conteudo, FILE *stream)
{
   fprintf( stream, "%s\t%d\n", conteudo.chave,
               conteudo.valor );
}

/****
 *
 * ExibeLista() Escreve em arquivo os contedos dos ns de
 *              uma rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - a tabela
 *
 * Retorno: Nada
 *
 ****/
void ExibeLista(tArvoreFunil arvore)
{
   FILE *stream;

      /* Tenta abrir o arquivo para escrita em modo texto */
   stream = fopen(NOME_ARQUIVO_CHAVES, "w");

      /* Se arquivo no foi aberto, aborta o programa */
   ASSEGURA(stream,
            "Impossivel criar arquivo para escrita de chaves");

   printf("\nEscrevendo tabela em arquivo...");

      /* Cabealho de apresentao */
   fprintf( stream,
            "   Chave   \tIndice\n"
            "   =====   \t======\n");

      /* Apresenta cada estrutura numa linha separada */
   CaminhamentoInfixo( arvore, ExibeConteudo, stream );

   fclose(stream);

   printf( "\n*** Resultado escrito no arquivo %s *** \n",
           NOME_ARQUIVO_CHAVES );
}

/****
 *
 * LeArquivo(): L arquivo de binrio e armazena as
 *              chaves e os ndices dos registro numa
 *              rvore binria de busca
 *
 * Parmetros:
 *      arqBin (entrada) - nome do arquivo binrio
 *     *arvore (sada) - arvore que conter as chaves e
 *                       os ndices dos registros lidos
 *
 * Retorno: Stream associado ao arquivo aberto para
 *          leitura e escrita
 *
 ****/
FILE *LeArquivo( const char *arqBin, tArvoreFunil *arvore )
{
   FILE      *stream;
   tConteudo  umElemento;
   tRegistro  umRegistro;
   int        indiceReg = 0;

      /* Tenta abrir o arquivo binrio para leitura e escrita */
   stream = fopen(arqBin, "rb+");

      /* Se o arquivo no pode ser aberto, aborta o programa */
   ASSEGURA(stream, "Arquivo nao pode ser aberto");

      /* L cada registro do arquivo e acrescenta */
      /* sua chave (CEP) e seu ndice na rvore   */
   while (1) {
         /* L um registro no arquivo */
      fread( &umRegistro, sizeof(umRegistro), 1, stream);

         /* Verifica se ocorreu erro ou o */
         /* final do arquivo foi atingido */
      if (ferror(stream) || feof(stream)) {
         break;
      }

         /* A chave no  armazenada como string no registro. */
         /* Portanto, no se pode usar strcpy().              */
      strncpy(umElemento.chave, umRegistro.CEP, TAM_CEP);

         /* Transforma a chave num string. Isso no  */
         /* essencial, mas facilita o processamnto.   */
      umElemento.chave[TAM_CEP] = '\0';

         /* Armazena no novo elemento da tabela o */
         /* ndice do registro no arquivo e, ao   */
         /* mesmo tempo, incrementa esse ndice   */
      umElemento.valor = indiceReg++;

         /* Acrescenta o novo elemento  tabela */
      InsereNoArvoreFunil(arvore, &umElemento);
   }

      /* Se ocorreu erro de leitura no arquivo, fecha */
      /* o arquivo e aborta o programa                */
   if (ferror(stream)) {
      fclose(stream);
      ASSEGURA(0, "Erro de leitura de arquivo ");
   }

   return stream;
}

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
static void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\nEscolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t[%d] %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * AtualizaArquivoBin(): Cria um novo arquivo binrio contendo
 *                       cada registro de outro arquivo binrio
 *                       cuja chave no faa parte de uma arvore
 *                       de registros que devem ser removidos
 *
 * Parmetros:
 *      nomeArquivo (entrada) - nome do arquivo binrio original
 *      removidos (entrada) - arvore contendo as estruturas que
 *                            NO sero escritas no novo arquivo
 *
 * Retorno: Nenhum
 *
 ****/
void AtualizaArquivoBin( const char *nomeArquivo,
                         tArvoreFunil removidos )
{
   char      ar[L_tmpnam]; /* Nome de um arquivo temporrio */
   FILE     *stream, /* Stream associado ao arquivo original */
            *streamTmp; /* Stream associado ao temporrio */
   tRegistro umRegistro; /* Um registro do arquivo */
   tChave    umaChave;
   int       ocorreuErro = 0; /* Indicar se ocorreu erro */

      /* Tenta abrir o arquivo original para leitura */
   ASSEGURA( stream = fopen(nomeArquivo, "rb"),
             "Nao foi possivel atualizar arquivo" );

      /* Obtm o nome de um arquivo temporrio */
   tmpnam(ar);

      /* Tenta abrir o arquivo temporrio */
      /* para escrita em modo binrio     */
   ASSEGURA(streamTmp = fopen(ar, "wb"),
            "Nao foi possivel abrir arquivo temporario");

      /* O lao encerra quando no for */
      /* mais possvel ler um registro */
   while (1) {
         /* L um registro no arquivo de entrada */
      fread(&umRegistro, sizeof(umRegistro), 1, stream);

         /* Testa se final do arquivo de entrada foi */
         /* atingido ou ocorreu erro de leitura      */
      if (feof(stream) || ferror(stream)) {
         break; /* Processamento encerrado */
      }

         /* Cria uma chave (string). O campo CEP */
         /* do registro no  string.            */
      strncpy(umaChave, umRegistro.CEP, TAM_CEP);
      umaChave[TAM_CEP] = '\0';

         /* Se a chave do registro lido no faz parte da  */
         /* arvore de removidos, escreve-o no novo arquivo */
      if (!BuscaArvoreFunil(&removidos, umaChave)) {
            /* Escreve o registro lido no arquivo de sada */
         fwrite(&umRegistro, sizeof(umRegistro), 1, streamTmp);

            /* Verifica se ocorreu erro de escrita */
         if (ferror(streamTmp)) {
            break;
         }
      }
   }

      /* Verifica se ocorreu erro de leitura ou escrita */
   if ( ferror(stream) || ferror(streamTmp) ) {
      ocorreuErro = 1;
   }

      /* Fecha os arquivos */
   fclose(stream);
   fclose(streamTmp);

      /* Se ocorreu erro de leitura ou */
      /* escrita, aborta o programa    */
   ASSEGURA(!ocorreuErro, "Erro na atualizacao do arquivo");

      /* Tenta apagar o arquivo original. Se isso */
      /* no for possvel, aborta o programa      */
   ASSEGURA( !remove(nomeArquivo),
             "Erro na atualizacao do arquivo" );

      /* Tenta rebatizar o arquivo temporrio */
      /* com o nome do arquivo original       */
   ocorreuErro = rename(ar, nomeArquivo);

   if (ocorreuErro) {
      printf( "\n\t>>> O resultado da atualizacao se encontra"
              "no arquivo \"%s\". Nao foi possivel "
              "rebatiza-lo.\n", ar );
      ASSEGURA( 0, "Erro em batismo de arquivo" );
   }
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/********************* Funo main() **********************/

/****
 *
 * main(): Implementa operaes sobre tabela implementada
 *         como rvore afunilada contendo chaves de registros
 *         do tipo tCEP
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tArvoreFunil arvore, /* Tabela de busca */
                removidos; /* Chaves dos registros a ser  */
                           /* removidos quando o arquivo  */
                           /* de registros for atualizado */
   tRegistro   umRegistro;    /* Armazenam um registro */
   tConteudo   conteudo, /* Contedo de um n da rvore */
              *ptrConteudo; /* Ponteiro para o contedo */
                            /* de um n da arvore       */
   const char *opcoes[] = { "Consulta dados de um CEP",
                            "Acrescenta um CEP",
                            "Remove um CEP",
                            "Apresenta a arvore de CEPs",
                            "Menor chave",
                            "Maior chave",
                            "Encerra o programa"
                          };
   int         op, /* Opo escolhida pelo usurio */
               nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]);
   char        umCEP[TAM_CEP + 1]; /* Uma chave de busca */
   FILE       *stream; /* Stream associado ao  */
                       /* arquivo de registros */
   int         primeiroByte, /* Indica a posio do */
                             /* primeiro byte de um */
                             /* registro no arquivo */
               atualizarArq = 0; /* Indica se o arquivo */
                                 /* de registros ser   */
                                 /* atualizado          */

   MedidaDeTempo();
   printf("\nCriando arvore...");

      /* Inicia a rvore que armazena a tabela de busca */
   IniciaArvoreFunil(&arvore);

      /* Inicia a rvore que armazena os ndices */
      /* dos registros removidos                 */
   IniciaArvoreFunil(&removidos);

      /* L o contedo do arquivo de  */
      /* dados e o armazena na rvore */
   stream = LeArquivo(NOME_ARQUIVO_BIN, &arvore);

   printf("\n...arvore criada com sucesso\n");
   MedidaDeTempo();

#if DEPURACAO == 0

   int nRegistros, /* Nmero de registros no arquivo  */
       nNos,       /* Tamanho da tabela de busca      */
       prof;       /* Profundidade da rvore de busca */

   MedidaDeTempo();
   printf("\nChecando a consistencia do programa...");

      /* Calcula o nmero de registros do arquivo */
   nRegistros = NumeroDeRegistros( stream, sizeof(tRegistro) );

      /* Calcula o nmero de ns na rvore */
   nNos = NumeroNosArvoreFunil(arvore);

      /* Calcula a profundidade da rvore */
   prof = ProfundidadeArvoreFunil(arvore);

   printf( "\n\t>>> Numero de registros no arquivo: %d",
           nRegistros );
   printf("\n\t>>> Numero de nos na arvore: %d", nNos);
   printf("\n\t>>> Profundidade da arvore: %d", prof);

   ASSEGURA( nNos == nRegistros,
             "This program is fucking wrong!!!" );

   printf("\n...consistencia OK");
   MedidaDeTempo();

#endif

      /* O lao a seguir encerra quando o usurio */
      /* escolher a opo de encerramento         */
   while (1) {
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("1234567");

         /* Verifica se o usurio quer encerrar o programa */
      if (op == '7') { /* Encerra o programa */
         fclose(stream);

            /* A tabela de busca no  mais necessria */
         DestroiArvoreBin(arvore); /* BUGGY??? */

         break; /* Sada do lao */
      }

         /* Processa as demais opes */
      switch (op) {
         case '1': /* Consulta dados de um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Procura o CEP na tabela de busca */
            ptrConteudo = BuscaArvoreFunil(&arvore, umCEP);

               /* Se a chave for encontrada, utiliza-se o */
               /* ndice associado a ela para recuperar   */
               /* o registro correspondente no arquivo    */
            if (!ptrConteudo) {
               printf("\n>>> CEP nao foi encontrado\n");
            } else {
                  /* Utilizando o ndice do registro, */
                  /* calcula-se a posio no arquivo  */
                  /* do primeiro byte do registro     */
               primeiroByte = ptrConteudo->valor *
                              sizeof(tRegistro);

                  /* Move o apontador de posio para o   */
                  /* primeiro byte do registro. Se isso   */
                  /* no for possvel, aborta o programa. */
               ASSEGURA( !fseek(stream, primeiroByte, SEEK_SET),
                         "Erro de posicionamento em arquivo" );

                  /* L o registro no arquivo */
               fread( &umRegistro, sizeof(umRegistro), 1,
                      stream );

                  /* Se ocorreu erro de leitura aborta */
               ASSEGURA(!ferror(stream), "Erro de leitura");

                  /* Apresenta o registro ao usurio */
               ExibeRegistro(&umRegistro);

               putchar('\n'); /* Embelezamento */
            }

            break;

         case '2': /* Acrescenta um registro de CEP */
               /* L dados do novo CEP */
            LeRegistro(&umRegistro, NULL);

               /* O valor do campo 'numero' corresponde  */
               /* posio do registro no arquivo de CEPs  */
               /* (isso  coisa dos Correios)             */
            umRegistro.numero = NumeroNosArvoreFunil(arvore) + 1;

               /* Cria a chave a ser inserida na tabela de */
               /* busca. Na tabela, a chave  um string    */
               /* mas no registro CEP no  string         */
            strncpy(conteudo.chave, umRegistro.CEP, TAM_CEP);

               /* Torna a chave um string */
            conteudo.chave[TAM_CEP] = '\0';

               /* Acrescenta o ndice do registro ao contedo */
            conteudo.valor = umRegistro.numero;

               /* Insere novo contedo  tabela de busca */
            InsereNoArvoreFunil(&arvore, &conteudo);

               /* Move apontador de posio para o final */
               /* do arquivo. Se isso no for possvel,  */
               /* aborta o programa.                     */
            ASSEGURA( !fseek(stream, 0, SEEK_END),
                      "Erro de posicionamento em arquivo");

               /* Escreve o novo registro no final do arquivo */
            fwrite(&umRegistro, sizeof(umRegistro), 1, stream);

               /* Se ocorreu erro de escrita, aborta */
            ASSEGURA(!ferror(stream),
                     "Erro de escrita em arquivo ");

            printf("\n>>> Acrescimo bem sucedido\n");

            break;

         case '3': /* Remove um CEP */
               /* L um CEP introduzido pelo usurio */
            LeMatricula("CEP", umCEP, TAM_CEP + 1);

               /* Se a chave for encontrada, remove-a da  */
               /* tabela de busca e acrescenta-se um novo */
               /* n na rvore de removidos. A remoo    */
               /* do registro do arquivo s ser efetuada */
               /* ao final do programa.                   */
            if (!RemoveNoArvoreFunil(&arvore, umCEP)) {
                  /* Cria um novo contedo e acrescenta-o */
                  /*  rvore de removidos                */
               strcpy(conteudo.chave, umCEP);
               InsereNoArvoreFunil(&removidos, &conteudo);

                  /* Arquivo de registros precisar */
                  /* ser atualizado                 */
               atualizarArq = 1;

               printf("\n>>> Remocao bem sucedida\n");
            } else {
               printf("\n>>> CEP nao encontrado\n");
            }

            break;

         case '4': /* Apresenta a arvore na tela */
            ExibeLista(arvore);
            break;

         case '5': /* Apresenta a menor chave */
               /* Obtm e exibe a menor chave */
            ptrConteudo = MenorChave(arvore);

            printf( "\n\t>>> Menor chave: %s\n",
                    ptrConteudo->chave );

            break;

         case '6': /* Apresenta a maior chave */
               /* Obtm e exibe a maior chave */
            ptrConteudo = MaiorChave(arvore);

            printf( "\n\t>>> Maior chave: %s\n",
                    ptrConteudo->chave );

            break;

         default: /* O programa no deve chegar at aqui */
            printf("\nEste programa contem um erro\n");
            return 1;
      }
   }

      /* Atualiza arquivo binrio se for necessrio */
   if (atualizarArq) {
      MedidaDeTempo();
      printf("\nAtualizando arquivo...");

         /* Atualiza arquivo binrio */
      AtualizaArquivoBin(NOME_ARQUIVO_BIN, removidos);

      printf("\n... Atualizacao bem sucedida");
      MedidaDeTempo();
   }

      /* Despede-se do usurio */
   printf( "\n\t>>> Obrigado por usar este programa\n\n");

   return 0;
}

/******************* Exemplo de Execuo ******************



**********************************************************/

