/****
 *
 * Arquivo ArvoreMulti.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 20/04/2004
 * ltima modificao: 20/07/2008
 *
 * Descrio: Implementao das funes que executam operaes
 *            usando rvores multidirecionais de busca em
 *            memria principal.
 *
 * NOTA: rvores multidirecionais de busca devem ser
 *       implementadas em memria secundria. Esta
 *       implementao  meramente didtica!
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h> /* Funes de alocao dinmica e outras */
#include <stdio.h>  /* Funes de E/S */
#include <string.h>      /* Funes strXXX() e memXXX() */
#include "ArvoreMulti.h" /* Interface deste mdulo      */
#include "Registros.h"   /* Funes de processamento    */
                         /* de registros                */
#include "leitura.h"     /* Funes de leitura via teclado */
#include "ASSEGURA.h"    /* Macro ASSEGURA e outras macros */
                         /* de tratamento de exceo       */

/******************** Funes de Locais *******************/

/****
 *
 * BuscaEmNo(): Faz uma busca (sequencial) em um n de uma
 *              rvore multidirecional
 *
 * Parmetros:
 *      chave (entrada) - a chave de busca
 *      no (entrada) - ponteiro para o n no qual ser feita
 *                     a busca
 *
 * Retorno: O ndice da primeira chave encontrada dentro do
 *          n que  maior do que ou igual a chave de busca.
 *          Se tal chave no for encontrada, o nmero de chaves
 *
 ****/
static int BuscaEmNo(tChave chave, const tNoMulti *no)
{
   int i,
       nChaves = no->nFilhos - 1; /* No  necessrio, */
                                  /* mas facilita      */

   for (i = 0; i < nChaves; ++i)
      if ( chave <= no->chaves[i].chave )
            /* Encontrada uma chave maior do */
            /* que ou igual  chave de busca */
         return i;

      /* A chave de busca  maior do que */
      /* qualquer chave no n            */
   return nChaves;
}

/****
 *
 * EncontraNo(): Procura uma chave na rvore e retorna um
 *               ponteiro para o n que contm a chave, se
 *               ela for encontrada, ou deveria cont-la,
 *               se ela no for encontrada
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *      chave (entrada) - a chave de busca
 *     *posicao (sada) - posio da primeira chave no n que
 *                         maior do que ou igual a chave de
 *                        busca
 *     *encontrado (sada) - informar se a chave foi
 *                           encontrada ou no
 *
 * Retorno: Ponteiro para o n contendo a chave de busca, se
 *          ela for encontrada; ou onde ela estaria, se ela
 *          no for encontrada
 *
 ****/
static tNoMulti *EncontraNo( tArvoreMulti arvore, tChave chave,
                             int *posicao, int *encontrado )
{
   tNoMulti *p, *q;
   int       i;

   p = arvore; /* Ponteiro que desce a rvore */
   q = NULL;   /* Ponteiro que desce atrs de p */

   while (p) {
      i = BuscaEmNo(chave, p);
      q = p;

            /* A primeira comparao  necessria */
            /* para evitar acesso a uma posio   */
            /* invlida no array chaves[]         */
      if( (i < p->nFilhos - 1) &&
          (chave == p->chaves[i].chave) ) {
         *encontrado = 1;
         *posicao = i;

            /* A chave foi encontrada */
         return p; /* No h mais o que fazer */
      }

      p = p->filhos[i]; /* Desce mais um nvel na rvore */
   }

   /*                                             */
   /* A chave no foi encontrada. Ela deveria ser */
   /* inserida na posio i do n apontado por q  */
   /*                                             */

   *encontrado = 0;
   *posicao = i;

   return q;
}

/****
 *
 * InsereEmFolha(): Insere uma chave na folha incompleta
 *                  especificada
 *
 * Parmetros:
 *      pNo (entrada/sada) - ponteiro para a folha onde
 *                            ser feita a insero
 *      posicao (entrada) - posio da chave no array de
 *                          chaves do n
 *      chaveEIndice (entrada) - a chave (e seu respectivo
 *                               ndice) que ser inserida
 *
 * Retorno: Nada
 *
 ****/
static void InsereEmFolha( tNoMulti *pNo, int posicao,
                           tChaveIndice chaveEIndice )
{
   int i;

   ASSEGURA( pNo->nFilhos < G,
             "Tentativa de insercao em no' e' completo" );

      /* Abre espao para a nova chave */
   for (i = pNo->nFilhos - 1; i > posicao; --i)
      pNo->chaves[i] = pNo->chaves[i - 1];

   pNo->chaves[posicao] = chaveEIndice; /* Insere a chave */

   ++pNo->nFilhos; /* Nasceu mais um filho */
}

/****
 *
 * ConstroiNo(): Cria um n de uma rvore multidirecional
 *               contendo uma nica chave
 *
 * Parmetros:
 *     *chaveIndice (entrada) - um par chave/ndice a ser
 *                              armazenado no n
 *
 * Retorno: Ponteiro para a n criado
 *
 ****/
static tNoMulti *ConstroiNo(const tChaveIndice *chaveIndice)
{
   tNoMulti *pNo;
   int       i;

   pNo = malloc(sizeof(tNoMulti));
   ASSEGURA(pNo, "Nao foi possivel alocar um no");

      /* A chave ser a primeira do n */
   pNo->chaves[0] = *chaveIndice;

      /* O nmero de filhos num n de uma rvore  */
      /* multidirecional  sempre igual ao nmero */
      /* de chaves mais um                        */
   pNo->nFilhos = 2;

      /* Inicia todos os possveis filhos do n com NULL */
   for (i = 0; i < G; ++i)
      pNo->filhos[i] = NULL;

   return pNo;
}

/****
 *
 * CompactaNo(): Compacta um n de uma rvore multidirecional
 *               de busca aps uma operao de remoo de chave
 *
 * Parmetros:
 *      pNo (entrada) - ponteiro para o n que ser compactado
 *      posicao (entrada) - posio da chave removida
 *
 * Retorno: Nada
 *
 ****/
static void CompactaNo(tNoMulti *pNo, int posicao)
{
   int i;

      /* A chave removida deve ter pelo menos um filho vazio */
   ASSEGURA(!pNo->filhos[posicao] || !pNo->filhos[posicao + 1],
            "Nenhum dos filhos da chave removida e' nulo");

      /* Primeiro, movem-se as chaves */
   for (i = posicao; i < pNo->nFilhos - 2; ++i)
      pNo->chaves[i] = pNo->chaves[i + 1];

      /* A movimentao de filhos depende de qual dos filhos  */
      /* da chave removida  vazio. Se o filho esquerdo no   */
      /* for vazio, a movimentao comea uma posio adiante */
   if (pNo->filhos[posicao])
      posicao++;

      /* Move os filhos do n */
   for (i = posicao; i < pNo->nFilhos - 1; ++i)
      pNo->filhos[i] = pNo->filhos[i + 1];

      /* O nmero de filhos (e de chaves) diminuiu */
   --pNo->nFilhos;
}

/****
 *
 * MaiorInteiro(): Encontra o maior elemento de um array de
 *                 inteiros do tipo int
 *
 * Parmetros:
 *      ar (entrada) - o array
 *      nElementos (entrada) - nmero de elementos do array
 *
 * Retorno: O maior valor encontrado em ar[]
 *
 ****/
static int MaiorInteiro(const int ar[], int nElementos)
{
   int i, maior = ar[0];

   for (i = 1; i < nElementos; ++i) {
      if (maior < ar[i]) {
         maior = ar[i];
      }
   }

   return maior;
}

/******************** Funes de Globais ******************/

/****
 *
 * BuscaMulti(): Faz uma busca numa rvore multidirecional
 *
 * Parmetros:
 *      chave (entrada) - a chave de busca
 *      arvore (entrada) - ponteiro para a raiz da rvore na
 *                         qual ser feita a busca
 *
 * Retorno: ndice do registro contendo a chave de busca,
 *          ou -1, se tal chave no for encontrada
 *
 ****/
int BuscaMulti(tArvoreMulti arvore, tChave chave)
{
   int i;

      /* Desce na rvore realizando uma busca em um  */
      /* n de cada nvel. A busca prossegue at que */
      /* a chave ou um filho nulo seja encontrado    */
   while (arvore) { /*  */
         /* Faz busca num n no nvel corrente */
      i = BuscaEmNo(chave, arvore);

         /* A primeira comparao  necessria para evitar  */
         /* acesso a uma posio invlida no array chaves[] */
      if ( i < arvore->nFilhos - 1 &&
           (chave == arvore->chaves[i].chave) )
            /* A chave foi encontrada. Retorna */
            /* o ndice do registro            */
         return arvore->chaves[i].indice;

         /* A busca continua no prximo nvel */
      arvore = arvore->filhos[i];
   }

   return -1; /* A chave no foi encontrada */
}

/****
 *
 * InsereMulti(): Insere uma chave numa rvore de busca
 *                multidirecional armazenada em memria principal
 *
 * Parmetros:
 *     *arvore (entrada/sada) - ponteiro para a rvore
 *     *chaveEIndice (entrada/sada) - par chave/ndice que ser
 *                                     inserido
 *
 * Retorno: 1, se tudo ocorrer bem; 0, se ocorrer algum erro
 *
 * Observao: A chave  considerada primria; i.e., chaves
 *             repetidas no so permitidas na rvore.
 *
 ****/
int InsereMulti(tArvoreMulti *arvore, tChaveIndice *chaveEIndice)
{
   tNoMulti *pNo, *pNovoNo;
   int       encontrado, /* Indicar se a chave */
                         /* foi encontrada      */
             indiceDaChave; /* ndice da chave no n onde ela */
                            /*  encontrada ou inserida       */

   if (!*arvore) { /* A rvore ainda no existe */
         /* Constri a raiz da rvore */
      *arvore = ConstroiNo(chaveEIndice);

      return 1;
   }

      /* Procura a chave na rvore */
   pNo = EncontraNo( *arvore, chaveEIndice->chave,
                     &indiceDaChave, &encontrado );

      /* Se a chave foi encontrada, no h mais nada */
      /* a fazer, pois ela  considerada primria    */
   if (encontrado) {
      return 0;
   }

      /* Numa rvore multidirecional descendente, */
      /* se um n  incompleto, ele  uma folha   */
   if (pNo->nFilhos < G) {
         /* O n  uma folha incompleta e no */
         /*  necessrio criar um novo n     */
      InsereEmFolha(pNo, indiceDaChave, *chaveEIndice);

      return 1; /* Nada mais a fazer */
   }

      /****************************************************/
      /* Neste ponto, resta lidar com o caso em que o n  */
      /* no qual dever ser realizada a insero  um n  */
      /* completo. Nesse caso,  necessrio criar um novo */
      /* n na posio indicada por 'indiceDaChave' e     */
      /* inserir o par chave/ndice nesse novo n.        */
      /****************************************************/

      /*  preciso criar um novo n com uma nica chave */
   pNovoNo = ConstroiNo(chaveEIndice);
   pNo->filhos[indiceDaChave] = pNovoNo;

   return 1;
}

/****
 *
 * RemoveChaveMulti(): Remove uma chave especificada da rvore e seu
 *                respectivo registro do arquivo
 *
 * Parmetros:
 *     *arvore (entrada/sada) - ponteiro para a raiz da rvore
 *      chave (entrada) - a chave que ser removida
 *
 * Retorno: 1, se a remoo for bem sucedida;
 *          0, em caso contrrio
 *
 ****/
int RemoveChaveMulti(tArvoreMulti *arvore, tChave chave)
{
   tNoMulti    *p, *q;
   int          i, iPai;
   tChaveIndice sucessoraEIndice, *pSucessoraEIndice;

   p = *arvore; /* Ponteiro usado para descer na rvore   */
   q = NULL;    /* Ponteiro que segue um nvel acima de p */

      /* A busca inicia na raiz e prossegue at que se */
      /* encontre a chave a ser removida ou um n nulo */
   while (p) {
         /* Faz busca num n no nvel corrente */
      i = BuscaEmNo(chave, p);

          /* Verifica se a chave foi encontrada */
      if ( i < p->nFilhos - 1 &&
           chave == p->chaves[i].chave )
         break; /* Chave foi encontrada */

      q = p;

         /* ndice do filho do penltimo n */
         /* visitado que aponta para p      */
      iPai = i;

         /* A busca continua no prximo nvel */
      p = p->filhos[i];
   }

      /* Verifica se a chave foi encontrada */
   if (!p)
      return 0; /* A chave no foi encontrada */

   /*********************************************************/
   /* Neste ponto sabe-se que p aponta para o n que contm */
   /* a chave que ser removida e que a posio dessa chave */
   /* nesse n  i. O ponteiro q est apontando para o n   */
   /* que tem p como filho na posio 'iPai'.               */
   /*********************************************************/

      /* Verifica quantos filhos a chave a ser removida possui */
   if (!p->filhos[i] || !p->filhos[i+1]) {
      /* A chave tem, no mximo, um filho */

         /* Verifica se o n s tem uma chave */
      if(p->nFilhos == 2) {
         /*                                     */
         /* A ltima chave do n ser removida. */
         /* Logo esse n tambm ser removido.  */
         /*                                     */

            /* Se o n a ser removido no */
            /* tem pai, trata-se da raiz  */
         if (!q) {
            free(p); /* Libera a raiz */
            *arvore = NULL; /* A rvore acabou-se */
            return 1; /* Game over */
         }

            /* Verifica quantos filhos o n */
            /* a ser removido possui        */
         if(p->filhos[i])
               /* O n tem filho esquerdo. O pai desse filho */
               /* passar a ser o pai do n a ser removido   */
            q->filhos[iPai] = p->filhos[i];
         else if(p->filhos[i + 1])
               /* O n tem filho direito. O pai desse filho */
               /* passar a ser o pai do n a ser removido  */
            q->filhos[iPai] = p->filhos[i + 1];
         else
               /* O n no tem nenhum filho. O pai do */
               /* n a ser removido passa a ser nulo  */
            q->filhos[iPai] = NULL;

         free(p); /* O n j pode ser liberado */
      } else /* H mais de uma chave no n */
            /* A funo CompactaNo() completa a remoo */
         CompactaNo(p,i);

      return 1;
   }

   /***************************************************/
   /* Neste ponto, sabe-se que a chave a ser removida */
   /* possui dois filhos no vazios. Nesse caso,      */
   /* substitui-se a chave a ser removida por sua     */
   /* sucessora imediata e remove-se essa sucessora   */
   /***************************************************/

      /* A chave sucessora imediata da chave a ser removida */
      /*  a menor chave da subrvore cuja raiz  o filho   */
      /* direito da chave a ser removida                    */
   pSucessoraEIndice = MenorChaveMulti(p->filhos[i + 1]);

      /* Se a chave possui dois filhos, ela */
      /* deve ter uma chave sucessora       */
   ASSEGURA(pSucessoraEIndice,
            "RemoveChaveMulti(): Nao existe chave sucessora");

      /* Guarda a chave sucessora com respectivo */
      /* ndice para posterior substituio da   */
      /* chave que ser removida                 */
   sucessoraEIndice = *pSucessoraEIndice;

      /* Agora, chama-se esta funo recursivamente     */
      /* para remover a chave sucessora. Haver apenas  */
      /* uma chamada recursiva porque a chave sucessora */
      /* possui, no mximo, um filho.                   */
   if (!RemoveChaveMulti(arvore, pSucessoraEIndice->chave))
      return 0;

      /* Substitui a chave a ser removida por sua sucessora */
   p->chaves[i] = sucessoraEIndice;

   return 1;
}

/****
 *
 * CaminhamentoInfixoMulti(): Executa um caminhamento em ordem
 *                       infixa numa rvore de busca
 *                       multidirecional exibindo todas as
 *                       chaves da rvore em ordem crescente
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *      stream (entrada) - stream associado ao arquivo onde sero
 *                         armazenadas as chaves visitadas
 *
 * Retorno: Nada
 *
 * Observao: Um caminhamento envolve executar uma operao
 *             sobre cada chave na rvore numa determinada
 *             ordem. A ordem de caminhamento apresentada aqui
 *              uma generalizao do caminhamento em ordem
 *             infixa para rvores binrias e a operao
 *             executada sobre cada chave  simplesmente a
 *             exibio dela. Este caminhamento  o mais til
 *             para rvores de busca porque permite acessar
 *             as chaves em ordem crescente, mas no h nenhuma
 *             dificuldade em generalizar outros caminhamentos
 *             para rvores binrias nem o tipo de operao
 *             realizada sobre as chaves.
 *
 ****/
void CaminhamentoInfixoMulti(tArvoreMulti arvore, FILE *stream)
{
   int i;

      /* O caminhamento encerra quando */
      /* se encontra um n nulo        */
   if (arvore) {
      for (i = 0; i < arvore->nFilhos-1; ++i) {
            /* Caminha no filho esquerdo */
            /* da chave de ndice i      */
         CaminhamentoInfixoMulti(arvore->filhos[i], stream);

            /* Visita a chave de ndice i */
         fprintf(stream, "%d\n", arvore->chaves[i].chave);
      }
         /* Caminha no filho mais  direita do n */
      CaminhamentoInfixoMulti( arvore->filhos[arvore->nFilhos-1],
                           stream );
   }
}

/****
 *
 * MenorChaveMulti(): Encontra a menor chave de uma rvore
 *               multidirecional, que  a primeira chave
 *               do n mais  esquerda da rvore
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: a menor chave da rvore
 *
 ****/
tChaveIndice *MenorChaveMulti(tArvoreMulti arvore)
{
   if (!arvore)
      return NULL; /* Se no h rvore, no h chave */

      /* Desce na rvore sempre seguindo o */
      /* filho mais  esquerda de cada n  */
   while (arvore->filhos[0])
      arvore = arvore->filhos[0];

   return &arvore->chaves[0];
}

/****
 *
 * MaiorChaveMulti(): Encontra a maior chave de uma rvore
 *               multidirecional, que  a ltima chave
 *               do n mais  direita da rvore
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: a maior chave da rvore
 *
 ****/
tChaveIndice *MaiorChaveMulti(tArvoreMulti arvore)
{
   if (!arvore)
      return NULL; /* Se no h rvore, no h chave */

   while (arvore->filhos[arvore->nFilhos-1])
         /* Vai descendo  direita */
      arvore = arvore->filhos[arvore->nFilhos-1];

   return &arvore->chaves[arvore->nFilhos-2];
}

/****
 *
 * AlturaMulti(): Calcula a altura de uma rvore
 *                      multidirecional de busca
 *
 * Parmetros:
 *     arvore (entrada) - a rvore cuja altura ser calculada
 *
 * Retorno: A referida altura
 *
 ****/
int AlturaMulti(tArvoreMulti arvore)
{
   int  i,
        alturas[G] = {0}; /* Array que armazena as alturas */
                          /* das subrvores de um n       */

   if (arvore != NULL) {
         /* Recursivamente, calcula-se a altura */
         /* da subrvore esquerda de cada chave       */
      for (i = 0; i < arvore->nFilhos - 1; ++i) {
         alturas[i] = 1 + AlturaMulti(arvore->filhos[i]);
      }

         /* Agora, calcula-se a altura da */
         /* subrvore direita da ltima chave   */
      alturas[i] = 1 + AlturaMulti(arvore->filhos[arvore->nFilhos - 1]);

         /* A altura de uma rvore  a maior        */
         /* altura encontrada entre suas subrvores */
      return MaiorInteiro(alturas, G);
   } else {
      return 0; /* rvore vazia tem altura zero */
   }
}

/****
 *
 * NumeroDeChavesMulti(): Calcula o nmero de chaves de uma
 *                   rvore multidirecional de busca
 *
 * Parmetros:
 *     arvore (entrada) - a rvore cujo nmero de chaves
 *                        ser determinado
 *
 * Retorno: O referido nmero de chaves
 *
 ****/
int NumeroDeChavesMulti(tArvoreMulti arvore)
{
   int  i,
        nChaves = 0;

      /* Visita cada n da rvore em ordem infixa at */
      /* que a posio inicial do caminhamento assuma */
      /* um valor invlido                            */
   if (arvore != NULL) {
         /* Conta as chaves deste n */
      nChaves = nChaves + arvore->nFilhos - 1;

         /* Agora, visita-se recursivamente a subrvore */
         /* esquerda de cada chave contando-se o nmero */
         /* de chaves da raiz de cada subrvore         */
      for (i = 0; i < arvore->nFilhos - 1; ++i) {
         nChaves = nChaves +
                   NumeroDeChavesMulti(arvore->filhos[i] );
      }

         /* Agora, contam-se as chaves da subrvore */
         /* direita da ltima chave                 */
      nChaves = nChaves +
                NumeroDeChavesMulti(arvore->filhos[arvore->nFilhos - 1]);
   }

   return nChaves;
}

/****
 *
 * NumeroDeNosMulti(): faz um caminhamento numa rvore de busca
 *                multidirecional contando o nmero de ns
 *                visitados
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: Nmero de ns da rvore
 *
 ****/
int NumeroDeNosMulti(tArvoreMulti arvore)
{
   int i, nNos = 0;

   if (arvore) {
      nNos++;

      for (i = 0; i < arvore->nFilhos; ++i)
         nNos = nNos + NumeroDeNosMulti(arvore->filhos[i]);
   }

   return nNos;
}
