/****
 *
 * Arquivo ArvoreB.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 25/04/2004
 * ltima modificao: 21/07/2008
 *
 * Descrio: Implementao das funes que executam
 *            operaes usando rvores B armazenadas
 *            em arquivo
 *
 ****/

/*********************** Includes *************************/

#include <string.h>    /* Funo strxxx() e memxxx()  */
#include "ArvoreB.h"   /* Interface deste mdulo      */
#include "Registros.h" /* Funes de processamento    */
                       /* de registros                */
#include "Pilha.h"     /* Mdulo que gerencia a pilha */
#include "leitura.h"   /* Funes de leitura via teclado */
#include "ASSEGURA.h"  /* Macro ASSEGURA                 */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/******************** Funes de Locais *******************/

/****
 *
 * IniciaNoMultiMS(): Inicia um n de uma rvore multidirecional armazenada
 *                    em arquivo
 *
 * Parmetros:
 *      chaveIndice (entrada) - um par chave/ndice a ser armazenado no n
 *      pNo (sada) - endereo do n criado
 *
 * Retorno: Nada
 *
 * Observao: O primeiro parmetro poder ser NULL. Nesse caso,
 *             ser criado um n vazio (i.e., sem nenhuma chave)
 *
 ****/
static void IniciaNoMultiMS( const tChaveIndice *chaveIndice, tNoMultiMS *pNo )
{
   int i;

   if (chaveIndice)
      pNo->chaves[0] = *chaveIndice; /* A chave ser a primeira do n */

      /* O nmero de filhos num n de uma rvore     */
      /* multidirecional  igual ao nmero de chaves */
      /* mais um, a no ser que no haja chave       */
   pNo->nFilhos = chaveIndice ? 2 : 0;

      /* Inicia todos os possveis filhos do n */
   for (i = 0; i < G; ++i)
         /* Um novo n ainda no possui filhos */
      pNo->filhos[i] = POSICAO_NULA;
}

/****
 *
 * LeNoMultiMS(): Faz a leitura de um n de uma rvore B armazenada em arquivo
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *                         contendo a rvore onde ser
 *                         feita a leitura
 *      pos (entrada) - posio no arquivo onde ser feita
 *                      a leitura
 *     *no (sada) - ponteiro para o n que conter o
 *                   resultado da leitura
 *
 * Retorno: Nada
 *
 ****/
static void LeNoMultiMS(FILE *stream, int pos, tNoMultiMS *no)
{
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, LeNoMultiMS));
   ASSEGURA(pos >= 0, ERRO_POSICAO(LeNoMultiMS));

      /* Tenta mover o apontador de arquivo para o local  */
      /* de leitura; se no conseguir, aborta o programa  */
   MoveApontador(stream, sizeof(tNoMultiMS)*pos, SEEK_SET);

      /* Tenta ler o n */
   fread(no, sizeof(tNoMultiMS), 1, stream);

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), ERRO_FREAD(LeNoMultiMS) );
}

/****
 *
 * EscreveNoMultiMS(): Atualiza o contedo de um n no arquivo que contm a rvore
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm a rvore
 *      pos (entrada) - posio do n no arquivo
 *                      que contm a rvore
 *     *pNo (entrada) - n que ser atualizado
 *
 * Retorno: Nada
 *
 ****/
static void EscreveNoMultiMS(FILE *stream, int pos, const tNoMultiMS *pNo)
{
      /* Tenta mover o apontador de posio     */
      /* para o local onde o n est armazenado */
   MoveApontador(stream, sizeof(*pNo)*pos, SEEK_SET);

    fwrite(pNo, sizeof(*pNo), 1, stream); /* Escreve o n no arquivo */

      /* Se ocorreu erro, aborta o programa */
   ASSEGURA(!ferror(stream), ERRO_FWRITE(EscreveNoMultiMS));
}

/****
 *
 * BuscaEmNoMultiMS(): Tenta encontrar dentro de um n uma chave que
 *                     seja maior do que ou igual a uma dada chave
 *
 * Parmetros:
 *      chave (entrada) - a chave
 *      no (entrada) - ponteiro para o n no qual ser feita
 *                     a busca
 *
 * Retorno: O ndice da primeira chave do n que  maior do
 *          que ou igual a chave de busca. Se tal chave no
 *          for encontrada, retorna-se o nmero de chaves
 *
 ****/
static int BuscaEmNoMultiMS(tChave chave, const tNoMultiMS *no)
{
   int i,
       nChaves = no->nFilhos - 1; /* Nmero de chaves do n */

      /* Procura no n uma chave que seja maior */
      /* do que ou igual  chave de busca       */
   for (i = 0; i < nChaves; ++i)
         /* Verifica se tal chave foi encontrada */
      if (chave <= no->chaves[i].chave)
         return i;   /* A chave foi encontrada */

      /* A chave de busca  maior do que qualquer chave do n */
   return nChaves; /* A chave NO foi encontrada */
}

/****
 *
 * EncontraCaminhoB(): Procura um o n que contm uma chave,
 *                    se ela for encontrada, ou deveria cont-
 *                    la, se ela no for encontrada. O caminho
 *                    da raiz at tal n encontrado  empilhado
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que
 *                         contm a rvore
 *      chave (entrada) - a chave de busca
 *      raiz (entrada) - posio no arquivo da raiz da rvore
 *      posicaoDaChaveNoNo (sada) - ponteiro para uma varivel
 *                                   que conter o ndice da
 *                                   chave no n
 *      encontrado (sada) - ponteiro para uma varivel que
 *                           indicar se a chave foi encontrada
 *     *pilha (sada) - pilha que armazenar o caminho de ns
 *                      visitados
 *
 * Retorno: Nada
 *
 ****/
static void EncontraCaminhoB( FILE *stream, tChave chave, int raiz,
                              int *posicaoDaChaveNoNo,
                              int *encontrado, tPilha *pilha)
{
   int        p,  /* Conter a posio desejada */
              indiceNo; /* ndice de um n entre seus irmos */
   tNoMultiMS  umNo;
   tNoCaminhoB caminho;

      /* Verifica se o stream  NULL */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, EncontraCaminhoB));

      /* Tenta mover o apontador de arquivo para seu incio */
  MoveApontador(stream, 0, SEEK_SET);

   p = raiz; /* Comea busca na raiz da rvore */
   indiceNo = -1;  /* A raiz no tem irmos */
   *encontrado = 0;  /* A chave ainda no foi encontrada */
   CriaPilha(pilha); /* Inicia a pilha */

      /* Desce na rvore at encontrar a chave de busca ou */
      /* atingir uma folha na qual a chave no se encontra */
   while (p != POSICAO_NULA) {
         /* L um n da rvore na posio p do arquivo */
      LeNoMultiMS(stream, p, &umNo);

      /***                                ***/
      /*** Cria o item que ser empilhado ***/
      /***                                ***/

         /* Armazena o n no item da pilha */
      caminho.no = umNo;
         /* Armazena a posio do n */
      caminho.endereco = p;
         /* Armazena seu ndice entre irmos */
      caminho.pos = indiceNo;

      Empilha(caminho, pilha); /* Faz o empilhamento */

         /* Tenta encontrar a chave no n corrente */
      indiceNo = BuscaEmNoMultiMS(chave, &umNo);

         /* Verifica se a chave foi encontrada */
      if( ( indiceNo < umNo.nFilhos - 1) &&
            chave == umNo.chaves[indiceNo].chave ) {
         *encontrado = 1;  /* A chave foi encontrada */
         *posicaoDaChaveNoNo = indiceNo;

         return;
      }

         /* A chave ainda no foi encontrada. */
         /* Desce um nvel na rvore          */
      p = umNo.filhos[indiceNo];
   }

      /* A chave no foi encontrada. 'posicaoDaChaveNoNo' */
      /* conter o ndice no ltimo n visitado (folha)   */
      /* no qual a chave dever ser inserida              */
   *posicaoDaChaveNoNo = indiceNo;
}

/****
 *
 * InsereEmNoB(): Insere um par chave/ndice e seu respectivo
 *               filho direito num n incompleto
 *
 * Parmetros:
 *     *pNo (entrada/sada) - n no qual ser feita a insero
 *      pos (entrada) - posio no array de chaves do n na
 *                      qual a chave ser inserida
 *      chaveEIndice (entrada) - par chave/ndice que ser
 *                               inserido
 *      pFilho (entrada) - ponteiro para o filho a ser
 *                         inserido  direita do novo par
 *
 * Retorno: Nada
 *
 ****/
static void InsereEmNoB( tNoMultiMS *pNo, int pos,
                         tChaveIndice chaveEIndice, int pFilho )
{
   int i;

      /* Verifica se o n  vlido */
   ASSEGURA( pNo, "Tentativa de insercao em no' nulo" );

      /* Verifica se a posio de insero  vlida */
   ASSEGURA(pos <= pNo->nFilhos, ERRO_POSICAO(InsereEmNoB));

      /* Abre espao para a nova chave */
   for (i = pNo->nFilhos - 1; i > pos; --i){
      pNo->filhos[i + 1] = pNo->filhos[i];
      pNo->chaves[i] = pNo->chaves[i - 1];
   }

      /* Insere o novo par chave/ndice e seu filho direito */
   pNo->chaves[pos] = chaveEIndice; /* Insere o par */
   pNo->filhos[pos + 1] = pFilho; /* Insere seu filho direito */

      /* O nmero de filhos (e de chaves) foi acrescido de um */
   pNo->nFilhos++;
}

/****
 *
 * CopiaChavesB(): Copia chaves de um n para outro
 *
 * Parmetros:
 *      pNo1 (entrada) - ponteiro para o n cujas chaves
 *                       sero copiadas
 *      pNo2 (sada) - ponteiro para o n que receber as
 *                     chaves copiadas
 *      indiceI (entrada) - ndice da primeira chave a
 *                          ser copiada
 *      indiceF (entrada) - ndice da ltima chave a ser
 *                          copiada
 *
 * Retorno: Nada
 *
 ****/
static void CopiaChavesB( const tNoMultiMS *pNo1, tNoMultiMS *pNo2,
                         int indiceI, int indiceF )
{
   int i,
       nChaves; /* Nmero de chaves a copiar. Essa varivel */
                /* no  essencial, mas facilita a operao */

   nChaves = indiceF - indiceI + 1;

      /* Verifica se o nmero de chaves  vlido */
   ASSEGURA( nChaves < G, "Chaves demais para copiar" );

      /* Idem */
   ASSEGURA( nChaves > 0,
             "ERRO: Nao ha' chaves para copiar" );

      /* Copia chaves e filhos do n apontado por */
      /* pNo1 para o n apontado por pNo2         */
   for (i = 0; i < nChaves; ++i) {
      pNo2->chaves[i] = pNo1->chaves[indiceI + i];
      pNo2->filhos[i] = pNo1->filhos[indiceI + i];
   }

      /* Faltou copiar o ltimo filho */
   pNo2->filhos[nChaves] = pNo1->filhos[indiceI + nChaves];

      /* Atualiza o nmero de chaves */
      /* do n que recebeu a doao  */
   pNo2->nFilhos = nChaves + 1;
}

/****
 *
 * DivideNoB(): Insere uma chave num n completo, criando um
 *             novo n para a metade superior das chaves e
 *             mantendo a metade inferior das chaves no n
 *             dividido. A chave intermediria  guardada
 *             para posterior insero num n que fica num
 *             nvel acima.
 *
 * Parmetros:
 *     *pNoDiv (entrada/sada) - n que ser dividido
 *      pos (entrada) - posio onde a nova chave ser
 *                      inserida se esta for menor do
 *                      que a chave intermediria
 *      chave (entrada) - a chave (e seu respectivo valor) que
 *                        ser inserida num dos ns resultantes
 *                        da diviso ou num n um nvel acima
 *      pFilhoDireita (entrada) - posio no arquivo do n que
 *                                ser o filho  direita da
 *                                chave inserida
 *     *pNoDireita (sada) - posio no arquivo do novo n no
 *                           qual sero colocadas as chaves
 *                           superiores do n dividido
 *     *chaveMeio (sada) - a chave do meio (i.e., aquela que
 *                          subir para o n-pai) do n dividido
 *      stream (entrada) - stream associado ao arquivo que
 *                         contm a rvore
 *
 * Retorno: Nada
 *
 * Observaes:
 *    1. Se o grau da rvore for par, a rvore ter tendncia
 *        esquerda; i.e., o n dividido ter uma chave a mais
 *       do que o novo n
 *
 *    2. IMPORTANTE: O n apontado por pNoDiv deve ser
 *                   atualizado no arquivo que contm a rvore
 *                   pela funo que chama a presente funo.
 *
 ****/
static void DivideNoB( tNoMultiMS *pNoDiv, int pos,
                       tChaveIndice chave, int pFilhoDireita,int *pNoDireita,
                       tChaveIndice *chaveMeio, FILE *stream )
{
   static const int meio = G/2; /* Evita recalcular esse */
//   static const int meio = (G - 1)/2; /* Evita recalcular esse */
                                      /* valor a cada chamada  */
                                      /* desta funo          */
   tNoMultiMS        noDireita; /* Novo n que ser      */
                               /* acrescentado  rvore */

      /* Inicia o novo n que receber as chaves superiores */
   IniciaNoMultiMS(NULL, &noDireita);

      /* Efetua a insero com a devida diviso de ns */
   if (pos > meio) { /* Caso 1: A nova chave pertence */
                     /* ao novo n da direita         */
         /* Copia metade menos uma chave  */
         /* do n dividido para o novo n */
      CopiaChavesB(pNoDiv, &noDireita, meio + 1, G - 2);

         /* Insere a nova chave no novo n da direita */
      InsereEmNoB( &noDireita, pos - meio - 1, chave,
                  pFilhoDireita );

         /* Ajusta o nmero de ns do n dividido */
      pNoDiv->nFilhos = meio + 1;

         /* A chave que subir  a ltima chave do */
         /* n que foi dividido depois da diviso  */
      *chaveMeio = pNoDiv->chaves[meio];
   } else if (pos == meio) { /* Caso 2: A nova chave      */
                             /* aquela que subir um nvel */
         /* Copia metade das chaves do */
         /* n dividido para o novo n */
      CopiaChavesB(pNoDiv, &noDireita, meio, G - 2);

         /* Ajusta o nmero de ns do n dividido */
      pNoDiv->nFilhos = meio + 1;

         /* O filho direito da nova chave passa */
         /* a ser o primeiro filho do novo n   */
      noDireita.filhos[0] = pFilhoDireita;

         /* A chave que subir  a nova chave */
      *chaveMeio = chave;
   } else { /* Caso 3: A nova chave pertence */
            /* ao n que ser dividido       */
         /* Copia metade das chaves do */
         /* n dividido para o novo n */
      CopiaChavesB(pNoDiv, &noDireita, meio, G - 2);

         /* Ajusta o nmero de ns do n dividido */
      pNoDiv->nFilhos = meio + 1;

         /* Insere a nova chave no n que foi dividido */
      InsereEmNoB(pNoDiv, pos, chave, pFilhoDireita);

         /* A chave que subir  a ltima chave do */
         /* n que foi dividido depois da diviso  */
      *chaveMeio = pNoDiv->chaves[meio];

         /* Corrige o nmero de filhos do n dividido, pois    */
         /* sua ltima chave subir para o prximo nvel acima */
      --pNoDiv->nFilhos;
   }

      /* O novo n ser armazenado ao final do arquivo; */
      /* portanto, sua posio no arquivo ser igual ao */
      /* nmero de registros (ns) do arquivo (NB: a    */
      /* indexao do arquivo comea em zero; por isso, */
      /* no se acrescenta 1 a esse valor).             */
   *pNoDireita = NumeroDeNosB(stream);

      /* Atualiza n no arquivo */
   EscreveNoMultiMS(stream, *pNoDireita, &noDireita);
}

/****
 *
 * JuntaNosB(): Combina um n com outros ns de uma rvore B
 *             quando ele fica com grau abaixo do mnimo
 *
 * Parmetros:
 *      pNoN (entrada/sada) - ponteiro para o n que ser
 *                             combinado com outros
 *      posNo (entrada) - posio do n entre seus irmos
 *      endNo (entrada) - posio do n no arquivo que armazena a
 *                        rvore
 *     *pilha (entrada/sada) - pilha que armazena os ns
 *                              encontrados no caminho da raiz
 *                              at o n que ser combinado
 *                              (sem inclu-lo)
 *      stream (entrada) - stream associado ao arquivo que
 *                         contm a rvore
 *
 * Retorno: Nada
 *
 * Observaes:
 *    1. Considerando as figuras apresentadas no texto, pNoN 
 *       um ponteiro para o n N, pNoE  a posio no arquivo que
 *       contm a rvore do n E (aqui denominado noE) e pNoD  a
 *       posio no mesmo arquivo do n D (aqui denominado noD).
 *
 *    2. Quando se faz referncia a uma varivel que contm uma
 *       posio em arquivo, 'apontar para' no tem o mesmo
 *       significado dessa mesma expresso usada com ponteiros.
 *       I.e., aqui, 'apontar para' significa conter a posio
 *       de um n no arquivo que o armazena.
 *
 *    3. V. figuras e algoritmo no texto.  muito difcil
 *       acompanhar esta funo sem examinar cuidadosamente
 *       essas figuras.
 *
 ****/
static void JuntaNosB( tNoMultiMS *pNoN, int posNo, int endNo,
                       tPilha* pilha, FILE *stream )
{
   static const int minimo = G/2; /* Essa varivel no   */
                                  /* essencial, mas faci- */
                                  /* lita a implementao */
   tNoCaminhoB   caminho; /* Armazenar cada item desempilhado */
   tNoMultiMS    noE, /* N E das figuras do texto */
                noD; /* N D das figuras do texto */
   int          pNoE = POSICAO_NULA, /* Posio em arquivo de noE */
                pNoD = POSICAO_NULA, /* Posio em arquivo de noD */
                i;
   tChaveIndice sucessora; /* A chave sucessora  qual */
                           /* se refere o algoritmo    */

      /* Passo 1 do algoritmo */
   if (pNoN->nFilhos >= minimo) {
      return; /* N no precisa ser combinado com outro */
   }

   /*                                                     */
   /* Neste ponto, sabe-se que o n apontado por pNoN tem */
   /* um nmero de filhos menor do que o mnimo desejado  */
   /*                                                     */

      /* A pilha no pode estar vazia aqui */
   ASSEGURA(!PilhaVazia(*pilha), "Pilha vazia em JuntaNosB()");

      /* O item no topo da pilha contm informaes */
      /* sobre o pai do n apontado por pNoN        */
   caminho = Desempilha(pilha);

      /* Se o n a ser combinado no  o primeiro */
      /* filho, ele tem um irmo esquerdo         */
   if(posNo > 0)
         /* Faz pNoE apontar para o irmo    */
         /* esquerdo do n apontado por pNoN */
      pNoE = caminho.no.filhos[posNo - 1];

      /* Se o n a ser combinado no  o ltimo */
      /* filho, ele tem um irmo direito        */
   if(posNo < caminho.no.nFilhos - 1)
         /* Faz pNoD apontar para o irmo   */
         /* direito do n apontado por pNoN */
      pNoD = caminho.no.filhos[posNo + 1];

      /* O n a ser combinado deve ter pelo menos um  */
      /* irmo mais prximo ou, ento, ele deve ser a */
      /* raiz que no tem irmos                      */
   ASSEGURA( pNoE != POSICAO_NULA || pNoD != POSICAO_NULA ||
             PilhaVazia(*pilha),
             "Erro em JuntaNosB(): o no' nao possui irmao" );

      /* Passo 2: Verifique se o irmo esquerdo (apontado */
      /* por pNoE) do n apontado por pNoN tem um nmero  */
      /* de chaves maior do que o mnimo                  */

      /* Verifica se o n apontado por pNoN tem irmo esquerdo */
   if (pNoE != POSICAO_NULA) {
         /* L o irmo esquerdo noE */
      LeNoMultiMS(stream, pNoE, &noE);

         /* Verifica se o irmo esquerdo pode doar uma chave */
      if (noE.nFilhos > minimo) {
            /* Passo 2.1: Insira a sucessora da ltima */
            /* chave do n noE no incio das chaves do */
            /* n apontado por pNoN                    */

            /* Obtm a sucessora que se encontra no n-pai */
         sucessora = caminho.no.chaves[posNo-1];

             /* Abre espao para o novo filho */
         for (i = pNoN->nFilhos - 1; i >= 0; --i)
            pNoN->filhos[i + 1] = pNoN->filhos[i];

             /* Abre espao para a nova chave */
         for (i = pNoN->nFilhos - 2; i >= 0; --i)
            pNoN->chaves[i + 1] = pNoN->chaves[i];

            /* Coloca a sucessora em seu devido lugar */
         pNoN->chaves[0] = sucessora;

            /* Passo 2.2: Faa o filho esquerdo da primeira */
            /* chave de pNoN apontar para o filho direito   */
            /* da ltima chave de noE                       */
         pNoN->filhos[0] = noE.filhos[noE.nFilhos - 1];

            /* Passo 2.3: Substitua a sucessora da ltima */
            /* chave do n noE por essa ltima chave      */
         caminho.no.chaves[posNo - 1] = noE.chaves[noE.nFilhos - 2];

            /* O n-pai que foi alterado  uma cpia daquele */
            /* que se encontra na rvore. Logo  preciso     */
            /* atualizar o n correspondente na rvore.      */
         EscreveNoMultiMS(stream, caminho.endereco, &caminho.no);

            /* Passo 2.4: Decremente o nmero de filhos de noE */
         noE.nFilhos--;

            /* O n noE precisa ser atualizado em arquivo */
         EscreveNoMultiMS(stream, pNoE, &noE);

            /* Passo 2.5: Incremente o nmero de */
            /* filhos do n apontado por pNoN    */
         pNoN->nFilhos++;

            /* O n apontado por pNoN precisa */
            /* ser atualizado em arquivo      */
         EscreveNoMultiMS(stream, endNo, pNoN);

            /* Passo 2.6: Encerre */
         return;
      }
   }

      /* Passo 3: Verifique se o irmo direito (apontado */
      /* por pNoD) do n apontado por pNoN tem um nmero */
      /* de chaves maior do que o mnimo                 */
   if (pNoD != POSICAO_NULA) {
         /* L o irmo direito noD */
      LeNoMultiMS(stream, pNoD, &noD);

         /* Verifica se o irmo direito */
         /* pode fazer doao de chaves */
      if (noD.nFilhos > minimo) {
            /* Passo 3.1: Acrescente a sucessora da */
            /* ltima chave do n apontado por pNoN */
            /* ao final das chaves desse n         */

            /* Obtm a sucessora que se encontra no n-pai */
         sucessora = caminho.no.chaves[posNo];

            /* Coloca a sucessora em seu devido lugar */
         pNoN->chaves[pNoN->nFilhos - 1] = sucessora;

            /* Passo 3.2: Faa o filho direito da ltima    */
            /* chave de pNoN apontar para o filho esquerdo  */
            /* da primeira chave de noD. Note que o nmero  */
            /* de filhos de pNoN ainda no foi incrementado */
         pNoN->filhos[pNoN->nFilhos] = noD.filhos[0];

            /* Passo 3.3: Substitua a sucessora da ltima chave */
            /* do n pNoN pela primeira chave do n noD         */
         caminho.no.chaves[posNo] = noD.chaves[0];

             /* Passo 3.4: Desloque cada chave e filho */
             /* do n noD uma posio para trs        */
         for (i = 1; i <= noD.nFilhos - 1; ++i)
            noD.filhos[i - 1] = noD.filhos[i];

         for (i = 1; i <= noD.nFilhos - 2; ++i)
            noD.chaves[i - 1] = noD.chaves[i];

            /* O n-pai que foi alterado  uma cpia daquele */
            /* que se encontra na rvore. Logo  preciso     */
            /* atualizar o n correspondente na rvore.      */
         EscreveNoMultiMS(stream, caminho.endereco, &caminho.no);

            /* Passo 3.5: Decremente o nmero */
            /* de filhos do n noD            */
         noD.nFilhos--;

            /* Passo 3.6: Incremente o nmero de */
            /* filhos do n apontado por pNoN    */
         pNoN->nFilhos++;

            /* O n noD precisa ser atualizado em arquivo */
         EscreveNoMultiMS(stream, pNoD, &noD);

            /* O n apontado por pNoN precisa */
            /* ser atualizado em arquivo      */
         EscreveNoMultiMS(stream, endNo, pNoN);

            /* Passo 3.7: Encerre */
         return;
      }
   }

   /*                                                     */
   /* Passo 4: Se nenhum irmo vizinho do n apontado por */
   /* pNoN tiver um nmero de ns maior do que o mnimo   */
   /*                                                     */

      /* Passo 4.1: Verifique se o n apontado */
      /* por pNoN tem irmo direito (pNoD) com */
      /* nmero de filhos igual ao mnimo      */
   if (pNoD != POSICAO_NULA) {
         /* L o irmo direito noD */
      LeNoMultiMS(stream, pNoD, &noD);

         /* Garante que o nmero de filhos de noD  igual ao   */
         /* mnimo. Se esse nmero fosse maior do que o mnimo */
         /* ter-se-ia o Caso 3. Se esse nmero fosse menor do  */
         /* que o mnimo, no se teria uma rvore B.           */
      ASSEGURA( noD.nFilhos == minimo,
                "Numero de chaves em no' direito nao e' minimo" );

         /* Passo 4.1.1: Insira a chave sucessora (S) da  */
         /* ltima chave do n apontado por pNoN ao final */
         /* das chaves do n apontado por pNoN            */

         /* Obtm a sucessora */
      sucessora = caminho.no.chaves[posNo];

         /* Coloca a sucessora em seu devido lugar */
      pNoN->chaves[pNoN->nFilhos - 1] = sucessora;

         /* Passo 4.1.2: Faa o filho direito da ltima  */
         /* chave de pNoN apontar para  o filho esquerdo */
         /* da primeira chave de noD. Note que o nmero  */
         /* de filhos de pNoN ainda no foi incrementado */
      pNoN->filhos[pNoN->nFilhos] = noD.filhos[0];

         /* O nmero de chaves e filhos do */
         /* n apontado por pNoN aumentou  */
      ++pNoN->nFilhos;

         /* Passo 4.1.3: Desloque cada chave do pai do n   */
         /* apontado por pNoN e seu filho direito, a partir */
         /* da chave sucessora, para uma posio anterior   */
      for (i = posNo + 1; i < caminho.no.nFilhos - 1; ++i)
         caminho.no.filhos[i] = caminho.no.filhos[i + 1];

      for (i = posNo; i < caminho.no.nFilhos - 2; ++i)
         caminho.no.chaves[i] = caminho.no.chaves[i + 1];

         /* Passo 4.1.4: Decremente o nmero de filhos */
         /* do pai do n apontado por pNoN             */
      --caminho.no.nFilhos;

         /*  preciso atualizar o n */
         /* correspondente na rvore */
      EscreveNoMultiMS(stream, caminho.endereco, &caminho.no);

         /* Passo 4.1.5: Acrescente todas as chaves e filhos */
         /* direitos correspondentes do n noD ao final das  */
         /* chaves e filhos do n apontado por pNoN          */
      for ( i = 0; i < noD.nFilhos - 1; ++i)
         pNoN->filhos[pNoN->nFilhos + i] = noD.filhos[i + 1];

      for ( i = 0; i <= noD.nFilhos - 2; ++i)
         pNoN->chaves[pNoN->nFilhos + i - 1] = noD.chaves[i];

         /* Passo 4.1.6: Atualize o nmero de */
         /* filhos do n apontado por pNoN    */
      pNoN->nFilhos += noD.nFilhos;

         /* Passo 4.1.7: Libere o n noD. No h como liberar */
         /* um n no interior de um arquivo. Ento, a melhor  */
         /* opo  indicar que esse  um n removido atribu- */
         /* indo zero ao nmero de filhos.                    */
      noD.nFilhos = 0;

         /* O n noD precisa ser atualizado em arquivo */
      EscreveNoMultiMS(stream, pNoD, &noD);

         /* O n apontado por pNoN precisa */
         /* ser atualizado em arquivo      */
      EscreveNoMultiMS(stream, endNo, pNoN);

         /* Passo 4.1.8: Ajuste o pai do n       */
         /* apontado por pNoN usando JuntaNosB() */
      JuntaNosB( &caminho.no, caminho.pos, caminho.endereco,
                pilha, stream );

      return;
   }

      /* Passo 4.2: Verifique se o n apontado */
      /* por pNoN tem irmo esquerdo (noE) com */
      /* nmero de filhos igual ao mnimo      */
   if (pNoE != POSICAO_NULA) {
         /* L o irmo esquerdo noE */
      LeNoMultiMS(stream, pNoE, &noE);

         /* Garante que o nmero de filhos de noE  igual ao   */
         /* mnimo. Se esse nmero fosse maior do que o mnimo */
         /* ter-se-a o Caso 2. Se esse nmero fosse menor do  */
         /* que o mnimo, no se teria uma rvore B.           */
      ASSEGURA( noE.nFilhos == minimo,
                "Numero de chaves em no' esquerdo nao e' minimo" );

         /* Passo 4.2.1: Insira ao final das chaves de noE */
         /* a chave sucessora da ltima chave de noE       */

         /* Obtm a sucessora */
      sucessora = caminho.no.chaves[posNo - 1];

         /* Coloca a sucessora em seu devido lugar */
      noE.chaves[noE.nFilhos - 1] = sucessora;

         /* Passo 4.2.2: Faa o filho direito da ltima  */
         /* chave de noE apontar para  o filho esquerdo  */
         /* da primeira chave de pNoN. Note que ainda    */
         /* no houve incremento no nmero de ns de noE */
      noE.filhos[noE.nFilhos] = pNoN->filhos[0];

         /* O nmero de chaves e filhos do n noE aumentou */
      ++noE.nFilhos;

         /* Passo 4.2.3: Desloque cada chave do pai do n   */
         /* apontado por pNoN e seu filho direito, a partir */
         /* da chave sucessora, para uma posio anterior   */
      for (i = posNo; i < caminho.no.nFilhos - 1; ++i)
         caminho.no.filhos[i] = caminho.no.filhos[i + 1];

      for (i = posNo - 1; i < caminho.no.nFilhos - 2; ++i)
         caminho.no.chaves[i] = caminho.no.chaves[i + 1];

         /* Passo 4.2.4: Decremente o nmero de filhos */
         /* do pai do n apontado por pNoN             */
      --caminho.no.nFilhos;

         /*  preciso atualizar o n correspondente na rvore */
      EscreveNoMultiMS(stream, caminho.endereco, &caminho.no);

         /* Passo 4.2.5: Acrescente todas as chaves e filhos */
         /* direitos correspondentes do n apontado por pNoN */
         /* ao final das chaves de noE                       */
      for (i = 0; i < pNoN->nFilhos - 1; ++i)
         noE.filhos[noE.nFilhos + i] = pNoN->filhos[i + 1];

      for (i = 0; i <= pNoN->nFilhos - 2; ++i)
         noE.chaves[noE.nFilhos + i - 1] = pNoN->chaves[i];

         /* Passo 4.2.6: Atualize o nmero de filhos de noE */
      noE.nFilhos += pNoN->nFilhos;

         /* Passo 4.2.7: Libere o n apontado por pNoN. */
         /* V. comentrio acima sobre liberao de ns. */
      pNoN->nFilhos = 0;

         /* O n noE precisa ser atualizado em arquivo */
      EscreveNoMultiMS(stream, pNoE, &noE);

         /* O n apontado por pNoN precisa */
         /* ser atualizado em arquivo      */
      EscreveNoMultiMS(stream, endNo, pNoN);

         /* Passo 4.2.8: Ajuste o pai do n     */
         /* apontado por pNoE usando JuntaNosB() */
      JuntaNosB( &caminho.no, caminho.pos, caminho.endereco,
                pilha, stream );
   }
}

/****
 *
 * RemoveEmFolhaB(): Remove uma chave de uma folha de uma rvore B
 *
 * Parmetros:
 *     *pilha (entrada/sada) - pilha que armazena os ns
 *                              encontrados no caminho da raiz
 *                              at a folha (incluindo-a)
 *      stream (entrada) - stream associado ao arquivo que contm a rvore
 *
 * Retorno: Nada
 *
 ****/
static void RemoveEmFolhaB(tPilha *pilha, FILE *stream)
{
   tNoCaminhoB caminho; /* Um elemento da pilha que    */
                       /* representa o caminho de ns */
   int        i;

   ASSEGURA( !PilhaVazia(*pilha),
             "ERRO: Pilha vazia em RemoveEmFolhaB()" );

      /* No topo da pilha, encontra-se a folha */
      /* na qual a remoo ser efetuada       */
   caminho = Desempilha(pilha);

         /* Move as chaves */
   for (i = caminho.pos; i < caminho.no.nFilhos - 2; ++i)
      caminho.no.chaves[i] = caminho.no.chaves[i + 1];

      /* O nmero de filhos da folha foi reduzido */
   caminho.no.nFilhos--;

      /* Chama JuntaNosB() para completar o servio */
   JuntaNosB( &caminho.no, caminho.pos, caminho.endereco, pilha,
             stream );

      /* O contedo da pilha no  mais necessrio */
   EsvaziaPilha(pilha);
}

/******************** Funes de Globais ******************/

/****
 *
 * LeRaizB(): L num arquivo binrio a posio da raiz de uma
 *           rvore B armazenada em arquivo
 *
 * Parmetros:
 *     nomeArqRaiz (entrada) - nome do arquivo que armazena a
 *                             raiz da rvore
 *
 * Retorno: O valor lido no arquivo
 *
 ****/
int LeRaizB(const char *nomeArqRaiz)
{
   FILE *stream;
   int   raiz;

      /* Tenta abrir o arquivo que contm a */
      /* raiz da rvore para leitura apenas */
   stream = AbreArquivo(nomeArqRaiz, "rb");

   fread(&raiz, sizeof(raiz), 1, stream);

      /* Certifica-se que no ocorreu erro de leitura */
   ASSEGURA(!ferror(stream), ERRO_FREAD(LeRaizB));

      /* O arquivo no precisa mais estar aberto */
   FechaArquivo(stream, nomeArqRaiz);

   return raiz;
}

/****
 *
 * EscreveRaizB(): Escreve num arquivo binrio a posio da raiz
 *                de uma rvore B armazenada em arquivo
 *
 * Parmetros:
 *     nomeArqRaiz (entrada) - nome do arquivo que armazena a
 *                             raiz da rvore
 *
 * Retorno: Nada
 *
 ****/
void EscreveRaizB(const char *nomeArqRaiz, int raiz)
{
   FILE *stream;

      /* Tenta abrir o arquivo que contm a */
      /* raiz da rvore para escrita apenas */
   stream = AbreArquivo(nomeArqRaiz, "wb");

   fwrite(&raiz, sizeof(raiz), 1, stream);

      /* Certifica-se que no ocorreu erro de escrita */
   ASSEGURA(!ferror(stream), ERRO_FWRITE(EscreveRaizB));

      /* O arquivo no precisa mais estar aberto */
   FechaArquivo(stream, nomeArqRaiz);
}

/****
 *
 * BuscaB(): Faz uma busca numa rvore B armazenada em arquivo
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo no
 *                         qual a rvore reside
 *      raiz (entrada) - a posio da raiz no arquivo dela
 *      chave (entrada) - a chave de busca
 *
 * Retorno: ndice do registro no arquivo de registros
 *          se a chave for encontrada; -1, se a chave
 *          no for encontrada.
 *
 ****/
int BuscaB(FILE *stream, int raiz, tChave chave)
{
   tNoMultiMS no;   /* Armazenar cada n lido no arquivo */
   int       posNo = raiz, /* Posio de um n no arquivo,  */
                           /* com a busca comeando na raiz */
       i; /* ndice de uma chave num n */

      /* Desce na rvore at encontrar a    */
      /* chave de busca ou uma posio nula */
   while (posNo != POSICAO_NULA) {
         /* L o n na posio indicada */
      LeNoMultiMS(stream, posNo, &no);

         /* Tenta encontrar a chave no n recm-recuperado */
      i = BuscaEmNoMultiMS(chave, &no);

         /* Verifica se a chave foi encontrada */
      if (i < no.nFilhos - 1 && chave == no.chaves[i].chave)
         return no.chaves[i].indice; /* A chave foi encontrada */

      posNo = no.filhos[i]; /* Desce mais um nvel na rvore */
   }

   return POSICAO_NULA;  /* A chave no foi encontrada */
}

/****
 *
 * InsereB(): Insere um par chave/ndice numa rvore B armazenada em arquivo
 *
 * Parmetros:
 *     *posicaoRaiz (entrada/sada) - posio no arquivo da
 *                                    raiz da rvore
 *      chaveEIndice (entrada/sada) - ponteiro para um par
 *                                     chave/ndice que ser
 *                                     encontrado/inserido
 *     *indiceDaChave (sada) - ndice da chave no n no qual
 *                              ela se encontra ou no qual ela
 *                              ser inserida
 *      streamArvore (entrada) - stream associado ao arquivo
 *                               que contm a rvore
 *
 * Retorno: 1, se houver insero; 0, em caso contrrio
 *
 ****/
int InsereB( int *posicaoRaiz, tChaveIndice *chaveEIndice,
             FILE *streamArvore )
{
   tNoMultiMS    no; /* Novo n que poder ser inserido */
   int          encontrado, /* Indicar se a chave */
                            /* foi encontrada      */
                indiceDaChave, /* ndice da chave no n onde     */
                               /* ela for encontrada ou inserida */
                pFilhoDireito,
                pNoDireito;
   tPilha       pilha; /* Pilha que armazenar o caminho */
                       /* de ns visitados               */
   tNoCaminhoB   caminho; /* Armazenar cada item desempilhado */
   tChaveIndice chaveAInserir, chaveQueSobe;

      /* O stream que representa o arquivo     */
      /* que contm a rvore no pode ser NULL */
   ASSEGURA( streamArvore,
             ERRO_STREAM_NULL(streamArvore, InsereB) );

      /* Verifica se a rvore j existe */
   if (*posicaoRaiz == POSICAO_NULA) {
      /*                               */
      /* A rvore ainda no foi criada */
      /*                               */

         /* Cria a raiz  */
      IniciaNoMultiMS(chaveEIndice, &no);
      indiceDaChave = 0;
      *posicaoRaiz = 0;

         /* Armazena a raiz da rvore */
         /* na posio 0 do arquivo   */
      EscreveNoMultiMS(streamArvore, 0, &no);

      return 1;
   }

         /* Tenta encontrar o n que contm a  */
         /* chave e o caminho que leva at ele */
   EncontraCaminhoB( streamArvore, chaveEIndice->chave,
                    *posicaoRaiz, &indiceDaChave,
                    &encontrado, &pilha );

      /* Verifica se a chave foi encontrada */
   if (encontrado) {
         /* A chave foi encontrada. Obtm o valor do par   */
         /* chave/ndice que se encontra no topo da pilha. */
      caminho = Desempilha(&pilha);

         /* Atualiza-se o ndice do par 'chaveEIndice', */
         /* cujo endereo foi recebido  como parmetro. */
      chaveEIndice->indice =
         caminho.no.chaves[indiceDaChave].indice;

         /* Antes de retornar, esvazia-se a pilha */
      EsvaziaPilha(&pilha);

      return 0; /* No houve insero */
   }

      /* A diferena significativa entre esta funo */
      /* e a funo correspondente para rvores      */
      /* multidirecionais descendentes comea aqui   */

      /* A insero comea numa folha. Assim, o filho */
      /*  direita da chave a ser inserida  nulo     */
   pFilhoDireito = POSICAO_NULA;

      /* A primeira chave a ser inserida  */
      /* aquela recebida como parmetro    */
   chaveAInserir = *chaveEIndice;

      /* Percorre o caminho inverso at encontrar */
      /* a raiz ou um n incompleto para insero */
   while (!PilhaVazia(pilha)) {
         /* Obtm informaes sobre o n corrente */
      caminho = Desempilha(&pilha);

         /* Verifica se  possvel inserir neste n */
      if (caminho.no.nFilhos < G) {
            /* H espao para insero neste n */
         InsereEmNoB( &caminho.no, indiceDaChave,
                     chaveAInserir, pFilhoDireito );

            /* Atualiza o n na rvore */
         EscreveNoMultiMS(streamArvore, caminho.endereco, &caminho.no);

            /* O contedo da pilha no  mais necessrio */
         EsvaziaPilha(&pilha);

         return 1; /* Servio concludo */
      }

         /* O n corrente  completo e precisa ser dividido */
      DivideNoB( &caminho.no, indiceDaChave, chaveAInserir,
                pFilhoDireito, &pNoDireito, &chaveQueSobe,
                streamArvore );

         /* Atualiza o n na rvore */
      EscreveNoMultiMS(streamArvore, caminho.endereco, &caminho.no);

         /* O filho  direita da chave que subir */
         /* para o nvel superior ser o novo n  */
         /* criado no nvel atual                 */
      pFilhoDireito = pNoDireito;
      indiceDaChave = caminho.pos;
      chaveAInserir = chaveQueSobe;
   }

   /**************************************************/
   /* Neste ponto, sabe-se que o ltimo caminho      */
   /* desempilhado continha uma raiz completa, que   */
   /* j foi dividida no lao while. Resta criar uma */
   /* nova raiz contendo a chave que deve subir um   */
   /* nvel tendo como filho direito o n dividido   */
   /* e  esquerda a antiga raiz aps a diviso      */
   /**************************************************/

      /* Cria a nova raiz */
   IniciaNoMultiMS(&chaveQueSobe, &no);

      /* O filho da esquerda da nova raiz  o n  */
      /* que restou da raiz antiga aps a diviso */
   no.filhos[0] = caminho.endereco;

      /* O filho da direita da nova raiz  */
      /* o novo n resultante da diviso   */
   no.filhos[1] = pNoDireito;

      /* A nova raiz ser acrescentada ao final do arquivo */
      /* que armazena a rvore. Portanto, sua posio     */
      /* igual ao nmero corrente de ns.                  */
   *posicaoRaiz = NumeroDeNosB(streamArvore);

      /* Atualiza raiz no arquivo */
   EscreveNoMultiMS(streamArvore, *posicaoRaiz, &no);

   return 1; /* Insero bem sucedida */
}

/****
 *
 * RemoveChaveB(): Remove a chave especificada de uma rvore
 *                B armazenada em arquivo
 *
 * Parmetros:
 *      *streamArvore (entrada) - stream associado ao arquivo
 *                                que armazena a rvore
 *      *raiz (entrada/sada) - posio no arquivo da raiz da rvore
 *       chave (entrada) - a chave que ser removida
 *
 * Retorno: 1, se a remoo for bem sucedida; 0, em caso contrrio
 *
 * Descrio do algoritmo: v. texto
 *
 ****/
int RemoveChaveB(FILE *streamArvore, int *raiz, tChave chave)
{
   int          encontrado, /* A chave foi encontrada? */
                indiceDaChave, /* ndice da chave no n no */
                         /* qual ela est ou deveria estar */
                iEntreIrmaos, /* O ndice de um n */
                              /* entre seus irmos */
                p, /* Posio de um n no arquivo */
                   /* que armazena a rvore       */
                pNoDaRemocao; /* Posio em arquivo do n que */
                             /* contm a chave a ser removida */
   tPilha       pilha, /* Pilha que armazenar o caminho */
                       /* de ns visitados               */
                pilhaAux; /* Uma pilha auxiliar */
   tNoCaminhoB   caminho; /* Um elemento da pilha que    */
                         /* representa o caminho de ns */
   tNoMultiMS    no; /* Um n da rvore */

      /* Se a rvore for vazia, no h chave para remover */
   if (*raiz == POSICAO_NULA)
      return 0; /* No h chave para remover */

      /* Encontra o caminho da raiz at o n */
      /* que contm a chave a ser removida   */
   EncontraCaminhoB( streamArvore, chave, *raiz,
                     &indiceDaChave, &encontrado, &pilha );

      /* Se a chave no for encontrada, no h mais nada a  */
      /* fazer. Mas, antes de retornar, esvazia-se a pilha. */
   if (!encontrado) {
      EsvaziaPilha(&pilha);

      return 0; /* No h chave para remover */
   }

   /***                                           ***/
   /** A chave foi encontrada e informaes sobre **/
   /** o n que a contm esto no topo da pilha   **/
   /***                                          ***/

      /* Obtm o elemento no topo da */
      /* pilha sem desempilh-lo     */
   caminho = ElementoTopo(pilha);

      /* Verifica se o n no qual ser */
      /* feita a remoo  uma folha   */
   if (caminho.no.filhos[0] == POSICAO_NULA) {
         /* O n  uma folha. A funo RemoveEmFolhaB() */
         /* completa o servio                         */
      RemoveEmFolhaB(&pilha, streamArvore);

      return 1; /* Servio concludo */
   }

   /*******************************************************/
   /* A chave a ser removida no se encontra numa folha.  */
   /*  necessrio substitu-la por sua sucessora e,      */
   /* depois, remover esta sucessora. Em seguida, ser    */
   /* feito um caminhamento, empilhando os ns no caminho */
   /* que leva at a chave sucessora.                     */
   /*******************************************************/

      /* Guarda a posio em arquivo */
      /* do n que contm a chave    */
   pNoDaRemocao = caminho.endereco;

      /****                                    ***/
      /* Ser feita uma descida na rvore usando */
      /* p para indicar um n no nvel corrente  */
      /****                                    ***/

      /* Atribui a p a posio do filho  */
      /* direito da chave a ser removida */
   p = caminho.no.filhos[indiceDaChave + 1];
   iEntreIrmaos = indiceDaChave + 1;

   ASSEGURA( p != POSICAO_NULA, "ERRO: p nao deveria "
             "ser POSICAO_NULA em RemoveChaveB()" );

      /* Segue descendo at encontrar */
      /* uma subrvore esquerda vazia */
   while (p != POSICAO_NULA) {
         /* L o n na posio indicada por p */
      LeNoMultiMS(streamArvore, p, &no);

         /* Armazena num item da pilha o n mais        */
         /* recentemente visitado, sua posio no       */
         /* arquivo da rvore e seu ndice entre irmos */
      caminho.no = no;
      caminho.endereco = p;
      caminho.pos = iEntreIrmaos;

      Empilha(caminho, &pilha); /* Faz o empilhamento */

         /* Desce mais um nvel  esquerda da primeira chave */
      p = no.filhos[0];

      iEntreIrmaos = 0; /* A primeira chave tem ndice zero */
   }

   /****                                             ****/
   /* A chave sucessora da chave a ser removida  a     */
   /* primeira chave do ltimo n lido. Essa chave      */
   /* sucessora deve substituir a chave a ser removida. */
   /* Portanto,  preciso substituir a chave a ser      */
   /* removida no n que se encontra na pilha.          */
   /****                                             ****/

      /* Cria uma pilha auxiliar para guardar os n */
      /* que se encontram sobre o n a ser alterado */
   CriaPilha(&pilhaAux);

      /* Desempilha at encontrar o n que ser alterado */
   do {
      caminho = Desempilha(&pilha);
      Empilha(caminho, &pilhaAux);
   } while (caminho.endereco != pNoDaRemocao);

      /* Substitui a chave a ser removida por sua sucessora */
   caminho.no.chaves[indiceDaChave] = no.chaves[0];

      /* Atualiza n no arquivo */
   EscreveNoMultiMS(streamArvore, caminho.endereco, &caminho.no);

      /* Reconstri a pilha */
   while (!PilhaVazia(pilhaAux)) {
      caminho = Desempilha(&pilhaAux);
      Empilha(caminho, &pilha);
   }

      /* Resta remover a chave sucessora na folha */
      /* que se encontra no topo da pilha         */
   RemoveEmFolhaB(&pilha, streamArvore);

   return 1; /* Remoo bem sucedida */
}

/****
 *
 * CaminhamentoInfixoB(): Executa um caminhamento em ordem infixa
 *                       numa rvore B escrevendo num arquivo de
 *                       texto todas chaves em ordem crescente
 *
 * Parmetros:
 *      streamArvore (entrada) - stream associado ao arquivo
 *                               que armazena a rvore
 *      posNo (entrada) - posio no arquivo que contm
 *                        o n inicial do caminhamento
 *      streamChaves (entrada)- stream associado ao arquivo no
 *                              qual sero escritas as chaves
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixoB( FILE *streamArvore, int posNo,
                         FILE *streamChaves )
{
   int  i;
   tNoMultiMS  umNo;

      /* Nenhum dos streams recebidos pode ser NULL */
   ASSEGURA(streamArvore,
            ERRO_STREAM_NULL(streamArvore, CaminhamentoInfixoB));
   ASSEGURA(streamChaves,
            ERRO_STREAM_NULL(streamChaves, CaminhamentoInfixoB));

      /* Visita cada n da rvore em ordem infixa at   */
      /* que a posio 'posNo' assuma um valor invlido */
   if (posNo != POSICAO_NULA) {
         /* L o n na posio 'posNo' do */
         /* arquivo que contm a rvore   */
      LeNoMultiMS(streamArvore, posNo, &umNo);

         /* Para cada n, caminha-se recursivamente na */
         /* subrvore esquerda de cada chave e depois  */
         /* escreve-se essa chave no arquivo de texto  */
      for (i = 0; i < umNo.nFilhos-1; ++i) {
         CaminhamentoInfixoB( streamArvore, umNo.filhos[i],
                             streamChaves );
            /* Visita a chave de ndice i */
         fprintf(streamChaves, "%ld\n", umNo.chaves[i].chave);
      }

         /* Agora caminha-se recursivamente na */
         /* subrvore direita da ltima chave  */
      CaminhamentoInfixoB( streamArvore,
                          umNo.filhos[umNo.nFilhos-1],
                          streamChaves );
   }
}

/****
 *
 * MenorChaveB(): Encontra a menor chave com a respectiva
 *               posio do registro correspondente numa
 *               rvore B
 *
 * Parmetros:
 *      streamArvore (entrada) - stream associado ao arquivo
 *                               que contm a rvore
 *      raiz (entrada) - posio da raiz da rvore no arquivo
 *                       que contm a rvore
 *
 * Retorno: O par chave/ndice que contm a menor chave da rvore
 *
 ****/
tChaveIndice MenorChaveB(FILE *streamArvore, int raiz)
{
   tNoMultiMS umNo; /* Armazena um n */
   int pos = raiz; /* Armazena a posio de um n, */
                   /* comeando pela raiz          */

      /* Verifica se o stream que representa a rvore  vlido */
   ASSEGURA( streamArvore,
             ERRO_STREAM_NULL(streamArvore, MenorChaveB) );

      /* Verifica se a raiz da rvore  vlida */
   ASSEGURA( raiz >= 0, ERRO_POSICAO(MenorChaveB) );

      /* Encontra o n mais  esquerda na rvore */
   do {
         /* L o n cuja posio no arquivo */
         /*  indicada por 'pos'            */
      LeNoMultiMS(streamArvore, pos, &umNo);

         /* Desce-se at o filho mais  esquerda deste n */
      pos = umNo.filhos[0];
   } while (pos != POSICAO_NULA);

      /* Retorna o par chave/ndice que tem a menor chave */
   return umNo.chaves[0];
}

/****
 *
 * MaiorChaveB(): Encontra a maior chave com a respectiva
 *               posio do registro correspondente numa
 *               rvore B
 *
 * Parmetros:
 *      streamArvore (entrada) - stream associado ao arquivo
 *                               que contm a rvore
 *
 * Retorno: A posio do registro que contm a maior chave
 *          da rvore
 *
 ****/
tChaveIndice MaiorChaveB(FILE *streamArvore, int raiz)
{
   tNoMultiMS umNo; /* Armazena um n */
   int pos = raiz; /* Armazena a posio de um n, */
                   /* comeando pela raiz          */

      /* Verifica se o stream que representa a rvore  vlido */
   ASSEGURA( streamArvore,
             ERRO_STREAM_NULL(streamArvore, MaiorChaveB) );

      /* Verifica se a raiz da rvore  vlida */
   ASSEGURA( raiz >= 0, ERRO_POSICAO(MaiorChaveB) );

      /* Encontra o n mais  direita na rvore */
   do {
         /* L o n cuja posio no arquivo */
         /*  indicada por 'pos'            */
      LeNoMultiMS(streamArvore, pos, &umNo);

         /* Passa para o filho mais  direita deste n */
      pos = umNo.filhos[umNo.nFilhos - 1];
   } while (pos != POSICAO_NULA);

      /* Retorna o par chave/ndice que tem a maior chave */
   return umNo.chaves[umNo.nFilhos - 2];
}

/****
 *
 * NumeroDeNosB(): Calcula o nmero de ns de uma rvore
 *                B armazenada em arquivo
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *                         contendo a rvore
 *
 * Retorno: O nmero de ns da rvore
 *
 ****/
int NumeroDeNosB(FILE *stream)
{
   int       nNos = 0;
   tNoMultiMS umNo;

      /* O stream que contm a rvore no pode ser NULL */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, NumeroDeNosB) );

      /* Tenta mover o apontador de posio */
      /* do arquivo para o seu incio       */
   MoveApontador(stream, 0, SEEK_SET);

      /* Acessa sequencialmente cada n da rvore */
      /* contando o nmero de chaves em cada n   */
   while (1) {
         /* L um n da rvore */
      fread(&umNo, sizeof(umNo), 1, stream);

         /* Se ocorreu erro ou o final do arquivo */
         /* foi atingido, encerra o lao          */
      if (ferror(stream) || feof(stream))
         break;

         /* Verifica se o n no foi removido */
      if (umNo.nFilhos > 0)
         ++nNos;
   }

      /* Verifica se houve erro de leitura */
   ASSEGURA(!ferror(stream),ERRO_FREAD(NumeroDeNosB));

   return nNos; /* Servio completo */
}

/****
 *
 * NumeroDeChavesB(): Calcula o nmero de chaves de uma
 *                   rvore B armazenada em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que
 *                        armazena a rvore cujo nmero de
 *                        chaves ser determinado
 *
 * Retorno: O referido nmero de chaves
 *
 ****/
int NumeroDeChavesB(FILE *stream)
{
   int       nChaves = 0;
   tNoMultiMS umNo;

      /* O stream que contm a rvore no pode ser NULL */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, NumeroDeChavesB) );

      /* Tenta mover o apontador de posio do arquivo para o seu incio */
   MoveApontador(stream, 0, SEEK_SET);

      /* Acessa sequencialmente cada n da rvore */
      /* contando o nmero de chaves em cada n   */
   while (1) {
         /* L um n da rvore */
      fread(&umNo, sizeof(umNo), 1, stream);

         /* Se ocorreu erro ou o final do arquivo */
         /* foi atingido, encerra o lao          */
      if (ferror(stream) || feof(stream))
         break;

         /* Conta o nmero de chaves do n corrente */
      if (umNo.nFilhos > 0)
         nChaves += umNo.nFilhos - 1;
   }

      /* Verifica se houve erro de leitura */
   ASSEGURA(!ferror(stream),ERRO_FREAD(NumeroDeChavesB));

   return nChaves; /* Servio completo */
}

/****
 *
 * AlturaB(): Calcula a altura de uma rvore B
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que
 *                        armazena a rvore cujo nmero de
 *                        chaves ser determinado
 *     raiz (entrada) - a posio da raiz da rvore no
 *                      arquivo que a contm
 *
 * Retorno: A referida altura
 *
 ****/
int AlturaB(FILE *stream, int raiz)
{
   int       altura = 0;
   tNoMultiMS no;

      /* Se a rvore estiver vazia, sua altura  zero */
   if (raiz == POSICAO_NULA)
      return 0;

      /* O stream que contm a rvore no pode ser NULL */
   ASSEGURA( stream, ERRO_STREAM_NULL(stream, AlturaB) );

      /* A altura de uma rvore B corresponde ao nvel de qualquer folha, j que  */
      /* todas elas esto no mesmo  nvel. Como a folha mais fcil de encontrar  */
      /* aquela mais  esquerda na rvore,  isso que faz o lao a seguir.        */
   do {
      LeNoMultiMS(stream, raiz, &no);

      altura++; /* Visitou-se mais um n */

      raiz = no.filhos[0]; /* Desce sempre pelo filho esquerdo do n corrente */
   } while (raiz != POSICAO_NULA);

   return altura;
}

