/****
 *
 * Arquivo TabelaCuco.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2008
 * ltima modificao: 08/06/2010
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso aberto
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include <math.h>     /* Funes strxxx() e memxxx() */
#include "TabelaCuco.h"  /* Interface deste mdulo      */
#include "Registros.h"  /* Processamento de registros  */

/********************* Funes Locais *********************/

/****
 *
 * RedimensionaCuco(): Redimensiona uma tabela de disperso cuco
 *
 * Parmetros:
 *     tabela (entrada/sada) - a tabela de disperso
 *     chaveEIndice (entrada) - o par chave/ndice que ser inserido
 *                              aps o redimensionamento da tabela
 *
 * Retorno: Nada
 *
 ****/
static void RedimensionaCuco(tTabelaCuco *tabela, tCEP_Ind chaveEIndice)
{
   tTabelaCuco novaTabela; /* A nova tabela */
   int         i,
               teste, /* Resultado de uma insero */
               novoTamanho; /* Tamanho da nova tabela */

      /* O novo tamanho ser 50% maior do que o antigo valor */
   novoTamanho = 1.5*tabela->tam;

      /* Cria uma nova tabela de disperso cuco */
   CriaTabelaCuco(&novaTabela, novoTamanho, tabela->fD1, tabela->fD2);

      /* Insere as chaves da antiga tabela na nova tabela */
   for (i = 0; i < tabela->tam; ++i) {
         /* Verifica se a posio corrente da primeira tabela contm uma chave */
      if (tabela->tab1[i].status == OCUPADO) {
            /* Existe uma chave nesta posio */
         teste = InsereCuco( &novaTabela, tabela->tab1[i].chaveEIndice );
            /* Esta insero no pode falhar */
         ASSEGURA( teste, "Falha de insercao em redimensionamento" );
      }

         /* Verifica se a posio corrente da segunda tabela contm uma chave */
      if (tabela->tab2[i].status == OCUPADO) {
            /* Existe uma chave nesta posio */
         teste = InsereCuco( &novaTabela, tabela->tab2[i].chaveEIndice );
            /* Esta insero no pode falhar */
         ASSEGURA( teste, "Falha de insercao em redimensionamento" );
      }
   }

      /* Neste ponto, todas as chaves da antiga tabela foram inseridas, */
      /* de modo que a tabela antiga pode ser extinta                   */
   DestroiTabelaCuco(tabela);

   *tabela = novaTabela; /* Substitui a tabela antiga com a nova tabela */

    teste = InsereCuco(tabela, chaveEIndice); /* Insere a chave que restou */

      /* Esta insero no pode falhar */
    ASSEGURA(teste, "Falha de insercao em redimensionamento");
}

/********************* Funes Globais ********************/

/****
 *
 * CriaTabelaCuco(): Cria e inicializa uma tabela de disperso cuco
 *
 * Parmetros:
 *     tabelas (sada) - a tabela cuco criada e iniciada
 *     nElementos (entrada) - nmero de posies da tabela
 *                            de disperso
 *     fD1, fD2 (entrada) - as funes de disperso utilizadas
 *
 * Retorno: Nada
 *
 ****/
void CriaTabelaCuco( tTabelaCuco *tabelas, int nElementos,
                     tFDispersaoCuco fD1, tFDispersaoCuco fD2 )
{
   int i;

      /* Aloca espao para as duas tabelas */
   tabelas->tab1 = calloc(nElementos, sizeof(tColetorCuco));
   tabelas->tab2 = calloc(nElementos, sizeof(tColetorCuco));
   ASSEGURA( tabelas->tab1 && tabelas->tab2,
             "\nImpossivel alocar a tabela de disperso\n" );

      /* Inicia as duas tabelas */
   for (i = 0; i < nElementos; ++i) {
         /* Todos os elementos esto inicialmente desocupados */
      tabelas->tab1[i].status = VAZIO;
      tabelas->tab2[i].status = VAZIO;
   }

      /* Inicia as funes de disperso */
   tabelas->fD1 = fD1;
   tabelas->fD2 = fD2;

      /* Inicia o nmero de chaves nas tabelas */
   tabelas->nChaves = 0;
}

/****
 *
 * BuscaCuco(): Executa uma busca simples numa tabela de disperso cuco
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      chave (entrada) - a chave de busca
 *
 * Retorno: ndice do registro no arquivo de registros,
 *          se a chave for encontrada; -1, em caso contrrio
 *
 ****/
int BuscaCuco( tTabelaCuco tabela, tCEP chave )
{
   int pos; /* Essa varivel no  estritamente necessria, mas */
            /* ela facilita a escrita das expresses seguintes  */

   pos = tabela.fD1(chave, tabela.tam);

      /* Verifica se a chave se encontra na primeira tabela */
   if ( tabela.tab1[pos].status == OCUPADO &&
        !strcmp(tabela.tab1[pos].chaveEIndice.chave, chave) )
      return tabela.tab1[pos].chaveEIndice.indice; /* Chave encontrada */

   /* A chave ainda no foi encontrada. Verifica */
   /* se ela se encontra na segunda tabela.      */

   pos = tabela.fD2(chave, tabela.tam);

   if ( tabela.tab2[pos].status == OCUPADO &&
        !strcmp(tabela.tab2[pos].chaveEIndice.chave, chave) )
      return tabela.tab2[pos].chaveEIndice.indice; /* Chave encontrada */

   return -1; /* A chave no foi encontrada em nenhuma das tabelas */
}

/****
 *
 * InsereCuco(): Faz insero numa tabela de disperso cuco
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela de disperso
 *      chaveEIndice (entrada) - o par chave/ndice a ser inserido
 *
 * Retorno: 1, se houver insero; 0, em caso contrrio
 *
 ****/
int InsereCuco(tTabelaCuco *tabela, tCEP_Ind chaveEIndice)
{
   int          i,
                pos,
                maxDesalojamentos;
   tCEP_Ind     aux;

      /* Verifica se a chave j existe na tabela */
   if (BuscaCuco(*tabela, chaveEIndice.chave) >= 0)
      return 0; /* Chave j existe na tabela */

      /* Calcula o valor limite de desalojamentos. O valor   */
      /* nlog n  sugerido pelos criadores da disperso cuco */
   maxDesalojamentos = tabela->nChaves*log2(tabela->nChaves);

      /* Tenta efetuar a insero numa das tabelas */
   for (i = 0; i < maxDesalojamentos; ++i) {
      /*                                  */
      /* Tenta inserir na primeira tabela */

      pos = tabela->fD1(chaveEIndice.chave, tabela->tam);

      if (tabela->tab1[pos].status == VAZIO) {
         /*                                                                */
         /* Foi encontrado um espao vazio na primeira tabela. Efetua-se a */
         /* insero e retorna-se  indicando o sucesso da operao         */
         /*                                                                */

            /* Insere o par chave/ndice */
         tabela->tab1[pos].chaveEIndice = chaveEIndice;

         tabela->tab1[pos].status = OCUPADO; /* Esta posio passa a ser ocupada */

         ++tabela->nChaves; /* O nmero de chaves aumentou */

         return 1;
      } else {
         /*                                                                     */
         /* No foi encontrado um espao vazio.  preciso desalojar a chave que */
         /* se encontra nessa posio e tentar inseri-la na segunda tabela      */
         /*                                                                     */

         aux = tabela->tab1[pos].chaveEIndice; /* Guarda chave a ser desalojada */
            /* Armazena a nova chave no lugar da chave desalojada */
         tabela->tab1[pos].chaveEIndice = chaveEIndice;
            /* Tentar-se-a inserir a chave desalojada na segunda tabela */
         chaveEIndice = aux;
      }

      /*                                  */
      /* Tenta inserir na segunda tabela */
      /*                                  */

      pos = tabela->fD2(chaveEIndice.chave, tabela->tam);

      if (tabela->tab2[pos].status == VAZIO) {
         /*                                                             */
         /* Foi encontrado um espao vazio na segunda tabela. Efetua-se */
         /* a insero e retorna-se indicando o sucesso da operao     */
         /*                                                             */

            /* Insere o par chave/ndice */
         tabela->tab2[pos].chaveEIndice = chaveEIndice;

            /* Esta posio passa a ser ocupada */
         tabela->tab2[pos].status = OCUPADO;

         ++tabela->nChaves; /* O nmero de chaves aumentou */

         return 1;
      } else {
         /*                                                                     */
         /* No foi encontrado um espao vazio.  preciso desalojar a chave que */
         /* se encontra nessa posio e tentar inseri-la na primeira tabela     */
         /*                                                                     */

            /* Guarda a chave a ser desalojada */
         aux = tabela->tab2[pos].chaveEIndice;
            /* Armazena a nova chave no lugar da chave desalojada */
         tabela->tab2[pos].chaveEIndice = chaveEIndice;
            /* Tentar-se- inserir a chave desalojada na primeira tabela */
         chaveEIndice = aux;
      }
   }

   /*                                                                   */
   /* Se o lao terminou sem que houvesse retorno, considera-se que     */
   /* o processo entrou em repetio  sem fim (de acordo com o critrio */
   /* especificado). Portanto, a tabela ser reconstruda e a chave     */
   /* restante ser inserida nessa nova tabela.                         */
   /*                                                                   */

   RedimensionaCuco(tabela, chaveEIndice);

   return 1;
}

/****
 *
 * RemoveCuco(): Remove uma chave de uma tabela de disperso cuco
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela de disperso
 *      chave (entrada) - a chave de busca
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveCuco(tTabelaCuco *tabela, tCEP chave)
{
   int pos; /* Essa varivel no  estritamente necessria, mas */
            /* ela facilita a escrita das expresses seguintes  */

   pos = tabela->fD1(chave, tabela->tam);

      /* Tenta efetuar a remoo na primeira tabela */
   if ( tabela->tab1[pos].status == OCUPADO &&
        !strcmp(tabela->tab1[pos].chaveEIndice.chave, chave) ) {
      tabela->tab1[pos].status = VAZIO; /* Chave encontrada */

      --tabela->nChaves;

      return 1;
   }

   /*                                                                    */
   /* A chave ainda no foi removida. Tenta remov-la da segunda tabela. */
   /*                                                                    */

   pos = tabela->fD2(chave, tabela->tam);

   if ( tabela->tab2[pos].status == OCUPADO &&
        !strcmp(tabela->tab2[pos].chaveEIndice.chave, chave) ) {
      tabela->tab2[pos].status = VAZIO; /* Chave encontrada */

      --tabela->nChaves;

      return 1;
   }

   return 0;  /* A chave no foi encontrada */
}

/****
 *
 * NChavesCuco(): Calcula o nmero de chaves na tabela de disperso
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *
 * Retorno: O nmero de chaves na tabela de disperso
 *
 ****/
int NChavesCuco(tTabelaCuco tabela)
{
   int i, nChaves = 0;

   for (i = 0; i < tabela.tam; ++i) {
      if (tabela.tab1[i].status == OCUPADO)
         ++nChaves;

      if (tabela.tab2[i].status == OCUPADO)
         ++nChaves;
   }

   ASSEGURA(tabela.nChaves == nChaves, "\nErro na contagem de numero de chaves\n");

   return nChaves;
}

/****
 *
 * NElementosVaziosCuco(): Calcula o nmero de elementos vazios
 *                     na tabela de disperso
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *
 * Retorno: O nmero de elementos vazios na tabela de disperso
 *
 ****/
int NElementosVaziosCuco(tTabelaCuco tabela)
{
   int i, n = 0;

   for (i = 0; i < tabela.tam; ++i) {
      if (tabela.tab1[i].status == VAZIO)
         ++n;

      if (tabela.tab2[i].status == VAZIO)
         ++n;
   }

   return n;
}

/****
 *
 * DestroiTabelaCuco(): Libera o espao ocupado por uma tabela
 *                  de disperso com endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *
 * Retorno: Nada
 *
 ****/
void DestroiTabelaCuco(tTabelaCuco *tabela)
{
      /* Libera o espao ocupado por cada array */
   free(tabela->tab1);
   free(tabela->tab2);

      /* Os arrays no so mais vlidos */
   tabela->tab1 = NULL;
   tabela->tab2 = NULL;
}
