/****
 *
 * Ttulo: Filtro de Bloom
 *
 * Seo: 7.8.2
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2016
 * ltima modificao: 16/04/2016
 *
 * Descrio: Filtro de Bloom
 *
 * Entrada: No h
 *
 * Sada: v. Resultado do Programa
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* printf()   */
#include <stdlib.h> /* Miscelnea */
#include <limits.h> /* CHAR_BIT   */

/************************* Macros *************************/

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

   /* Tamanho do filtro em bytes */
#define TAM_FILTRO 40

/****************** Definies de Tipos *******************/

   /* Tipo de ponteiro para uma funo de disperso */
typedef unsigned int (*tFDispersao)(const char *);

   /* Tipo de n de lista encadeada que armazena */
   /* ponteiros para funes de disperso        */
typedef struct rotNoListaBloom {
           tFDispersao             fDispersao;
           struct rotNoListaBloom *proximo;
        } tNoListaBloom, *tListaBloom;

   /* Tipo de estrutura que representa um filtro de Bloom */
typedef struct {
           tListaBloom listaDispersao;
           void       *bits;
           size_t      tamanho;
        } tFiltroBloom, *tFiltroBloomPtr;

/************************* Aluses ************************/

extern tFiltroBloomPtr CriaFiltro(size_t tamanho);
extern void DestroiFiltro(tFiltroBloomPtr filtro);
extern void AcrescentaFuncaoAFiltro( tFiltroBloomPtr filtro,
                                     tFDispersao fDispersao );
extern void AcrescentaItemAFiltro(tFiltroBloomPtr filtro, const void *item);
extern int ItemEstaEmFiltro(tFiltroBloomPtr filtro, const void *item);

extern unsigned int DispersaoPolinomial(const char *chave);
extern unsigned int DispersaoRotativa(const char *chave);
extern unsigned int DispersaoJenkins(const char *chave);
extern unsigned int DispersaoDJB(const char *chave);
extern unsigned int DispersaoDJB2(const char *chave);
extern unsigned int DispersaoSAX(const char *chave);
extern unsigned int DispersaoFNV(const char *chave);

/***************** Definies de Funes ******************/

/****
 *
 * CriaFiltro(): Cria um novo filtro de Bloom sem nenhuma
 *               funo de disperso e tamanho especificado
 *
 * Parmetros:
 *     tamanho (entrada) - tamanho do filtro em bytes
 *
 * Retorno: Ponteiro para o filtro criado ou NULL se a estrutura
 *          ou o array de bits no forem alocados
 *
 ****/
tFiltroBloomPtr CriaFiltro(size_t tamanho)
{
   tFiltroBloomPtr filtroPtr = calloc(1, sizeof(tFiltroBloom));

      /* Garante que houve alocao */
   ASSEGURA(filtroPtr, "Alocao filtro fracassou");

   filtroPtr->tamanho = tamanho;
   filtroPtr->bits = calloc(1, tamanho);
   filtroPtr->listaDispersao = NULL;

      /* Garante que houve alocao do array de bits */
   ASSEGURA(filtroPtr->bits, "Alocao do array fracassou");

   return filtroPtr;
}

/****
 *
 * DestroiFiltro(): Libera o espao ocupado por um filtro de Bloom
 *
 * Parmetros:
 *     filtro (entrada/sada) - ponteiro para o filtro de Bloom
 *
 * Retorno: Nada
 *
 ****/
void DestroiFiltro(tFiltroBloomPtr filtro)
{
   tListaBloom p;

   if (filtro) {
         /*                                            */
         /* Libera espao ocupado pela lista encadeada */
         /*                                            */

         /* Faz p apontar para o incio da lista */
      p = filtro->listaDispersao;

         /* Libera cada n da lista */
      while (p) {
            /* Passa para o prximo n antes que o n corrente seja destrudo */
         filtro->listaDispersao = p->proximo;
         free(p); /* Libera o espao do n corrente */

         p = filtro->listaDispersao; /* Faz p apontar para o prximo n */
      }

      free(filtro->bits); /* Libera o array de bits */

      free(filtro); /* Libera a estrutura que representa o filto */
   }
}

/****
 *
 * AcrescentaFuncaoAFiltro(): Acrescenta uma funo de disperso ao filtro de Bloom
 *
 * Parmetros:
 *     filtro (entrada/sada) - ponteiro para o filtro de Bloom
 *     fDispersao (entrada) - funo de disperso que ser
 *                            acrescentada
 *
 * Retorno: Nada
 *
 ****/
void AcrescentaFuncaoAFiltro(tFiltroBloomPtr filtro, tFDispersao fDispersao)
{
   tListaBloom pNo = malloc(sizeof(tListaBloom *));

   ASSEGURA(pNo, "Impossivel alocar no de lista encadeada");

      /* Armazena no novo n o endereo da */
      /*funo recebido como parmetro     */
   pNo->fDispersao = fDispersao;

      /* O novo n apontar para o incio corrente da lista */
   pNo->proximo = filtro->listaDispersao;

      /* O incio da lista passa a apontar para o novo n */
   filtro->listaDispersao = pNo;
}

/****
 *
 * AcrescentaItemAFiltro(): Acrescenta um item a um filtro de Bloom
 *
 * Parmetros:
 *     filtro (entrada/sada) - ponteiro para o filtro de Bloom
 *     item (entrada) - item que ser acrescentado
 *
 * Retorno: Nada
 *
 ****/
void AcrescentaItemAFiltro(tFiltroBloomPtr filtro, const void *item)
{
   tListaBloom    pNo; /* Apontar para um n da lista */
                       /* de endereos de funes      */
   unsigned char *byte; /* Apontar para cada byte do    */
                        /* array que representa o filtro */
   unsigned int   dispersao, /* Armazenar um valor de disperso */
                  nBits; /* Nmero de bits do filtro */

   nBits = CHAR_BIT*filtro->tamanho; /* Calcula o nmero de bits do filtro */

      /* Faz 'byte' apontar para o primeiro byte  */
      /* do array que constitui o filtro de Bloom */
   byte = filtro->bits;

      /* Faz 'pNo' apontar para o primeiro n da lista que */
      /* contm os endereos das funes de disperso      */
   pNo = filtro->listaDispersao;

      /* Aplica cada funo de disperso cujo endereo se encontra na   */
      /* lista e liga o bit correspondente ao valor de disperso obtido */
   while (pNo) {
         /* Obtm o valor de disperso aplicando a funo */
         /* cujo endereo se encontra no n corrente      */
      dispersao = pNo->fDispersao(item);

         /* Assegura que o valor de disperso obtido  */
         /* menor do que o nmero de bits do filtro    */
      dispersao %= nBits;

         /* Liga o bit correspondente ao valor de disperso */
      byte[dispersao/CHAR_BIT] |= 1 << dispersao%CHAR_BIT;

      pNo = pNo->proximo; /* Passa para o prximo n da lista */
   }
}

/****
 *
 * ItemEstaEmFiltro(): Verifica se um item se encontra num filtro de Bloom
 *
 * Parmetros:
 *     filtro (entrada/sada) - ponteiro para o filtro de Bloom
 *     item (entrada) - item que ser acrescentado
 *
 * Retorno: 0, se o item nunca foi acrescentado ao filtro, ou
 *          1 se o item foi provavelmente acrescentado ao filtro
 *
 ****/
int ItemEstaEmFiltro(tFiltroBloomPtr filtro, const void *item)
{
   tListaBloom    pNo; /* Apontar para um n da lista */
                       /* de endereos de funes      */
   unsigned char *byte; /* Apontar para cada byte do    */
                        /* array que representa o filtro */
   unsigned int   dispersao, /* Armazenar um valor de disperso */
                  nBits; /* Nmero de bits do filtro */

   nBits = CHAR_BIT*filtro->tamanho; /* Calcula o nmero de bits do filtro */

      /* Faz 'byte' apontar para o primeiro byte  */
      /* do array que constitui o filtro de Bloom */
   byte = filtro->bits;

      /* Faz 'pNo' apontar para o primeiro n da lista que */
      /* contm os endereos das funes de disperso      */
   pNo = filtro->listaDispersao;

      /* Aplica cada funo de disperso cujo endereo se encontra na lista e */
      /* verifica se o bit correspondente ao valor de disperso obtido est   */
      /* ligado. O lao encerra quando  encontrado algum bit desligado ou    */
      /* quando todas as funes tiverem sido aplicadas sem que isso acontea */
   while (pNo) {
         /* Obtm o valor de disperso aplicando a funo */
         /* cujo endereo se encontra no n corrente      */
      dispersao = pNo->fDispersao(item);

         /* Assegura que o valor de disperso obtido  */
         /* menor do que o nmero de bits do filtro    */
      dispersao %= nBits;

         /* Verifica se o bit correspondente ao valor de disperso est ligado */
      if ( !(byte[dispersao/CHAR_BIT] & (1 << dispersao%CHAR_BIT)) )
         return 0; /* Item no se encontra no filtro */

      pNo = pNo->proximo;
   }

   return 1; /* Talvez o item faa parte do filtro */
}

/***************** Funes de Disperso *******************/

/****
 *
 * DispersaoPolinomial(): Calcula o valor de disperso de um
 *                        string usando o mtodo polinomial
 *
 * Parmetros:
 *      chave (entrada) - o string
 *
 * Retorno: Nada
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoPolinomial(const char *chave)
{
   unsigned int dispersao = 0,
                base = 2,
                primo = 31;

   while (*chave) {
      dispersao = (dispersao*base + *chave++)%primo;
   }

   return dispersao;
}

/****
 *
 * DispersaoRotativa(): Calcula o valor de disperso de um
 *                      string usando o mtodo rotativo
 *
 * Parmetros:
 *      chave (entrada) - o string
 *
 * Retorno: Nada
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoRotativa(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
       dispersao = (dispersao << 4) ^
                   (dispersao >> 28) ^ (*chave++);
   }

   return dispersao;
}

/****
 *
 * DispersaoJenkins(): Implementao da funo de disperso JOAAT
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoJenkins(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao += *chave++;
      dispersao += (dispersao << 10);
      dispersao ^= (dispersao >> 6);
   }

   dispersao += (dispersao << 3);
   dispersao ^= (dispersao >> 11);
   dispersao += (dispersao << 15);

   return dispersao;
}

/****
 *
 * DispersaoDJB(): Implementao da funo de disperso DJB
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoDJB(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao = 33*dispersao + *chave++;
   }

   return dispersao;
}

/****
 *
 * DispersaoDJB2(): Implementao da funo de disperso DJB2
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoDJB2(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao = 33*dispersao ^ *chave++;
   }

   return dispersao;
}

/****
 *
 * DispersaoSAX(): Implementao da funo de disperso SAX
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoSAX(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao ^= (dispersao << 5) +
                   (dispersao >> 2) + *chave++;
   }

   return dispersao;
}

/****
 *
 * DispersaoFNV(): Implementao da funo de disperso FNV
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoFNV(const char *chave)
{
   unsigned int dispersao = 2166136261;

   while (*chave) {
      dispersao = (dispersao*16777619) ^ *chave++;
   }

   return dispersao;
}


/********************* Funo main() **********************/

/****
 *
 * main(): Testa o filtro de Bloom
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   tFiltroBloomPtr  filtro; /* O filtro de Bloom */
   int         consulta, /* Resultado de uma consulta */
               i;
   const char *itens[] = { "banana", "laranja", "jaca",
                           "pitomba", "oliveira", "coco"
                         };
   size_t      nItens = sizeof(itens)/sizeof(itens[0]);
   const char *s1 = "jabuticaba",
              *s2 = "umbu";

      /* Cria o filtro de Bloom */
   filtro = CriaFiltro(TAM_FILTRO);

      /* Acrescenta os endereos das funes */
      /* de disperso ao filtro              */
   AcrescentaFuncaoAFiltro(filtro, DispersaoPolinomial);
   AcrescentaFuncaoAFiltro(filtro, DispersaoRotativa);
   AcrescentaFuncaoAFiltro(filtro, DispersaoJenkins);
   AcrescentaFuncaoAFiltro(filtro, DispersaoDJB);
   AcrescentaFuncaoAFiltro(filtro, DispersaoDJB2);
   AcrescentaFuncaoAFiltro(filtro, DispersaoSAX);
   AcrescentaFuncaoAFiltro(filtro, DispersaoFNV);

      /* Acrescenta itens no filtro */
   for (i = 0; i < nItens; ++i)
      AcrescentaItemAFiltro(filtro, itens[i]);

      /* Consulta itens no filtro */
   for (i = 0; i < nItens; ++i) {
      consulta = ItemEstaEmFiltro(filtro, itens[i]);

      printf( "\t>>> \"%s\"%sfoi encontrado\n", itens[i],
              consulta ? " " : " nao " );
   }

      /*                                        */
      /* Consulta itens que no esto no filtro */
      /*                                        */

   consulta = ItemEstaEmFiltro(filtro, s1);
   printf( "\n\t>>> \"%s\"%sfoi encontrado\n", s1,
           consulta ? " " : " nao " );

   consulta = ItemEstaEmFiltro(filtro, s2);
   printf( "\t>>> \"%s\"%sfoi encontrado\n", s2,
           consulta ? " " : " nao " );

   return 0;
}

/****************** Resultado do Programa *****************

        >>> "banana" foi encontrado
        >>> "laranja" foi encontrado
        >>> "jaca" foi encontrado
        >>> "pitomba" foi encontrado
        >>> "oliveira" foi encontrado
        >>> "coco" foi encontrado

        >>> "jabuticaba" nao foi encontrado
        >>> "umbu" nao foi encontrado

**********************************************************/
