/****
 *
 * Arquivo DispersaoExt.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 06/09/2017
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso extensvel em memria
 *            secundria
 *
 ****/

/*********************** Includes *************************/

#include <limits.h>     /* Limites de tipos inteiros   */
#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include <math.h>       /* Funes strxxx() e memxxx() */
#include "DispersaoExt.h" /* Interface deste mdulo    */
#include "Registros.h"  /* Processamento de registros  */
#include "ASSEGURA.h"   /* Macro ASSEGURA              */

/********************* Funes Locais *********************/

/****
 *
 * ValorBit(): Retorna o valor de um bit de um valor inteiro
 *
 * Parmetros:
 *      valor (entrada) - valor que ter seu bit retornado
 *      pos (entrada) - posio do bit que ser retornado
 *
 * Retorno: O valor do bit
 *
 ****/
int ValorBit(int valor, int pos)
{
   return 1 & valor >> pos;
}

/****
 *
 * ObtemBitsDeChave(): Obtm os bits menos significativos de uma chave
 *
 * Parmetros:
 *      chave (entrada) - a chave da qual os bits sero obtidos
 *      nBits (entrada) - o nmero de bits que sero extrados da chave
 *
 * Retorno: O valor inteiro que representa esses bits
 *
 ****/
static int ObtemBitsDeChave(tChave chave, int nBits)
{
   int mascara = (1 << nBits) - 1;

   return chave & mascara;
}

/****
 *
 * NovoColetorDExt(): Cria um novo coletor num arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O ndice do coletor criado
 *
 ****/
static int NovoColetorDExt(FILE *stream)
{
   tColetorDExt coletor;
   int          indice;

   coletor.nRegistros = 0;
   coletor.profLocal = 0;

      /* Move o apontador de posio do arquivo para seu final */
   MoveApontador(stream, 0, SEEK_END);

      /* Obtm o ndice do coletor */
   indice = ftell(stream)/sizeof(coletor);

      /* Escreve o coletor no arquivo */
   fwrite(&coletor, sizeof(coletor), 1, stream);

   return indice;
}

/****
 *
 * InsereEmDiretorioDExt(): Insere num diretrio uma referncia para
 *                      um coletor de numa tabela de busca com
 *                      disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca
 *      pColetor (entrada) - endereo do coletor que ser inserido
 *      iColetor (entrada) - ndice do coletor no arquivo de
 *                           coletores
 *
 * Retorno: Nada
 *
 ****/
static void InsereEmDiretorioDExt(tTabelaDExt *tab, const tColetorDExt *pColetor,
                                  int iColetor)
{
   int    i, bits, pLocal;
   tChave chave;

      /* Obtm a chave do primeiro item do coletor */
   chave = ObtemChave(pColetor->registros);

   pLocal = pColetor->profLocal; /* Obtm a profundidade local do coletor */

      /* Duplica o tamanho do diretrio enquanto sua profundidade */
      /* global for menor do que a profundidade local do coletor  */
   while (tab->profGlobal < pLocal) {
         /* Duplica o tamanho do diretrio */
      tab->dir = realloc(tab->dir, 2*tab->tamDir*sizeof(long));
      ASSEGURA(tab->dir, "Impossivel alocar novo diretorio");

         /* Copia as referncias do antigo diretrio */
         /* para a metade final do novo diretrio    */
      memcpy(tab->dir + tab->tamDir, tab->dir, tab->tamDir*sizeof(int));

         /* Atualiza a profundidade global e o tamanho do diretrio */
      ++tab->profGlobal;
      tab->tamDir = 2*tab->tamDir;
   }

   /***                                                                     ***/
   /* Se a profundidade local do novo coletor for menor do que a profundidade */
   /* global, ser preciso atualizar mais de uma referncia do diretrio      */
   /***                                                                     ***/

      /* Obtm os bits menos significativos coincidentes no coletor */
   bits = ObtemBitsDeChave(chave, pLocal);

      /* Corrige as referncias do diretrio para o novo coletor */
   for (i = 0; i < tab->tamDir; ++i)
      if (ObtemBitsDeChave(i, pLocal) == bits)
         tab->dir[i] = iColetor;
}

/****
 *
 * PodeDividirColetorDExt(): Verifica se um coletor pode ser dividido
 *
 * Parmetros:
 *     coletor (entrada) - o coletor que ser examinado
 *
 * Retorno: 1, se o coletor pode ser dividido; 0, em caso contrrio
 *
 ****/
static int PodeDividirColetorDExt(const tColetorDExt *coletor)
{
   tChave chave;
   int    i;

      /* Esta funo deve ser chamada apenas quando o coletor est repleto */
   ASSEGURA( coletor->nRegistros == M,
             "PodeDividirColetorDExt() nao deveria ter sido chamada" );

   chave = ObtemChave(coletor->registros); /* Obtm a chave do primeiro registro */

   for (i = 1; i < coletor->nRegistros; ++i)
      if ( chave != ObtemChave(coletor->registros + i) )
         return 1; /* Coletor pode ser dividido */

      /* As chaves deste coletor so todas iguais e, */
      /* portanto, ele no pode ser dividido         */
   return 0;
}

/****
 *
 * InsereColetorVazioDExt(): Insere num diretrio uma referncia para
 *                       um coletor vazio de numa tabela de busca
 *                       com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca
 *      pColetor (entrada) - endereo do coletor que deu origem ao
 *                           coletor vazio
 *      iColetor (entrada) - ndice do coletor que deu origem ao
 *                           coletor vazio no arquivo de coletores
 *      iVazio (entrada) - ndice do coletor vazio no arquivo de
 *                         coletores
 *
 * Retorno: Nada
 *
 ****/
static void InsereColetorVazioDExt( tTabelaDExt *tab, const tColetorDExt *pColetor,
                                    int iColetor, int iVazio )
{
   int    i, bits, pLocal;
   tChave chave;

   pLocal = pColetor->profLocal; /* Obtm a profundidade local do coletor */

      /* Duplica o tamanho do diretrio enquanto sua profundidade global */
      /* for menor do que a profundidade local do coletor                */
   while (tab->profGlobal < pLocal) {
         /* Duplica o tamanho do diretrio */
      tab->dir = realloc(tab->dir, 2*tab->tamDir*sizeof(int));
      ASSEGURA(tab->dir, "Impossivel alocar novo diretorio");

         /* Copia as referncias do antigo diretrio */
         /* para a metade final do novo diretrio    */
      memcpy( tab->dir + tab->tamDir, tab->dir,
              tab->tamDir*sizeof(int) );

         /* Atualiza a profundidade global e o tamanho do diretrio */
      ++tab->profGlobal;
      tab->tamDir = 2*tab->tamDir;
   }

      /* Obtm a chave do primeiro item do coletor que deu origem ao coletor vazio */
   chave = ObtemChave(pColetor->registros);

      /* Obtm os bits menos significativos coincidentes no */
      /* coletor que deu origem ao coletor vazio            */
   bits = ObtemBitsDeChave(chave, pLocal);

      /* Corrige as referncias para o coletor vazio */
   for (i = 0; i < tab->tamDir; ++i)
      if (tab->dir[i] == iColetor && ObtemBitsDeChave(i, pLocal) != bits)
         tab->dir[i] = iVazio;
}

/****
 *
 * DivideColetorDExt(): Divide um coletor de uma tabela de busca
 *                     com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca que contm a
 *                            referncia para o coletor que ser dividido
 *      coletor (entrada/sada) - o coletor que ser dividido
 *      iColetor (entrada) - ndice do coletor que ser dividido
 *                           no arquivo de coletores
 *      stream (entrada) - stream associado ao arquivo contendo os coletores
 *
 * Retorno: Nada
 *
 ****/
static void DivideColetorDExt( tTabelaDExt *tab, tColetorDExt *coletor,
                               int iColetor, FILE *stream )
{
   int          i, iNovo, chave, bits;
   tColetorDExt novo;

      /* Cria um novo coletor e obtm seu ndice em arquivo */
   iNovo = NovoColetorDExt(stream);

   LeColetorDExt(stream, iNovo, &novo); /* L o novo coletor em arquivo */

      /* Divide registros do antigo coletor com o novo coletor */
   while (coletor->nRegistros == M) {
         /* Zera os registros do antigo e do novo coletor */
      coletor->nRegistros = 0;
      novo.nRegistros = 0;

         /* Distribui registros entre o antigo e o novo coletor */
      for (i = 0; i < M; i++) {
         /* Para cada item do antigo coletor, verifica-se  */
         /* se o bit de ordem profLocal + 1 de sua chave   */
         /*  0 ou 1. Se ele for 0, o respectivo item fica */
         /* no antigo coletor. Caso contrrio, esse item   */
         /* vai para o novo coletor.                       */

            /* Obtm a chave do item corrente */
         chave = ObtemChave(coletor->registros + i);

            /* Obtm profLocal + 1 bits menos significativos */
         bits = ObtemBitsDeChave(chave, coletor->profLocal + 1);

            /* Obtm o valor do bit mais significativo    */
            /* dessa sequncia de bits e testa esse valor */
         if (ValorBit(bits, coletor->profLocal) == 0)
               /* Este item fica no antigo coletor */
            coletor->registros[coletor->nRegistros++] = coletor->registros[i];
         else
               /* Este item vai para o novo coletor */
            novo.registros[novo.nRegistros++] = coletor->registros[i];
      }

         /* O novo e o antigo coletor tero a mesma profundidade local, */
         /* que  a profundidade local do antigo coletor mais um        */
      novo.profLocal = ++coletor->profLocal;

         /* Se no ficou nenhum item no coletor antigo,  preciso  */
         /* repetir o processo usando uma profundidade local maior */
      if (novo.nRegistros == M)
            /* Copia os registros do novo coleto de volta para o antigo coletor */
         *coletor = novo;
      else if (novo.nRegistros == 0) {
            /* Verifica se o antigo coletor pode ser dividido.  */
            /* Se ele no pode ser dividido, o novo registro    */
            /* deve ser inserido no novo coletor que dever se  */
            /* tornar um coletor excedente. Como esta implemen- */
            /* tao no lida com esse caso, o programa ser    */
            /* abortado se o coletor no puder ser dividido.    */
         ASSEGURA( PodeDividirColetorDExt(coletor),
                   "Erro: O coletor nao pode ser dividido\n" );

            /* Insere uma referncia para o novo coletor vazio no diretrio */
         InsereColetorVazioDExt(tab, coletor, iColetor, iNovo);

            /* Escreve o coletor vazio no arquivo */
         EscreveColetorDExt(stream, iNovo, &novo);

            /* Cria um novo coletor e e o l em arquivo */
         iNovo = NovoColetorDExt(stream);
         LeColetorDExt(stream, iNovo, &novo);
      }
   }

      /* Escreve os dois coletores no arquivo */
   EscreveColetorDExt(stream, iColetor, coletor);
   EscreveColetorDExt(stream, iNovo, &novo);

      /* Insere uma referncia para o novo coletor no diretrio */
   InsereEmDiretorioDExt(tab, &novo, iNovo);
}

/********************* Funes Globais ********************/

/****
 *
 * LeColetorDExt(): L um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo
 *                        os coletores no qual ser feita a leitura
 *     pos (entrada) - posio no arquivo na qual ser feita a
 *                     leitura
 *    *coletor (sada) - ponteiro para o coletor que conter o
 *                       resultado da leitura
 *
 * Retorno: Nada
 *
 ****/
void LeColetorDExt(FILE *stream, int pos, tColetorDExt *coletor)
{
      /* Verifica a validade do stream */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, LeColetorDExt));

      /* Tenta mover o apontador de arquivo para o local */
      /* de leitura; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tColetorDExt)*pos, SEEK_SET);

      /* Efetua a leitura */
   fread(coletor, sizeof(tColetorDExt), 1, stream);

      /* Certifica-se que no houve erro de leitura */
   ASSEGURA(!ferror(stream),  "Erro de leitura em LeColetorDExt()");
}

/****
 *
 * EscreveColetorDExt(): Escreve um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo os
 *                        coletores no qual ser feita a escrita
 *     pos (entrada) - posio no arquivo no qual ser feita a
 *                     escrita
 *    *coletor (entrada) - coletor que ser escrito no arquivo
 *
 * Retorno: Nada
 *
 ****/
void EscreveColetorDExt(FILE *stream, int pos, const tColetorDExt *coletor)
{
      /* Verifica se o stream  vlido */
   ASSEGURA(stream, "Erro em EscreveColetorDExt(): stream e' NULL");

      /* Tenta mover o apontador de arquivo para o local */
      /* de escrita; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tColetorDExt)*pos, SEEK_SET);

      /* Efetua a escrita */
   fwrite(coletor, sizeof(tColetorDExt), 1, stream);

      /* Certifica-se que no houve erro */
   ASSEGURA( !ferror(stream), "Erro de escrita em EscreveColetorDExt()" );
}

/****
 *
 * IniciaTabelaDExt(): Inicia uma tabela de busca com disperso extensvel
 *
 * Parmetros:
 *      tabela (sada) - a tabela que ser iniciada
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: Nada
 *
 ****/
void IniciaTabelaDExt(tTabelaDExt *tabela, FILE *stream)
{
      /*                                                 */
      /* A tabela  iniciada com um tamanho de diretrio */
      /* igual a 1 que aponta para um coletor vazio      */
      /*                                                 */

    tabela->profGlobal = 0;
    tabela->nReg = 0;
    tabela->tamDir = 1;

    tabela->dir = malloc(tabela->tamDir*sizeof(int));

    tabela->dir[0] = NovoColetorDExt(stream);
}

/****
 *
 * BuscaDExt(): Efetua uma busca numa tabela de busca com disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *      tab (entrada) - a tabela
 *      chave (entrada) - a chave de busca
 *
 * Retorno: O ndice do registro no arquivo, se ele for encontrado;
 *          -1, em caso contrrio
 *
 * Observao: Os registros so ordenados em cada coletor, de
 *             modo que busca binria poderia ser utilizada
 *
 ****/
tRegistroMEC *BuscaDExt( FILE *stream, const tTabelaDExt *tab, tChave chave,
                         tRegistroMEC *reg )
{
   int      iColetor, i;
   tColetorDExt coletor;

      /* Obtm ndice do coletor que contm o ndice do registro */
   iColetor = tab->dir[ObtemBitsDeChave(chave, tab->profGlobal)];

      /* L o coletor que contm o ndice do registro */
   LeColetorDExt(stream, iColetor, &coletor);

      /* Efetua uma busca sequencial pelo */
      /* registro em memria principal    */
   for (i = 0; i < coletor.nRegistros; i++)
         /* Verifica se a chave foi encontrada */
      if (chave == ObtemChave(coletor.registros + i)) {
            /* Chave encontrada */
         *reg = coletor.registros[i]; /* Chave encontrada */
         return reg;
      }

   return NULL; /* A chave no foi encontrada */
}

/****
 *
 * InsereDExt(): Insere um item numa tabela de busca com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela na qual ser efetuada a insero
 *      item (entrada) - o item que ser inserido
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: Nada
 *
 * Observao: Para facilitar a implementao, o nmero mximo
 *             de registros em qualquer coletor  M - 1, de modo
 *             que sempre h espao para insero
 *
 ****/
void InsereDExt( tTabelaDExt *tab, const tRegistroMEC *item, FILE *stream )
{
   int          i, j, iColetor;
   tChave       chave;
   tColetorDExt coletor;

   chave = ObtemChave(item); /* Obtm a chave do item */

      /* Obtm o ndice do coletor */
   iColetor = tab->dir[ObtemBitsDeChave(chave, tab->profGlobal)];


   LeColetorDExt(stream, iColetor, &coletor); /* L o coletor em arquivo */

      /* O nmero de registros num coletor deve ser menor do que M */
   ASSEGURA( coletor.nRegistros < M, "Excesso de registros em coletor" );

      /* Como os registros so ordenado,  precico encontrar */
      /* a posio de insero do novo item no coletor   */
   for (i = 0; i < coletor.nRegistros; i++)
      if ( chave < ObtemChave(coletor.registros + i) )
         break;

      /* Abre espao no coletor para o novo item */
   for (j = coletor.nRegistros; j > i; j--)
      coletor.registros[j] = coletor.registros[j - 1];

   coletor.registros[i] = *item; /* Insere o novo item no coletor */

      /* O nmero de registros no coletor aumentou */
   coletor.nRegistros++;

   ++tab->nReg; /* O nmero de registros na tabela tambm aumentou */

      /* Verifica se o coletor ficou repleto */
   if (coletor.nRegistros == M)
         /* O coletor ficou repleto e  preciso dividi-lo */
      DivideColetorDExt(tab, &coletor, iColetor, stream);
   else
         /* O coletor foi alterado e  necessrio reescrev-lo no arquivo */
      EscreveColetorDExt(stream, iColetor, &coletor);
}

/****
 *
 * RemoveDExt(): Remove um item de uma tabela de disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de disperso
 *      chave (entrada) - a chave do item que ser removido
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveDExt(tTabelaDExt *tab, tChave chave, FILE *stream)
{
   int          i, j, iColetor;
   tColetorDExt coletor;

      /* Obtm o ndice do coletor */
   iColetor = tab->dir[ObtemBitsDeChave(chave, tab->profGlobal)];

   LeColetorDExt(stream, iColetor, &coletor); /* L o coletor em arquivo */

      /* Efetua uma busca sequencial pelo item em memria principal */
   for (i = 0; i < coletor.nRegistros; i++)
         /* Verifica se a chave foi encontrada */
      if (chave == ObtemChave(coletor.registros + i))
         break;

      /* Verifica se a chave foi encontrada */
   if (i >= coletor.nRegistros)
      return 0;  /* A chave no foi encontrada */

      /* O item a ser removido se encontra na posio i.  necessrio mover uma */
      /* posio para trs todos os registros que se encontram  sua frente.    */
   for (j = i; j < coletor.nRegistros - 1; ++j)
      coletor.registros[j] = coletor.registros[j + 1];

   --coletor.nRegistros; /* O nmero de registros no coletor diminuiu */

   --tab->nReg; /* O nmero de registros na tabela tambm diminuiu */

      /* Verifica se o coletor ficou vazio */
   if (coletor.nRegistros == 0)
         /* O coletor ficou vazio. Talvez o tamanho do        */
         /* diretrio possa ser reduzido, mas  possvel que, */
         /* mesmo assim, essa operao no valha a pena.      */
         /* Substituir a instruo vazia a seguir pela        */
         /* implementao de uma operao que reduza o        */
         /* tamanho do diretrio  deixada como exerccio.    */
      ;

      /* O coletor foi alterado e  preciso reescrev-lo no arquivo */
   EscreveColetorDExt(stream, iColetor, &coletor);

   return 1;  /* Remoo bem-sucedida */
}

/****
 *
 * NRegistrosDExt(): Calcula o nmero de registros numa tabela
 *                     de disperso extensvel
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de registros na tabela de disperso
 *
 ****/
int NRegistrosDExt(const tTabelaDExt *tabela, FILE *stream)
{
   int          nRegistros = 0;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      nRegistros += umColetor.nRegistros;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NRegistrosDExt" );

      /* Checa a consistncia */
   ASSEGURA( nRegistros == tabela->nReg, "Erro de consistencia" );

   return nRegistros;
}

/****
 *
 * NColetoresDExt(): Calcula o nmero de coletores numa tabela
 *                   de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores na tabela de disperso
 *
 ****/
int NColetoresDExt(FILE *stream)
{
   int          n = 0;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresDExt" );

   return n;
}

/****
 *
 * NColetoresVaziosDExt(): Calcula o nmero de coletores vazios
 *                         numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores vazios
 *
 ****/
int NColetoresVaziosDExt(FILE *stream)
{
   int      n = 0;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (!umColetor.nRegistros)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresVaziosDExt" );

   return n;
}

/****
 *
 * NColetoresRepletosDExt(): Calcula o nmero de coletores repletos
 *                           numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores repletos
 *
 ****/
int NColetoresRepletosDExt(FILE *stream)
{
   int      n = 0;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nRegistros == M)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresRepletosDExt" );

   return n;
}

/****
 *
 * NColetoresOcupadosDExt(): Calcula o nmero de coletores ocupados
 *                           numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores ocupados
 *
 ****/
int NColetoresOcupadosDExt(FILE *stream)
{
   int      n = 0;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nRegistros)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresOcupadosDExt" );

   return n;
}

/****
 *
 * ProfLocalMinDExt(): Calcula a menor profundidade local
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: A profundidade local mnima
 *
 ****/
int ProfLocalMinDExt(FILE *stream)
{
   int      min;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMinDExt" );

   min = umColetor.profLocal;

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.profLocal < min)
         min = umColetor.profLocal;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em ProfLocalMinDExt" );

   return min;
}

/****
 *
 * ProfLocalMaxDExt(): Calcula a maior profundidade local
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: A profundidade local mxima
 *
 ****/
int ProfLocalMaxDExt(FILE *stream)
{
   int          max;
   tColetorDExt umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

      /* L um coletor */
   fread(&umColetor, sizeof(umColetor), 1, stream);

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em ProfLocalMaxDExt" );

   max = umColetor.profLocal;

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.profLocal > max)
         max = umColetor.profLocal;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em ProfLocalMaxDExt" );

   return max;
}

