/****
 *
 * Ttulo do Programa: Tabela de Disperso Extensvel
 *
 * Seo: 8.2.6
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 25/08/2017
 * ltima modificao: 28/08/2017
 *
 * Descrio Geral: Este programa demonstra vrias operaes
 *                  utilizando tabela de busca com disperso
 *                  extensvel armazenada em arquivo
 *
 * Entrada: Opes escolhidas atravs de menu
 *
 * Sada: Resultados das operaes escolhidas pelo usurio
 *        (v. exemplo de execuo do programa no final
 *        deste arquivo)
 *
 * Arquivos:
 *     ASSEGURA.h - macro de tratamento de exceo usadas pelo
 *                  programa
 *     DispersaoExt.h e - interface e implementao de operaes
 *     DispersaoExt.c     operaes sobre tabelas de disperso
 *                        extensvel
 *     Registros.h e - interface e implementao de operaes
 *     Registros.c     sobre registros armazenados em arquivos
 *     main.c (este arquivo) - contm main() e outras funes
 *                             que testam as operaes sobre
 *                             tabelas de disperso extensvel
 *
 * Arquivo de dados: "MiniCensoMEC2012.bin" (para testes) ou
 *                   "CensoMEC2012.bin"
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>        /* Funes de Entrada/Sada       */
#include <string.h>       /* Funes strXXX() e memXXX()    */
#include "DispersaoExt.h" /* Operaes com tabela de busca  */
#include "Registros.h"    /* Funes de processamento       */
                          /* de registros                   */
#include "leitura.h"      /* Funes de leitura via teclado */
#include "ASSEGURA.h"     /* Macro ASSEGURA                 */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo de texto contendo os registros */
#define NOME_ARQUIVO_REGS "../CensoMEC2012.txt"

   /* Nome do arquivo de registros */
//#define NOME_ARQ_BIN "CensoMEC2012.bin"
#define NOME_ARQ_BIN "MiniCensoMEC2012.bin"

   /* Nome do arquivo contendo os registros indexados */
#define NOME_ARQ_TAB "ColetoresExt.dat"

   /* Nome do arquivo contendo o diretrio */
#define NOME_ARQ_DIR "Diretorio.dat"

   /* Nome inicial do arquivo de texto   */
   /* contendo as chaves em cada coletor */
#define NOME_ARQ_CHAVES_DEBUG "ChavesEmColetoresDebug"

   /* Nome inicial do arquivo de texto contendo o diretrio */
#define NOME_ARQ_DIR_DEBUG "DiretorioDebug"

   /* Nome do arquivo de texto contendo registros em formato legvel */
#define NOME_ARQ_DEBUG   "RegistrosDebug.txt"

/**************** Compilao Condidicional ****************/

#define DEPURACAO 0

/******************** Funes de Locais *******************/

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
static void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\n>>> Escolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t[%d] %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * ConstroiTabela(): L arquivo de dados (binrio) contendo os
 *                   registros e constri outro arquivo contendo os
 *                   coletores de uma tabela de disperso extensvel
 *
 * Parmetros:
 *     streamReg (entrada) - stream associado ao arquivo de
 *                           registros
 *     streamTab (entrada) - stream associado ao arquivo que
 *                           conter os coletores
 *
 * Retorno: Endereo de uma estrutura contendo o diretrio da
 *          tabela de disperso extensvel
 *
 ****/
tTabelaDExt *ConstroiTabela( tTabelaDExt *tabela, FILE *streamReg,
                             FILE *streamTab )
{
   tRegistroMEC    registro;
   tChaveIndice chaveIndice;
   int          indice = 0;

   IniciaTabelaExt(tabela, streamTab);

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(streamReg);

   while (1) {
         /* L um registro */
      fread(&registro, sizeof(registro), 1, streamReg);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(streamReg) || ferror(streamReg))
         break;

      chaveIndice.chave = registro.codAlunoCurso;
      chaveIndice.indice = indice++;

      InsereExt(tabela, &chaveIndice, streamTab);
   }

      /* Processamento terminado. Verifica se */
      /* ocorreu algum erro de leitura        */
   ASSEGURA( !ferror(streamReg),
             "Erro de leitura em ConstroiTabela" );

   return tabela;
}

/****
 *
 * ChecaTabela(): Verifica se a tabela de busca j existe. Se ela
 *                j existir e o usurio desejar reconstrui-la ou
 *                se ela no existir, ela ser construda.
 *
 * Parmetros:
 *     nomeArqRegistros (entrada) - nome do arquivo de registros
 *     nomeArqTabela (entrada) - nome do arquivo que armazena a
 *                               tabela
 *     *streamRegistros (entrada/sada) - stream associado ao
 *                                        arquivo de registros
 *     *streamTabela (entrada/sada) - stream associado ao
 *                                     arquivo que armazena
 *                                     a tabela
 *
 * Retorno: Nada
 *
 ****/
void ChecaTabela( tTabelaDExt *tabela,
                  const char *nomeArqRegistros,
                  const char *nomeArqTabela,
                  FILE **streamRegistros, FILE **streamTabela )
{
   int op; /* Opo do usurio */

      /* Tenta abrir o arquivo de registros */
   *streamRegistros = AbreArquivo(nomeArqRegistros, "r+b");

      /* Verifica se o arquivo que contm a tabela j existe */
   if (ArquivoExiste(nomeArqTabela)) {
      /*                                */
      /* Tabela j existe. Se o usurio */
      /* desejar, ela ser reconstruda */
      /*                                */

         /* Oferece a opo ao usurio */
      printf( "\n\n\t>>> Tabela ja existe. "
              "Deseja substitui-la (s/n)? " );
      op = LeOpcao("sSnN");

         /* Verifica se o usurio deseja reconstruir a tabela */
      if ( op == 'n' || op == 'N') { /* Ele no deseja isso   */
            /* Tenta abrir o arquivo que contm a tabela */
         *streamTabela = AbreArquivo(nomeArqTabela, "r+b");

         return; /* Servio completo */
      }
   }

   /*                                        */
   /* Neste ponto, ou a tabela ainda no     */
   /* existe ou usurio deseja reconstrui-la */
   /*                                        */

      /* Cria o arquivo e checa se abertura foi OK */
   *streamTabela = AbreArquivo(nomeArqTabela, "w+b");

      /* Constri a tabela */
   ConstroiTabela(tabela, *streamRegistros, *streamTabela);
}

/****
 *
 * GravaDiretorio(): Armazena o diretrio de uma tabela de busca
 *                   extensvel em arquivo
 *
 * Parmetros:
 *     nomeArqDiretorio (entrada) - nome do arquivo que
 *                                  armazenar o diretrio
 *     tabela (entrada) - ponteiro para a estrutura que contm o
 *                        diretrio
 *
 * Retorno: Nada
 *
 ****/
void GravaDiretorio( const char *nomeArqDiretorio,
                     const tTabelaDExt *tabela )
{
   FILE *stream = AbreArquivo(nomeArqDiretorio, "wb");

   fwrite(&tabela->profGlobal, sizeof(int), 1, stream);
   fwrite(&tabela->tamDir, sizeof(int), 1, stream);
   fwrite(&tabela->nEl, sizeof(int), 1, stream);
   fwrite(tabela->dir, sizeof(int), tabela->tamDir, stream);

   ASSEGURA( !ferror(stream),
             "Erro de escrita em arquivo de diretorio" );

   FechaArquivo(stream, nomeArqDiretorio);
}

/****
 *
 * LeDiretorio(): Armazena o diretrio de uma tabela de busca
 *                   extensvel em arquivo
 *
 * Parmetros:
 *     nomeArqDiretorio (entrada) - nome do arquivo que
 *                                  armazenar o diretrio
 *     tabela (sada) - ponteiro para a estrutura que contm o
 *                      diretrio
 *
 * Retorno: Nada
 *
 ****/
void LeDiretorio( const char *nomeArqDiretorio,
                  tTabelaDExt *tabela )
{
   FILE *stream = AbreArquivo(nomeArqDiretorio, "rb");

   fread(&tabela->profGlobal, sizeof(int), 1, stream);
   fread(&tabela->tamDir, sizeof(int), 1, stream);
   fread(&tabela->nEl, sizeof(int), 1, stream);
   fread(tabela->dir, sizeof(int), tabela->tamDir, stream);

   ASSEGURA( !ferror(stream),
             "Erro de leitura em arquivo de diretorio" );

   FechaArquivo(stream, nomeArqDiretorio);
}

/****
 *
 * InfoTabela(): Apresenta informaes gerais sobre a tabela
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *    *streamTab (entrada) - stream associado ao arquivo que
 *                           contm os coletores
 *    *streamRegs (entrada) - stream associado ao arquivo que
 *                            contm os registros
 *
 * Retorno: Nada
 *
 ****/
void InfoTabela( const tTabelaDExt* tabela, FILE *streamTab,
                 FILE *streamRegs )
{
   int       nRegs;
   tRegistroMEC registro;

      /* Verifica se os streams so vlidos */
   ASSEGURA( streamTab && streamRegs,
             "Um stream e' NULL em InfoTabela" );

   printf("\n\n\t\t>>>  Informacoes sobre coletores  <<<\n");

   printf("\n\t>>> Tamanho do bloco lido ou escrito: %d bytes", TB);
   printf("\n\t>>> Tamanho de um registro: %d bytes", TR);
   printf("\n\t>>> Tamanho de um campo do tipo int: %d bytes", TI);
   printf("\n\t>>> Numero maximo de registros num coletor: %d", M);
   printf("\n\t>>> Tamanho de um coletor: %d bytes",
          (int) sizeof(tColetor));
   printf( "\n\t>>> Espaco desperdicado em cada coletor: %d",
           (int) (TB - sizeof(tColetor)) );

   printf("\n\n\t\t>>>  Informacoes sobre a tabela  <<<\n");

      /* Calcula o nmero de registros no arquivo original */
   nRegs = NumeroDeRegistros(streamRegs, sizeof(registro));

   printf("\n\t>>> Numero de registros no arquivo original: %d",
          nRegs);

      /* Calcula o nmero de registros na tabela */
   nRegs = NRegistrosTabExt(tabela, streamTab);

   printf("\n\t>>> Numero de registros na tabela: %d", nRegs);

   printf( "\n\n\t>>> Tamanho do diretorio: %d",
           tabela->tamDir );

   printf( "\n\t>>> Profundidade global: %d",
           tabela->profGlobal );

   printf( "\n\n\t>>> Numero total de coletores: %d",
           NColetores(streamTab) );
   printf( "\n\t>>> Numero de coletores vazios: %d",
           NColetoresVazios(streamTab) );
   printf( "\n\t>>> Numero de coletores ocupados: %d",
           NColetoresOcupados(streamTab) );
   printf( "\n\t>>> Numero de coletores repletos: %d\n",
           NColetoresRepletos(streamTab) );

   printf( "\n\t>>> Profundidade local minima: %d",
           ProfLocalMin(streamTab) );
   printf( "\n\t>>> Profundidade local maxima: %d\n",
           ProfLocalMax(streamTab) );
}

/****
 *
 * CriaArquivoRegistros(): Verifica se o arquivo de registros j
 *                         existe. Se ela j existir e o usurio
 *                         desejar reconstrui-lo ou se ele no
 *                         existir, ele ser construdo.
 *
 * Parmetros:
 *     nomeArqTexto (entrada) - nome do arquivo texto de registros
 *     nomeArqBin (entrada) - nome do arquivo binrio de registros
 *
 * Retorno: Stream associado ao arquivo binrio de registros
 *
 ****/
FILE *CriaArquivoRegistros( const char *nomeArqTexto,
                            const char *nomeArqBin )
{
   int   op, /* Opo do usurio */
         nRegistros, /* Nmero de registros no arquivo */
         maxRegistros; /* Nmero mximo de registros no arquivo */

      /* Verifica se o arquivo binrio de registros j existe */
   if (ArquivoExiste(nomeArqBin)) {
      /*                                */
      /* Arquivo j existe. Se o usurio */
      /* desejar, ele ser reconstrudo */
      /*                                */

         /* Oferece a opo ao usurio */
      op = LeSimNao( "Arquivo de registros ja' existe. "
                      "Deseja substitui-lo" );

         /* Verifica se o usurio deseja reconstruir o arquivo */
      if (op == 'n') /* Ele no deseja isso   */
         return AbreArquivo(nomeArqBin, "r+b");
   }

   /*                                        */
   /* Neste ponto, ou o arquivo ainda no    */
   /* existe ou usurio deseja reconstrui-lo */
   /*                                        */

   printf("\n\t>>> Digite o numero maximo de registros (0 = TODOS): ");
   maxRegistros = LeNatural();

   printf("\nCriando arquivo...\n");
   fflush(stdout);

      /* Cria o arquivo binrio */
   nRegistros = CriaArquivoBin( nomeArqTexto, nomeArqBin,
                                maxRegistros );

      /* Apresenta o resultado da converso */
   printf( "\n\t>>> Foram armazenados %d registros no arquivo"
           "\n\t>>> \"%s\"\n", nRegistros, nomeArqBin );

      /* Abre e retorna stream assoaciado ao arquivo binrio */
   return AbreArquivo(nomeArqBin, "r+b");
}

/****
 *
 * EscreveChavesColetor(): Escreve num stream as chaves de um coletor
 *
 * Parmetros:
 *     *coletor (entrada) - o coletor
 *      stream (entrada) - o stream no qual ocorrer a escrita
 *
 * Retorno: Nada
 *
 ****/
void EscreveChavesColetor(const tColetor *coletor, FILE *stream)
{
   int i;

   fprintf(stream, "\nChaves = {%ld", coletor->elementos[0].chave);

   for (i = 1; i < coletor->nElementos; ++i)
      fprintf(stream, ", %ld", coletor->elementos[i].chave);

   fprintf(stream, "}");
}

/****
 *
 * ExibeColetor(): Escreve o contedo de um coletor num stream
 *
 * Parmetros:
 *      *coletor (entrada) - o coletor
 *      stream (entrada) - o stream no qual ocorrer a escrita
 *
 * Retorno: Nada
 *
 ****/
void ExibeColetor( const tColetor *coletor, FILE *stream )
{
   fprintf(stream, "\nNumero de chaves: %d", coletor->nElementos);

   EscreveChavesColetor(coletor, stream);
}

/****
 *
 * CriaArquivoDeChavesEmColetores(): Apresenta as chaves em cada
 *                                   coletor de uma tabela de
 *                                   disperso esttica arquivo
 *                                   de texto
 *
 * Parmetros:
 *      streamTab (entrada) - stream associado ao arquivo que
 *                            contm a tabela
 *      nomeIni (entrada) - nome inicial do arquivo que conter as
 *                          informaes
 *
 * Retorno: Nada
 *
 ****/
void CriaArquivoDeChavesEmColetores( FILE *streamTab,
                                     const char *nomeIni )
{
   static int ordem = 0;
   int        i = 0,
              nColetores = 0,
              nChaves = 0;
   tColetor   coletor;
   char      *nomeArq;
   FILE      *streamCol; /* Stream de texto que contm o resultado */

      /* O limite de arquivos criados por seo  99 */
   if (++ordem > 99) {
      printf("\n\t\a>>> Ja' criei arquivos demais <<<\n");
      return;
   }

      /* Aloca espao para o nome do arquivo de ns: */
      /* 1 (terminal) + 2 (ordem) + 4 (".txt") = 7   */
   nomeArq = malloc(strlen(nomeIni) + 7);
   ASSEGURA( nomeArq,
             "Nao foi possivel criar nome de arquivo de coletores" );

      /* Cria o nome completo do arquivo */
   sprintf(nomeArq, "%s%.2d%s", nomeIni, ordem, ".txt");

      /* Tenta criar o aquivo */
   ASSEGURA( streamCol = fopen(nomeArq, "w"),
             "Nao foi possivel criar arquivo de coletores" );

      /* A leitura deve comear no incio   */
      /* do arquivo que contm os coletores */
   rewind(streamTab);

      /* L cada coletor no arquivo que os contm e escreve */
      /* as informaes sobre eles num arquivo de texto     */
   while (1) {
      fread(&coletor, sizeof(coletor), 1, streamTab);

      if (feof(streamTab) || ferror(streamTab))
         break;

      fprintf(streamCol, "%d.\n", i++);

      ExibeColetor(&coletor, streamCol);

      nChaves += coletor.nElementos;

      ++nColetores;

      fprintf(streamCol, "\n\n----------\n\n");
   }

   fprintf( streamCol, ">>> Numero de coletores: %d", nColetores );
   fprintf( streamCol, "\n>>> Max chaves em coletor (M): %d",
            (int) M );

   fprintf( streamCol, "\n>>> Numero total de chaves: %d\n\n",
            nChaves );

   fclose(streamCol);

   printf("\n>> Resultado escrito no arquivo: \"%s\"\n", nomeArq);

   free(nomeArq);
}

/****
 *
 * CriaArquivoDeDiretorio(): Apresenta o diretrio de uma tabela
 *                           de disperso extensvel num arquivo
 *                           de texto
 *
 * Parmetros:
 *      tabela (entrada) - stream associado ao arquivo que
 *                            contm a tabela
 *      nomeIni (entrada) - nome inicial do arquivo que conter as
 *                          informaes
 *
 * Retorno: Nada
 *
 ****/
void CriaArquivoDeDiretorio(tTabelaDExt *tab, const char *nomeIni)
{
   static int ordem = 0;
   int        i;
   char      *nomeArq;
   FILE      *streamCol; /* Stream de texto que contm o resultado */

      /* O limite de arquivos criados por seo  99 */
   if (++ordem > 99) {
      printf("\n\t\a>>> Ja' criei arquivos demais <<<\n");
      return;
   }

      /* Aloca espao para o nome do arquivo de ns: */
      /* 1 (terminal) + 2 (ordem) + 4 (".txt") = 7   */
   nomeArq = malloc(strlen(nomeIni) + 7);
   ASSEGURA( nomeArq,
             "Nao foi possivel criar nome de arquivo de coletores" );

      /* Cria o nome completo do arquivo */
   sprintf(nomeArq, "%s%.2d%s", nomeIni, ordem, ".txt");

      /* Tenta criar o aquivo */
   ASSEGURA( streamCol = fopen(nomeArq, "w"),
             "Nao foi possivel criar arquivo de coletores" );

   fprintf( streamCol, ">>> Profundidade global: %d",
            tab->profGlobal );
   fprintf( streamCol, "\n>>> Tamanho do diretrio: %d",
            tab->tamDir );
   fprintf( streamCol, "\n>>> Nmero de registros: %d\n",
            tab->nEl );
   fprintf(streamCol, "\n--------------------------------\n");

   fprintf(streamCol, "\n    *** Diretorio ***\n");

      /* Escreve os elementos do diretrio */
   for (i = 0; i < tab->tamDir; ++i) {
      fprintf( streamCol, "\t%d ---> %d\n", i, tab->dir[i] );
   }

   fclose(streamCol);

   printf("\n>> Resultado escrito no arquivo: \"%s\"\n", nomeArq);

   free(nomeArq);
}

/****
 *
 * EscreveRegistrosDebug(): Escreve o contedo de um arquivo
 *                          binrio de registros num arquivo de
 *                          texto de modo legvel
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 ****/
void EscreveRegistrosDebug(FILE *streamReg, const char *nomeArqTxt)
{
   int     i = 0, /* ndice do registro no arquivo */
           ocorreuErro = 0;
   tRegistroMEC  reg;
   FILE   *streamTxt;

   streamTxt = AbreArquivo(nomeArqTxt, "w");

   MoveApontador(streamReg, 0L, SEEK_SET);

   while (1) {
      fread(&reg, sizeof(reg), 1, streamReg);

      if (feof(streamReg) || ferror(streamReg))
         break;

      fprintf(streamTxt, "%d.\n\n", i++);

      if (ferror(streamTxt))
         break;

      EscreveRegistro(&reg, streamTxt);

      fprintf(streamTxt, "\n\n----------\n\n");

      if (ferror(streamTxt))
         break;
   }

   ocorreuErro = ferror(streamReg) || ferror(streamTxt);

   FechaArquivo(streamTxt, nomeArqTxt);

   if (ocorreuErro) {
      fclose(streamReg);
      printf("\n\t\a>>> Ocorreu erro de leitura ou escrita\n");
      exit(1);
   }
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa operaes com tabela de disperso extensvel
 *
 * Parmetros: nenhum
 *
 * Retorno: 0
 *
 ****/
int main(void)
{
   tRegistroMEC   umRegistro;
   const char *opcoes[] = { "Consulta dados de um aluno",    /* 1 */
                            "Insere um aluno",               /* 2 */
                            "Remove um aluno",               /* 3 */
                            "Escreve coletores (depuracao)", /* 4 */
                            "Escreve diretorio (depuracao)", /* 5 */
                            "Escreve registros (depuracao)", /* 6 */
                            "Encerra o programa"             /* 7 */
                          };
   int          op,  /* Opo do usurio  */
                nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]),
                sizeofLong = (int) sizeof(long),
                nRegs = 0, /* Nmero de registros no arquivo */
                indice;
   tChave       umaChave;  /* Uma chave de busca */
   FILE        *streamReg, /* Stream associado ao  */
                           /* arquivo de registros */
               *streamTab; /* Stream associado ao       */
                           /* arquivo contendo a tabela */
   tTabelaDExt  tabela;
   tChaveIndice chaveIndice;

   printf("\n\t>>> sizeof(long) = %d", sizeofLong);

   if (sizeofLong >= 8) {
      printf("... OK\n");
   } else {
      printf("\n\t>>> Este programa pode nao funcionar adequadamente "
             "\n\t>>> quando sizeof(long) e' menor do que 8");
      op = LeSimNao("Deseja prosseguir");

      if (op == 'n') {
         printf("\n\t>>> Adquira um compilador decente e recompile "
                "\n\t>>> o programa. Bye.");
         return 1;
      }
   }

      /* Abre ou cria o arquivo de registros */
   streamReg = CriaArquivoRegistros(NOME_ARQUIVO_REGS, NOME_ARQ_BIN);

   MedidaDeTempo("Criando tabela...");

      /* Inicia a tabela B e obtm a posio da */
      /* raiz da tabela no arquivo que a contm */
//   ChecaTabela( NOME_ARQ_BIN, NOME_ARQ_TAB,
//                &streamReg, &streamTab );

         /* Tenta abrir o arquivo de registros */
   streamReg = AbreArquivo(NOME_ARQ_BIN, "r+b");

      /* Cria o arquivo e checa se abertura foi OK */
   streamTab = AbreArquivo(NOME_ARQ_TAB, "w+b");

      /* Constri a tabela */
   ConstroiTabela(&tabela, streamReg, streamTab);

   MedidaDeTempo("Tabela criada com sucesso");

   InfoTabela(&tabela, streamTab, streamReg);

      /* Calcula o nmero de registros */
   nRegs = NumeroDeRegistros(streamReg, sizeof(umRegistro));

   while (1) { /* Lao principal do programa */
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("1234567");

         /* Verifica se o usurio deseja sair do programa */
      if (op == '7')
         break; /* Encerra o lao */

      if (op <= '3') {
         printf("\n\t>>> Digite o codigo aluno/curso: ");
         umaChave = LeNatural();
      }

      switch (op) {
         case '1': /* Procura o registro que */
                   /* possui uma dada chave  */

               /* Tenta encontrar o registro  */
               /* que contm a chave de busca */
            indice = BuscaExt( streamTab, &tabela, umaChave );

               /* Verifica se a chave foi encontrada */
            if (indice < 0)
               printf("\n\t>>> A chave nao foi encontrada\n");
            else {
               LeRegistroEmArquivo(&umRegistro, streamReg, indice);
               printf( "\n\n\t>>> Registro com a chave %ld <<<\n\n",
                       umaChave );
               ExibeRegistro(&umRegistro);
               putchar('\n'); /* Embelezamento */
               putchar('\n');
            }

            break;

         case '2': /* Insere um novo registro, se a */
                   /* chave dada ainda no existe   */
               /* Recebe do usurio os dados do  */
               /* registro e insere-o no arquivo */
            chaveIndice.chave = umaChave;
            chaveIndice.indice = nRegs++;
            InsereRegistroEmArquivo(umaChave, streamReg);

            InsereExt(&tabela, &chaveIndice, streamTab);

            printf("\n\t>>> Insercao bem-sucedida");

            break;

         case '3': /* Remove um registro */
            if (RemoveExt(&tabela, umaChave, streamTab))
               printf( "\n\t>>> Remocao bem sucedida\n");
            else
               printf( "\n\t>>> A remocao MAL sucedida\n");

            break;

         case '4': /* Apresenta as chaves em cada coletor */
                   /* num arquivo de texto (depurao)    */
            CriaArquivoDeChavesEmColetores( streamTab,
                                            NOME_ARQ_CHAVES_DEBUG );

            break;

         case '5': /* Apresenta o diretrio num    */
                   /* arquivo de texto (depurao) */
            CriaArquivoDeDiretorio(&tabela, NOME_ARQ_DIR_DEBUG);

            break;

         case '6': /* Apresenta os registros num arquivo */
                   /* de texto para depurao             */
            EscreveRegistrosDebug(streamReg, NOME_ARQ_DEBUG);
               printf( "\n\t>>> Os registros foram escritos no "
                       "arquivo: \"%s\"\n", NOME_ARQ_DEBUG );

            break;
      } /* switch */
   } /* while */

      /* Fecha o arquivo que contm os coletores */
   FechaArquivo(streamTab, NOME_ARQ_TAB);

      /* Fecha o arquivo que contm os registros */
   FechaArquivo(streamReg, NOME_ARQ_BIN);

      /* Armazena o diretrio em arquivo */
   GravaDiretorio(NOME_ARQ_DIR, &tabela);

   printf("\nObrigado por usar este programa. Bye.\n\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

        >>> sizeof(long) = 8... OK

        >>> Digite o numero maximo de registros (0 = TODOS): 0

Criando arquivo...

        >>> Foram armazenados 9565483 registros no arquivo
        >>> "MiniCensoMEC2012.bin"

Criando tabela...

        >>> Tabela criada com sucesso

        *** Tempo gasto na operacao: 250.00 segundos ***


                >>>  Informacoes sobre coletores  <<<

        >>> Tamanho do bloco lido ou escrito: 4096 bytes
        >>> Tamanho de um registro: 16 bytes
        >>> Tamanho de um campo do tipo int: 4 bytes
        >>> Numero maximo de registros num coletor: 255
        >>> Tamanho de um coletor: 4088 bytes
        >>> Espaco desperdicado em cada coletor: 8 bytes

                >>>  Informacoes sobre a tabela  <<<

        >>> Numero de registros no arquivo original: 9565483
        >>> Numero de registros na tabela: 9565483

        >>> Tamanho do diretorio: 65536
        >>> Profundidade global: 16

        >>> Numero total de coletores: 65536
        >>> Numero de coletores vazios: 0
        >>> Numero de coletores ocupados: 65536
        >>> Numero de coletores repletos: 0

        >>> Profundidade local minima: 16
        >>> Profundidade local maxima: 16

>>> Escolha uma das opcoes a seguir:

        [1] Consulta dados de um aluno
        [2] Insere um aluno
        [3] Remove um aluno
        [4] Encerra o programa

        >>>


**********************************************************/

