/****
 *
 * Exemplo: Casamento de Strings Usando o Algoritmo
 *          de Boyer e Moore (BM)
 *
 * Seo: 9.4.5
 *
 ****/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/****************** Definies de Macros ******************/

#define TAM_ALFABETO 256

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

#define MAX(a, b) (((a) >= (b)) ? (a) : (b))

#define DEBUG 1

/***************** Definies de Funes ******************/

/****
 *
 * ExibeTabelaCar(): Apresenta o contedo de um array de
 *                   ints indexados por caracteres
 *
 * Parmetros:
 *      ar (entrada) - o array
 *      tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada apenas para depurao
 *
 ****/
void ExibeTabelaCar(const int ar[], int tam)
{
   int i;

   for (i = 0; i < tam; ++i) {
      if (ar[i] >= 0)
         printf("\ttab[%c] = %d\n", i, ar[i]);
   }

   printf("\ttab[%c] = %d\n", '*', -1);
}

/****
 *
 * ExibeTabela(): Apresenta o contedo de um array de ints
 *
 * Parmetros:
 *      ar (entrada) - o array
 *      tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada apenas para depurao
 *
 ****/
void ExibeTabela(const int ar[], int tam)
{
   int i;

   for (i = 0; i < tam; ++i)
      printf("\ttab[%d] = %d\n", i, ar[i]);
}

/****
 *
 * TabelaMauCaractereBM(): Atribui valores para os elementos da
 *                       tabela de maus caracteres usada pelo
 *                       algoritmo BM
 *
 * Parmetros:
 *      p (entrada) - string que representa o padro
 *      m (entrada) - tamanho do string que representa o padro
 *
 * Retorno: Endereo da tabela
 *
 ****/
int *TabelaMauCaractereBM(const char *p, int m)
{
   int *tab,
        caractere,
        i;

      /* O nmero de elementos do array que representa a tabela de */
      /* maus caracteres  igual ao tamanho do alfabeto utilizado  */
   tab = calloc(TAM_ALFABETO, sizeof(int));

   ASSEGURA(tab, "Impossivel alocar tabela de maus caracteres");

      /* Inicialmente, supe-se que nenhum caractere do     */
      /* alfabeto faz parte do padro e portanto o elemento */
      /* correspondente a cada caractere  iniciado com -1  */
   for (i = 0; i < TAM_ALFABETO; ++i)
      tab[i] = -1;

      /* Atribui valores aos elementos correspondentes */
      /* aos caracteres que fazem parte do padro      */
   for (i = 0; i < m; ++i) {
         /* Obtm o caractere que est na posio i do padro */
      caractere = (int) p[i];

         /* O valor da tabela correspondente ao */
         /* caractere  a sua posio no padro */
      tab[caractere] = i;
   }

   return tab;
}

/****
 *
 * TabelaSufixosBM(): Cria a tabela de bons sufixos do algoritmo BM
 *
 * Parmetros:
 *      p (entrada) - string que representa o padro
 *      m (entrada) - tamanho do string que representa o padro
 *
 * Retorno: Endereo da tabela
 *
 ****/
int *TabelaSufixosBM(const char *p, int m)
{
   int  i = m,
        j = m + 1,
       *bons, /* Tabela de bons sufixos */
       *aux;  /* Tabela auxiliar */

      /* Aloca espao para os arrays bons[] */
      /* e aux[] e testa suas alocaes     */
   bons = calloc(m + 1, sizeof(int));
   ASSEGURA(bons, "Impossivel alocar array bons[]");
   aux = calloc(m + 1, sizeof(int));
   ASSEGURA(aux, "Impossivel alocar array aux[]");

               /**********/
               /* Caso 1 */
               /**********/

      /* O sufixo que inicia na posio m no possui   */
      /* borda. Portanto, aux[m] recebe o valor m + 1. */
   aux[m] = m + 1;

   while (i > 0) {
      while (j <= m && p[i - 1] != p[j - 1]) {
            /* O salto correspondente  posio j   */
            /* armazenado em s[j], desde que o valor */
            /* de s[j] seja diferente de 0           */
         if (bons[j] == 0)
            bons[j] = j - i;

         j = aux[j]; /* Passa para a prxima borda do padro */
      }

         /* Passa para o prximo sufixo */
      --i;
      --j;

         /* Cada elemento do array aux[] contm a    */
         /* posio inicial da maior borda do sufixo */
         /* do padro que inicia na posio i        */
      aux[i] = j;
   }

               /**********/
               /* Caso 2 */
               /**********/

      /* Armazena em j a posio inicial da maior    */
      /* borda do padro, que  armazenada em aux[0] */
   j = aux[0];

      /* Armazena o valor de aux[0] em cada elemento de s[]  */
      /* que  igual a 0. Quando o sufixo do padro se torna */
      /* menor do que aux[0], mantm-se a prxima borda mais */
      /* larga do padro (i.e., aux[j]).                     */
   for (i = 0; i <= m; i++) {
      if (bons[i] == 0)
         bons[i] = j;

      if (i == j)
         j = aux[j];
   }

   free(aux); /* A tabela auxiliar no  mais necessria */

   return bons;
}

/****
 *
 * CasamentoBM(): Verifica se um padro ocorre num texto
 *                usando o algoritmo BM
 *
 * Parmetros:
 *      t (entrada) - string que representa o texto
 *      p (entrada) - string que representa o padro
 *
 * Retorno: A posio do padro no texto, se ele for
 *          encontrado; -1, em caso contrrio.
 *
 ****/
int CasamentoBM(const char *t, const char *p)
{
   int i = 0,
       j,
       caractere,
       n = strlen(t), /* Tamanho do texto  */
       m = strlen(p), /* Tamanho do padro */
      *maus, /* Tabela de maus caracteres */
      *bons; /* Tabela de bons sufixos    */

      /* Obtm a tabela de maus caracteres */
   maus = TabelaMauCaractereBM(p, m);

      /* Obtm a tabela de bons sufixos */
   bons = TabelaSufixosBM(p, m);

      /* Compara caracteres do padro com caracteres do  */
      /* texto da direita para a esquerda. Quando ocorre */
      /* uma discordncia entre caracteres, o padro    */
      /* deslocado pelo maior valor dentre aqueles       */
      /* encontrados nos arrays bons[] e maus[].         */
   while (i <= n - m) {
      j = m - 1;

      while (j >= 0 && p[j] == t[i + j])
         j--;

      if (j < 0) {
         break; /* O padro foi encontrado no texto */
      } else {
         caractere = (int) t[i + j];
         i += MAX(bons[j + 1], j - maus[caractere]);
      }
   }

   free(bons); /* O array bons[] no  mais necessrio */
   free(maus); /* O array maus[] no  mais necessrio */

      /* Verifica se o padro foi encontrado no texto */
   if (i <= n - m)
      return i; /* O padro foi encontrado */
   else
      return -1; /* O padro no foi encontrado */
}

/****
 *
 * main(): Testa a funo CasamentoBM()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   char *texto = "abacaabaccabacabaabb";
//   char *texto = "abacabccabaababbca";
   char *padrao = "abacab";
//   char *padrao = "abab";
   int   posicao;

   printf("\n\t>>>> Usando Boyer e Moore <<<\n");

   posicao = CasamentoBM(texto, padrao);

   if (posicao >= 0) {
      printf( "\n\"%s\" encontrado em \"%s\" na posicao %d\n",
              padrao, texto, posicao );
   } else {
      printf( "\n\"%s\" nao foi encontrado em \"%s\"\n",
              padrao, texto );
   }

   return 0;
}

/****************** Resultado do Programa *****************

        >>>> Usando Boyer e Moore <<<

        >>>> Tabela de Maus Caracteres <<<

        tab[a] = 4
        tab[b] = 5
        tab[c] = 3
        tab[*] = -1

        >>>> Tabela Auxiliar <<<

        tab[0] = 4
        tab[1] = 5
        tab[2] = 6
        tab[3] = 6
        tab[4] = 6
        tab[5] = 6
        tab[6] = 7

        >>>> Tabela de Sufixos (Caso 1) <<<

        tab[0] = 0
        tab[1] = 0
        tab[2] = 0
        tab[3] = 0
        tab[4] = 0
        tab[5] = 0
        tab[6] = 1

        >>>> Tabela de Sufixos (Caso 2) <<<

        tab[0] = 4
        tab[1] = 4
        tab[2] = 4
        tab[3] = 4
        tab[4] = 4
        tab[5] = 6
        tab[6] = 1

"abacab" encontrado em "abacaabaccabacabaabb" na posicao 10

**********************************************************/
