/****
 *
 * Exemplo: Casamento de Strings Usando o Algoritmo
 *          de Rabin e Karp (RK)
 *
 * Seo: 9.6.3
 *
 ****/

#include <stdio.h>
#include <string.h>
#include <limits.h>

/****
 *
 * DigitalRabin(): Calcula de modo correto e eficiente o valor de
 *                 disperso de um string (verso junior)
 *
 * Parmetros:
 *      s (entrada) - o string
 *      b (entrada) - a base numrica utilizada
 *      q (entrada) - o nmero primo utilizado
 *
 * Retorno: O valor de disperso almejado
 *
 ****/
int DigitalRabin(const char *s, int b, int q)
{
   int i, grau,
       valor = 0;

   grau = strlen(s) - 1; /* Grau do polinmio */

   for (i = 0; i <= grau; ++i)
      valor = (valor*b + s[i])%q;

   return valor%q; /* Retorna o valor de disperso */
}

/****
 *
 * DigitalRabin2(): Calcula de modo correto e eficiente o valor de
 *                  disperso de um string (verso senior)
 *
 * Parmetros:
 *      s (entrada) - o string
 *      b (entrada) - a base numrica utilizada
 *      q (entrada) - o nmero primo utilizado
 *
 * Retorno: O valor de disperso almejado
 *
 ****/
int DigitalRabin2(const char *s, int b, int q)
{
   int valor = 0;

   while (*s)
      valor = (valor*b + *s++)%q;

   return valor%q;  /* Retorna o valor de disperso */
}

/****
 *
 * DigitalRabinErrada(): Calcula de modo incorreto e ineficiente
 *                       o valor de disperso de um string
 *
 * Parmetros:
 *      s (entrada) - o string
 *      b (entrada) - a base numrica utilizada
 *      q (entrada) - o nmero primo utilizado
 *
 * Retorno: O valor de disperso almejado
 *
 * Observao: Esta funo est propositalmente
 *             implementada incorretamente
 *
 ****/
int DigitalRabinErrada(const char *s, int b, int q)
{
   int i,   /* ndice de um caractere do string */
       grau, /* Grau do polinmio */
       potencia = 1, /* Base elevada a uma potncia */
       valor = 0; /* Valor de disperso a ser retornado */

   grau = strlen(s) - 1; /* Calcula o grau do polinmio */

      /* Obtm o valor do polinmio */
   for (i = grau; i >= 0; --i) {
      valor = valor + s[i]*potencia;
      potencia *= b;
   }

   return valor%q; /* Retorna o valor de disperso */
}

/****
 *
 * main(): Testa funes de disperso para o algoritmo KR
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int   dispersao,
         base = 128,
//         primo = 1009;
         primo = 33554393;
   char *s1 = "bola",
        *s2 = "loba",
        *s3 = "anticonstitucionalissimamente";

   printf("\n>>> Digital de Rabin 1 <<<\n");
   printf( ">>> Base = %d, Primo = %d, INT_MAX = %d <<<\n",
           base, primo, INT_MAX );

   dispersao = DigitalRabin(s1, base, primo);
   printf( "\n>>> Valor de dispersao de \"%s\": %d\n", s1,
           dispersao );

   dispersao = DigitalRabin(s2, base, primo);
   printf( ">>> Valor de dispersao de \"%s\": %d\n", s2,
           dispersao );

   dispersao = DigitalRabin(s3, base, primo);
   printf( ">>> Valor de dispersao de \"%s\": %d\n", s3,
           dispersao );

   printf("\n\n>>> Digital de Rabin 2 <<<\n");
   printf( ">>> Base = %d, Primo = %d, INT_MAX = %d <<<\n",
           base, primo, INT_MAX );

   dispersao = DigitalRabin2(s1, base, primo);
   printf( "\n>>> Valor de dispersao de \"%s\": %d\n", s1,
           dispersao );

   dispersao = DigitalRabin2(s2, base, primo);
   printf( ">>> Valor de dispersao de \"%s\": %d\n", s2,
           dispersao );

   dispersao = DigitalRabin2(s3, base, primo);
   printf( ">>> Valor de dispersao de \"%s\": %d\n", s3,
           dispersao );

   printf("\n\n>>> Digital de Rabin 3 (Errada!) <<<\n");
   printf( ">>> Base = %d, Primo = %d, INT_MAX = %d <<<\n",
           base, primo, INT_MAX );

   dispersao = DigitalRabinErrada(s1, base, primo);
   printf( "\n>>> Valor de dispersao de \"%s\": %d\n", s1,
           dispersao );

   dispersao = DigitalRabinErrada(s2, base, primo);
   printf( ">>> Valor de dispersao de \"%s\": %d\n", s2,
           dispersao );

   dispersao = DigitalRabinErrada(s3, base, primo);
   printf( ">>> Valor de dispersao de \"%s\": %d\n", s3,
           dispersao );

   return 0;
}

/****************** Resultado do Programa *****************

>>> Digital de Rabin 1 <<<
>>> Base = 128, Primo = 33554393, INT_MAX = 2147483647 <<<

>>> Valor de dispersao de "bola": 6027083
>>> Valor de dispersao de "loba": 26997323
>>> Valor de dispersao de "anticonstitucionalissimamente": 32133178


>>> Digital de Rabin 2 <<<
>>> Base = 128, Primo = 33554393, INT_MAX = 2147483647 <<<

>>> Valor de dispersao de "bola": 6027083
>>> Valor de dispersao de "loba": 26997323
>>> Valor de dispersao de "anticonstitucionalissimamente": 32133178


>>> Digital de Rabin 3 (Errada!) <<<
>>> Base = 128, Primo = 33554393, INT_MAX = 2147483647 <<<

>>> Valor de dispersao de "bola": 6027083
>>> Valor de dispersao de "loba": 26997323
>>> Valor de dispersao de "anticonstitucionalissimamente": -21252146

**********************************************************/
