/****
 *
 * Exemplo: Casamento de Strings Usando o Algoritmo
 *          de Rabin e Karp (RK)
 *
 * Seo: 9.6.3
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/************************ Macros **************************/

//#define PRIMO 33554393
#define PRIMO 32887
#define BASE  256

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

/***************** Definies de Funes ******************/

/****
 *
 * CasamentoKR(): Verifica se um padro ocorre num texto
 *                usando o algoritmo de Karp e Rabin (KR)
 *
 * Parmetros:
 *      t (entrada) - string que representa o texto
 *      p (entrada) - string que representa o padro
 *      b (entrada) - a base numrica utilizada
 *      q (entrada) - o nmero primo utilizado
 *
 * Retorno: A posio do padro no texto, se ele for
 *          encontrado; -1, em caso contrrio.
 *
 ****/
int CasamentoKR(const char *t, const char *p, int b, int q)
{
   int dispersaoPadrao = 0, /* Valor de disperso do padro */
       dispersaoTexto = 0, /* Valor de disperso de cada janela */
       maiorPotencia = 1,   /* Valor da base elevado a m - 1 */
       i, /* ndice da primeira ocorrncia */
          /* do padro no texto encontrada */
       m = strlen(p), /* Tamanho do padro */
       n = strlen(t); /* Tamanho do texto  */

      /* Se o tamanho do texto for menor do que o tamanho */
      /* do padro, o padro jamais ser encontrado       */
   if (n < m)
      return -1;

      /* Calcula a maior potncia do polinmio */
   for (i = 0; i < m - 1; ++i)
      maiorPotencia = maiorPotencia*b%q;

      /* Calcula o valor de disperso do padro */
      /* e da primeira janela do texto          */
   for (i = 0; i < m; ++i) {
      dispersaoPadrao = (dispersaoPadrao*b + p[i])%q;
      dispersaoTexto = (dispersaoTexto*b + t[i])%q;
   }

      /* Efetua a busca pelo padro no texto */
   for (i = 0; i < n; ++i) {
         /* Se os valores de disperso do padro e da */
         /* janela coincidem, verifica se os strings  */
         /* correspondentes realmente so iguais      */
      if (dispersaoPadrao == dispersaoTexto && !memcmp(p, t+i, m))
         return i; /* Padro encontrado no texto */

         /* Calcula o valor de disperso da prxima janela */
      dispersaoTexto = ( b*(dispersaoTexto - t[i]*maiorPotencia) +
                         t[i + m] )%q;

         /* Corrige o valor da disperso se ele ficou negativo */
      if (dispersaoTexto < 0)
         dispersaoTexto += q;
   }

      /* Se ainda no houve retorno, o padro no foi encontrado */
   return -1;
}

/****
 *
 * EhPrimo(): Verifica se um nmero  primo
 *
 * Parmetros:
 *      n (entrada) - o nmero a ser testado
 *
 * Retorno: 1, se o nmero for primo; 0, em caso contrrio.
 *
 ****/
int EhPrimo(int n)
{
   int i;

   if (n < 2) /* Por definio, no existe primo menor do que 2*/
      return 0;

   for (i = 2; i <= n/2; ++i)
         /* Se for encontrado um divisor, o nmero no  primo */
      if (!(n%i))
         return 0;

   return 1; /* O nmero  primo */
}

/****
 *
 * MaiorPrimo(): Encontra o maior nmero primo menor do que ou
 *               igual a um dado nmero natural
 *
 * Parmetros:
 *      max (entrada) - valor mximo do nmero primo a ser
 *                      encontrado
 *
 * Retorno: 1, se o nmero for primo; 0, em caso contrrio.
 *
 ****/
int MaiorPrimo(int max)
{
   int i;

      /* Encontra o maior nmero primo menor */
      /* do que ou igual ao parmetro        */
   for (i = max; i > 2; --i)
      if (EhPrimo(i))
         return i;

   return 2; /* 2  o menor nmero que existe */
}

/****
 *
 * main(): Testa a funo CasamentoKR()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   char *texto = "caabaccabacabaabbabacaabaccabacabaabbabaca"
                 "abaccacbbaaabbbcbbccabaaaabbcbcbbc";
   char *padrao = "abacaabaccabacabaabbabacaabaccacbba";
   int   posicao,
         b = BASE,
         q = PRIMO,
         qMax = INT_MAX/(BASE*BASE - BASE);

      /* Teste preventivo de overflow */
   if (q > qMax) {
      qMax = MaiorPrimo(qMax);
      printf("\n\t\a>>> Ocorrera' overflow!\n"
             "\n\t>>> O maior valor de q deveria ser %d\n", qMax);
      return 1;
   }

   printf("\n>>>> Usando Algoritmo KR <<<\n");

   posicao = CasamentoKR(texto, padrao, b, q);

   if (posicao >= 0)
      printf( "\n>>> \"%s\" encontrado em \"%s\" na posicao %d\n",
              padrao, texto, posicao );
   else
      printf( "\n>>> \"%s\" nao foi encontrado em \"%s\"\n",
              padrao, texto );

   return 0;
}

/****************** Resultado do Programa *****************

>>>> Usando Algoritmo KR <<<

>>> b = 32      q = 65075243    bq = 2082407776 <<<

>>> "abaca" encontrado em "caabaccabacabaabbabacaabaccabacabaabbabacaabacc" na posicao 7

**********************************************************/
