/****
 *
 * Exemplo: rvore Trie Bsica
 *
 * Seo: 6.4.1
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "Trie.h"     /* Interface deste mdulo */
#include "ASSEGURA.h" /* Macro ASSEGURA         */

/****************** Definies de Macros ******************/

   /* Obtm o ndice de um caractere num array */
   /* que representa os caracteres do alfabeto */
#define INDICE_CARACTERE(c) ((c) - 'a')

   /* Libera um n e torna nulo o   */
   /* ponteiro usado para liber-lo */
#define FREE(p) do {free(p); p = NULL;} while(0)

/********************* Funes Locais *********************/

/****
 *
 * NovoNoTrieMachado(): Cria um n de uma trie
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo do n criado, se houve alocao; NULL, em caso contrrio
 *
 ****/
static tNoTrieMachado *NovoNoTrieMachado(void)
{
   tNoTrieMachado *pNo;
   int      i;

   pNo = malloc(sizeof(*pNo));

   if (pNo) {
         /* Por enquanto, o n no  final */
      IniciaListaSE(&pNo->valores);

         /* Inicia cada filho do n com NULL */
      for (i = 0; i < TAM_ALFABETO; i++)
         pNo->filhos[i] = NULL;
   }

   return pNo;
}

/****
 *
 * EhNoFinalTrieMachado(): Verifica se um n de uma trie  folha
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *
 * Retorno: 1, se o n for folha;
 *          0, em caso contrrio
 *
 ****/
static int EhNoFinalTrieMachado(const tNoTrieMachado *pNo)
{
   return !EstaVaziaLSE(pNo->valores);
}

/****
 *
 * EhNoVazioTrieMachado(): Verifica se um n de uma trie est vazio
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *
 * Retorno: 1, se o n estiver vazio;
 *          0, em caso contrrio
 *
 ****/
static int EhNoVazioTrieMachado(const tNoTrieMachado *pNo)
{
   int i;

   for(i = 0; i < TAM_ALFABETO; i++)
      if(pNo->filhos[i])
         return 0;

   return 1;
}

/****
 *
 * RemoveTrieMachadoAux(): Funo auxiliar de remoo de chaves
 *
 * Parmetros:
 *      pNo (entrada/sada) - endereo do n que poder
 *                            ser removido
 *      chave (entrada) - chave a ser removida
 *      nivel (entrada) - ???
 *      tam (entrada) - ???
 *
 * Retorno: 1, se o n for livre;
 *          0, em caso contrrio
 *
 ****/
static int RemoveTrieMachadoAux( tNoTrieMachado *pNo, const char *chave,
                                 int nivel, int tam )
{
   int indice; /* ndice de um caractere no array de filhos de um n  */

   if (pNo) {
         /* Verifica se o o nvel corrente  igual ao tamanho da chave */
      if(nivel == tam) {
            /* Se o n indicar o final da chave,  ele deixar de s-lo */
         if(EhNoFinalTrieMachado(pNo)) {
            DestroiListaSE(&pNo->valores); /* N deixa de ser final de string */

               /* Se o n estiver vazio, ele precisa ser removido */
            if(EhNoVazioTrieMachado(pNo))
               return 1;

            return 0; /* N no deve ser removido */
         }
      } else { /* Caso recursivo */
            /* Obtm o ndice do caractere corrente da chave */
         indice = INDICE_CARACTERE(chave[nivel]);

            /* Certifica-se que o ndice obtido  vlido */
         ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                   "Indice invalido em RemoveTrieMachadoAux()" );

         if(RemoveTrieMachadoAux( pNo->filhos[indice], chave,  nivel + 1, tam)) {
            FREE(pNo->filhos[indice]); /* N precisa ser removido */

               /* Sobe a rvore recursivamente removendo */
               /* os ns que devem ser removidos         */
            return !EhNoFinalTrieMachado(pNo) && EhNoVazioTrieMachado(pNo);
         }
      }
   }

   return 0; /* pNo no aponta para nenhum n e no h mais o que fazer */
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaTrieMachado(): Inicia uma trie contendo apenas uma raiz vazia
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo da raiz da trie
 *
 ****/
tTrieMachado IniciaTrieMachado(void)
{
   tNoTrieMachado *pNo;

   pNo = NovoNoTrieMachado();

   ASSEGURA(pNo, "Impossivel alocar no' da arvore trie");

   return pNo;
}

/****
 *
 * InsereTrieMachado(): Insere uma nova chave (string) na trie
 *
 * Parmetros:
 *      raiz (entrada/sada) - endereo da raiz da trie
 *      chave (entrada) - chave que ser inserida
 *      valor (entrada) - valor associado  chave inserida
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. Se a chave j existir, um novo n ser inserido na
 *        lista encadeada apontada pelo campo 'valores'. Esse
 *        n conter o ndice da linha e o ndice da palavra na linha.
 *     2. O ltimo n visitado indica o final de um string,
 *        de modo que se a chave era apenas prefixo de alguma
 *        outra chave na trie, ele passar a ser uma chave
 *
 ****/
void InsereTrieMachado( tNoTrieMachado *raiz, const char *chave,
                        const tConteudo *valor )
{
   int             indice;/* ndice de um caractere no array de filhos de um n */
   tNoTrieMachado *p = raiz; /* Ponteiro usado para descer na trie */

   for (; *chave; ++chave) {
      indice = INDICE_CARACTERE(*chave); /* Obtm o ndice do caractere corrente da chave */

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em InsereTrieMachado()" );

         /* Verifica se o caractere corrente existe no presente n */
      if (!p->filhos[indice])
            /* O referido caractere no existe e  preciso */
            /* criar mais um filho para o corrente n      */
         p->filhos[indice] = NovoNoTrieMachado();

      p = p->filhos[indice]; /* Passa para o prximo nvel da trie */
   }

      /* O ltimo n visitado deve indicar o final de uma chave. Neste */
      /* caso, armazena-se o valor  associado a essa chave na lista.   */
   InsereNoLSE(&p->valores, *valor);
}

/****
 *
 * BuscaTrieMachado(): Efetua uma busca numa trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *      chave (entrada) - a chave de busca
 *
 * Retorno: O endereo da lista encadeada contendo os valores
 *          associado  chave, se ela for encontrada;
 *          NULL, em caso contrrio
 *
 ****/
tListaSE BuscaTrieMachado(tNoTrieMachado *raiz, const char *chave)
{
   int             indice; /* ndice de um caractere no */
                           /* array de filhos de um n  */
   tNoTrieMachado *p = raiz; /* Ponteiro usado para descer na trie */

   for (; *chave && p; ++chave) {
         /* Obtm o ndice do caractere corrente da chave */
      indice = INDICE_CARACTERE(*chave);

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em Busca()" );

         /* Se o caractere corrente da chave no existe   */
         /* no presente n, a chave no faz parte da trie */
      if (!p->filhos[indice])
         return NULL;

      p = p->filhos[indice]; /* Passa para o prximo nvel da trie */
   }

      /* Se p aponta para um n vlido, esse n aponta */
      /* para a lista que contm os valores procurados */
   if (p)
      return p->valores;

   return NULL; /* A chave no foi encontrada */
}

/****
 *
 * RemoveTrieMachado(): Remove uma chave de um trie
 *
 * Parmetros:
 *      *pTrie (entrada/sada) - endereo da raiz da trie
 *      chave (entrada) - chave a ser removida
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  apenas uma interface para
 *             a funo RemoveTrieMachadoAux() que realmente realiza
 *             a tarefa de remoo
 *
 ****/
void RemoveTrieMachado(tTrieMachado *pTrie, const char *chave)
{
   int tam = strlen(chave);

      /* Se a chave no for um string vazio,    */
      /* chama a funo Remove() para remov-la */
   if(tam > 0)
      RemoveTrieMachadoAux(*pTrie, chave, 0, tam);
}

/****
 *
 * NumeroDeChavesTrieMachado(): Calcula o nmero de chaves de uma trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *
 * Retorno: O nmero de chaves da trie
 *
 ****/
int NumeroDeChavesTrieMachado(tNoTrieMachado *raiz)
{
   int cont = 0, /* Armazena o nmero de chaves */
       i;

      /* Trie vazia no contm nenhuma chave */
   if (!raiz)
      return 0;

      /* Se o n for final, encontrou-se uma chave */
   if (raiz->valores)
      cont++;

      /* Conta o nmero de chaves em cada subrvore */
      /* do n corrente que no  vazia             */
   for (i = 0; i < TAM_ALFABETO; ++i)
      if (raiz->filhos[i])
         cont += NumeroDeChavesTrieMachado(raiz->filhos[i]);

   return cont;
}
