/****
 *
 * Ttulo: Maior Prefixo Comum (MPC) a um Conjunto de Strings
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 05/04/2016
 * ltima modificao: 05/04/2016
 *
 * Descrio: Encontra o maior prefixo comum a um conjunto de
 *            strings (MPC) usando trie
 *
 * Entrada: No h
 *
 * Sada: Mensagens apresentando os resultados das operaes
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/***************** Constantes Simblicas ******************/

   /* Tamanho do alfabeto */
#define TAM_ALFABETO 'z' - 'a' + 1

   /* A maior palavra da lngua portuguesa possui */
   /* 46 letras de acordo com a Wikipdia         */
#define TAM_MAIOR_PALAVRA 46 + 1

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

   /* Resulta no ndice de um caractere num array */
   /* que representa os caracteres do alfabeto    */
#define INDICE_CARACTERE(c) ((c) - 'a')

   /* Libera um n e torna nulo o   */
   /* ponteiro usado para liber-lo */
#define FREE(p) do {free(p); p = NULL;} while(0)

/****************** Definies de Tipos *******************/

   /* Tipo de n da trie */
typedef struct rotNoTrie {
           struct rotNoTrie *filhos[TAM_ALFABETO];
           int               ehFinal;
        } tNoTrie, *tTrie;

/************************* Aluses ************************/

extern tNoTrie *NovoNoDeTrie(void);
extern tTrie IniciaTrie(void);
extern void InsereEmTrie(tNoTrie *raiz, const char *chave);
extern int BuscaEmTrie(tNoTrie *raiz, const char *chave);
extern int EhNoFinalDeTrie(const tNoTrie *pNo);
extern int EhNoVazioDeTrie(const tNoTrie *pNo);
extern int RemoveEmTrieAux(tNoTrie *pNo, const char *chave, int nivel, int tam);
extern void RemoveEmTrie(tTrie pTrie, const char *chave);
extern int NumeroDeChavesEmTrie(tNoTrie *raiz);
extern int NumeroDeFilhosDeNoDeTrie(const tNoTrie *pNo, int *indice);
extern char *MPC(const tNoTrie *raiz);

/***************** Definies de Funes ******************/

/****
 *
 * NumeroDeFilhosDeNoDeTrie(): Calcula o nmero de filhos de um n de uma trie
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *      indice (sada) - armazenar o ndice do ltimo
 *                       filho do n que no  NULL
 *
 * Retorno: o nmero de filhos do n
 *
 ****/
int NumeroDeFilhosDeNoDeTrie(const tNoTrie *pNo, int *indice)
{
   int i,
       nFilhos = 0; /* Armazenar o nmero de filhos */

   for (i = 0; i < TAM_ALFABETO; ++i)
      if (pNo->filhos[i]) {
         nFilhos++; /* Mais um filho encontrado */

         *indice = i; /* Atualiza o parmetro indice */
      }

   return nFilhos;
}

/****
 *
 * MPC(): Encontra o maior prefixo comum (MPC) de uma trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *
 * Retorno: Endereo do string que armazena o tal prefixo
 *
 * Observao: Como a funo retorna o endereo de um array
 *             de durao fixa, a cada chamada dela, esse
 *             array pode ser sobrescrito
 *
 ****/
char *MPC(const tNoTrie *raiz)
{
   static char    prefixo[TAM_MAIOR_PALAVRA] = {0};
   const tNoTrie *p = raiz;
   int            indice,
                  tamPrefixo = 0;

      /* Tenta encontrar a primeira bifurcao da trie */
   while (NumeroDeFilhosDeNoDeTrie(p, &indice) == 1 && !p->ehFinal) {
      p = p->filhos[indice];

         /* Certifica-se que no haver corrupo de memria */
      ASSEGURA(tamPrefixo < TAM_MAIOR_PALAVRA, "A Wikipedia estava errada");

         /* Acrescenta o caractere encontrado */
         /* ao array que armazena o prefixo   */
      prefixo[tamPrefixo] = 'a' + indice;

      ++tamPrefixo; /* O tamanho do prefixo aumentou */
   }

   return prefixo;
}

/****
 *
 * NovoNoDeTrie(): Cria um n de uma trie
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo do n criado, se houve alocao; NULL, em caso contrrio
 *
 ****/
tNoTrie *NovoNoDeTrie(void)
{
   tNoTrie *pNo;
   int      i;

   pNo = malloc(sizeof(*pNo));

   if (pNo) {
      pNo->ehFinal = 0; /* Por enquanto, o n no  final */

         /* Inicia cada filho do n com NULL */
      for (i = 0; i < TAM_ALFABETO; i++)
         pNo->filhos[i] = NULL;
   }

   return pNo;
}

/****
 *
 * IniciaTrie(): Inicia uma trie contendo apenas uma raiz vazia
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo da raiz da trie
 *
 ****/
tTrie IniciaTrie(void)
{
   tNoTrie *pNo;

   pNo = NovoNoDeTrie();

   ASSEGURA(pNo, "Impossivel alocar no' da trie");

   return pNo;
}

/****
 *
 * InsereEmTrie(): Insere uma nova chave (string) numa trie
 *
 * Parmetros:
 *      raiz (entrada/sada) - endereo da raiz da trie
 *      chave (entrada) - chave que ser inserida
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. Se a chave j existir, ela no ser inserida
 *     2. O ltimo n visitado indica o final de um string,
 *        de modo que se a chave era apenas prefixo de alguma
 *        outra chave na trie, ele passar a ser uma chave
 *
 ****/
void InsereEmTrie(tNoTrie *raiz, const char *chave)
{
   int      indice; /* ndice de um caractere no */
                    /* array de filhos de um n  */
   tNoTrie *p = raiz; /* Ponteiro usado para descer na trie */

   for (; *chave; ++chave) {
         /* Obtm o ndice do caractere corrente da chave */
      indice = INDICE_CARACTERE(*chave);

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em InsereEmTrie()" );

         /* Verifica se o caractere corrente existe no presente n */
      if (!p->filhos[indice])
            /* O referido caractere no existe e  preciso */
            /* criar mais um filho para o corrente n      */
         p->filhos[indice] = NovoNoDeTrie();

      p = p->filhos[indice]; /* Passa para o prximo nvel da trie */
   }

   p->ehFinal = 1; /* O ltimo n visitado deve indicar o final de um string */
}

/****
 *
 * BuscaEmTrie(): Efetua uma busca numa trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *      chave (entrada) - a chave de busca
 *
 * Retorno: 1, se a chave for encontrada;
 *          0, em caso contrrio
 *
 ****/
int BuscaEmTrie(tNoTrie *raiz, const char *chave)
{
   int      indice; /* ndice de um caractere no */
                    /* array de filhos de um n  */
   tNoTrie *p = raiz; /* Ponteiro usado para descer na trie */

   for (; *chave && p; ++chave) {
         /* Obtm o ndice do caractere corrente da chave */
      indice = INDICE_CARACTERE(*chave);

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em BuscaEmTrie()" );

         /* Se o caractere corrente da chave no existe   */
         /* no presente n, a chave no faz parte da trie */
      if (!p->filhos[indice])
         return 0;

         /* Passa para o prximo nvel da trie */
      p = p->filhos[indice];
   }

      /* Se o lao encerrou porque p assumiu NULL, */
      /* a chave no se encontra na trie           */
   if (!p) {
      return 0;
   }

      /* A chave est presente na trie se o */
      /* ltimo n visitado for um n final */
   return p->ehFinal;
}

/****
 *
 * EhNoFinalDeTrie(): Verifica se um n de uma trie indica
 *              o final de um string
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *
 * Retorno: 1, se o n indicar o final de um string;
 *          0, em caso contrrio
 *
 ****/
int EhNoFinalDeTrie(const tNoTrie *pNo)
{
   return pNo->ehFinal;
}

/****
 *
 * EhNoVazioDeTrie(): Verifica se um n de uma trie est vazio
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *
 * Retorno: 1, se o n estiver vazio;
 *          0, em caso contrrio
 *
 ****/
int EhNoVazioDeTrie(const tNoTrie *pNo)
{
   int i;

      /* Verifica se o n possui algum filho */
   for(i = 0; i < TAM_ALFABETO; i++) {
         /* Se o n tiver pelo menos um */
         /* filho, ele no est vazio   */
      if(pNo->filhos[i])
         return 0; /* N no est vazio */
   }

      /* N no possui nenhum filho */
   return 1;
}

/****
 *
 * RemoveEmTrieAux(): Funo auxiliar de remoo de chaves
 *
 * Parmetros:
 *      pNo (entrada/sada) - endereo do n que poder ser removido
 *      chave (entrada) - chave a ser removida
 *      nivel (entrada) - nvel de remoo
 *      tam (entrada) - tamanho da chave a ser removida
 *
 * Retorno: 1, se o n estiver vazio;
 *          0, em caso contrrio
 *
 ****/
int RemoveEmTrieAux(tNoTrie *pNo, const char *chave, int nivel, int tam)
{
   int indice; /* ndice de um caractere no array de filhos de um n */

   if (pNo) {
         /* Verifica se o o nvel corrente  igual ao tamanho da chave */
      if (nivel == tam) {
            /* Se o n indicar o final da chave, ele deixar de s-lo */
         if (pNo->ehFinal) {
            pNo->ehFinal = 0; /* N deixa de ser final de string */

               /* Se o n estiver vazio, ele precisa ser removido */
            if(EhNoVazioDeTrie(pNo))
               return 1;

            return 0; /* N no deve ser removido */
         }
      } else {
            /* Obtm o ndice do caractere corrente da chave */
         indice = INDICE_CARACTERE(chave[nivel]);

            /* Certifica-se que o ndice obtido  vlido */
         ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                   "Indice invalido em RemoveEmTrieAux()" );

         if( RemoveEmTrieAux( pNo->filhos[indice], chave, nivel + 1, tam ) ) {
            FREE(pNo->filhos[indice]); /* N precisa ser removido */

               /* Sobe a rvore recursivamente removendo */
               /* os ns que devem ser removidos         */
            return !EhNoFinalDeTrie(pNo) && EhNoVazioDeTrie(pNo);
         }
      }
   }

   return 0; /* pNo no aponta para nenhum n e no h mais o que fazer */
}

/****
 *
 * RemoveEmTrie(): Remove uma chave de um trie
 *
 * Parmetros:
 *      pTrie (entrada) - raiz da trie
 *      chave (entrada) - chave a ser removida
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  apenas uma interface para
 *             a funo RemoveEmTrieAux() que realmente realiza
 *             a tarefa de remoo
 *
 ****/
void RemoveEmTrie(tTrie pTrie, const char *chave)
{
   int tam = strlen(chave);

      /* Se a chave no for um string vazio, chama a funo */
      /* RemoveEmTrieAux() para remov-la                   */
   if(tam > 0)
      RemoveEmTrieAux(pTrie, chave, 0, tam);
}

/****
 *
 * NumeroDeChavesEmTrie(): Calcula o nmero de chaves de uma trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *
 * Retorno: O nmero de chaves da trie
 *
 ****/
int NumeroDeChavesEmTrie(tNoTrie *raiz)
{
   int cont = 0, /* Armazena o nmero de chaves */
       i;

      /* Trie vazia no contm nenhuma chave */
   if (!raiz)
      return 0;

      /* Se o n for final, encontrou-se uma chave */
   if (raiz->ehFinal)
      cont++;

      /* Conta o nmero de chaves em cada subrvore */
      /* do n corrente que no  vazia             */
   for (i = 0; i < TAM_ALFABETO; ++i)
      if (raiz->filhos[i])
         cont += NumeroDeChavesEmTrie(raiz->filhos[i]);

   return cont;
}

/****
 *
 * ResultadoBuscaEmTrie(): Efetua uma busca numa trie e apresenta o resultado
 *
 * Parmetros:
 *      trie (entrada) - a trie
 *      chave (entrada) - a chave a ser procurada
 *
 * Retorno: Nada
 *
 ****/
void ResultadoBuscaEmTrie(tTrie trie, const char *chave)
{
   int resultado;

      /* Efetua a busca */
   resultado = BuscaEmTrie(trie, chave);

      /* Apresenta o resultado da busca */
   printf( "\n>>> A chave \"%s\" %s presente na trie", chave,
           resultado ? "esta'" : "NAO esta'");
}

/****
 *
 * Indices(): Apresenta os caracteres do alfabeto
 *            usado e seus ndices
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada apenas para depurao
 *
 ****/
void Indices(void)
{
   int i;

   printf("TAM_ALFABETO = %d\n", TAM_ALFABETO);

   for (i = 0; i <= 255; ++i) {
      if (islower(i)) {
         printf("\n%c ---> %2d ---> %2d", i, i, INDICE_CARACTERE(i));
      }
   }
}

/****
 *
 * main(): Encontra o maior prefixo comum (MPC) a um
 *         conjunto de strings usando uma trie
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   const char *chaves[] = { "rata", "rato", "ratazana", "ratao", "ratinho" };
   tTrie       raiz = IniciaTrie();
   int         i,
               nChaves;

   raiz = IniciaTrie(); /* Inicia a trie */

   nChaves = sizeof(chaves)/sizeof(chaves[0]);

      /* Constri a trie */
   for (i = 0; i < nChaves; ++i)
      InsereEmTrie(raiz, chaves[i]);

      /* Efetua buscas na trie e apresenta os resultados */
   ResultadoBuscaEmTrie(raiz, "rato");
   ResultadoBuscaEmTrie(raiz, "ratazana");
   ResultadoBuscaEmTrie(raiz, "romano");
   ResultadoBuscaEmTrie(raiz, "ratinho");

   printf("\n\n>>> Numero de chaves na trie: %d\n", NumeroDeChavesEmTrie(raiz));

   printf("\n>>> Maior prefixo na trie: %s\n", MPC(raiz));

   RemoveEmTrie(raiz, "rato");
   RemoveEmTrie(raiz, "ratinho");

   printf("\n>>> Removida a chave \"%s\"", "rato");
   printf("\n>>> Removida a chave \"%s\"\n", "ratinho");

      /* Efetua buscas na trie e apresenta os resultados */
   ResultadoBuscaEmTrie(raiz, "rato");
   ResultadoBuscaEmTrie(raiz, "ratinho");

   printf("\n\n>>> Numero de chaves na trie: %d\n", NumeroDeChavesEmTrie(raiz));

   printf("\n>>> Maior prefixo na trie: %s\n", MPC(raiz));

   return 0;
}

/****************** Resultado do Programa *****************

>>> A chave "rato" esta' presente na trie
>>> A chave "ratazana" esta' presente na trie
>>> A chave "romano" NAO esta' presente na trie
>>> A chave "ratinho" esta' presente na trie

>>> Numero de chaves na trie: 5

>>> Maior prefixo na trie: rat

>>> Removida a chave "rato"
>>> Removida a chave "ratinho"

>>> A chave "rato" NAO esta' presente na trie
>>> A chave "ratinho" NAO esta' presente na trie

>>> Numero de chaves na trie: 3

>>> Maior prefixo na trie: rata

**********************************************************/
