/****
 *
 * Arquivo: Huffman.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 06/10/2017
 * ltima modificao: 06/10/2017
 *
 * Descrio: Implementao da codificao de Huffman
 *
 ****/

/*********************** Includes *************************/

#include <string.h>   /* memmove()                */
#include <stdlib.h>   /* Alocao dinmica        */
#include <limits.h>   /* Limites de tipos inteiros */
#include <ctype.h>    /* Classificao de caracteres */
#include "Tipos.h"    /* Tipos usados pelo programa  */
#include "Heap.h"     /* Interface do mdulo Heap  */
#include "Huffman.h"  /* Interface deste mdulo    */
#include "Arquivos.h" /* Processamento de arquivos */
#include "Bits.h"     /* Processamento de bits     */
#include "ASSEGURA.h" /* Macro ASSEGURA()          */

/****************** Diretiva de Depurao *****************/

#define DEPURACAO 0

/********************* Funes Locais *********************/

/****
 *
 * ComparaNosHeapHuff(): Compara dois elementos do heap
 *
 * Parmetros:
 *      e1, e2 (entrada) - os elementos
 *
 * Retorno:
 *     0, se e1->byteFreq.frequencia == e2->byteFreq.frequencia
 *     < 0, se e1->byteFreq.frequencia < e2->byteFreq.frequencia
 *     > 0, se e1->byteFreq.frequencia > e2->byteFreq.frequencia
 *
 ****/
static int ComparaNosHeapHuff(const void *e1, const void *e2)
{
   tNoHeapHuff elemento1 = (tNoHeapHuff)e1, elemento2 = (tNoHeapHuff)e2;

   ASSEGURA(e1 && e2, "Elemento nulo recebido");

   return elemento1->byteFreq.frequencia - elemento2->byteFreq.frequencia;
}

/****
 *
 * ExibeCodigosHuff(): Exibe o cdigo associado a cada byte de uma
 *                 lista de cdigos cannicos
 *
 * Parmetros:
 *     listaCanonica (entrada) - a lista cannica de cdigos
 *     tamLista (entrada) - tamanho da lista
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada em teste e depurao
 *
 ****/
static void ExibeCodigosHuff(tNoListaCanHuff *listaCanonica, int tamLista)
{
   int   i,
         tamanho = 0;

      /* Exibe cabealho da apresentao na tela */
   printf("Byte  Tamanho  Codificacao\n");
   printf("----- -------- ----------------\n");

      /* Apresenta os cdigos na tela  */
   for(i = 0; i < tamLista; i++) {
         /* Verifica a validade do cdigo */
      ASSEGURA(listaCanonica[i].nBits > 0, "Tamanho de codigo deveria ser positivo");

         /* Exibe informaes sobre o valor do byte corrente */
//      printf( "0x%02X  %02d       ", listaCanonica[i].valor,
      printf("%c  %02d       ", listaCanonica[i].valor, listaCanonica[i].nBits);

         /* Escreve os bits do cdigo */
      for(tamanho = 0; tamanho < listaCanonica[i].nBits; tamanho++)
         if (ConsultaBitEmArray(listaCanonica[i].cod, tamanho))
            putchar('1');
         else
            putchar('0');

      putchar('\n');
   }
}

/****
 *
 * ConstroiNoArvoreHuff(): Constri um n de uma rvore binria
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo do n criado
 *
 ****/
static tNoArvoreHuff *ConstroiNoArvoreHuff(void)
{
   tNoArvoreHuff *pNo;

   pNo = malloc(sizeof(tNoArvoreHuff)); /* Tenta alocar o novo n */

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(pNo, "Nao foi possivel alocar no' de arvore");

      /* Este n ainda no tem filhos */
   pNo->filhoEsquerda = NULL;
   pNo->filhoDireita = NULL;

   return pNo;
}

/****
 *
 * ReconstroiArvoreHuff(): Reconstri uma rvore de codificao a
 *                     partir de uma lista de cdigos
 *
 * Parmetros:
 *       lista (entrada) - a lista de cdigos
 *       tamLista (entrada) - tamanho da lista
 *
 * Retorno: Endereo da raiz da rvore
 *
 ****/
static tArvoreHuff ReconstroiArvoreHuff(tNoListaCanHuff lista[], int tamLista)
{
   tNoArvoreHuff *raiz, /* Raiz da rvore a ser criada */
                 *p, /* Ponteiro usado para descer a rvore    */
                 *q; /* Segue o ponteiro p um nvel acima dele */
   int            i, j;

   raiz = ConstroiNoArvoreHuff(); /* Cria a raiz da rvore */

      /* Para cada valor de byte da lista, desce-se a rvore criando ou */
      /* visitando ns de acordo com o cdigo associado ao byte         */
   for (i = 0; i < tamLista; ++i) {
         /* A descida comea sempre pela raiz */
      q = NULL;
      p = raiz;

         /* Acessa cada bit do cdigo associado ao byte corrente e desce a */
         /* rvore de acordo com seu valor: se o bit for 0, desce-se pela  */
         /* esquerda; se o bit for 1, desce-se pela direita                */
      for (j = 0; j < lista[i].nBits; ++j) {
         q = p; /* Guarda o endereo do n corrente antes de descer */

         if (ConsultaBitEmArray(lista[i].cod, j)) {
            p = p->filhoDireita; /* O bit  1 e a descida  pela direita */

               /* Se o n no existir, cria-se um n que ser filho direito de q */
            if (!p) {
               p = ConstroiNoArvoreHuff();
               q->filhoDireita = p; /* Inclui novo n na rvore */
            }
         } else {
            p = p->filhoEsquerda; /* O bit  0 e a descida  pela esquerda */

               /* Se o n no existir, cria-se um */
               /* n que ser filho esquerdo de q */
            if (!p) {
               p = ConstroiNoArvoreHuff();
               q->filhoEsquerda = p;/* Inclui novo n na rvore */
            }
         }
      }

         /* Neste ponto, p deve estar apontando para uma folha */
      ASSEGURA( !p->filhoEsquerda && !p->filhoDireita, "Deveria ser folha!" );

      p->byteFreq.byte = lista[i].valor; /* Armazena o valor do byte na folha */
   }

   return raiz;
}

/****
 *
 * ExibeHeapHuff(): Exibe o contedo do heap
 *
 * Parmetros:
 *     heap (entrada) - o heap
 *
 * Retorno: Nada
 *
 * Observao: Usada apenas em depurao
 *
 ****/
static void ExibeHeapHuff(const tHeapHuff *heap)
{
   int i;
   double freqTotal = 0.0;

   for (i = 0; i < heap->nItens; ++i) {
      printf( "\n [%02d]\tbyte = 0x%02X\tfrequencia = %.2f", i,
              heap->itens[i]->byteFreq.byte,
              heap->itens[i]->byteFreq.frequencia );
      freqTotal += heap->itens[i]->byteFreq.frequencia;
   }

   printf( "\n\n\t>>> Frequencia total: %.2f\n\n", freqTotal);
}

/****
 *
 * CriaHeapHuff(): Cria um heap contendo ns de uma rvore que armazenam
 *                 valores de bytes e suas respectivas frequncias
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo no qual os bytes sero lidos
 *     nBytes (sada) - nmero total de bytes lidos no arquivo
 *
 * Retorno: Endereo do heap criado
 *
 ****/
static tHeapHuff *CriaHeapHuff(FILE *stream, int *nBytes)
{
   tHeapHuff     *pHeap; /* Ponteiro para o heap que ser criado */
   tNoArvoreHuff *ptrNovoNo; /* Apontar para o novo n alocado */
   int           *byteFreq, /* Apontar para um array contendo os */
                            /* nmeros de ocorrncias dos bytes   */
                  umByte, /* Armazena um byte */
                  tamArquivo = 0, /* Nmero de bytes do arquivo */
                  i;

      /* Aloca um array para conter as ocorrncias dos bytes */
   byteFreq = calloc(UCHAR_MAX + 1, sizeof(int));
   ASSEGURA(byteFreq, "Array de ocorrencias nao foi alocado");

   rewind(stream); /* Move o apontador de posio para o incio do arquivo */

      /* Obtm o nmero de ocorrncias de cada byte */
   while (1) {
      umByte = fgetc(stream); /* L um byte */

         /* Verifica se ocorreu erro */
      ASSEGURA(!ferror(stream), "Erro de leitura de arquivo");

      if (feof(stream))
         break; /* No h mais bytes a serem lidos */

         /* Verifica se o programa pode lidar a frequncia de ocorrncia */
         /* desse byte. A ausncia desse teste causar overflow.         */
      ASSEGURA(byteFreq[umByte] < INT_MAX,
               "Este programa nao suporta um arquivo tao grande");

         /* Incrementa o nmero de ocorrncias do ltimo byte lido */
      ++byteFreq[umByte];

         /* Verifica se o programa pode lidar com o tamanho do */
         /* arquivo. A ausncia desse teste causar overflow.  */
      ASSEGURA(tamArquivo < INT_MAX,
               "Este programa nao suporta um arquivo tao grande");

      ++tamArquivo; /* Mais um byte foi lido */
   }

      /* Aloca a varivel que armazenar o heap */
   ASSEGURA(pHeap = malloc(sizeof(*pHeap)), "Heap nao alocado");

   IniciaHeapHuff(pHeap); /* Inicia o heap */

      /* Constri o heap */
   for (i = 0; i < UCHAR_MAX; ++i) {
         /* Cria um n para cada valor de byte encontrado no arquivo */
      if (byteFreq[i]) {
         ptrNovoNo = ConstroiNoArvoreHuff(); /* Tenta alocar um novo n */

            /* Armazena o byte corrente e sua frequncia no novo n */
         ptrNovoNo->byteFreq.byte = i;
         ptrNovoNo->byteFreq.frequencia = (double) 100*byteFreq[i]/tamArquivo;

            /* Insere o novo n no heap */
         InsereEmHeapHuff(pHeap, ptrNovoNo, ComparaNosHeapHuff);
      }
   }

   free(byteFreq); /* O array de ocorrncias no  mais necessrio */

   *nBytes = tamArquivo; /* Atualiza o segundo parmetro antes de retornar */

   return pHeap; /* Retorna o endereo do heap */
}

/****
 *
 * DestroiArvoreHuff(): Destri uma rvore binria; i.e., libera o
 *                  espao ocupado por cada n da rvore
 *
 * Parmetros:
 *      arvore (entrada e sada) - ponteiro para a raiz da rvore
 *                                 binria a ser destruda
 *
 * Retorno: Nada
 *
 ****/
static void DestroiArvoreHuff(tArvoreHuff arvore)
{
   if (arvore) {
      DestroiArvoreHuff(arvore->filhoEsquerda); /* Destri subrvore esquerda */

      DestroiArvoreHuff(arvore->filhoDireita); /* Destri subrvore direita */

      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * ObtemTamanhosHuff(): Obtm os tamanhos dos cdigos associados aos
 *                  bytes armazenados nas folhas de uma rvore de
 *                  codificao de Huffman
 *
 * Parmetros:
 *      raiz (entrada) - ponteiro para a raiz da rvore que contm a codificao
 *      tabela[] (sada) - tabela de consulta que conter os
 *                         tamanhos dos cdigos cannicos
 *      n (entrada) - nvel da rvore na qual a visitao dos ns ser iniciada
 *
 * Retorno: Nada
 *
 * Observao: Na primeira chamada desta funo, o valor do
 *             ltimo parmetro deve ser 0
 *
 ****/
static void ObtemTamanhosHuff(tArvoreHuff raiz, int tabela[], int n)
{
      /* A rvore no pode estar vazia */
   ASSEGURA(raiz, "A raiz da arvore e' nula");

      /* Se for possvel, caminha na subrvore esquerda do prximo nvel */
   if (raiz->filhoEsquerda)
      ObtemTamanhosHuff(raiz->filhoEsquerda, tabela, n + 1);

      /* Se for possvel, caminha na subrvore direita do prximo nvel */
   if (raiz->filhoDireita)
      ObtemTamanhosHuff(raiz->filhoDireita, tabela, n + 1);

      /* Se o n corrente for uma folha, ele  */
      /* armazena um valor de byte do arquivo */
   if (!raiz->filhoEsquerda && !raiz->filhoDireita)
         /* Armazena no ndice da tabela correspondente ao valor do byte       */
         /* armazenado nessa folha o tamanho que foi acumulado no caminhamento */
      tabela[raiz->byteFreq.byte] = n;
}

/****
 *
 * ComparaTamanhosHuff(): Funo de comparao usada por qsort() para
 *                    ordenar a lista de cdigos cannicos por
 *                    tamanhos de cdigos e por valores de bytes
 *
 * Parmetros:
 *     e1, e2 (entrada) - ponteiros para elementos de uma lista cannica
 *
 * Retorno:
 *      1 se e1 > e2
 *     -1 se e1 < e2
 *
 * Observaes:
 *      1. Se os tamanhos forem iguais, os valores dos bytes sero
 *         usados para desempatar
 *      2. Esta funo nunca retorna zero
 *
 ****/
static int ComparaTamanhosHuff(const void *e1, const void *e2)
{
   int nBits1 = ((tNoListaCanHuff *)e1)->nBits,
       nBits2 = ((tNoListaCanHuff *)e2)->nBits,
       valor1 = ((tNoListaCanHuff *)e1)->valor,
       valor2 = ((tNoListaCanHuff *)e2)->valor;

   if (nBits1 > nBits2)
      return 1; /* e1 > e2 */
   else if (nBits1 < nBits2)
      return -1; /* e1 < e2 */

      /* Se ainda no houve retorno, os tamanhos so iguais e */
      /* o maior elemento ser aquele que tiver o maior valor */
   if (valor1 > valor2)
      return 1;  /* e1 > e2 */
   else if (valor1 < valor2)
      return -1;  /* e1 < e2 */

      /* J deveria ter havido retorno */
   ASSEGURA(0, "Erro: houve empate em ComparaTamanhosHuff()");
}

/****
 *
 * ObtemCodigosCanonicosHuff(): Atribui um cdigo cannico para cada
 *                          byte de uma lista de bytes ordenados
 *                          de acordo com seus tamanhos de cdigos e seus valores
 *
 * Parmetros:
 *     lista (entrada e sada ) - a lista ordenada de bytes
 *     nItens (entrada) - nmero de elementos da lista
 *
 * Retorno: Nada
 *
 ****/
static void ObtemCodigosCanonicosHuff(tNoListaCanHuff lista[], int nItens)
{
   int   i;
   tByte tamanho;
   tByte codigo[MAX_BYTES] = {0};

      /* Obtm o tamanho do cdigo armazenado no primeiro elemento da lista */
   tamanho = lista[0].nBits;

      /* Atribui cdigos aos elementos da lista */
   for(i = 0; i < nItens; ++i) {
         /* Verifica se o tamanho do cdigo  vlido */
      ASSEGURA(lista[i].nBits, "Tamanho de codigo e' zero");

      IncrementaBits(codigo); /* Soma 1 ao cdigo do byte anterior */

         /* Ajusta o cdigo se seu tamanho for menor */
         /* do que o tamanho do cdigo anterior      */
      if (lista[i].nBits > tamanho) {
         DeslocaBitsEsquerda(codigo, lista[i].nBits - tamanho);
         tamanho = lista[i].nBits;
      }

         /* Copia o cdigo obtido para o respectivo */
         /* elemento da lista e alinha-o  esquerda */
      memmove(lista[i].cod, codigo, MAX_BYTES);
      DeslocaBitsEsquerda(lista[i].cod, MAX_BITS - tamanho);
   }
}

/****
 *
 * ExibeTamanhosCodigosHuff(): Exibe os tamanhos dos cdigos
 *
 * Parmetros:
 *     tamanhos (entrada) - tabela de consulta contendo os tamanhos dos cdigos
 *     nItens (entrada) - nmero de elementos da tabela
 *
 * Retorno: Nada
 *
 * Observao: Usada apenas em depurao
 *
 ****/
void ExibeTamanhosCodigosHuff(int *tamanhos, int nItens)
{
   int i;

   for (i = 0; i < nItens; ++i)
      if (tamanhos[i])
         printf( "0x%02X\t%d\n",i, tamanhos[i]);
}

/****
 *
 * CriaListaCanonicaHuff(): Constri uma lista de cdigos cannicos
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para raiz da rvore a partir
 *                        da qual a lista ser criada
 *     nItens (entrada) - nmero de elementos que a lista ter
 *
 * Retorno: Endereo da lista de cdigos cannicos
 *
 ****/
static tNoListaCanHuff *CriaListaCanonicaHuff(tArvoreHuff arvore, int nItens)
{
   tNoListaCanHuff *listaCanonica;
   int             *tamanhos, i, j;

      /* Aloca um array para conter os tamanhos dos */
      /* cdigos associados aos valores de bytes    */
   tamanhos = calloc(UCHAR_MAX + 1, sizeof(int));
   ASSEGURA(tamanhos, "Array de tamanhos nao foi alocado");

      /* Obtm os tamanhos dos cdigos dos bytes */
   ObtemTamanhosHuff(arvore, tamanhos, 0);

      /* Aloca a lista de cdigos cannicos */
   listaCanonica = calloc(nItens, sizeof(tNoListaCanHuff));
   ASSEGURA(listaCanonica, "Lista canonica nao foi alocada");

      /* Copia os tamanhos dos cdigos para a lista de cdigos cannicos */
   for (i = 0, j = 0; i < UCHAR_MAX + 1; ++i)
         /* A lista cannica contm apenas tamanhos no nulos */
      if (tamanhos[i]) {
         listaCanonica[j].valor = i;
         listaCanonica[j].nBits = tamanhos[i];
         ++j;
      }

   free(tamanhos); /* O array que contm os tamanhos no  mais necessrio */

      /* Ordena a lista de cdigos cannicos por */
      /* tamanho de cdigo e valor de byte       */
   qsort( listaCanonica, nItens, sizeof(tNoListaCanHuff), ComparaTamanhosHuff );

      /* Obtm os cdigos cannicos */
   ObtemCodigosCanonicosHuff(listaCanonica, nItens);

   return listaCanonica;
}

/****
 *
 * EscreveCabecalhoHuff(): Escreve o cabealho da codificao no
 *                     incio do arquivo de sada
 *
 * Parmetros:
 *     cabecalho (entrada) - ponteiro para o cabealho que ser escrito
 *     stream (entrada) - stream associado ao arquivo de sada
 *
 * Retorno: Nada
 *
 ****/
static void EscreveCabecalhoHuff(const tCabecalhoHuff *cabecalho, FILE *stream)
{
      /* Escreve o nmero de elementos da lista */
   fwrite( &cabecalho->tamLista, sizeof(cabecalho->tamLista), 1, stream );

      /* Escreve a lista de cdigos cannicos */
   fwrite(cabecalho->lista, sizeof(tNoListaCanHuff), cabecalho->tamLista, stream);
}

/****
 *
 * LeCabecalhoHuff(): L o cabealho da codificao num arquivo de entrada codificado
 *
 * Parmetros:
 *     cabecalho (sada) - ponteiro para o cabealho que ser lido
 *     stream (entrada) - stream associado ao arquivo de sada
 *
 * Retorno: Endereo da lista de cdigos cannicos lida no
 *          arquivo ou NULL, se ocorrer algum erro
 *
 ****/
static tNoListaCanHuff *LeCabecalhoHuff(tCabecalhoHuff *cabecalho, FILE *stream)
{
      /* L o nmero de elementos da lista */
   fread( &cabecalho->tamLista, sizeof(cabecalho->tamLista), 1, stream );

   if (ferror(stream))
      return NULL; /* Ocorreu erro de leitura */

      /* Aloca espao para a lista de cdigos cannicos */
   cabecalho->lista = calloc( cabecalho->tamLista, sizeof(tNoListaCanHuff) );
   if(!cabecalho->lista)
      return NULL; /* Alocao falhou */

      /* L a lista de cdigos cannicos */
   fread(cabecalho->lista, sizeof(tNoListaCanHuff), cabecalho->tamLista, stream);

   if (ferror(stream))
      return NULL; /* Ocorreu erro de leitura */

   return cabecalho->lista;
}

/****
 *
 * ObtemByteHuff(): Obtm o valor de um byte usando o cdigo associado a ele
 *
 * Parmetros:
 *     arvore (entrada) - rvore de codificao
 *     codigo (entrada) - array de bytes contendo a codificao
 *                        de (pelo menos) um byte
 *     tamCodigo (entrada) - nmero de bytes no array
 *    *indice (entrada e sada) - ndice do primeiro bit a ser examinado
 *
 * Retorno: O valor do byte se for encontrada uma folha da rvore
 *          de codificao durante uma descida nessa rvore; em caso contrrio, -1
 *
 ****/
static int ObtemByteHuff( tArvoreHuff arvore, tByte *codigo, int tamCodigo,
                          int *indice )
{
   static tNoArvoreHuff *pNoInicial = NULL; /* N no qual a descida */
                                            /* na rvore ser iniciada */
   int                   byte; /* Byte que ser retornado */

   if (!pNoInicial)
      pNoInicial = arvore;

      /* Desce a rvore at encontrar uma folha de */
      /* acordo com o valor de cada bit do cdigo  */
   for ( ; *indice < tamCodigo; ++*indice) {
      if (ConsultaBitEmArray(codigo, *indice))
            /* Bit = 1: desce pela direita */
         pNoInicial = pNoInicial->filhoDireita;
      else
            /* Bit = 0: desce pela esquerda */
         pNoInicial = pNoInicial->filhoEsquerda;

         /* Verifica se j se chegou a uma folha */
      if (!pNoInicial->filhoEsquerda && !pNoInicial->filhoDireita){
         byte = pNoInicial->byteFreq.byte; /* Um byte foi decodificado */

         pNoInicial = NULL; /* A prxima descida na rvore comear na raiz */

         return byte;
      }
   }

      /* Se ainda no houve retorno, no se atingiu uma folha */
      /* e a busca ser reiniciada no n no qual ela parou    */
   return -1;
}

/********************* Funes Globais *********************/

/****
 *
 * CodificaHuff(): Codifica um arquivo seguindo o algoritmo de
 *                 codificao de Huffman
 *
 * Parmetros:
 *    *arqEntrada (entrada) - nome do arquivo que ser codificado
 *    *arqSaida (entrada) - nome do arquivo que conter a codificao
 *
 * Retorno: Nada
 *
 ****/
void CodificaHuff(const char *arqEntrada, const char *arqSaida)
{
   FILE          *streamE, /* Associado ao arquivo a ser codificado */
                 *streamS; /* Associado ao arquivo resultante */
   tHeapHuff     *pHeap; /* Apontar para o heap que armazena     */
                         /* temporariamente os ns da codificao */
   tNoArvoreHuff *no1, /* Ponteiro para o n com a */
                       /* menor frequncia no heap */
                 *no2, /* Ponteiro para o n com a segunda */
                       /* menor frequncia no heap         */
                 *noNovo; /* Ponteiro para um novo n da rvore */
                          /* que ter no1 e no2 como filhos     */
   tNoListaCanHuff *lista; /* Lista que conter os cdigos cannicos */
   tCabecalhoHuff   cabecalho; /* Cabealho da codificao */
   int              i, j,
                    c, /* Armazenar um byte lido */
                    bits, /* Empacotar os bits */
                    nFolhas, /* Nmero de valores de bytes distintos no arquivo */
                    contaBits, /* Conta os bits de um byte */
                    tamArquivo; /* Nmero de bytes do arquivo */

   streamE = AbreArquivo(arqEntrada, "rb"); /* Abre o arquivo que ser codificado */

      /* Cria o heap que armazena temporariamente os ns da rvore de codificao */
   pHeap = CriaHeapHuff(streamE, &tamArquivo);

      /* O nmero de folhas da rvore de codificao ser igual ao tamanho */
      /* inicial do heap. Esse valor  igual ao nmero de valores de bytes */
      /* distintos encontrados no arquivo.                                 */
   nFolhas = TamanhoHeapHuff(pHeap);

      /* Constri a rvore de codificao */
   while (!HeapVazioHuff(pHeap)) {
         /* Remove o n com a menor frequncia do heap */
      no1 = RemoveMinHeapHuff(pHeap, ComparaNosHeapHuff);

         /* Se o heap ficou vazio, o ltimo n removido  */
         /* a raiz da rvore que representa a codificao */
      if (HeapVazioHuff(pHeap))
         break; /* Construo de rvore concluda */

         /* Remove o n com a segunda menor frequncia do heap  */
      no2 = RemoveMinHeapHuff(pHeap, ComparaNosHeapHuff);

      noNovo = ConstroiNoArvoreHuff(); /* Tenta alocar um novo n */

         /* Armazena a frequncia do novo n. O contedo do campo 'byte' no */
         /* tem importncia porque este n no  uma folha.                  */
      noNovo->byteFreq.frequencia = no1->byteFreq.frequencia +
                                    no2->byteFreq.frequencia;

         /* O filho esquerdo do novo n  o primeiro n removido do heap */
         /* e o filho direito  o segundo n removido do heap            */
      noNovo->filhoEsquerda = no1;
      noNovo->filhoDireita = no2;

      InsereEmHeapHuff(pHeap, noNovo, ComparaNosHeapHuff); /* Insere o n no heap */
   }

         /*                                      */
         /* Neste ponto, a rvore de codificao */
         /* j foi concluda e sua raiz  no1    */
         /*                                      */

      /* Obtm a lista de cdigos cannicos */
   lista = CriaListaCanonicaHuff(no1, nFolhas);

      /* Nem o heap nem a rvore so mais necessrios */
   DestroiHeapHuff(pHeap);
   free(pHeap);
   DestroiArvoreHuff(no1);

   streamS = AbreArquivo(arqSaida, "wb"); /* Abre o arquivo de sada */

      /* Cria o cabealho da codificao */
   cabecalho.tamLista = nFolhas;
   cabecalho.lista = lista;

      /* Escreve o cabealho da codificao no arquivo de sada */
   EscreveCabecalhoHuff(&cabecalho, streamS);

   bits = 0; /* Zera o empacotador de bits */
   contaBits = 0; /* Inicia a contagem de bits */

      /* Faz o apontador de posio do arquivo apontar para o seu primeiro byte */
   rewind(streamE);

      /* L cada byte no arquivo de entrada, codifica-o usando a lista */
      /* de cdigos e escreve a codificao no arquivo de sada        */
   while(1) {
      c = fgetc(streamE); /* L um byte no arquivo de entrada */

         /* Verifica se ocorreu erro */
      ASSEGURA(!ferror(streamE), "Erro de leitura");

      if (feof(streamE))
         break; /* Acabou a leitura */

         /* Encontra o valor do byte lido na lista de cdigos. */
      for (i = 0; i < nFolhas; ++i)
         if (lista[i].valor == c)
            break;

         /* O valor do byte deve ter sido encontrado */
      ASSEGURA(i < nFolhas, "Valor de byte nao encontrado");

         /* Empacota os bits do cdigo deste byte */
      for(j = 0; j < lista[i].nBits; ++j) {
         contaBits++; /* Mais um bit neste cdigo */

            /* Armazena bit na varivel que empacota os bits */
         bits = (bits << 1) | ConsultaBitEmArray(lista[i].cod, j);

            /* Se j houver um byte na varivel de          */
            /* empacotamento, escreve-o no arquivo de sada */
         if (contaBits == CHAR_BIT) {
            fputc(bits, streamS); /* Escreve byte no arquivo */

               /* Verifica se ocorreu erro de escrita */
            ASSEGURA(!ferror(streamS), "Erro de escrita");

            contaBits = 0; /* Reinicia a contagem de bits */
         }
      }
   }

      /* Escreve os bits que restaram no ltimo empacotamento */
   if (contaBits) {
      bits <<= CHAR_BIT - contaBits; /* Alinha bits  esquerda */
      fputc(bits, streamS); /* Escreve-os no arquivo */
      ASSEGURA(!ferror(streamS), "Erro de escrita"); /* Testa */
   }

   free(lista); /* Libera o espao ocupado pela lista de cdigos */

      /* Fecha os arquivos */
   FechaArquivo(streamE, arqEntrada);
   FechaArquivo(streamS, arqSaida);
}

/****
 *
 * DecodificaHuff(): Decodifica um arquivo e escreve o resultado no
 *               arquivo especificado
 *
 * Parmetros:
 *     arqEntrada (entrada) - nome do arquivo a ser decodificado
 *     arqSaida (entrada) - nome do arquivo no qual o resultado
 *                          ser escrito
 *
 * Retorno: Nada
 *
 ****/
void DecodificaHuff(const char *arqEntrada, const char *arqSaida)
{
   tNoListaCanHuff *listaCanonica; /* Lista de cdigos cannicos */
   tByte            buffer[MAX_BYTES]; /* Buffer que armazenar cada leitura */
                                       /* efetuada no arquivo de entrada     */
   int              i,
                    byte; /* Decodificao de um byte */
   FILE            *streamE, /* Stream associado ao arquivo a ser decodificado */
                   *streamS; /* Associado ao arquivo resultante */
   tCabecalhoHuff   cabecalho; /* Cabealho da codificao */
   tArvoreHuff      arvore; /* Raiz da rvore de codificao */

   streamE = AbreArquivo(arqEntrada, "rb"); /* Tenta abrir o arquivo de entrada */

   streamS = AbreArquivo(arqSaida, "wb"); /* Tenta abrir o arquivo de sada */

      /* L o cabealho no arquivo de entrada */
   listaCanonica = LeCabecalhoHuff(&cabecalho, streamE);

      /* Verifica se ocorreu erro na leitura do cabealho */
   if (!listaCanonica) {
         /* Fecha os arquivos e aborta o programa */
      FechaArquivo(streamE, arqEntrada);
      FechaArquivo(streamS, arqSaida);
      ASSEGURA(0, "Impossivel ler lista de codigos");
   }

      /* Constri a rvore de codificao usando a lista de cdigos cannicos */
   arvore = ReconstroiArvoreHuff(listaCanonica, cabecalho.tamLista);

      /* Decodifica o arquivo de entrada */
   while(1) {
         /* L um array de bytes no arquivo codificado. O tamanho desse array */
         /*  igual ao nmero mximo de bytes que pode constituir um cdigo   */
      fread(buffer, sizeof(buffer), 1, streamE);

         /* Verifica se houve erro de leitura */
      ASSEGURA(!ferror(streamE), "Erro de leitura");

         /* Verifica se a leitura acabou */
      if (feof(streamE))
         break;

         /* Decodifica os bytes lidos e armazenados no buffer */
         /* e escreve o resultado no arquivo de sada         */
      for (i = 0; i < MAX_BITS; ++i) {
            /* Obtm um byte da decodificao do buffer */
         byte = ObtemByteHuff( arvore, buffer, MAX_BITS, &i );

            /* Se um byte foi obtido, escreve-o no arquivo de sada; caso */
            /* contrrio, ele dever ser completado na prxima leitura    */
         if (byte > 0)
            fputc(byte, streamS);
      }
   }

      /* Fecha arquivos */
   FechaArquivo(streamE, arqEntrada);
   FechaArquivo(streamS, arqSaida);

   free(listaCanonica); /* A lista de cdigos cannicos no  mais necessria */
}
