/****
 *
 * Arquivo: Eventos.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 06/10/2017
 * ltima modificao: 06/10/2017
 *
 * Descrio: Implementao de eventos
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>   /* Funes de Entrada/Sada */
#include <stdlib.h>  /* Funes de alocao dinmica */
#include <time.h>    /* Funo time() para medida de tempo */
#include <math.h>    /* Funo log() */
#include "Heap.h"    /* Operaes sobre heaps */
#include "Eventos.h" /* Interface deste mdulo */
#include "Macros.h"  /* Macros usadas pelo programa */

/***************** Definies de Funes ******************/

/****
 *
 * ApresentaHora(): Apresenta uma hora e os minutos que passam dela
 *
 * Parmetros:
 *      hora (entrada) - a hora
 *      min (entrada) - os minutos
 *
 * Retorno: Nada
 *
 ****/
void ApresentaHora(int hora, int min)
{
   printf("\t>>> %.2d:%.2d", hora + min/60, min%60);
}

/****
 *
 * ComparaEventos(): Compara dois eventos
 *
 * Parmetros:
 *      e1, e2 (entrada) - os eventos
 *
 * Retorno:
 *     0, se e1->instante == e2->instante
 *     < 0, se e1->instante < e2->instante
 *     > 0, se e1->instante > e2->instante
 *
 ****/
int ComparaEventos(const void *e1, const void *e2)
{
   tEvento *evento1 = (tEvento *)e1,
           *evento2 = (tEvento *)e2;

   ASSEGURA(e1 && e2, "Evento nulo recebido");

   return evento1->instante - evento2->instante;
}

/****
 *
 * CriaEvento(): Cria um evento
 *
 * Parmetros:
 *      tipoDeEvento (entrada) - o tipo de evento
 *      instante (entrada) - o instante do evento
 *      nPessoas (entrada) - nmero de pessoas do evento
 *
 * Retorno: Endereo do evento criado
 *
 ****/
tEvento *CriaEvento( tTipoDeEvento tipoDeEvento, int instante,
                     int nPessoas )
{
   tEvento *evento;

   evento = malloc(sizeof(tEvento));
   ASSEGURA(evento, "Impossivel alocar evento");

   evento->instante = instante;
   evento->nPessoas = nPessoas;
   evento->tipoDeEvento = tipoDeEvento;

   return evento;
}

/****
 *
 * NAleatorio(): Retorna um nmero aleatrio entre os valores
 *               recebidos como parmetros
 *
 * Parmetros:
 *      m, n (entrada) - valores que definem o intervalo de
 *                       nmeros sorteados
 *
 * Retorno: O nmero aleatrio gerado
 *
 * Observaes:
 *      1.  indiferente se m <= n ou n <= m
 *      2. Quando chamada pela primeira vez, esta funo
 *         alimenta o gerador de nmeros aleatrios
 *
 ****/
int NAleatorio(int m, int n)
{
   static int primeiraChamada = 1; /* Esta varivel checa se a  */
                                   /* funo est sendo chamada */
                                   /* pela primeira vez e deve  */
                                   /* ter durao fixa          */

      /* Se esta for a primeira chamada da funo, */
      /* alimenta o gerador de nmeros aleatrios  */
   if (primeiraChamada) {
         /* Alimenta o gerador de nmeros aleatrios */
      srand(time(NULL));

         /* A prxima chamada no ser mais a primeira */
      primeiraChamada = 0;
   }

      /* Leva em considerao o fato de */
      /* n poder ser menor do que m     */
   if (n < m) {
      return rand()%(m - n + 1) + n;
   }

      /* m  menor do que ou igual a n */
   return rand()%(n - m + 1) + m;
}

/****
 *
 * DRand(): Retorna um nmero aleatrio entre 0.0 e 1.0
 *
 * Parmetros: Nenhum
 *
 * Retorno: Um nmero aleatrio entre 0.0 e 1.0
 *
 ****/
double DRand(void)
{
   static int primeiraChamada = 1; /* Esta varivel checa se a  */
                                   /* funo est sendo chamada */
                                   /* pela primeira vez e deve  */
                                   /* ter durao fixa          */

      /* Se esta for a primeira chamada da funo, */
      /* alimenta o gerador de nmeros aleatrios  */
   if (primeiraChamada) {
         /* Alimenta o gerador de nmeros aleatrios */
      srand(time(NULL));

         /* A prxima chamada no ser mais a primeira */
      primeiraChamada = 0;
   }

   return (double)rand()/(double)RAND_MAX ;
}

/****
 *
 * IntervaloExponencial(): Retorna um intervalo de tempo usando
 *                         uma distribuio exponencial
 *
 * Parmetros:
 *     intervaloMedio (entrada) - o intervalo mdio decorrido
 *                                entre duas chegadas de clientes
 *
 * Retorno: Um nmero aleatrio de acordo com a descrio
 *
 ****/
double IntervaloExponencial(double intervaloMedio)
{
   return -intervaloMedio*log(1 - DRand());
}

/****
 *
 * ProcessaChegada(): Processa um evento de chegada verificando
 *                    se a lanchonete pode acomodar mais clientes
 *
 * Parmetros:
 *      evento (entrada) - evento de chegada
 *      lanchonete (entrada e sada) - a lanchonete
 *
 * Retorno: 1, se os clientes forem acomodados;
 *          0, se eles desistirem
 *
 * Observao: A lanchonete s atende clientes sentados e no h
 *             fila de espera
 *
 ****/
int ProcessaChegada(const tEvento *evento, tLanchonete *lanchonete)
{
      /* Este deve ser um evento de chegada */
   ASSEGURA( evento->tipoDeEvento == CHEGADA,
             "O evento deveria ser de chegada" );

      /* Descreve o evento na tela */
   ApresentaHora(H_ABERTURA, evento->instante);
   printf( "\t%d %s", evento->nPessoas,
           evento->nPessoas > 1 ? "fregueses chegam"
                                : "fregues chega");

      /* Verifica se os novos clientes ficam ou desistem */
   if (evento->nPessoas < lanchonete->nCadeiras) {
         /* Apresenta na tela o que ocorre */
      printf( " e %s\n", evento->nPessoas > 1 ? "sentam" : "senta");

         /* O nmero de cadeiras disponveis diminui */
      lanchonete->nCadeiras -= evento->nPessoas;

      return 1; /* Clientes ficaram */
   } else {
      printf( ".\n\t\t\tNao ha' espaco e %s embora.\n",
              evento->nPessoas > 1 ? "eles vao" : "ele vai" );

      return 0; /* Clientes foram embora */
   }
}

/****
 *
 * ProcessaPedido(): Atende o pedido de um grupo de clientes
 *
 * Parmetros:
 *      evento (entrada) - evento de pedido
 *      nSanduiches (entrada) - nmero de sanduches que sero
 *                              servidos
 *      lanchonete (entrada e sada) - a lanchonete
 *
 * Retorno: Nada
 *
 ****/
void ProcessaPedido( const tEvento *evento, int nSanduiches,
                     tLanchonete *lanchonete )
{
      /* Este evento deve ser de pedido */
   ASSEGURA( evento->tipoDeEvento == PEDIDO,
             "O evento deveria ser de pedido" );

      /* Informa o que acontece */
   ApresentaHora(H_ABERTURA, evento->instante);
   printf( "\tAtendido pedido de %d %s\n", nSanduiches,
           nSanduiches > 1 ? "sanduiches" : "sanduiche");

      /* Atualiza o faturamento da lanchonete */
   lanchonete->faturamento += PRECO*nSanduiches;
}

/****
 *
 * ProcessaSaida(): Processa um evento de sada
 *
 * Parmetros:
 *      evento (entrada) - evento de sada
 *      lanchonete (entrada e sada) - a lanchonete
 *
 * Retorno: Nada
 *
 ****/
void ProcessaSaida(const tEvento *evento, tLanchonete *lanchonete)
{
      /* Este evento deve ser de sada */
   ASSEGURA( evento->tipoDeEvento == SAIDA,
             "O evento deveria ser de saida" );

      /* Informa o que acontece */
    ApresentaHora(H_ABERTURA, evento->instante);
    printf( "\tGrupo de %d %s sai\n", evento->nPessoas,
            evento->nPessoas > 1 ? "pessoas" : "pessoa" );

         /* O nmero de cadeiras disponveis aumenta */
    lanchonete->nCadeiras += evento->nPessoas;
}

/****
 *
 * DespachaEvento(): Despacha um evento
 *
 * Parmetros:
 *      evento (entrada e sada) - o evento que ser despachado
 *      lanchonete (entrada e sada) - a lanchonete
 *      heap (entrada e sada) - a fila de eventos
 *
 * Retorno: Nada
 *
 ****/
void DespachaEvento( tEvento *evento,
                     tLanchonete *lanchonete, tHeap *heap )
{
   tEvento *novoEvento;
   int      i;

      /* Processa o evento de acordo com seu tipo */
   switch (evento->tipoDeEvento) {
      case CHEGADA:
            /* Verifica se existem cadeiras para o grupo de     */
            /* clientes. Se houver disponibilidade, os clientes */
            /* ficam; caso contrrio, eles desistem.            */
         if (ProcessaChegada(evento, lanchonete)) {
               /* Os clientes foram acomodados. Cria um novo   */
               /* evento de pedido e processa-o imediatamente. */
            novoEvento = CriaEvento( PEDIDO, evento->instante +
                                IntervaloExponencial(ATENDIMENTO),
                                     evento->nPessoas );
            DespachaEvento(novoEvento, lanchonete, heap);
         }

         free(evento); /* Esse evento no  mais necessrio */

         break;

      case SAIDA:
            /* Chama a funo responsvel por    */
            /* processamento de eventos de sada */
         ProcessaSaida(evento, lanchonete);

         free(evento); /* Esse evento no  mais necessrio */

         break;

      case PEDIDO:
            /* Cada pessoa do grupo pede um certo nmero */
            /* de sanduches (entre 1 e MAX_SAND)        */
         for (i = 0; i < evento->nPessoas; i++)
            ProcessaPedido( evento, NAleatorio(1, MAX_SAND),
                            lanchonete );

            /* Cria um evento de sada e insere-o na fila */
         novoEvento = CriaEvento( SAIDA, evento->instante +
                                IntervaloExponencial(PERMANENCIA),
                                  evento->nPessoas);
         InsereEmHeap(heap, novoEvento, ComparaEventos);

         free(evento); /* Esse evento no  mais necessrio */

         break;

      default:
         printf("\nErro indeterminado no programa\n");
         exit(1);
   }
}
