/****
 *
 * Exemplo: BucketSort
 *
 * Seo: 11.4.2
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <time.h>   /* Funes de cronometragem       */
#include <limits.h> /* Limites inteiros               */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL 20

   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL 30

   /* Nmero de coletores */
#define N_COLETORES 8

   /* Nmero de bits mais significativos */
   /* que sero obtidos de cada chave    */
#define N_BITS_SIG  3

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/****************** Definies de Tipos *******************/

   /* Tipo de um coletor, que  uma lista simplesmente encadeada */
typedef struct rotNoLSE {
           unsigned         conteudo;/* Contedo efetivo do n    */
           struct rotNoLSE *proximo; /* Prximo elemento da lista */
        } tNoListaSE, *tBucket;

/************************* Aluses ************************/

extern void BucketSort(unsigned tabela[], unsigned nElem);
extern unsigned* CriaTabela( unsigned *nElementos, unsigned min, unsigned max );
extern void ApresentaTabela( const unsigned *tabela, unsigned nElementos );
extern void LiberaTabela(unsigned **tabela);
extern int EstaOrdenada(unsigned tabela[], unsigned nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * MSBits(): Obtm os bits mais significativos de uma chave do tipo
 *           unsigned char
 *
 * Parmetros:
 *      chave (entrada) - a chave cujos bits sero extrados
 *      bits (entrada) - o nmero de bits mais significativos que
 *                       sero extrados
 *
 * Retorno: O valor inteiros resultante dos bits extrados
 *
 * Observaes:
 *     1. O valor mximo da chave deve ser UCHAR_MAX (i.e.,
 *        o maior valor do tipo unsigned char)
 *
 *     2. O maior valor do nmero de bits mais significativos
 *        deve ser CHAR_BIT (normalmente, 8)
 *
 ****/
unsigned MSBits(unsigned chave, unsigned bits)
{
   ASSEGURA(chave <= UCHAR_MAX, "A chave e' grande demais");
   ASSEGURA(bits <= CHAR_BIT, "Numero de bits e' grande demais");

      /* UCHAR_MAX mascara os primeiros bits de um valor do    */
      /* tipo char, de modo que se o valor da chave for maior  */
      /* do que um valor do tipo char, o restante  descartado */
   return (chave & UCHAR_MAX) >> (CHAR_BIT - bits);
}

/****
 *
 * ExibeLista(): Exibe na tela o contedo do tipo int de cada n
 *               de uma lista encadeada
 *
 * Parmetros:
 *      lista (entrada) - a lista a ser exibida
 *
 * Retorno: Nada
 *
 ****/
void ExibeLista(tBucket lista)
{
   if (!lista) {
      printf("\tVAZIO");
      return;
   }

   printf("\t%d", lista->conteudo);
   lista = lista->proximo;

   while (lista) {
       printf(" -> %d", lista->conteudo);
       lista = lista->proximo;
   }
}

/****
 *
 * ApresentaColetores(): Apresenta na tela o contedo de cada
 *                       coletor usado em BucketSort
 *
 * Parmetros:
 *      coletores (entrada) - array de coletores
 *      nColetores (entrada) - nmero de coletores
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada para depurao
 *
 ****/
void ApresentaColetores(tBucket coletores[], unsigned nColetores)
{
   unsigned i;

   for (i = 0; i < nColetores; ++i) {
      printf("\n\n>>> Coletor %u <<<\n", i);
      ExibeLista(coletores[i]);
   }
}

/****
 *
 * DestroiColetor(): Libera o espao ocupado pelos ns de
 *                   uma lista encadeada, tornando-a vazia
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro para a lista
 *                             simplesmente encadeada que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiColetor(tBucket *lista)
{
   tBucket p; /* Aponta para o prximo n a ser liberado */

   if (!*lista) /* Verifica se a lista est vazia */
      return; /* Lista vazia no precisa ser destruda */

   p = *lista; /* Faz p apontar para o incio da lista */

      /* Visita cada n da lista liberando-o */
   do {
         /* Passa para o prximo n antes que o n corrente */
         /* seja destrudo. No importa que o ponteiro para */
         /*o incio da lista seja alterado. Afinal, este    */
         /* ponteiro ser mesmo anulado.                    */
      *lista = (*lista)->proximo;

      free(p); /* Libera o espao do n corrente */

         /* Faz p apontar para o prximo n, cujo */
         /* endereo est armazenado em *lista    */
      p = *lista;
   } while (p);

      /* Neste ponto, a lista ficou vazia, pois o  */
      /* ltimo valor assumido por *lista foi NULL */
}

/****
 *
 * InsereEmColetor(): Insere um novo elemento num coletor de modo
 *                    que ele permanea ordenado
 *
 * Parmetros:
 *      *coletor (entrada e sada) - o coletor (lista encadeada no
 *                                   qual ocorrer a insero)
 *      conteudo (entrada) - contedo do novo elemento
 *
 * Retorno: Nada
 *
 ****/
void InsereEmColetor(tBucket *coletor, unsigned conteudo)
{
   tBucket p = *coletor,
            q = NULL,
            novoNo;

   ASSEGURA( (novoNo = malloc(sizeof(tNoListaSE))),
             "Impossivel alocar no' de coletor" );

      /* Preenche o contedo do novo n */
   novoNo->conteudo = conteudo;

      /* Se o coletor estiver vazio, insere o n em seu incio */
   if (!*coletor) {
      novoNo->proximo = NULL;
      *coletor = novoNo;
      return;
   }

      /* Encontra o local de insero */
   while (p && p->conteudo <= conteudo) {
      q = p;
      p = p->proximo;
   }

      /* O novo n apontar para o n que contm a primeira  */
      /* chave maior do que a chave recebida como parmetreo */
   novoNo->proximo = p;

   if (!q) /* Insero ser no incio da lista */
      *coletor = novoNo;
   else /* N apontado por q ser o antecessor do novo n */
      q->proximo = novoNo;
}

/****
 *
 * BucketSort(): Ordena uma tabela usando BucketSort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void BucketSort(unsigned tabela[], unsigned nElem)
{
   tBucket  coletores[N_COLETORES]; /* Array de coletores */
   unsigned iColetor, i, j;

      /* Inicia os coletores */
   for (i = 0; i < N_COLETORES; i++)
      coletores[i] = NULL;

      /* Armazena as chaves em ordem nos coletores */
   for (i = 0; i < nElem; i++) {
         /* Obtm o ndice do coletor */
      iColetor = MSBits(tabela[i], N_BITS_SIG);

         /* Insere no devido coletor */
      InsereEmColetor(coletores + iColetor, tabela[i]);
   }

      /* Depurao */
   ApresentaColetores(coletores, N_COLETORES);

      /* Coloca as chaves de volta na tabela */
   for (i = 0, j = 0; i < N_COLETORES; i++) {
      while (coletores[i]) {
         tabela[j++] = coletores[i]->conteudo;
         coletores[i] = coletores[i]->proximo;
      }

         /* Libera o espao ocupado pelo coletor corrente */
      DestroiColetor(coletores + i);
   }
}

/****
 *
 * CriaTabela(): Cria uma tabela (array de inteiros) aleatoriamente
 *
 * Parmetros:
 *     nElementos (sada) - nmero de elementos na tabela criada
 *     min (entrada) - nmero mnimo de elementos na tabela criada
 *     max (entrada) - nmero mximo de elementos na tabela criada
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
unsigned* CriaTabela( unsigned *nElementos, unsigned min,
                      unsigned max )
{
   unsigned *tabela, nElem, i;

      /* Gera nmero de elementos da tabela */
   nElem = rand()%(max - min + 1) + min;

   ASSEGURA( nElem >= min && nElem <= max,
             "Numero de elementos fora do intervalo" );

   tabela = calloc(nElem, sizeof(int));
   ASSEGURA( tabela, "Nao foi possivel alocar a tabela" );

      /* Gera valores dos elementos entre 0 e UCHAR_MAX */
   for (i = 0; i < nElem; ++i)
      tabela[i] = (int) rand()%UCHAR_MAX;

   *nElementos = nElem;

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const unsigned *tabela, unsigned nElementos)
{
   unsigned i;

   printf("\nNumero de elementos: %u", nElementos);

   if (nElementos <= 0)
      return;

   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%u, ", tabela[i]);

   printf("%u}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser
 *                               liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(unsigned **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em
 *                 ordem crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(unsigned tabela[], unsigned nElementos)
{
   unsigned i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando BucketSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   unsigned *tabela; /* Ponteiro para as tabela utilizada */
   unsigned  nElementos;

   srand((int) time(NULL));

   tabela = CriaTabela(&nElementos, MIN_EL, MAX_EL);

   printf("\n\n>>>>>>>>>>>>>>>> BucketSort <<<<<<<<<<<<<<<<\n");

   printf("\n*********** Antes da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);

   BucketSort(tabela, nElementos);

   if (!EstaOrdenada(tabela, nElementos)) {
      printf("\n\n\a>>> Erro: A tabela NAO foi ordenada <<<\n");
      return 1;
   }

   printf("\n\n*********** Depois da Ordenacao ***********\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   LiberaTabela(&tabela);

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>>> BucketSort <<<<<<<<<<<<<<<<

*********** Antes da Ordenacao ************

Numero de elementos: 25
{33, 33, 156, 138, 90, 50, 228, 161, 41, 102, 64, 157, 170, 25, 187, 45, 153, 0, 81, 112, 51, 70, 134, 231, 122}

>>> Coletor 0 <<<
        0 -> 25

>>> Coletor 1 <<<
        33 -> 33 -> 41 -> 45 -> 50 -> 51

>>> Coletor 2 <<<
        64 -> 70 -> 81 -> 90

>>> Coletor 3 <<<
        102 -> 112 -> 122

>>> Coletor 4 <<<
        134 -> 138 -> 153 -> 156 -> 157

>>> Coletor 5 <<<
        161 -> 170 -> 187

>>> Coletor 6 <<<
        VAZIO

>>> Coletor 7 <<<
        228 -> 231

*********** Depois da Ordenacao ***********

Numero de elementos: 25
{0, 25, 33, 33, 41, 45, 50, 51, 64, 70, 81, 90, 102, 112, 122, 134, 138, 153, 156, 157, 161, 170, 187, 228, 231}

**********************************************************/
