/****
 *
 * Ttulo: Ordenao externa de arquivos usando intercalao binria
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2016
 * ltima modificao: 15/05/2018
 *
 * Descrio:
 *
 * Entrada: Um arquivo contendo registros sem ordenao
 *
 * Sada: O arquivo ordenado
 *
 ****/

/************************ Includes ************************/

#include <stdio.h>        /* Entrada e sada     */
#include <string.h>       /* strcpy() e strchr() */
#include <stdlib.h>       /* qsort()             */
#include "IntercalaBin.h" /* Interface deste mdulo */
#include "Fila.h"         /* Fila encadeada */
#include "Registros.h" /* Funes de processamento de registros */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/********************* Funes Locais *********************/

/****
 *
 * Compara(): Compara dois registros
 *
 * Parmetros:
 *      e1, e2 (entrada) - os dois registros
 *
 * Retorno:
 *     * 0, se os registros so iguais
 *     * > 0, se o primeiro registro for maior do que o segundo
 *     * < 0, se o primeiro registro for menor do que o segundo
 *
 ****/
static int Compara(const void *e1, const void* e2)
{
   tRegistro *r1 = (tRegistro *)e1,
             *r2 = (tRegistro *)e2;

   return r1->chave - r2->chave;
}

/****
 *
 * CopiaRestoArquivo(): Copia contedo de um arquivo para
 *                      outro byte a byte a partir da posio
 *                      em que se encontra o apontador de posio
 *                      de cada arquivo
 *
 * Parmetros:
 *      streamEntrada (entrada) - stream associado ao arquivo
 *                                que ser copiado
 *      streamSaida (entrada) - stream associado ao arquivo
 *                              que receber a cpia
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. O stream que ser lido deve estar aberto num modo
 *        que permita leitura e o stream que ser escrito
 *        deve estar aberto num modo que permite escrita.
 *
 *     2. O apontador de posio do arquivo lido no  movido
 *        para o incio do arquivo, de modo que o primeiro byte
 *        a ser copiado  aquele indicado pela posio corrente
 *        desse apontador quando esta funo  chamada.
 *
 ****/
static void CopiaRestoArquivo( FILE *streamEntrada,
                               FILE *streamSaida )
{
   int c; /* Armazenar cada byte lido e escrito  */

      /* O lao encerra quando houver tentativa de leitura   */
      /* alm do final do arquivo de entrada ou ocorrer erro */
      /* de leitura ou escrita em qualquer dos arquivos      */
   while (1) {
         /* L um byte no arquivo de entrada */
      c = fgetc(streamEntrada);

         /* Testa se final do arquivo de entrada foi */
         /* atingido ou ocorreu erro de leitura      */
      if (feof(streamEntrada) || ferror(streamEntrada)) {
         break; /* Processamento encerrado */
      }

         /* Escreve o byte lido no arquivo de sada */
      fputc(c, streamSaida);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamSaida)) {
         break;
      }
   }

      /* Certifica que no houve erro de leitura ou escrita */
   ASSEGURA(!ferror(streamEntrada),
            "Erro de leitura ao tentar copiar arquivo");
   ASSEGURA(!ferror(streamSaida),
            "Erro de escrita ao tentar copiar arquivo");
}

/****
 *
 * IntercalaDoisArquivos(): Intercala dois arquivos e armazena
 *                          o resultado num terceiro arquivo
 *
 * Parmetros:
 *      fila (sada) - fila na qual sero armazenadas
 *                      informaes sobre o arquivo que
 *                      contm o resultado da intercalao
 *      arq1, arq2 (entrada) - nomes e streams associados aos
 *                             arquivos que sero intercalados
 *
 * Retorno: Nada
 *
 ****/
static void IntercalaDoisArquivos( tFila *fila, tArquivo arq1,
                                   tArquivo arq2 )
{
   tArquivo  resultado; /* Nome e stream associados ao    */
                        /* arquivo que contm o resultado */
                        /* da intercalao                */
   tRegistro reg1, /* Um registro do primeiro arquivo */
             reg2; /* Um registro do segundo arquivo  */
   int       le1 = 1, /* Indica se um registro deve */
                      /* ser lido do arquivo 1      */
             le2 = 1; /* Indica se um registro deve */
                      /* ser lido do arquivo 2      */

      /* Abre os arquivos que sero intercalados */
   arq1.stream = AbreArquivo(arq1.nome, "rb");
   arq2.stream = AbreArquivo(arq2.nome, "rb");

      /* Cria nome do arquivo que conter */
      /* o resultado da intercalao      */
   ASSEGURA( NomeTemp(resultado.nome, INI_ARQ_TMP, MAX_ARQ_TMP),
             "Impossivel criar nome de arquivo" );

#if DEPURACAO == 1
   printf("\n\t>>> Intercalando arquivos \"%s\" e \"%s\" no "
          "arquivo \"%s\"", arq1.nome, arq2.nome, resultado.nome);
#endif

      /* Tenta abrir o arquivo que conter o resultado */
   resultado.stream = AbreArquivo(resultado.nome, "wb");

      /* Intercala os arquivos */
   while (1) {
      if (le1) {
            /* L um registro do primeiro arquivo */
            /* e checa se ocorreu erro            */
         fread(&reg1, sizeof(reg1), 1, arq1.stream);
         ASSEGURA( !ferror(arq1.stream), "Erro de leitura" );
      }

      if (le2) {
            /* L um registro do segundo arquivo */
            /* e checa se ocorreu erro           */
         fread(&reg2, sizeof(reg2), 1, arq2.stream);
         ASSEGURA( !ferror(arq2.stream), "Erro de leitura" );
      }

         /* Se o primeiro arquivo foi exaurido, copia os */
         /* registros restantes do segundo arquivo para  */
         /* o arquivo resultante e encerra o lao        */
      if (feof(arq1.stream)) {
         CopiaRestoArquivo(arq2.stream, resultado.stream);
         break;
      }

         /* Se o segundo arquivo foi exaurido, copia os  */
         /* registros restantes do primeiro arquivo para */
         /* o arquivo resultante e encerra o lao        */
      if (feof(arq2.stream)) {
         CopiaRestoArquivo(arq1.stream, resultado.stream);
         break;
      }

         /* Copia o menor dos dois registros para */
         /* o arquivo resultante da intercalao  */
      if (Compara(&reg1, &reg2) < 0) {
         fwrite(&reg1, sizeof(reg1), 1, resultado.stream);
         le1 = 1; /* L o prximo registro do arquivo 1 */
         le2 = 0; /* No l o prximo registro do arquivo 2 */
      } else {
         fwrite(&reg2, sizeof(reg2), 1, resultado.stream);
         le1 = 0; /* No l o prximo registro do arquivo 1 */
         le2 = 1; /* L o prximo registro do arquivo 2 */
      }
   }

      /* Fecha todos os arquivos */
   FechaArquivo(resultado.stream, resultado.nome);
   FechaArquivo(arq1.stream, arq1.nome);
   FechaArquivo(arq2.stream, arq2.nome);

      /* Apaga os dois arquivos que foram intercalados, */
      /* pois eles no so mais necessrios             */
   ASSEGURA(!remove(arq1.nome),
            "Impossivel remover arquivo 1");
   ASSEGURA(!remove(arq2.nome),
            "Impossivel remover arquivo 2");

      /* Acrescenta o resultado da intercalao */
   Acrescenta(fila, &resultado);
}

/********************* Funes Globais ********************/

/****
 *
 * CriaSeriesOrdenadas(): Cria sries de arquivos ordenados
 *
 * Parmetros:
 *      nomeArq (entrada) - nome do arquivo que origina as sries
 *
 * Retorno: Fila contendo informaes sobre os arquivos
 *          que contm as sries
 *
 ****/
tFila CriaSeriesOrdenadas(const char *nomeArq)
{
   tFila    fila; /* Fila contendo informaes sobre */
                  /* os arquivos que contm as sries */
   tRegistro *registros; /* Ponteiro para um array que armazena */
                         /* uma srie em memria principal *    */
   tArquivo  arqTmp; /* Nome e stream associados */
                     /* a um arquivo temporrio  */
   int       nRegistros; /* Nmero de registros numa srie */
   FILE     *stream; /* stream associado ao arquivo */
                     /* que origina as sries       */

      /* Inicia a fila */
   CriaFila(&fila);

      /* Aloca espao para o array. Um array desse */
      /* tamanho no pode ser alocado na pilha     */
   registros = calloc(MAX_REGISTROS, sizeof(*registros));
   ASSEGURA( registros, "Buffer nao foi alocado" );

      /* Tenta abrir o arquivo de dados */
   stream = AbreArquivo(nomeArq, "rb");

      /* L o arquivo a ser ordenado e cria as sries */
   while (1) {
         /* L os registros da srie corrente */
      nRegistros = fread( registros, sizeof(*registros),
                          MAX_REGISTROS, stream );

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA( !ferror(stream), "Erro de leitura em arquivo" );

         /* Encerra o lao se nenhum registro foi lido */
      if (!nRegistros)
         break;

         /* Ordena a srie */
      qsort(registros, nRegistros, sizeof(*registros), Compara);

         /* Cria nome do arquivo que conter a srie corrente */
      ASSEGURA(NomeTemp(arqTmp.nome, INI_ARQ_TMP, MAX_ARQ_TMP),
               "Impossivel criar nome de arquivo temporario");

#if DEPURACAO == 1
      printf("\n\t>>> Criando serie \"%s\"...", arqTmp.nome);
#endif

         /* Tenta abrir o arquivo que conter a srie */
      arqTmp.stream = AbreArquivo(arqTmp.nome, "wb");

         /* Escreve os registros ordenados */
         /* da srie corrente no arquivo   */
      fwrite( registros, sizeof(*registros), nRegistros,
              arqTmp.stream );

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA( !ferror(arqTmp.stream),
                "Erro de escrita em arquivo temporario" );

         /* O nmero de arquivos abertos pode ser */
         /* maior do que FOPEN_MAX, de modo que  */
         /* necessrio fechar os arquivos         */
      FechaArquivo(arqTmp.stream, arqTmp.nome);

         /* Acrescenta na fila informaes sobre  */
         /* o arquivo que contm a srie corrente */
      Acrescenta(&fila, &arqTmp);
   }

      /* Fecha o arquivo original */
   FechaArquivo(stream, nomeArq);

      /* Libera o espao ocupado pelo array */
   free(registros);

   return fila;
}

/****
 *
 * IntercalaBin(): Intercala sries de arquivos ordenados cujas
 *                 informaes esto armazenadas numa fila
 *
 * Parmetros:
 *      fila (entrada/sada) - fila contendo informaes
 *                             sobre as sries
 *      arqNome (entrada) - nome do arquivo que originou as
 *                          sries
 *
 * Retorno: Nada
 *
 ****/
void IntercalaBin(tFila *fila, const char *arqNome)
{
   tArquivo arq1, arq2;

      /* Se a fila estiver vazia, no h o que fazer */
   if (FilaVazia(fila))
      return;

      /* Intercala os arquivos que se encontram */
      /* na fila at esvazi-la                 */
   while (1) {
         /* Obtm o primeiro arquivo a ser intercalado */
      arq1 = Retira(fila);

         /* Se a fila ficou vazia, no h mais */
         /* intercalaes a serem feitas       */
      if (FilaVazia(fila)) {
            /* Rebatiza o ltimo arquivo temporrio na */
            /* fila com o nome do arquivo original    */
         ASSEGURA( !rename(arq1.nome, arqNome),
                   "Impossivel rebatizar arquivo" );

         return; /* Servio concludo */
      }

         /* Obtm o segundo arquivo a ser intercalado */
      arq2 = Retira(fila);

         /* Intercala os dois arquivos desempilhados */
      IntercalaDoisArquivos(fila, arq1, arq2);
   }
}
