/****
 *
 * Ttulo: Ordenao externa de arquivos por intercalao binria
 *
 * Seo: 12.2.2
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2017
 * ltima modificao: 16/04/2017
 *
 * Descrio: Demonstra ordenao externa de arquivos por
 *            intercalao binria
 *
 * Entrada: Arquivo "CensoMEC2012.bin"
 *
 * Sada: Arquivo "CensoMEC2012_Ord.bin"
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>        /* Funes de Entrada/Sada */
#include <string.h>       /* Processamento de strings */
#include <stdlib.h>       /* Funes de alocao dinmica e */
                          /* gerao de nmeros aleatrios  */
#include <time.h>         /* Funo time() para alimentar o */
                          /* gerador de nmeros aleatrios  */
#include "IntercalaBin.h" /* Interface do mdulo de ordenao */
#include "Fila.h"         /* Interface do tipo tFila  */
#include "Registros.h" /* Funes de processamento de registros */
#include "Leitura.h"   /* Funes de leitura via teclado   */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/***************** Definies de Funes ******************/

/****
 *
 * CriaArquivoRegistros(): Constri o arquivo binrio de registros
 *
 * Parmetros:
 *     nomeArqTexto (entrada) - nome do arquivo texto de registros
 *     nomeArqBin (entrada) - nome do arquivo binrio de registros
 *
 * Retorno: Stream associado ao arquivo binrio de registros
 *
 ****/
void CriaArquivoRegistros( const char *nomeArqTexto,
                           const char *nomeArqBin )
{
   int nRegistros, /* Nmero de registros no arquivo */
       maxRegistros; /* Nmero mximo de registros no arquivo */

   printf("\n\t>>> Digite o numero maximo de registros (0 = TODOS): ");
   maxRegistros = LeNatural();

   printf("\nCriando arquivo...\n");
   fflush(stdout);

      /* Cria o arquivo binrio */
   nRegistros = CriaArquivoBin( nomeArqTexto, nomeArqBin,
                                maxRegistros );

      /* Apresenta o resultado da converso */
   printf( "\n\t>>> Foram armazenados %d registros no arquivo"
           "\n\t>>> \"%s\"\n", nRegistros, nomeArqBin );
}

/****
 *
 * TestaOrdenacao(): Verifica se um arquivo est ordenado em ordem
 *                   crescente
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que ser
 *                         testado
 *
 * Retorno: 1, se o arquivo estiver ordenado; 0, em caso contrrio
 *
 ****/
int TestaOrdenacao(const char *nomeArq)
{
   tRegistro registroAtual, registroAnterior;
   int       teste;
   FILE     *stream;

   stream = AbreArquivo(nomeArq, "rb");

      /* L o primeiro registro do arquivo */
   fread(&registroAnterior, sizeof(registroAnterior), 1, stream);
   ASSEGURA( !ferror(stream), "Erro de leitura" );

      /* L os demais registros e testa se eles esto ordenados */
   while (1) {
         /* L o prximo registro */
      fread(&registroAtual, sizeof(registroAtual), 1, stream);
      ASSEGURA( !ferror(stream), "Erro de leitura" );

      if (feof(stream))
         break; /* Leitura acabou */

         /* Compara a chave do registro atual */
         /* com a chave do registro anterior  */
      teste = registroAtual.chave - registroAnterior.chave;

         /* Verifica se as duas ltimas chaves */
         /* esto em ordem crescente           */
      if (teste < 0) {
         FechaArquivo(stream, nomeArq);
         return 0;
      }

         /* A chave anterior passa a ser a chave corrente */
      memcpy( &registroAnterior, &registroAtual,
              sizeof(registroAtual) );
   }

   FechaArquivo(stream, nomeArq);

      /* Se ainda no houve retorno, o arquivo est ordenado */
   return 1;
}

/****
 *
 * main(): Efetua ordenao externa de arquivos usando intercalao
 *         binria
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tFila fila;
   int   teste;

      /* Cria o arquivo binrio de registro se ele no for encontrado */
   if (!ArquivoExiste(NOME_ARQUIVO_BIN)) {
      MedidaDeTempo("Criando arquivo de registros...");

      CriaArquivoRegistros( NOME_ARQUIVO_REGS, NOME_ARQUIVO_BIN );

      MedidaDeTempo("...Arquivo criado");
   }

   MedidaDeTempo("\nCriando series ordenadas...\n");

      /* L o arquivo original e cria as sries */
   fila = CriaSeriesOrdenadas(NOME_ARQUIVO_BIN);

   MedidaDeTempo("\n\n...series criadas com sucesso");

   MedidaDeTempo("\nIntercalando series...\n");

   IntercalaBin(&fila, NOME_ARQUIVO_ORD);

   MedidaDeTempo("\n\n...series intercaladas com sucesso");

   MedidaDeTempo("Testando ordenacao do arquivo...");

   teste = TestaOrdenacao(NOME_ARQUIVO_ORD);

   if (teste)
      MedidaDeTempo("...Ordenacao esta' OK\n");
   else
      MedidaDeTempo("\a...O arquivo nao esta ordenado\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 600000

Criando arquivo...

        >>> Foram armazenados 600000 registros no arquivo
        >>> "../CensoMEC2012.bin"

>>> ...Arquivo criado

        *** Tempo gasto na operacao: 11.00 segundos ***

>>> Criando series ordenadas...

        >>> Criando serie "_tmp001"...
        >>> Criando serie "_tmp002"...
        >>> Criando serie "_tmp003"...
        >>> Criando serie "_tmp004"...
        >>> Criando serie "_tmp005"...
        >>> Criando serie "_tmp006"...
        >>> Criando serie "_tmp007"...
        >>> Criando serie "_tmp008"...

>>>...series criadas com sucesso

        *** Tempo gasto na operacao:  3.00 segundos ***

>>> Intercalando series...

        >>> Intercalando arquivos "_tmp001" e "_tmp002" no arquivo "_tmp009"
        >>> Intercalando arquivos "_tmp003" e "_tmp004" no arquivo "_tmp010"
        >>> Intercalando arquivos "_tmp005" e "_tmp006" no arquivo "_tmp011"
        >>> Intercalando arquivos "_tmp007" e "_tmp008" no arquivo "_tmp012"
        >>> Intercalando arquivos "_tmp009" e "_tmp010" no arquivo "_tmp013"
        >>> Intercalando arquivos "_tmp011" e "_tmp012" no arquivo "_tmp014"
        >>> Intercalando arquivos "_tmp013" e "_tmp014" no arquivo "_tmp015"

>>> ...series intercaladas com sucesso

        *** Tempo gasto na operacao: 26.00 segundos ***

>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK

        *** Tempo gasto na operacao:  3.00 segundos ***


Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 1200000

Criando arquivo...

        >>> Foram armazenados 1200000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao: 29.00 segundos ***


>>>
Criando series ordenadas...


        >>> Criando serie "_tmp001"...
        >>> Criando serie "_tmp002"...
        >>> Criando serie "_tmp003"...
        >>> Criando serie "_tmp004"...
        >>> Criando serie "_tmp005"...
        >>> Criando serie "_tmp006"...
        >>> Criando serie "_tmp007"...
        >>> Criando serie "_tmp008"...
        >>> Criando serie "_tmp009"...
        >>> Criando serie "_tmp010"...
        >>> Criando serie "_tmp011"...
        >>> Criando serie "_tmp012"...
        >>> Criando serie "_tmp013"...
        >>> Criando serie "_tmp014"...
        >>> Criando serie "_tmp015"...

>>>

...series criadas com sucesso

        *** Tempo gasto na operacao:  9.00 segundos ***


>>>
Intercalando series...


        >>> Intercalando arquivos "_tmp001" e "_tmp002" no arquivo "_tmp016"
        >>> Intercalando arquivos "_tmp003" e "_tmp004" no arquivo "_tmp017"
        >>> Intercalando arquivos "_tmp005" e "_tmp006" no arquivo "_tmp018"
        >>> Intercalando arquivos "_tmp007" e "_tmp008" no arquivo "_tmp019"
        >>> Intercalando arquivos "_tmp009" e "_tmp010" no arquivo "_tmp020"
        >>> Intercalando arquivos "_tmp011" e "_tmp012" no arquivo "_tmp021"
        >>> Intercalando arquivos "_tmp013" e "_tmp014" no arquivo "_tmp022"
        >>> Intercalando arquivos "_tmp015" e "_tmp016" no arquivo "_tmp023"
        >>> Intercalando arquivos "_tmp017" e "_tmp018" no arquivo "_tmp024"
        >>> Intercalando arquivos "_tmp019" e "_tmp020" no arquivo "_tmp025"
        >>> Intercalando arquivos "_tmp021" e "_tmp022" no arquivo "_tmp026"
        >>> Intercalando arquivos "_tmp023" e "_tmp024" no arquivo "_tmp027"
        >>> Intercalando arquivos "_tmp025" e "_tmp026" no arquivo "_tmp028"
        >>> Intercalando arquivos "_tmp027" e "_tmp028" no arquivo "_tmp029"

>>>

...series intercaladas com sucesso

        *** Tempo gasto na operacao:  1.13 minutos ***


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao:  6.00 segundos ***


>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 2400000

Criando arquivo...

        >>> Foram armazenados 2400000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao: 39.00 segundos ***


>>>
Criando series ordenadas...


        >>> Criando serie "_tmp001"...
        >>> Criando serie "_tmp002"...
        >>> Criando serie "_tmp003"...
        >>> Criando serie "_tmp004"...
        >>> Criando serie "_tmp005"...
        >>> Criando serie "_tmp006"...
        >>> Criando serie "_tmp007"...
        >>> Criando serie "_tmp008"...
        >>> Criando serie "_tmp009"...
        >>> Criando serie "_tmp010"...
        >>> Criando serie "_tmp011"...
        >>> Criando serie "_tmp012"...
        >>> Criando serie "_tmp013"...
        >>> Criando serie "_tmp014"...
        >>> Criando serie "_tmp015"...
        >>> Criando serie "_tmp016"...
        >>> Criando serie "_tmp017"...
        >>> Criando serie "_tmp018"...
        >>> Criando serie "_tmp019"...
        >>> Criando serie "_tmp020"...
        >>> Criando serie "_tmp021"...
        >>> Criando serie "_tmp022"...
        >>> Criando serie "_tmp023"...
        >>> Criando serie "_tmp024"...
        >>> Criando serie "_tmp025"...
        >>> Criando serie "_tmp026"...
        >>> Criando serie "_tmp027"...
        >>> Criando serie "_tmp028"...
        >>> Criando serie "_tmp029"...
        >>> Criando serie "_tmp030"...

>>>

...series criadas com sucesso

        *** Tempo gasto na operacao: 20.00 segundos ***


>>>
Intercalando series...

        >>> Intercalando arquivos "_tmp001" e "_tmp002" no arquivo "_tmp031"
        >>> Intercalando arquivos "_tmp003" e "_tmp004" no arquivo "_tmp032"
        >>> Intercalando arquivos "_tmp005" e "_tmp006" no arquivo "_tmp033"
        >>> Intercalando arquivos "_tmp007" e "_tmp008" no arquivo "_tmp034"
        >>> Intercalando arquivos "_tmp009" e "_tmp010" no arquivo "_tmp035"
        >>> Intercalando arquivos "_tmp011" e "_tmp012" no arquivo "_tmp036"
        >>> Intercalando arquivos "_tmp013" e "_tmp014" no arquivo "_tmp037"
        >>> Intercalando arquivos "_tmp015" e "_tmp016" no arquivo "_tmp038"
        >>> Intercalando arquivos "_tmp017" e "_tmp018" no arquivo "_tmp039"
        >>> Intercalando arquivos "_tmp019" e "_tmp020" no arquivo "_tmp040"
        >>> Intercalando arquivos "_tmp021" e "_tmp022" no arquivo "_tmp041"
        >>> Intercalando arquivos "_tmp023" e "_tmp024" no arquivo "_tmp042"
        >>> Intercalando arquivos "_tmp025" e "_tmp026" no arquivo "_tmp043"
        >>> Intercalando arquivos "_tmp027" e "_tmp028" no arquivo "_tmp044"
        >>> Intercalando arquivos "_tmp029" e "_tmp030" no arquivo "_tmp045"
        >>> Intercalando arquivos "_tmp031" e "_tmp032" no arquivo "_tmp046"
        >>> Intercalando arquivos "_tmp033" e "_tmp034" no arquivo "_tmp047"
        >>> Intercalando arquivos "_tmp035" e "_tmp036" no arquivo "_tmp048"
        >>> Intercalando arquivos "_tmp037" e "_tmp038" no arquivo "_tmp049"
        >>> Intercalando arquivos "_tmp039" e "_tmp040" no arquivo "_tmp050"
        >>> Intercalando arquivos "_tmp041" e "_tmp042" no arquivo "_tmp051"
        >>> Intercalando arquivos "_tmp043" e "_tmp044" no arquivo "_tmp052"
        >>> Intercalando arquivos "_tmp045" e "_tmp046" no arquivo "_tmp053"
        >>> Intercalando arquivos "_tmp047" e "_tmp048" no arquivo "_tmp054"
        >>> Intercalando arquivos "_tmp049" e "_tmp050" no arquivo "_tmp055"
        >>> Intercalando arquivos "_tmp051" e "_tmp052" no arquivo "_tmp056"
        >>> Intercalando arquivos "_tmp053" e "_tmp054" no arquivo "_tmp057"
        >>> Intercalando arquivos "_tmp055" e "_tmp056" no arquivo "_tmp058"
        >>> Intercalando arquivos "_tmp057" e "_tmp058" no arquivo "_tmp059"

>>>

...series intercaladas com sucesso

        *** Tempo gasto na operacao:  3.02 minutos ***


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao: 12.00 segundos ***


>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 2400000

Criando arquivo...

        >>> Foram armazenados 2400000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao: 39.00 segundos ***


>>>
Criando series ordenadas...


        >>> Criando serie "_tmp001"...
        >>> Criando serie "_tmp002"...
        >>> Criando serie "_tmp003"...
        >>> Criando serie "_tmp004"...
        >>> Criando serie "_tmp005"...
        >>> Criando serie "_tmp006"...
        >>> Criando serie "_tmp007"...
        >>> Criando serie "_tmp008"...
        >>> Criando serie "_tmp009"...
        >>> Criando serie "_tmp010"...
        >>> Criando serie "_tmp011"...
        >>> Criando serie "_tmp012"...
        >>> Criando serie "_tmp013"...
        >>> Criando serie "_tmp014"...
        >>> Criando serie "_tmp015"...
        >>> Criando serie "_tmp016"...
        >>> Criando serie "_tmp017"...
        >>> Criando serie "_tmp018"...
        >>> Criando serie "_tmp019"...
        >>> Criando serie "_tmp020"...
        >>> Criando serie "_tmp021"...
        >>> Criando serie "_tmp022"...
        >>> Criando serie "_tmp023"...
        >>> Criando serie "_tmp024"...
        >>> Criando serie "_tmp025"...
        >>> Criando serie "_tmp026"...
        >>> Criando serie "_tmp027"...
        >>> Criando serie "_tmp028"...
        >>> Criando serie "_tmp029"...
        >>> Criando serie "_tmp030"...

>>>

...series criadas com sucesso

        *** Tempo gasto na operacao: 20.00 segundos ***


>>>
Intercalando series...


        >>> Intercalando arquivos "_tmp001" e "_tmp002" no arquivo "_tmp031"
        >>> Intercalando arquivos "_tmp003" e "_tmp004" no arquivo "_tmp032"
        >>> Intercalando arquivos "_tmp005" e "_tmp006" no arquivo "_tmp033"
        >>> Intercalando arquivos "_tmp007" e "_tmp008" no arquivo "_tmp034"
        >>> Intercalando arquivos "_tmp009" e "_tmp010" no arquivo "_tmp035"
        >>> Intercalando arquivos "_tmp011" e "_tmp012" no arquivo "_tmp036"
        >>> Intercalando arquivos "_tmp013" e "_tmp014" no arquivo "_tmp037"
        >>> Intercalando arquivos "_tmp015" e "_tmp016" no arquivo "_tmp038"
        >>> Intercalando arquivos "_tmp017" e "_tmp018" no arquivo "_tmp039"
        >>> Intercalando arquivos "_tmp019" e "_tmp020" no arquivo "_tmp040"
        >>> Intercalando arquivos "_tmp021" e "_tmp022" no arquivo "_tmp041"
        >>> Intercalando arquivos "_tmp023" e "_tmp024" no arquivo "_tmp042"
        >>> Intercalando arquivos "_tmp025" e "_tmp026" no arquivo "_tmp043"
        >>> Intercalando arquivos "_tmp027" e "_tmp028" no arquivo "_tmp044"
        >>> Intercalando arquivos "_tmp029" e "_tmp030" no arquivo "_tmp045"
        >>> Intercalando arquivos "_tmp031" e "_tmp032" no arquivo "_tmp046"
        >>> Intercalando arquivos "_tmp033" e "_tmp034" no arquivo "_tmp047"
        >>> Intercalando arquivos "_tmp035" e "_tmp036" no arquivo "_tmp048"
        >>> Intercalando arquivos "_tmp037" e "_tmp038" no arquivo "_tmp049"
        >>> Intercalando arquivos "_tmp039" e "_tmp040" no arquivo "_tmp050"
        >>> Intercalando arquivos "_tmp041" e "_tmp042" no arquivo "_tmp051"
        >>> Intercalando arquivos "_tmp043" e "_tmp044" no arquivo "_tmp052"
        >>> Intercalando arquivos "_tmp045" e "_tmp046" no arquivo "_tmp053"
        >>> Intercalando arquivos "_tmp047" e "_tmp048" no arquivo "_tmp054"
        >>> Intercalando arquivos "_tmp049" e "_tmp050" no arquivo "_tmp055"
        >>> Intercalando arquivos "_tmp051" e "_tmp052" no arquivo "_tmp056"
        >>> Intercalando arquivos "_tmp053" e "_tmp054" no arquivo "_tmp057"
        >>> Intercalando arquivos "_tmp055" e "_tmp056" no arquivo "_tmp058"
        >>> Intercalando arquivos "_tmp057" e "_tmp058" no arquivo "_tmp059"

>>>

...series intercaladas com sucesso

        *** Tempo gasto na operacao:  3.02 minutos ***


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao: 12.00 segundos ***
uly@DESKTOP-TJNTIGG:~/testes/Secao12_2_2_ALT> ./Secao12_2_2_ALT.exe


>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 4800000

Criando arquivo...

        >>> Foram armazenados 4800000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao:  1.28 minutos ***


>>>
Criando series ordenadas...


        >>> Criando serie "_tmp001"...
        >>> Criando serie "_tmp002"...
        >>> Criando serie "_tmp003"...
        >>> Criando serie "_tmp004"...
        >>> Criando serie "_tmp005"...
        >>> Criando serie "_tmp006"...
        >>> Criando serie "_tmp007"...
        >>> Criando serie "_tmp008"...
        >>> Criando serie "_tmp009"...
        >>> Criando serie "_tmp010"...
        >>> Criando serie "_tmp011"...
        >>> Criando serie "_tmp012"...
        >>> Criando serie "_tmp013"...
        >>> Criando serie "_tmp014"...
        >>> Criando serie "_tmp015"...
        >>> Criando serie "_tmp016"...
        >>> Criando serie "_tmp017"...
        >>> Criando serie "_tmp018"...
        >>> Criando serie "_tmp019"...
        >>> Criando serie "_tmp020"...
        >>> Criando serie "_tmp021"...
        >>> Criando serie "_tmp022"...
        >>> Criando serie "_tmp023"...
        >>> Criando serie "_tmp024"...
        >>> Criando serie "_tmp025"...
        >>> Criando serie "_tmp026"...
        >>> Criando serie "_tmp027"...
        >>> Criando serie "_tmp028"...
        >>> Criando serie "_tmp029"...
        >>> Criando serie "_tmp030"...
        >>> Criando serie "_tmp031"...
        >>> Criando serie "_tmp032"...
        >>> Criando serie "_tmp033"...
        >>> Criando serie "_tmp034"...
        >>> Criando serie "_tmp035"...
        >>> Criando serie "_tmp036"...
        >>> Criando serie "_tmp037"...
        >>> Criando serie "_tmp038"...
        >>> Criando serie "_tmp039"...
        >>> Criando serie "_tmp040"...
        >>> Criando serie "_tmp041"...
        >>> Criando serie "_tmp042"...
        >>> Criando serie "_tmp043"...
        >>> Criando serie "_tmp044"...
        >>> Criando serie "_tmp045"...
        >>> Criando serie "_tmp046"...
        >>> Criando serie "_tmp047"...
        >>> Criando serie "_tmp048"...
        >>> Criando serie "_tmp049"...
        >>> Criando serie "_tmp050"...
        >>> Criando serie "_tmp051"...
        >>> Criando serie "_tmp052"...
        >>> Criando serie "_tmp053"...
        >>> Criando serie "_tmp054"...
        >>> Criando serie "_tmp055"...
        >>> Criando serie "_tmp056"...
        >>> Criando serie "_tmp057"...
        >>> Criando serie "_tmp058"...
        >>> Criando serie "_tmp059"...
        >>> Criando serie "_tmp060"...

>>>

...series criadas com sucesso

        *** Tempo gasto na operacao: 37.00 segundos ***


>>>
Intercalando series...


        >>> Intercalando arquivos "_tmp001" e "_tmp002" no arquivo "_tmp061"
        >>> Intercalando arquivos "_tmp003" e "_tmp004" no arquivo "_tmp062"
        >>> Intercalando arquivos "_tmp005" e "_tmp006" no arquivo "_tmp063"
        >>> Intercalando arquivos "_tmp007" e "_tmp008" no arquivo "_tmp064"
        >>> Intercalando arquivos "_tmp009" e "_tmp010" no arquivo "_tmp065"
        >>> Intercalando arquivos "_tmp011" e "_tmp012" no arquivo "_tmp066"
        >>> Intercalando arquivos "_tmp013" e "_tmp014" no arquivo "_tmp067"
        >>> Intercalando arquivos "_tmp015" e "_tmp016" no arquivo "_tmp068"
        >>> Intercalando arquivos "_tmp017" e "_tmp018" no arquivo "_tmp069"
        >>> Intercalando arquivos "_tmp019" e "_tmp020" no arquivo "_tmp070"
        >>> Intercalando arquivos "_tmp021" e "_tmp022" no arquivo "_tmp071"
        >>> Intercalando arquivos "_tmp023" e "_tmp024" no arquivo "_tmp072"
        >>> Intercalando arquivos "_tmp025" e "_tmp026" no arquivo "_tmp073"
        >>> Intercalando arquivos "_tmp027" e "_tmp028" no arquivo "_tmp074"
        >>> Intercalando arquivos "_tmp029" e "_tmp030" no arquivo "_tmp075"
        >>> Intercalando arquivos "_tmp031" e "_tmp032" no arquivo "_tmp076"
        >>> Intercalando arquivos "_tmp033" e "_tmp034" no arquivo "_tmp077"
        >>> Intercalando arquivos "_tmp035" e "_tmp036" no arquivo "_tmp078"
        >>> Intercalando arquivos "_tmp037" e "_tmp038" no arquivo "_tmp079"
        >>> Intercalando arquivos "_tmp039" e "_tmp040" no arquivo "_tmp080"
        >>> Intercalando arquivos "_tmp041" e "_tmp042" no arquivo "_tmp081"
        >>> Intercalando arquivos "_tmp043" e "_tmp044" no arquivo "_tmp082"
        >>> Intercalando arquivos "_tmp045" e "_tmp046" no arquivo "_tmp083"
        >>> Intercalando arquivos "_tmp047" e "_tmp048" no arquivo "_tmp084"
        >>> Intercalando arquivos "_tmp049" e "_tmp050" no arquivo "_tmp085"
        >>> Intercalando arquivos "_tmp051" e "_tmp052" no arquivo "_tmp086"
        >>> Intercalando arquivos "_tmp053" e "_tmp054" no arquivo "_tmp087"
        >>> Intercalando arquivos "_tmp055" e "_tmp056" no arquivo "_tmp088"
        >>> Intercalando arquivos "_tmp057" e "_tmp058" no arquivo "_tmp089"
        >>> Intercalando arquivos "_tmp059" e "_tmp060" no arquivo "_tmp090"
        >>> Intercalando arquivos "_tmp061" e "_tmp062" no arquivo "_tmp091"
        >>> Intercalando arquivos "_tmp063" e "_tmp064" no arquivo "_tmp092"
        >>> Intercalando arquivos "_tmp065" e "_tmp066" no arquivo "_tmp093"
        >>> Intercalando arquivos "_tmp067" e "_tmp068" no arquivo "_tmp094"
        >>> Intercalando arquivos "_tmp069" e "_tmp070" no arquivo "_tmp095"
        >>> Intercalando arquivos "_tmp071" e "_tmp072" no arquivo "_tmp096"
        >>> Intercalando arquivos "_tmp073" e "_tmp074" no arquivo "_tmp097"
        >>> Intercalando arquivos "_tmp075" e "_tmp076" no arquivo "_tmp098"
        >>> Intercalando arquivos "_tmp077" e "_tmp078" no arquivo "_tmp099"
        >>> Intercalando arquivos "_tmp079" e "_tmp080" no arquivo "_tmp100"
        >>> Intercalando arquivos "_tmp081" e "_tmp082" no arquivo "_tmp101"
        >>> Intercalando arquivos "_tmp083" e "_tmp084" no arquivo "_tmp102"
        >>> Intercalando arquivos "_tmp085" e "_tmp086" no arquivo "_tmp103"
        >>> Intercalando arquivos "_tmp087" e "_tmp088" no arquivo "_tmp104"
        >>> Intercalando arquivos "_tmp089" e "_tmp090" no arquivo "_tmp105"
        >>> Intercalando arquivos "_tmp091" e "_tmp092" no arquivo "_tmp106"
        >>> Intercalando arquivos "_tmp093" e "_tmp094" no arquivo "_tmp107"
        >>> Intercalando arquivos "_tmp095" e "_tmp096" no arquivo "_tmp108"
        >>> Intercalando arquivos "_tmp097" e "_tmp098" no arquivo "_tmp109"
        >>> Intercalando arquivos "_tmp099" e "_tmp100" no arquivo "_tmp110"
        >>> Intercalando arquivos "_tmp101" e "_tmp102" no arquivo "_tmp111"
        >>> Intercalando arquivos "_tmp103" e "_tmp104" no arquivo "_tmp112"
        >>> Intercalando arquivos "_tmp105" e "_tmp106" no arquivo "_tmp113"
        >>> Intercalando arquivos "_tmp107" e "_tmp108" no arquivo "_tmp114"
        >>> Intercalando arquivos "_tmp109" e "_tmp110" no arquivo "_tmp115"
        >>> Intercalando arquivos "_tmp111" e "_tmp112" no arquivo "_tmp116"
        >>> Intercalando arquivos "_tmp113" e "_tmp114" no arquivo "_tmp117"
        >>> Intercalando arquivos "_tmp115" e "_tmp116" no arquivo "_tmp118"
        >>> Intercalando arquivos "_tmp117" e "_tmp118" no arquivo "_tmp119"

>>>

...series intercaladas com sucesso

        *** Tempo gasto na operacao:  6.97 minutos ***


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao: 25.00 segundos ***


Criando series ordenadas...

   >>> Criando serie "_tmp001"...
   >>> Criando serie "_tmp002"...
   ...
   >>> Criando serie "_tmp119"...
   >>> Criando serie "_tmp120"...

...series criadas com sucesso

   *** Tempo gasto na operacao:  1.23 minutos ***

Intercalando series...

   >>> Intercalando arquivos "_tmp001" e "_tmp002" no arquivo "_tmp121"
   >>> Intercalando arquivos "_tmp003" e "_tmp004" no arquivo "_tmp122"
   ...
   >>> Intercalando arquivos "_tmp235" e "_tmp236" no arquivo "_tmp238"
   >>> Intercalando arquivos "_tmp237" e "_tmp238" no arquivo "_tmp239"

...series intercaladas com sucesso

   *** Tempo gasto na operacao: 16.17 minutos ***

(Tempo total: 1.23 + 16.17 = 17.4 minutos)

**********************************************************/
