/****
 *
 * Ttulo: Ordenao externa de arquivos usando intercalao binria
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2016
 * ltima modificao: 15/05/2018
 *
 * Descrio:
 *
 * Entrada: Um arquivo contendo registros sem ordenao
 *
 * Sada: O arquivo ordenado
 *
 ****/

/************************ Includes ************************/

#include <stdio.h>        /* Entrada e sada     */
#include <string.h>       /* strcpy() e strchr() */
#include <stdlib.h>       /* qsort()             */
#include "IntercalaMult.h" /* Interface deste mdulo */
#include "Registros.h" /* Funes de processamento de registros */
#include "Heap.h"      /* Heap usado na intercalao     */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/**************** Constantes Simblicas ****************/

   /* Tamanho de um registro */
#define TAM_REG sizeof(tRegistroMEC)

/************************ Macros ************************/

   /* Menor de dois valores */
#define MIN(a, b) ((a) <= (b)) ? (a) : (b)

/********************* Funes Locais *********************/

/****
 *
 * ComparaInts2(): Compara dois registros
 *
 * Parmetros:
 *      e1, e2 (entrada) - os dois registros
 *
 * Retorno:
 *     * 0, se os registros so iguais
 *     * > 0, se o primeiro registro for maior do que o segundo
 *     * < 0, se o primeiro registro for menor do que o segundo
 *
 ****/
static int ComparaInts2(const void *e1, const void* e2)
{
   tRegistroMEC *r1 = (tRegistroMEC *)e1,
                *r2 = (tRegistroMEC *)e2;

   return r1->chave - r2->chave;
}

/****
 *
 * SeriesEstaoVaziasMS(): Verifica se todos os registros j foram processados
 *
 * Parmetros:
 *      series (entrada) - lista indexada contendo informaes
 *                         sobre as sries
 *      nSeries (entrada) - nmero de sries
 *     *heap (entrada) - o heap usado na intercalao
 *
 * Retorno: 1, se os registros j foram todos processados;
 *          0, em caso contrrio
 *
 ****/
static int SeriesEstaoVaziasMS(tSerieMS *series, int nSeries, tHeapIM_Arq *heap)
{
   int i;

   for (i = 0; i < nSeries; ++i)
      if (series[i].nRegs > 0 && series[i].resto > 0)
         return 0;

   return HeapVazioIM_Arq(heap) ? 1 : 0;
}

/****
 *
 * CarregaBufferMS(): Preenche o contedo de um buffer com registros
 *                   lidos de um arquivo
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os registros
 *      buffer (entrada) - o buffer
 *      maxRegs (entrada) - nmero mximo de registros que sero lidos
 *     *pos (entrada/sada) - posio da prxima leitura no arquivo
 *
 * Retorno: O nmero de registros lidos
 *
 ****/
static int CarregaBufferMS( FILE *stream, tRegistroMEC *buffer,
                            int maxRegs, long *pos )
{
   int nRegistros; /* Nmero de registros lidos */

      /* Move o apontador de posio para o local de leitura */
   MoveApontador(stream, *pos, SEEK_SET);

      /* L os registros no arquivo */
   nRegistros = fread(buffer, TAM_REG, maxRegs, stream);
   ASSEGURA(!ferror(stream), "Erro de leitura");

   *pos = ObtemApontador(stream); /* Obtm a posio da prxima leitura */

   return nRegistros;
}

/********************* Funes Globais ********************/

/****
 *
 * CriaSeriesMS(): Cria sries (ordenadas)
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que origina as sries
 *      nomeS (entrada) - nome do arquivo que armazenar as sries
 *
 * Retorno: Nmero de sries criadas
 *
 ****/
int CriaSeriesMS(const char *nomeE, const char *nomeS)
{
   tRegistroMEC *registros; /* Ponteiro para um array que armazena */
                         /* uma srie em memria principal      */
   int        nRegistros,  /* Nmero de registros numa srie */
              nSeries = 0; /* Nmero de sries */
   FILE      *streamS, /* Stream associado ao arquivo */
                       /* que armazenar as sries    */
             *streamE; /* Stream associado ao arquivo original */

      /* Aloca espao para o array. Um array deste */
      /* tamanho no pode ser alocado na pilha     */
   registros = calloc(MAX_REGISTROS, TAM_REG);
   ASSEGURA( registros, "Buffer nao foi alocado" );

      /* Tenta criar o arquivo que conter as sries */
   streamS = AbreArquivo(nomeS, "wb");

      /* Tenta abrir o arquivo original */
   streamE = AbreArquivo(NOME_ARQUIVO_BIN, "rb");

      /* L o arquivo a ser ordenado e cria as sries */
   while (1) {
         /* L os registros da srie corrente */
      nRegistros = fread( registros, TAM_REG, MAX_REGISTROS,
                          streamE );

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA( !ferror(streamE), "Erro de leitura em arquivo" );

         /* Encerra o lao se nenhum registro foi lido */
      if (!nRegistros)
         break;

         /* Ordena os registros lidos */
      qsort(registros, nRegistros, TAM_REG, ComparaInts2);

         /* Escreve a srie corrente no arquivo */
      fwrite(registros, TAM_REG, nRegistros, streamS);

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA( !ferror(streamS),
                "Erro de escrita em arquivo de series" );

      ++nSeries; /* Atualiza o nmero de sries */
   }

      /* Fecha os arquivos */
   FechaArquivo(streamS, nomeS);
   FechaArquivo(streamE, nomeE);

      /* Libera o espao ocupado pelo array */
   free(registros);

   return nSeries;
}

/****
 *
 * IntercalaMultiArq(): Intercala sries usando intercalao multplice
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que contm as sries
 *      nomeS (entrada) - nome do arquivo que conter o resultado
 *                        da intercalao
 *      nSeries (entrada) - o nmero de sries que sero intercaladas
 *      tamSerie (entrada) - tamanho mximo de uma srie
 *
 * Retorno: Nada
 *
 ****/
void IntercalaMultiArq( const char *nomeE, const char *nomeS, int nSeries,
                        int tamSerie )
{
   tSerieMS     *series; /* Ponteiro para um array contendo */
                         /* informaes sobre as sries     */
   int           tamBuffer; /* Tamanho de um buffer de entrada ou sada */
   FILE         *streamS, /* Stream associado ao arquivo que */
                          /* conter os registros ordenados  */
                *streamE; /* Stream associado ao arquivo que contm as sries */
   tHeapIM_Arq   heap; /* Heap usado na ordenao dos elementos */
                       /* dos buffers de entrada                */
   int           i,
                 ibs = 0, /* ndice no qual o prximo registro  */
                          /* do buffer de sada ser armazenado */
                 iHeap = 0; /* ndice de um elemento do heap */
   tNoHeapIM_Arq raiz; /* Uma raiz do heap */
   tRegistroMEC *bufferS; /* Apontar para o buffer de sada */

   tamBuffer = MAX_REGISTROS/(nSeries + 1); /* Calcula o tamanho de cada buffer */

      /* O tamanho de um buffer no pode ser zero */
   ASSEGURA(tamBuffer, "O tamanho de cada buffer e' zero");

      /* Aloca espao para o buffer de sada */
   bufferS = calloc(tamBuffer, TAM_REG);
   ASSEGURA(bufferS, "Impossivel alocar buffer de saida");

   IniciaHeapIM_Arq(&heap, nSeries); /* Inicia o heap */

      /* Tenta abrir o arquivo contendo as sries que sero intercaladas */
   streamE = AbreArquivo(nomeE, "rb");

      /* Aloca um array que contm informaes sobre as sries */
   series = calloc(nSeries, sizeof(tSerieMS));
   ASSEGURA(series, "Impossivel alocar array de series");

      /* Aloca espao para cada buffer de entrada, preenche cada um */
      /* com o contedo inicial de sua respectiva srie e associa o */
      /* primeiro registro de cada buffer a um n do heap           */
   for (i = 0; i < nSeries; ++i) {
         /* Aloca espao para o buffer de entrada corrente */
      series[i].buffer = calloc(tamBuffer, TAM_REG);
      ASSEGURA( series[i].buffer, "Impossivel alocar um buffer de entrada" );

         /* Obtm a posio do primeiro byte da srie */
      series[i].pos = i*tamSerie*TAM_REG;

         /* Preenche o contedo do buffer de entrada corrente */
      series[i].nRegs = CarregaBufferMS( streamE, series[i].buffer,
                                         tamBuffer, &series[i].pos );

         /* Obtm o nmero de registros restantes na srie */
      series[i].resto = tamSerie - series[i].nRegs;

         /* Associa dados do primeiro registro do buffer corrente a um n do heap */
      heap.itens[iHeap].elemento = series[i].buffer;/* Endereo do registro */
      heap.itens[iHeap].ir = 0; /* Seu ndice no buffer */
      heap.itens[iHeap].is = i; /* ndice de sua srie */

      ++iHeap; /* Passa para o prximo elemento do heap */
   }

      /* Ordena o heap */
   for (i = (nSeries - 1)/2; i >= 0; --i)
      OrdenaHeapIM_Arq(&heap, i, ComparaInts2);

      /* Abre o arquivo que conter o resultado da ordenao */
   streamS = AbreArquivo(nomeS, "wb");

      /* Obtm o menor elemento do heap e o substitui pelo prximo elemento */
      /* da srie que o contm at que todas as sries estejam esgotadas    */
   while (!SeriesEstaoVaziasMS(series, nSeries, &heap)) {
      raiz = ObtemMinimoIM_Arq(&heap); /* Obtm o menor valor armazenado no heap */

         /* Verifica se o buffer de sada est repleto */
      if (ibs > tamBuffer - 1) {
            /* O buffer de sada est repleto e  necessrio  */
            /* descarreg-lo no stream que contm o resultado */
         fwrite( bufferS, TAM_REG, tamBuffer, streamS );
         ASSEGURA(!ferror(streamS), "Erro de escrita 1");

         ibs = 0; /* Reinicia o buffer de sada */
      }

         /* Armazena o menor valor do heap no buffer */
         /* que conter o resultado da intercalao  */
      bufferS[ibs++] = *raiz.elemento;

         /* O elemento que substituir a raiz pertence */
         /*  mesma srie  qual pertencia a raiz      */
      if (++raiz.ir < series[raiz.is].nRegs) {
            /* Obtm o prximo elemento da srie  qual a raiz pertencia */
         raiz.elemento = &series[raiz.is].buffer[raiz.ir];
      } else { /* Esse buffer est esgotado */
            /* Tenta renovar o buffer de entrada  */
         series[raiz.is].nRegs = CarregaBufferMS(streamE, series[raiz.is].buffer,
                                                 MIN(tamBuffer, series[raiz.is].resto),
                                                &series[raiz.is].pos);

            /* Calcula o restante da srie em arquivo */
         series[raiz.is].resto = series[raiz.is].nRegs
                               ? series[raiz.is].resto - series[raiz.is].nRegs
                               : 0;

            /* Se o buffer foi renovado, obtm os dados do seu primeiro registro */
         if (series[raiz.is].nRegs > 0) {
            raiz.elemento = series[raiz.is].buffer;
            raiz.ir = 0;
         } else {
            /* O buffer no foi renovado porque todos os registros de sua srie */
            /* j foram consumidos. A nova raiz ser o ltimo elemento do heap, */
            /* que passar a ter um elemento a menos.                           */

               /* Teste de consistncia */
            ASSEGURA(!series[raiz.is].resto, "Nao deveria haver mais registros");

            raiz = ReduzHeapIM_Arq(&heap); /* O heap ter um elemento a menos */
         }
      }

         /* Substitui a raiz do heap pelo novo elemento e reordena o heap */
      SubstituiMinimoIM_Arq(&heap, &raiz, ComparaInts2);
   }

      /* Verifica se restam registros no buffer de sada */
   if (ibs > 0) {
         /* Escreve os registros restantes no stream que contm o resultado */
      fwrite( bufferS, TAM_REG, ibs, streamS );
      ASSEGURA(!ferror(streamS), "Erro de escrita 2");
   }

      /* Libera os buffers de entrada */
   for (i = 0; i < nSeries; ++i)
      free(series[i].buffer);

   free(series); /* Libera array com informaes sobre as sries */

   free(bufferS); /* Libera o buffer de sada */

   FechaArquivo(streamE, nomeE); /* Fecha os arquivos */
   FechaArquivo(streamS, nomeS);
}
