/****
 *
 * Arquivo: Heap.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/09/2017
 * ltima modificao: 15/05/2018
 *
 * Descrio: Implementao de heap
 *
 ****/

/*********************** Includes *************************/

#include <string.h> /* memmove()                */
#include <stdio.h>  /* Funes de Entrada/Sada */
#include <stdlib.h> /* Alocao dinmica        */
#include "Heap.h"   /* Interface deste mdulo   */

/********************* Funes Locais *********************/

/****
 *
 * Troca(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis
 *                                     que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void Troca(void *var1, void *var2, size_t tam)
{
   char *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, (char *) var1, tam);
   memmove((char *) var1, (char *) var2, tam);
   memmove((char *) var2, p, tam);

   free(p);
}

/********************* Funes Globais ********************/

/****
 *
 * OrdenaHeap(): Restaura a propriedade de ordenao de um
 *               heap de mnimo aps ele ter sido alterado
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap
 *      indice (entrada) - ndice do elemento do heap a partir
 *                         do qual ser efetuada a restaurao
 *      Compara (entrada) - ponteiro para uma funo que compara
 *                          dois elementos de array armazenados
 *                          no heap
 *
 * Retorno: Nada
 *
 ****/
void OrdenaHeap(tHeap *heap, int indice, tFCompara Compara)
{
   int iEsq = 2*indice + 1, /* ndice do filho esquerdo */
       iDir = 2*indice + 2, /* ndice do filho direito  */
       iMenor = indice; /* ndice do menor elemento */
                        /* entre pai e filhos       */

      /* Compara o filho esquerdo com seu pai */
   if ( iEsq < heap->nItens &&
        Compara( heap->itens[iEsq].elemento,
                 heap->itens[indice].elemento ) < 0 )
         /* O filho esquerdo  menor do que o pai */
      iMenor = iEsq;

      /* Compara o filho direito com o   */
      /* menor entre seu pai e seu irmo */
   if ( iDir < heap->nItens &&
        Compara( heap->itens[iDir].elemento,
                 heap->itens[iMenor].elemento ) < 0 )
      iMenor = iDir;

      /* Se o n no for menor do que seus filhos, */
      /* troca-o de posio com o menor deles      */
   if (iMenor != indice) {
         /* O pai e o menor filho trocam de posio */
      Troca( &heap->itens[indice], &heap->itens[iMenor],
             sizeof(tNoHeap) );

         /* Reordena o heap a partir do filho que se tornou pai */
      OrdenaHeap(heap, iMenor, Compara);
   }
}

/****
 *
 * IniciaHeap(): Inicia um heap
 *
 * Parmetros:
 *      *heap (sada) - heap que ser iniciado
 *      nElementos (entrada) - nmero de elementos do heap
 *
 * Retorno: O primeiro parmetro
 *
 ****/
tHeap *IniciaHeap(tHeap *heap, int nElementos)
{
   heap->nItens = nElementos;
   heap->itens = calloc(nElementos, sizeof(tNoHeap));

   ASSEGURA(heap->itens, "Impossivel alocar heap");

   return heap;
}

/****
 *
 * DestroiHeap(): Libera o espao ocupado por um heap
 *
 * Parmetros: heap (entrada/sada) - o heap
 *
 * Retorno: Nada
 *
 ****/
void DestroiHeap(tHeap *heap)
{
      /* Libera o espao ocupado pelos ns do heap */
   free(heap->itens);
   heap->nItens = 0;
}

/****
 *
 * ObtemMinimo(): Obtm o menor elemento de um heap de mnimo
 *
 * Parmetros:
 *      heap (entrada) - ponteiro para o heap
 *
 * Retorno: O menor elemento do heap recebido como parmetro
 *
 ****/
tNoHeap ObtemMinimo(const tHeap *heap)
{
      /* O menor elemento de um heap de mnimo se acha na raiz */
   return heap->itens[0];
}

/****
 *
 * ReduzHeap(): Reduz o tamanho de um heap e retorna seu ltimo
 *              elemento
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap
 *
 * Retorno: O ltimo elemento do heap recebido como parmetro
 *
 ****/
tNoHeap ReduzHeap(tHeap *heap)
{
   return heap->itens[--heap->nItens];
}

/****
 *
 * SubstituiMinimo(): Substitui o menor elemento de um heap
 *                    de mnimo
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap que ter
 *                               seu menor elemento substitudo
 *      item (entrada) - ponteiro para o contedo do elemento que
 *                       substituir o menor elemento do heap
 *      Compara (entrada) - ponteiro para uma funo que compara
 *                          dois elementos de array armazenados
 *                          no heap
 *
 * Retorno: Nada
 *
 ****/
void SubstituiMinimo( tHeap *heap, const tNoHeap *item,
                      tFCompara Compara )
{
   heap->itens[0] = *item; /* Efetua a substituio */

      /* Talvez a substituio da raiz faa com que o heap */
      /* deixe de satisfazer sua propriedade de ordenao  */
   OrdenaHeap(heap, 0, Compara);
}
