/****
 *
 * Arquivo ArvoreBM.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 07/08/2012
 * ltima modificao: 15/05/2018
 *
 * Descrio: Implementao das funes que efetuam operaes de
 *            busca e insero usando rvore B+ armazenada em
 *            arquivo
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Funes de E/S                 */
#include "ArvoreBM.h"  /* Interface deste mdulo         */
#include "Pilha.h"     /* Pilha usada para armazenar     */
                       /* caminhadas na rvore           */
#include "Registros.h" /* Interface do mdulo Registros  */
#include "ASSEGURA.h"  /* Macro ASSEGURA                 */
#include "Leitura.h"   /* Funes de leitura via teclado */
#include "Macros.h"    /* Macros usados pelo programa    */
#include "Heap.h"      /* Heap usado na intercalao     */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 0

/******************** Funes de Locais *******************/

/****
 *
 * IniciaNoBM(): Inicia o tipo do n e o nmero de chaves (folhas)
 *               ou filhos (ns internos)
 *
 * Parmetros:
 *     tipoDoNo (entrada) - o tipo do n (folha/interno) que ser iniciado
 *    *pNo (sada) - o n que ser iniciado
 *
 * Retorno: Nada
 *
 ****/
void IniciaNoBM(tTipoDoNo tipoDoNo, tNoBM *pNo)
{
   int i;

   ASSEGURA( tipoDoNo == FOLHA || tipoDoNo == INTERNO,
             "Tipo de no nao e' folha nem interno em IniciaNoBM" );

   pNo->tipoDoNo = tipoDoNo;

   if (tipoDoNo == FOLHA) {
      pNo->conteudo.noFolha.nChaves = 0;
      pNo->conteudo.noFolha.proximaFolha = POSICAO_NULA;
   } else {
      pNo->conteudo.noInterno.nFilhos = 0;

         /* Inicia todos o possveis filhos do n */
      for (i = 0; i < G; ++i) {
         pNo->conteudo.noInterno.filhos[i] = POSICAO_NULA;
      }
   }
}

/****
 *
 * EscreveNoBM(): Escreve um n de uma rvore B+ em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo a
 *                        rvore no qual ser feita a escrita
 *     pos (entrada) - posio no arquivo no qual ser feita a
 *                     escrita
 *    *no (entrada) - n que ser escrito no arquivo
 *
 * Retorno: Nada
 *
 ****/
static void EscreveNoBM(FILE *stream, int pos, const tNoBM *no)
{
      /* Verifica se o stream  vlido */
   ASSEGURA(stream, "Erro em EscreveNoBM(): stream e' NULL");

      /* Tenta mover o apontador de arquivo para o local */
      /* de escrita; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tNoBM)*pos, SEEK_SET);

      /* Efetua a escrita */
   fwrite(no, sizeof(tNoBM), 1, stream);

      /* Certifica-se que no houve erro */
   ASSEGURA( !ferror(stream), ERRO_FWRITE(EscreveNoBM) );
}

/****
 *
 * AcrescentaNoBM(): Acrescenta um n de uma rvore B+ ao final do
 *                   arquivo que contm a rvore
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo a
 *                        rvore na qual ser feita a insero
 *    *no (entrada) - ponteiro para o n que ser escrito
 *
 * Retorno: Posio na qual foi feita a insero do n no arquivo
 *
 ****/
static int AcrescentaNoBM(FILE *stream, const tNoBM *no)
{
   int posicao;

      /* Garante que stream  vlido */
   ASSEGURA( stream, ERRO_STREAM_NULL(stream, AcrescentaNoBM) );

      /* Obtm a posio de insero */
   posicao = NumeroDeRegistros(stream, sizeof(tNoBM));

      /* Tenta mover o apontador de posio */
      /* de arquivo para o local de escrita */
   MoveApontador(stream, 0, SEEK_END);

   fwrite(no, sizeof(tNoBM), 1, stream); /* Efetua a escrita */

      /* Certifica-se que no houve erro */
   ASSEGURA(!ferror(stream), ERRO_FWRITE(AcrescentaNoBM));

   return posicao;
}

/****
 *
 * BuscaEmNoBM(): Faz uma busca sequencial dentro de um
 *                n (interno ou folha)
 *
 * Parmetros:
 *     chave (entrada) - a chave de busca
 *     no (entrada) - ponteiro para o n no qual ser feita a busca
 *    *encontrado (sada) - informar se a chave foi encontrada
 *
 * Retorno: O ndice da chave dentro do n se ela for encontrada
 *          uma chave maior do que ou igual a chave de busca,
 *          ou o nmero de chaves, se tal chave no for
 *          encontrada; i.e., se a chave de busca for maior
 *          do que qualquer chave do n.
 *
 ****/
static int BuscaEmNoBM( tChave chave, const tNoBM *no,
                        int *encontrado )
{
   int i, nChaves;

      /* Ponteiro 'no'  vlido? */
   ASSEGURA(no, "ERRO: No' e' NULL em BuscaEmNoBM()");

      /* Obtm o nmero de chaves do n */
   nChaves = no->tipoDoNo == INTERNO
             ? no->conteudo.noInterno.nFilhos - 1
             : no->conteudo.noFolha.nChaves;

      /* Efetua a busca de acordo com o tipo do n */
   if (no->tipoDoNo == INTERNO) {
      for (i = 0; i < nChaves; ++i)
         if (chave < no->conteudo.noInterno.chaves[i]) {
            *encontrado = 0;
            return i; /* Posio na qual a chave deveria estar */
         } else if ( chave == no->conteudo.noInterno.chaves[i] ) {
            *encontrado = 1;
            return i; /* Posio na qual a chave est */
         }
   } else {
      for (i = 0; i < nChaves; ++i) {
            /* Verifica se a chave foi encontrada */
         if (chave < no->conteudo.noFolha.chaves[i].chave) {
            *encontrado = 0;
            return i;/* Posio na qual a chave deveria estar */
         } else if ( chave ==
                     no->conteudo.noFolha.chaves[i].chave ) {
            *encontrado = 1;
            return i; /* Posio na qual a chave est */
         }
      }
   }

   /****                                              ****/
   /* Neste ponto, sabe-se que no foi encontrada uma    */
   /* chave no n maior do que ou igual  chave de busca */
   /****                                              ****/

   *encontrado = 0;

   return nChaves;
}

/****
 *
 * EncontraCaminhoBM(): Armazena numa pilha o caminho da raiz at
 *                      a folha na qual se encontra uma chave de
 *                      busca ou na qual ela deve ser inserida
 *
 * Parmetros:
 *     raiz (entrada) - ponteiro para a raiz da rvore
 *                      (armazenada em memria)
 *     streamArvore (entrada) - stream associado ao arquivo
 *                              que contm a rvore
 *     chave (entrada) - a chave de busca
 *    *posicao (sada) - posio da chave no ltimo n do caminho
 *    *encontrado (sada) - informa se a chave foi encontrada
 *                          (1) ou no (0)
 *    *pilha (sada) - pilha que armazenar o caminho de ns
 *                     visitados
 *
 * Retorno: Nada
 *
 ****/
static void EncontraCaminhoBM( FILE *streamArvore, tChave chave,
                               const tNoBM *raiz, int *posicao,
                               int *encontrado, tPilha *pilha )
{
   tNoBM      no; /* N usado para armazenar os ns visitados */
   int        indiceNo; /* ndice do n visitado */
                        /* entre seus irmos     */
   tNoCaminho noCaminho;  /* Elemento armazenado na pilha */
   int        posNo; /* Posio de um n no arquivo */

      /* Certifica-se que a raiz no  NULL */
   ASSEGURA(raiz, "ERRO: Arvore vazia em EncontraCaminhoBM");

      /* Certifica-se que o stream no  NULL */
   ASSEGURA( streamArvore,
             ERRO_STREAM_NULL(streamArvore, EncontraCaminhoBM) );

   no = *raiz;   /* O primeiro n visitado  a raiz da rvore */
   posNo = 0; /* A raiz  sempre mantida no incio do arquivo */
   indiceNo = 0;    /* A raiz no tem irmos */
   *encontrado = 0; /* A chave ainda no foi encontrada */

   CriaPilha(pilha); /* Inicia a pilha */

   while (no.tipoDoNo == INTERNO) {
         /***                                ***/
         /*** Cria o item que ser empilhado ***/
         /***                                ***/
      noCaminho.no = no;          /* Armazena o n,...           */
      noCaminho.endereco = posNo; /* sua posio no arquivo e... */
      noCaminho.pos = indiceNo;  /* seu ndice entre seus irmos */

         /* Empilha as informaes sobre o n corrente */
      Empilha(noCaminho, pilha);

         /* Encontra o ndice do filho da prxima descida */
      indiceNo = BuscaEmNoBM(chave, &no, encontrado);

         /* Se a chave foi encontrada num n interno, */
         /* a descida  pelo filho direito dele       */
      if (*encontrado)
         ++indiceNo;

         /* Desce at o prximo n */
      posNo = no.conteudo.noInterno.filhos[indiceNo];

         /* L o prximo n no arquivo */
      LeNoBM(streamArvore, posNo, &no);
   }

   /****                                                   ****/
   /* Neste ponto, sabe-se que 'no'  uma folha.              */
   /* Se a chave encontrar-se na rvore ser nessa folha.     */
   /* Caso contrrio,  nessa folha que ela deve ser inserida */
   /****                                                   ****/

      /* Se a chave for encontrada, 'posicao' apontar para o */
      /* ndice da chave na folha. Caso contrrio, apontar   */
      /* para o ndice na folha na qual a chave ser inserida */
   *posicao = BuscaEmNoBM(chave, &no, encontrado);

      /* Falta acrescentar a folha ao caminho de ns */
   noCaminho.no = no;          /* Armazena o n,...            */
   noCaminho.endereco = posNo; /* ... seu endereo e...        */
   noCaminho.pos = indiceNo;   /* seu ndice entre seus irmos */

   Empilha(noCaminho, pilha);
}

/****
 *
 * InsereEmNoInterno(): Insere uma chave e seu respectivo filho
 *                       direita num n interno incompleto
 *
 * Parmetros:
 *    *pNo (entrada/sada) - n no qual ser feita a insero
 *     pos (entrada) - posio no array de chaves do n no qual
 *                     a chave ser inserida
 *     chave (entrada) - a chave que ser inserida
 *     pFilho (entrada) - posio do filho a ser inserido 
 *                        direita da nova chave
 *
 * Retorno: Nada
 *
 * Observao: A atualizao do n modificado por esta funo
 *             no arquivo que contm a rvore fica sob
 *             responsabilidade de quem chama esta funo.
 *
 ****/
static void InsereEmNoInterno( tNoBM *pNo, int pos,
                               tChave chave, int pFilho )
{
   int i;

      /* Verifica se o ponteiro pNo  vlido */
   ASSEGURA(pNo,"ERRO: Tentativa de inserir chave em no' nulo");

      /* Verifica se pNo aponta para um n interno */
   ASSEGURA( pNo->tipoDoNo == INTERNO,
             "ERRO: No' e' folha em InsereEmNoInterno()" );

      /* Verifica se a posio de insero  vlida */
   ASSEGURA( pos >= 0 && pos < pNo->conteudo.noInterno.nFilhos,
             "ERRO: Posicao de insercao invalida "
             "em InsereEmNoInterno()" );

       /* Abre espao para a nova chave */
   for (i = pNo->conteudo.noInterno.nFilhos - 1; i > pos; --i) {
      pNo->conteudo.noInterno.filhos[i + 1] =
         pNo->conteudo.noInterno.filhos[i];
      pNo->conteudo.noInterno.chaves[i] =
         pNo->conteudo.noInterno.chaves[i - 1];
   }

      /* Insere a nova chave e seu filho direito */
   pNo->conteudo.noInterno.chaves[pos] = chave;
   pNo->conteudo.noInterno.filhos[pos + 1] = pFilho;

   pNo->conteudo.noInterno.nFilhos++; /* A prole aumentou */
}

/****
 *
 * InsereEmFolhaBM(): Insere uma chave e seu respectivo ndice
 *                    de registro numa folha incompleta
 *
 * Parmetros:
 *    *pNo (entrada/sada) - folha na qual ser feita a insero
 *     pos (entrada) - posio no array de chaves do n no qual
 *                     a chave ser inserida
 *     chaveIndice (entrada) - par chave/ndice que ser inserido
 *
 * Retorno: Nada
 *
 * Observao: A atualizao da folha modificada por esta
 *             funo no arquivo que contm a rvore fica
 *             sob responsabilidade de quem chama esta funo
 *
 ****/
static void InsereEmFolhaBM( tNoBM *pNo, int pos,
                             tChaveIndice chaveIndice )
{
   int i;

      /* Verifica se pNo aponta para uma folha */
   ASSEGURA( pNo && pNo->tipoDoNo == FOLHA,
             "Tentativa de inserir chave em folha invalida" );

      /* Verifica se a posio de insero  vlida */
   ASSEGURA( pos >= 0 && pos <= pNo->conteudo.noFolha.nChaves,
             "ERRO: Posicao de insercao invalida "
             "em InsereEmFolhaBM()" );

      /* Abre espao para a nova chave */
   for (i = pNo->conteudo.noFolha.nChaves; i > pos; --i)
              /* Chave aqui  estrutura! */
      pNo->conteudo.noFolha.chaves[i] =
         pNo->conteudo.noFolha.chaves[i - 1];

      /* Insere a nova chave */
   pNo->conteudo.noFolha.chaves[pos] = chaveIndice;

      /* O nmero de chaves aumentou */
   pNo->conteudo.noFolha.nChaves++;
}

/****
 *
 * CopiaChavesBM(): Copia chaves de um n interno para outro
 *
 * Parmetros:
 *    *pNo1 (entrada) - n que doar as chaves que sero copiadas
 *    *pNo2 (sada) - n que receber as chaves copiadas
 *     indiceI (entrada) - ndice da primeira chave a ser copiada
 *     indiceF (entrada) - ndice da ltima chave a ser copiada
 *
 * Retorno: Nada
 *
 * Observao: A atualizao dos ns modificados por esta
 *             funo no arquivo que contm a rvore fica
 *             sob responsabilidade de quem chama esta funo
 *
 ****/
static void CopiaChavesBM( const tNoBM *pNo1, tNoBM *pNo2,
                           int indiceI, int indiceF )
{
   int i,
       nChaves = indiceF - indiceI + 1; /* Nmero de chaves */
                                        /* a copiar         */

      /* Verifica se os dois ns so internos */
   ASSEGURA(pNo1->tipoDoNo==INTERNO && pNo2->tipoDoNo==INTERNO,
            "ERRO: Um no' e' interno e o outro e' folha em "
            "CopiaChavesBM()");

      /* Verifica se o nmero de chaves  vlido */
   ASSEGURA( nChaves < G,
             "Erro: Excesso de chaves em CopiaChavesBM()" );

      /* Idem */
   ASSEGURA(nChaves > 0,
            "Erro: Nao ha chaves para copiar em CopiaChavesBM()");

      /* Copia chaves e filhos do n apontado por */
      /* pNo1 para o n apontado por pNo2         */
   for (i = 0; i < nChaves; ++i) {
      pNo2->conteudo.noInterno.chaves[i] =
              pNo1->conteudo.noInterno.chaves[indiceI + i];
      pNo2->conteudo.noInterno.filhos[i] =
              pNo1->conteudo.noInterno.filhos[indiceI + i];
   }

      /* Falta copiar o ltimo filho */
   pNo2->conteudo.noInterno.filhos[nChaves] =
            pNo1->conteudo.noInterno.filhos[indiceI + nChaves];

      /* Atualiza o nmero de chaves do n que recebeu */
      /* a doao. A atualizao do doador (pNo1)     */
      /* feita por quem chama esta funo.             */
   pNo2->conteudo.noInterno.nFilhos = nChaves + 1;
}

/****
 *
 * CopiaChavesIndicesBM(): Copia chaves e respectivos ndices de
 *                         uma folha para outra
 *
 * Parmetros:
 *    *pNo1 (entrada) - folha que doar as chaves que sero
 *                      copiadas
 *    *pNo2 (sada) - folha que receber as chaves copiadas
 *     indiceI (entrada) - ndice da primeira chave a ser copiada
 *     indiceF (entrada) - ndice da ltima chave a ser copiada
 *
 * Retorno: Nada
 *
 * Observao: A atualizao das folhas modificadas por esta
 *             funo no arquivo que contm a rvore fica sob
 *             responsabilidade de quem chama esta funo
 *
 ****/
static void CopiaChavesIndicesBM( const tNoBM *pNo1, tNoBM *pNo2,
                                  int indiceI, int indiceF )
{
   int i,
       nChaves = indiceF - indiceI + 1; /* Nmero de chaves */
                                        /* a copiar         */

      /* Verifica se os dois ns so folhas */
   ASSEGURA( pNo1->tipoDoNo == FOLHA && pNo2->tipoDoNo == FOLHA,
            "ERRO: Um no' e' interno e o outro e' folha em "
            "CopiaChavesIndicesBM()");

      /* Verifica se o nmero de chaves  vlido */
   ASSEGURA(nChaves <= F,
            "Erro: Excesso de chaves em CopiaChavesIndicesBM()");

      /* Idem */
   ASSEGURA(nChaves > 0, "Erro: Sem chaves para copiar em "
            "CopiaChavesIndicesBM()");

      /* Copia chaves e ndices da folha */
      /* *pNo1 para a folha *pNo2        */
   for (i = 0; i < nChaves; ++i) {
      pNo2->conteudo.noFolha.chaves[i] =
              pNo1->conteudo.noFolha.chaves[indiceI + i];
   }

      /* Atualiza o nmero de chaves do n que recebeu a */
      /* doao. A atualizao do doador (pNo1)  feita  */
      /* por quem chama esta funo.                     */
   pNo2->conteudo.noFolha.nChaves = nChaves;
}

/****
 *
 * DivideNoInternoBM(): Insere uma chave num n interno completo de
 *                      uma rvore B+, criando um novo n para a
 *                      metade superior das chaves e mantendo a
 *                      metade inferior das chaves no n dividido
 *
 * Parmetros:
 *    *pNoDiv (entrada/sada) - n que ser dividido
 *     pos (entrada) - posio na qual a nova chave ser inserida
 *     chave (entrada) - a chave que ser inserida
 *     pFilhoDireita (entrada) - posio em arquivo do n que
 *                               ser o filho direito da chave inserida
 *    *noDireito (sada) - novo n no qual sero colocadas as
 *                         chaves superiores do n dividido
 *    *chaveQueSobe (sada) - a chave que subir para um n no
 *                            nvel imediatamente superior
 *
 * Retorno: Nada
 *
 * Observaes:
 *    1. Existe uma diferena fundamental entre esta funo e
 *       a funo DivideFolhaBM(): aqui, quando um n  dividido,
 *       a chave que sobe  "movida" para um n interno no
 *       prximo nvel acima, enquanto que na funo
 *       DivideFolhaBM() a chave que sobe  "copiada" de uma
 *       folha para um n interno.
 *    2. Se o grau da rvore for par, a rvore (i.e., cada n
 *       interno) ter tendncia  esquerda; i.e., o novo
 *       dividido ter uma chave a mais do que o novo n. No
 *       confunda isto com a tendncia das folhas que   direita
 *    3. A atualizao dos ns modificados por esta funo no
 *       arquivo que contm a rvore fica sob responsabilidade
 *       de quem chama esta funo.
 *
 ****/
static void DivideNoInternoBM( tNoBM *pNoDiv, int pos, tChave chave,
                               int pFilhoDireita, tNoBM *noDireito,
                               tChave *chaveQueSobe )
{
   static const int meio = (G - 1)/2; /* Evita recalcular este valor */
                                      /* acada chamada desta funo  */

   ASSEGURA( pNoDiv->tipoDoNo == INTERNO, "Erro: Tentativa de "
             "dividir folha usando DivideNoInternoBM()" );

      /* Inicia o novo n que receber as chaves superiores */
   IniciaNoBM(INTERNO, noDireito);

      /* Efetua a insero com a devida diviso de ns */
   if (pos > meio) { /* Caso 1: A nova chave pertence */
                     /* ao novo n direito            */
         /* Copia metade menos uma chave  */
         /* do n dividido para o novo n */
      CopiaChavesBM(pNoDiv, noDireito, meio + 1, G - 2);

         /* Insere a nova chave no novo n da direita */
      InsereEmNoInterno( noDireito, pos - meio - 1, chave,
                         pFilhoDireita );

         /* A chave que subir  a ltima chave do */
         /* n que foi dividido depois da diviso  */
      *chaveQueSobe = pNoDiv->conteudo.noInterno.chaves[meio];

         /* Ajusta o nmero de ns do n dividido */
      pNoDiv->conteudo.noInterno.nFilhos = meio + 1;
   } else if (pos == meio) { /* Caso 2: A nova chave      */
                             /* aquela que subir um nvel */
         /* Copia metade das chaves do */
         /* n dividido para o novo n */
      CopiaChavesBM(pNoDiv, noDireito, meio, G - 2);

         /* Ajusta o nmero de ns do n dividido */
      pNoDiv->conteudo.noInterno.nFilhos = meio + 1;

         /* O filho direito da nova chave passa */
         /* a ser o primeiro filho do novo n   */
      noDireito->conteudo.noInterno.filhos[0] = pFilhoDireita;

         /* A chave que subir  a nova chave */
      *chaveQueSobe = chave;
   } else { /* Caso 3: A nova chave pertence */
            /* ao n que ser dividido       */
         /* Copia metade das chaves do */
         /* n dividido para o novo n */
      CopiaChavesBM(pNoDiv, noDireito, meio, G - 2);

         /* Ajusta o nmero de ns do n dividido */
      pNoDiv->conteudo.noInterno.nFilhos = meio + 1;

         /* Insere a nova chave no novo n da direita */
      InsereEmNoInterno(pNoDiv, pos, chave, pFilhoDireita);

         /* A chave que subir  a ltima chave do */
         /* n que foi dividido depois da diviso  */
      *chaveQueSobe = pNoDiv->conteudo.noInterno.chaves[meio];

         /* Corrige o nmero de filhos do n dividido, pois    */
         /* sua ltima chave subir para o prximo nvel acima */
      --pNoDiv->conteudo.noInterno.nFilhos;
   }
}

/****
 *
 * DivideFolhaBM(): Insere um par chave/ndice numa folha e
 *                  divide este n, criando um novo n para a
 *                  metade superior das chaves e mantendo a
 *                  metade inferior das chaves no n dividido
 *
 * Parmetros:
 *    *pNoDiv (entrada/sada) - folha que ser dividida
 *     pos (entrada) - posio na qual a nova chave ser inserida
 *     chaveIndice (entrada) - par chave/ndice que ser inserido
 *    *noDireito (sada) - novo n no qual sero colocadas as
 *                         chaves superiores do n dividido
 *    *chaveQueSobe (sada) - a chave que ser copiada para o
 *                            n-pai do n dividido; o filho
 *                            direito desta chave ser noDireito
 *
 * Retorno: Nada
 *
 * Observaes:
 *    1. Existe uma diferena fundamental entre esta funo e
 *       a funo DivideNoInternoBM(): aqui, quando uma folha 
 *       dividida, a chave que sobe  "copiada" para um n
 *       interno, enquanto que na funo DivideNoInternoBM() a
 *       chave que sobe  "movida" de um n interno para outro
 *       um nvel acima.
 *    2. Se o nmero de chaves for par, as folhas tero tendncia
 *        direita; i.e., o novo n ter uma chave a mais do que
 *       o n dividido. No confunda isto com a tendncia da
 *       prpria rvore (i.e., de cada n interno) que  
 *       esquerda.
 *    3. A atualizao dos ns modificados por esta funo no
 *       arquivo que contm a rvore fica sob responsabilidade
 *       de quem chama esta funo.
 *    4. O encadeamento das folhas fica sob responsabilidade
 *       de quem chama esta funo.
 *
 ****/
static void DivideFolhaBM( tNoBM *pNoDiv, int pos,
                           tChaveIndice chaveIndice,
                           tNoBM *noDireito, tChave *chaveQueSobe )
{
   static const int meio = F/2;  /* Evita recalcular esses */
   static const int resto = F%2; /* valores a cada chamada */

      /* Certifica-se que pNoDiv aponta para uma folha */
   ASSEGURA( pNoDiv->tipoDoNo == FOLHA, "Erro: Tentativa de "
             "dividir um no' interno com DivideFolhaBM()" );

      /* Inicia o novo n que receber as chaves superiores */
   IniciaNoBM(FOLHA, noDireito);

      /* Divide o n e efetua a insero no devido n */
   if (pos > meio) { /* Caso 1: A nova chave pertencer */
                     /* ao novo n direito              */
         /* Copia metade menos uma chave para o novo n */
      CopiaChavesIndicesBM(pNoDiv, noDireito, meio + resto, F - 1);

         /* Ajusta o nmero de ns do   */
         /* nmero de n do n dividido */
      pNoDiv->conteudo.noFolha.nChaves = meio + resto;

         /* Insere a nova chave no novo n */
      InsereEmFolhaBM(noDireito, pos - meio - resto, chaveIndice);

         /* A chave que sobe para um n interno */
         /*  a primeira chave da folha direita */
      *chaveQueSobe = noDireito->conteudo.noFolha.chaves[0].chave;
   } else { /* Caso 2: A nova chave pertencer ao n dividido */
         /* Copia metade das chaves para o novo n */
      CopiaChavesIndicesBM( pNoDiv, noDireito, meio, F - 1 );

         /* Antes de inserir a chave no n dividido */
         /* deve-se reduzir seu nmero de chaves,   */
         /* pois neste instante ele est completo   */
      pNoDiv->conteudo.noFolha.nChaves = meio;

         /* Insere a nova chave no novo n */
      InsereEmFolhaBM(pNoDiv, pos, chaveIndice);

         /* A chave que sobe para um n interno */
         /*  a primeira chave da folha direita */
      *chaveQueSobe = noDireito->conteudo.noFolha.chaves[0].chave;
   }
}

/****
 *
 * JuntaNosBM(): Combina um n interno com outros ns internos
 *               de uma rvore B+ quando ele fica com grau
 *               abaixo do mnimo
 *
 * Parmetros:
 *      pNoN (entrada/sada) - ponteiro para o n que ser
 *                             combinado com outros
 *      posNo (entrada) - posio do n entre seus irmos
 *      endNo (entrada) - posio do n no arquivo
 *     *pilha (entrada/sada) - pilha que armazena os ns
 *                              encontrados no caminho da raiz
 *                              at o n que ser combinado
 *                              (sem inclu-lo)
 *      stream (entrada) - stream associado ao arquivo que
 *                         contm a rvore
 *
 * Retorno: Nada
 *
 * Observaes:
 *    1. Esta funo  semelhante  funo JuntaNos() para rvores B
 *
 *    2. Considerando as figuras apresentadas no texto, pNoN 
 *       um ponteiro para o n N, pNoE  a posio no arquivo que
 *       contm a rvore do n E (aqui denominado noE) e pNoD  a
 *       posio no mesmo arquivo do n D (aqui denominado noD).
 *
 *    3. Quando se faz referncia a uma varivel que contm uma
 *       posio em arquivo, 'apontar para' no tem o mesmo
 *       significado dessa mesma expresso usada com ponteiros.
 *       I.e., aqui, 'apontar para' significa conter a posio
 *       de um n no arquivo que o armazena.
 *
 *    4. V. figuras e algoritmo no texto.  muito difcil
 *       acompanhar esta funo sem examinar cuidadosamente
 *       essas figuras.
 *
 ****/
static void JuntaNosBM( tNoBM *pNoN, int posNo, int endNo,
                        tPilha* pilha, FILE *stream )
{
   tNoCaminho caminho; /* Armazenar cada item desempilhado */
   tNoBM      noE, /* N E das figuras do texto */
              noD; /* N D das figuras do texto */
   int        pNoE = POSICAO_NULA,/* Posio em arquivo de noE */
              pNoD = POSICAO_NULA,/* Posio em arquivo de noD */
              i;
   tChave     sucessora; /* A chave sucessora  qual */
                         /* se refere o algoritmo    */

      /* Passo 1 do algoritmo */
   if (pNoN->conteudo.noInterno.nFilhos >= TG) {
         /* N no precisa ser combinado com outro */
      return;
   }

   /*                                                     */
   /* Neste ponto, sabe-se que o n apontado por pNoN tem */
   /* um nmero de filhos menor do que o mnimo desejado  */
   /*                                                     */

      /* A pilha no pode estar vazia aqui */
   ASSEGURA(!PilhaVazia(*pilha), "Pilha vazia em JuntaNosBM()");

      /* O item no topo da pilha contm informaes */
      /* sobre o pai do n apontado por pNoN        */
   caminho = Desempilha(pilha);

      /* Se o n a ser combinado no  o primeiro */
      /* filho, ele tem um irmo esquerdo         */
   if(posNo > 0)
         /* Faz pNoE apontar para o irmo    */
         /* esquerdo do n apontado por pNoN */
      pNoE = caminho.no.conteudo.noInterno.filhos[posNo - 1];

      /* Se o n a ser combinado no  o ltimo */
      /* filho, ele tem um irmo direito        */
   if(posNo < caminho.no.conteudo.noInterno.nFilhos - 1)
         /* Faz pNoD apontar para o irmo   */
         /* direito do n apontado por pNoN */
      pNoD = caminho.no.conteudo.noInterno.filhos[posNo + 1];

      /* O n a ser combinado deve ter pelo menos um  */
      /* irmo mais prximo ou, ento, ele deve ser a */
      /* raiz que no tem irmos                      */
   ASSEGURA( pNoE != POSICAO_NULA || pNoD != POSICAO_NULA ||
             PilhaVazia(*pilha),
             "Erro em JuntaNosBM(): o no' nao possui irmao" );

      /* Passo 2: Verifique se o irmo esquerdo (apontado */
      /* por pNoE) do n apontado por pNoN tem um nmero  */
      /* de chaves maior do que o mnimo                  */

      /* Verifica se o n apontado por pNoN tem irmo esquerdo */
   if (pNoE != POSICAO_NULA) {
         /* L o irmo esquerdo noE */
      LeNoBM(stream, pNoE, &noE);

         /* Verifica se o irmo esquerdo pode doar uma chave */
      if (noE.conteudo.noInterno.nFilhos > TG) {
            /* Passo 2.1: Insira a sucessora da ltima */
            /* chave do n noE no incio das chaves do */
            /* n apontado por pNoN                    */

            /* Obtm a sucessora que se encontra no n-pai */
         sucessora = caminho.no.conteudo.noInterno.chaves[posNo-1];

             /* Abre espao para o novo filho */
         for (i = pNoN->conteudo.noInterno.nFilhos - 1; i >= 0; --i)
            pNoN->conteudo.noInterno.filhos[i + 1] =
               pNoN->conteudo.noInterno.filhos[i];

             /* Abre espao para a nova chave */
         for (i = pNoN->conteudo.noInterno.nFilhos - 2; i >= 0; --i)
            pNoN->conteudo.noInterno.chaves[i + 1] =
               pNoN->conteudo.noInterno.chaves[i];

            /* Coloca a sucessora em seu devido lugar */
         pNoN->conteudo.noInterno.chaves[0] = sucessora;

            /* Passo 2.2: Faa o filho esquerdo da primeira */
            /* chave de pNoN apontar para o filho direito   */
            /* da ltima chave de noE                       */
         pNoN->conteudo.noInterno.filhos[0] =
            noE.conteudo.noInterno.filhos
               [noE.conteudo.noInterno.nFilhos - 1];

            /* Passo 2.3: Substitua a sucessora da ltima */
            /* chave do n noE por essa ltima chave      */
         caminho.no.conteudo.noInterno.chaves[posNo - 1] =
             noE.conteudo.noInterno.chaves[
                     noE.conteudo.noInterno.nFilhos - 2 ];

            /* O n-pai que foi alterado  uma cpia daquele */
            /* que se encontra na rvore. Logo,  preciso    */
            /* atualizar o n correspondente na rvore.      */
         EscreveNoBM(stream, caminho.endereco, &caminho.no);

            /* Passo 2.4: Decremente o nmero de filhos de noE */
         noE.conteudo.noInterno.nFilhos--;

            /* O n noE precisa ser atualizado em arquivo */
         EscreveNoBM(stream, pNoE, &noE);

            /* Passo 2.5: Incremente o nmero de */
            /* filhos do n apontado por pNoN    */
         pNoN->conteudo.noInterno.nFilhos++;

            /* O n apontado por pNoN precisa */
            /* ser atualizado em arquivo      */
         EscreveNoBM(stream, endNo, pNoN);

            /* Passo 2.6: Encerre */
         return;
      }
   }

      /* Passo 3: Verifique se o irmo direito (apontado */
      /* por pNoD) do n apontado por pNoN tem um nmero */
      /* de chaves maior do que o mnimo                 */
   if (pNoD != POSICAO_NULA) {
         /* L o irmo direito noD */
      LeNoBM(stream, pNoD, &noD);

         /* Verifica se o irmo direito */
         /* pode fazer doao de chaves */
      if (noD.conteudo.noInterno.nFilhos > TG) {
            /* Passo 3.1: Acrescente a sucessora da */
            /* ltima chave do n apontado por pNoN */
            /* ao final das chaves desse n         */

            /* Obtm a sucessora que se encontra no n-pai */
         sucessora = caminho.no.conteudo.noInterno.chaves[posNo];

            /* Coloca a sucessora em seu devido lugar */
         pNoN->conteudo.noInterno.chaves
            [pNoN->conteudo.noInterno.nFilhos - 1] = sucessora;

            /* Passo 3.2: Faa o filho direito da ltima    */
            /* chave de pNoN apontar para o filho esquerdo  */
            /* da primeira chave de noD. Note que o nmero  */
            /* de filhos de pNoN ainda no foi incrementado */
         pNoN->conteudo.noInterno.filhos
            [pNoN->conteudo.noInterno.nFilhos] =
            noD.conteudo.noInterno.filhos[0];

            /* Passo 3.3: Substitua a sucessora da ltima chave */
            /* do n pNoN pela primeira chave do n noD         */
         caminho.no.conteudo.noInterno.chaves[posNo] =
               noD.conteudo.noInterno.chaves[0];

             /* Passo 3.4: Desloque cada chave e filho */
             /* do n noD uma posio para trs        */
         for (i = 1; i <= noD.conteudo.noInterno.nFilhos - 1; ++i)
            noD.conteudo.noInterno.filhos[i - 1] =
               noD.conteudo.noInterno.filhos[i];

         for (i = 1; i <= noD.conteudo.noInterno.nFilhos - 2; ++i)
            noD.conteudo.noInterno.chaves[i - 1] =
               noD.conteudo.noInterno.chaves[i];

            /* O n-pai que foi alterado  uma cpia daquele */
            /* que se encontra na rvore. Logo  preciso     */
            /* atualizar o n correspondente na rvore.      */
         EscreveNoBM(stream, caminho.endereco, &caminho.no);

            /* Passo 3.5: Decremente o nmero */
            /* de filhos do n noD            */
         noD.conteudo.noInterno.nFilhos--;

            /* Passo 3.6: Incremente o nmero de */
            /* filhos do n apontado por pNoN    */
         pNoN->conteudo.noInterno.nFilhos++;

            /* O n noD precisa ser atualizado em arquivo */
         EscreveNoBM(stream, pNoD, &noD);

            /* O n apontado por pNoN precisa */
            /* ser atualizado em arquivo      */
         EscreveNoBM(stream, endNo, pNoN);

            /* Passo 3.7: Encerre */
         return;
      }
   }

   /*                                                     */
   /* Passo 4: Se nenhum irmo vizinho do n apontado por */
   /* pNoN tiver um nmero de ns maior do que o mnimo   */
   /*                                                     */

      /* Passo 4.1: Verifique se o n apontado */
      /* por pNoN tem irmo direito (pNoD) com */
      /* nmero de filhos igual ao mnimo      */
   if (pNoD != POSICAO_NULA) {
         /* L o irmo direito noD */
      LeNoBM(stream, pNoD, &noD);

         /* Garante que o nmero de filhos de noD  igual ao   */
         /* mnimo. Se esse nmero fosse maior do que o mnimo */
         /* ter-se-a o Caso 3. Se esse nmero fosse menor do  */
         /* que o mnimo, no se teria uma rvore B+.          */
      ASSEGURA( noD.conteudo.noInterno.nFilhos == TG,
                "Numero de chaves em no' direito nao e' minimo" );

         /* Passo 4.1.1: Insira a chave sucessora (S) da  */
         /* ltima chave do n apontado por pNoN ao final */
         /* das chaves do n apontado por pNoN            */

         /* Obtm a sucessora */
      sucessora = caminho.no.conteudo.noInterno.chaves[posNo];

         /* Coloca a sucessora em seu devido lugar */
      pNoN->conteudo.noInterno.chaves
         [pNoN->conteudo.noInterno.nFilhos - 1] = sucessora;

         /* Passo 4.1.2: Faa o filho direito da ltima  */
         /* chave de pNoN apontar para  o filho esquerdo */
         /* da primeira chave de noD. Note que o nmero  */
         /* de filhos de pNoN ainda no foi incrementado */
      pNoN->conteudo.noInterno.filhos
         [pNoN->conteudo.noInterno.nFilhos] =
         noD.conteudo.noInterno.filhos[0];

         /* O nmero de chaves e filhos do */
         /* n apontado por pNoN aumentou  */
      ++pNoN->conteudo.noInterno.nFilhos;

         /* Passo 4.1.3: Desloque cada chave do pai do n    */
         /* apontado por pNoN e seu filho direito , a partir */
         /* da chave sucessora, para uma posio anterior    */
      for ( i = posNo + 1;
            i < caminho.no.conteudo.noInterno.nFilhos - 1; ++i )
         caminho.no.conteudo.noInterno.filhos[i] =
            caminho.no.conteudo.noInterno.filhos[i + 1];

      for ( i = posNo;
            i < caminho.no.conteudo.noInterno.nFilhos - 2; ++i )
         caminho.no.conteudo.noInterno.chaves[i] =
            caminho.no.conteudo.noInterno.chaves[i + 1];

         /* Passo 4.1.4: Decremente o nmero de filhos */
         /* do pai do n apontado por pNoN             */
      --caminho.no.conteudo.noInterno.nFilhos;

         /*  preciso atualizar o n */
         /* correspondente na rvore */
      EscreveNoBM(stream, caminho.endereco, &caminho.no);

         /* Passo 4.1.5: Acrescente todas as chaves e filhos */
         /* direitos correspondentes do n noD ao final das  */
         /* chaves e filhos do n apontado por pNoN          */
      for ( i = 0; i < noD.conteudo.noInterno.nFilhos - 1; ++i)
         pNoN->conteudo.noInterno.filhos
             [pNoN->conteudo.noInterno.nFilhos + i] =
               noD.conteudo.noInterno.filhos[i + 1];

      for ( i = 0; i <= noD.conteudo.noInterno.nFilhos - 2; ++i)
         pNoN->conteudo.noInterno.chaves
             [pNoN->conteudo.noInterno.nFilhos + i - 1] =
               noD.conteudo.noInterno.chaves[i];

         /* Passo 4.1.6: Atualize o nmero de */
         /* filhos do n apontado por pNoN    */
      pNoN->conteudo.noInterno.nFilhos +=
          noD.conteudo.noInterno.nFilhos;

         /* Passo 4.1.7: Libere o n noD. No h como liberar */
         /* um n no interior de um arquivo. Ento, a melhor  */
         /* opo  indicar que esse  um n vazio.           */
      noD.tipoDoNo = NO_VAZIO;

         /* O n noD precisa ser atualizado em arquivo */
      EscreveNoBM(stream, pNoD, &noD);

         /* O n apontado por pNoN precisa */
         /* ser atualizado em arquivo      */
      EscreveNoBM(stream, endNo, pNoN);

         /* Passo 4.1.8: Ajuste o pai do n       */
         /* apontado por pNoN usando JuntaNosBM() */
      JuntaNosBM( &caminho.no, caminho.pos, caminho.endereco,
                  pilha, stream );

      return;
   }

      /* Passo 4.2: Verifique se o n apontado */
      /* por pNoN tem irmo esquerdo (noE) com */
      /* nmero de filhos igual ao mnimo      */
   if (pNoE != POSICAO_NULA) {
         /* L o irmo esquerdo noE */
      LeNoBM(stream, pNoE, &noE);

         /* Garante que o nmero de filhos de noE  igual ao   */
         /* mnimo. Se esse nmero fosse maior do que o mnimo */
         /* ter-se-a o Caso 2. Se esse nmero fosse menor do  */
         /* que o mnimo, no se teria uma rvore B+.          */
      ASSEGURA( noE.conteudo.noInterno.nFilhos == TG,
                "Numero de chaves em no' esquerdo nao e' minimo" );

         /* Passo 4.2.1: Insira ao final das chaves de noE */
         /* a chave sucessora da ltima chave de noE       */

         /* Obtm a sucessora */
      sucessora = caminho.no.conteudo.noInterno.chaves[posNo - 1];

         /* Coloca a sucessora em seu devido lugar */
      noE.conteudo.noInterno.chaves
         [noE.conteudo.noInterno.nFilhos - 1] = sucessora;

         /* Passo 4.2.2: Faa o filho direito da ltima  */
         /* chave de noE apontar para  o filho esquerdo  */
         /* da primeira chave de pNoN. Note que ainda    */
         /* no houve incremento no nmero de ns de noE */
      noE.conteudo.noInterno.filhos
         [noE.conteudo.noInterno.nFilhos] =
         pNoN->conteudo.noInterno.filhos[0];

         /* O nmero de chaves e filhos do n noE aumentou */
      ++noE.conteudo.noInterno.nFilhos;

         /* Passo 4.2.3: Desloque cada chave do pai do n   */
         /* apontado por pNoN e seu filho direito, a partir */
         /* da chave sucessora, para uma posio anterior   */
      for ( i = posNo;
            i < caminho.no.conteudo.noInterno.nFilhos - 1; ++i)
         caminho.no.conteudo.noInterno.filhos[i] =
            caminho.no.conteudo.noInterno.filhos[i + 1];

      for ( i = posNo - 1;
            i < caminho.no.conteudo.noInterno.nFilhos - 2; ++i)
         caminho.no.conteudo.noInterno.chaves[i] =
            caminho.no.conteudo.noInterno.chaves[i + 1];

         /* Passo 4.2.4: Decremente o nmero de filhos */
         /* do pai do n apontado por pNoN             */
      --caminho.no.conteudo.noInterno.nFilhos;

         /*  preciso atualizar o n correspondente na rvore */
      EscreveNoBM(stream, caminho.endereco, &caminho.no);

         /* Passo 4.2.5: Acrescente todas as chaves e filhos */
         /* direitos correspondentes do n apontado por pNoN */
         /* ao final das chaves de noE                       */
      for (i = 0; i < pNoN->conteudo.noInterno.nFilhos - 1; ++i)
         noE.conteudo.noInterno.filhos
            [noE.conteudo.noInterno.nFilhos + i] =
            pNoN->conteudo.noInterno.filhos[i + 1];

      for (i = 0; i <= pNoN->conteudo.noInterno.nFilhos - 2; ++i)
         noE.conteudo.noInterno.chaves
            [noE.conteudo.noInterno.nFilhos + i - 1] =
            pNoN->conteudo.noInterno.chaves[i];

         /* Passo 4.2.6: Atualize o nmero de filhos de noE */
      noE.conteudo.noInterno.nFilhos +=
         pNoN->conteudo.noInterno.nFilhos;

         /* Passo 4.2.7: Libere o n apontado por pNoN. */
         /* V. comentrio acima sobre liberao de ns. */
      pNoN->tipoDoNo = NO_VAZIO;

         /* O n noE precisa ser atualizado em arquivo */
      EscreveNoBM(stream, pNoE, &noE);

         /* O n apontado por pNoN precisa */
         /* ser atualizado em arquivo      */
      EscreveNoBM(stream, endNo, pNoN);

         /* Passo 4.2.8: Ajuste o pai do n     */
         /* apontado por pNoE usando JuntaNosBM() */
      JuntaNosBM( &caminho.no, caminho.pos, caminho.endereco,
                  pilha, stream );
   }
}

/****
 *
 * AtualizaIndice(): Atualiza os ns internos de uma rvore B+
 *                   aps uma remoo de chave
 *
 * Parmetros:
 *       streamArvore (entrada) - stream associado ao arquivo
 *                                que armazena a rvore
 *      *pilha (entrada/sada) - pilha contendo o caminho de n
 *                               internos at a folha (sem inclu-
 *                               la) na qual ocorreu uma remoo
 *       removida (entrada) - chave que foi removida
 *       subst (entrada) - se no for NULL, aponta para uma chave
 *                         que substituir a chave removida caso
 *                         ela seja encontrada num n interno
 *       aRemover (entrada) - se no for NULL, aponta para uma
 *                            chave que ser removida do ndice
 *                            juntamente com seu filho direto
 *
 * Retorno: Nada
 *
 * Descrio do algoritmo: v. texto
 *
 ****/
static void AtualizaIndice( FILE *streamArvore, tPilha *pilha,
                            tChave removida, tChave *subst,
                            tChave *aRemover )
{
   int        i, j,
              nChaves; /* Nmero de chaves de um n */
   tNoCaminho caminho; /* Um elemento da pilha que    */
                       /* representa o caminho de ns */
   tPilha     pilhaAux; /* Pilha auxiliar */

      /* Se houver chave a remover e a substituir no     */
      /* ndice, ser necessrio usar uma pilha auxiliar */
   CriaPilha(&pilhaAux);

      /***                                 ***/
      /* A primeira parte desta funo lida  */
      /* com substituio de chave no ndice */
      /***                                 ***/

   while (subst && !PilhaVazia(*pilha)) {
      caminho = Desempilha(pilha);

      ASSEGURA( caminho.no.tipoDoNo == INTERNO,
                "Em AtualizaIndice(): No' deveria ser interno" );

         /* Obtm o nmero de chaves do n corrente */
      nChaves = caminho.no.conteudo.noInterno.nFilhos - 1;

         /* Procura pela chave a ser substituda no n corrente */
      for (i = 0; i < nChaves; ++i) {
         if ( removida ==
                  caminho.no.conteudo.noInterno.chaves[i] ) {
               /* A chave removida foi encontrada no */
               /* n corrente e ser substituda     */
            caminho.no.conteudo.noInterno.chaves[i] = *subst;

                  /* Atualiza n em arquivo */
               EscreveNoBM( streamArvore, caminho.endereco,
                            &caminho.no );

            subst = NULL; /* S h uma chave a substituir */

            break;
         }
      }

         /* Se haver remoo,  preciso empilhar */
         /* o n na pilha auxiliar                */
      if (aRemover) {
         Empilha(caminho, &pilhaAux);
      }
   }

      /***                               ***/
      /* A segunda parte desta funo lida */
      /* com remoo de chave no ndice    */
      /***                               ***/

      /* Se no h chave a remover no ndice, o */
      /* trabalho desta funo est encerrado   */
   if (!aRemover) {
      EsvaziaPilha(pilha); /* A pilha no ser necessria */
      return; /* Servio concludo */
   } else { /* Restaura a pilha, se for o caso */
      while (!PilhaVazia(pilhaAux)) {
         Empilha(Desempilha(&pilhaAux), pilha);
      }
   }

      /***                            ***/
      /* Esta funo s remove chave no */
      /* ltimo nvel de ns internos   */
      /***                            ***/

      /* Neste ponto, a pilha no pode estar vazia */
   ASSEGURA(!PilhaVazia(*pilha),
            "A pilha nao deveria estar vazia em AtualizaIndice()");

      /* Obtm informaes sobre o n interno do ltimo */
      /* nvel que se encontra no topo da pilha         */
   caminho = Desempilha(pilha);

      /* O n desempilhado deve ser interno */
   ASSEGURA( caminho.no.tipoDoNo == INTERNO,
             "Erro em AtualizaIndice(): No' deveria ser interno" );

      /* Obtm o nmero de chaves do n corrente */
   nChaves = caminho.no.conteudo.noInterno.nFilhos - 1;

      /* Procura a chave a ser removida no n corrente */
   for (i = 0; i < nChaves; ++i) {
      if ( *aRemover ==
               caminho.no.conteudo.noInterno.chaves[i] ) {
            /* A chave a ser removida foi encontrada. */
            /* Move cada chave uma posio para trs. */
         for (j = i; j < nChaves - 1; ++j)
            caminho.no.conteudo.noInterno.chaves[j] =
               caminho.no.conteudo.noInterno.chaves[j + 1];

            /* Faz o mesmo com cada filho direito */
         for ( j = i + 1;
               j < caminho.no.conteudo.noInterno.nFilhos - 1; ++j )
            caminho.no.conteudo.noInterno.filhos[j] =
               caminho.no.conteudo.noInterno.filhos[j + 1];

            /* O nmero de filhos do n foi reduzido */
         caminho.no.conteudo.noInterno.nFilhos--;

         aRemover = NULL; /* No h mais chave a remover */

         break;
      }
   }

      /* A chave deve ter sido encontrada e removida */
   ASSEGURA( !aRemover,
             "Chave a remover nao foi encontrada em no' interno" );

      /* Atualiza n em arquivo */
   EscreveNoBM( streamArvore, caminho.endereco, &caminho.no );

      /* Verifica se o n ficou descompensado */
   if (caminho.no.conteudo.noInterno.nFilhos < TG) {
         /* Chama JuntaNosBM() para completar o servio */
      JuntaNosBM( &caminho.no, caminho.pos, caminho.endereco,
                  pilha, streamArvore );
   } else {
      EsvaziaPilha(pilha);
   }
}

/****************** Funes de Ordenao ******************/

/****
 *
 * ComparaPares(): Compara dois pares chave/ndice que sero ordenados
 *
 * Parmetros:
 *      e1, e2 (entrada) - endereos dos pares que sero comparados
 *
 * Retorno:
 *     * 0, se os pares forem iguais
 *     * > 0, se o primeiro par for maior do que o segundo
 *     * < 0, se o primeiro par for menor do que o segundo
 *
 * Observao: Se um ponteiro  NULL, o registro para o qual
 *             ele aponta  considerado maior do que os demais
 *
 ****/
static int ComparaPares(const void *e1, const void* e2)
{
   tChaveIndice *p1 = (tChaveIndice *)e1,
                *p2 = (tChaveIndice *)e2;

   if (!p1 && !p2)
      return 0;
   else if (!p1)
      return 1;
   else if (!p2)
      return -1;

   return p1->chave - p2->chave;
}

/****
 *
 * SeriesEstaoVaziasMS2(): Verifica se todas os pares j foram processados
 *
 * Parmetros:
 *      series (entrada) - lista indexada contendo informaes
 *                         sobre as sries
 *      nSeries (entrada) - nmero de sries
 *     *heap (entrada) - o heap usado na intercalao
 *
 * Retorno: 1, se os pares j foram todos processados;
 *          0, em caso contrrio
 *
 ****/
static int SeriesEstaoVaziasMS2(tSerie *series, int nSeries, tHeapIM_Arq2 *heap)
{
   int i;

   for (i = 0; i < nSeries; ++i)
      if (series[i].nPares > 0 && series[i].resto > 0)
         return 0;

   return HeapVazioIM_Arq2(heap) ? 1 : 0;
}

/****
 *
 * CarregaBufferMS2(): Preenche o contedo de um buffer com pares
 *                     lidos de um arquivo
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os pares
 *      buffer (entrada) - o buffer
 *      maxRegs (entrada) - nmero mximo de pares que sero lidos
 *     *pos (entrada/sada) - posio da prxima leitura no arquivo
 *
 * Retorno: O nmero de pares lidos
 *
 ****/
static int CarregaBufferMS2( FILE *stream, tChaveIndice *buffer,
                             int maxRegs, long *pos )
{
   int nPares; /* Nmero de pares lidos */

      /* Move o apontador de posio para o local de leitura */
   MoveApontador(stream, *pos, SEEK_SET);

      /* L os pares no arquivo */
   nPares = fread(buffer, TAM_PAR, maxRegs, stream);
   ASSEGURA(!ferror(stream), "Erro de leitura");

      /* Obtm a posio da prxima leitura */
   *pos = ftell(stream);

   return nPares;
}

/****
 *
 * CriaSeriesMS2(): Cria sries (ordenadas)
 *
 * Parmetros:
 *      nomeE (entrada) - stream associado ao arquivo que origina as sries
 *      nomeS (entrada) - nome do arquivo que armazenar as sries
 *
 * Retorno: Nmero de sries criadas
 *
 ****/
static int CriaSeriesMS2(FILE *streamE, const char *nomeS)
{
   tChaveIndice *pares; /* Ponteiro para um array que armazena */
                        /* uma srie em memria principal      */
   int           nPares,  /* Nmero de pares numa srie */
                 nSeries = 0; /* Nmero de sries */
   FILE         *streamS; /* Stream associado ao arquivo */
                          /* que armazenar as sries    */

      /* Aloca espao para o array. Um array deste */
      /* tamanho no pode ser alocado na pilha     */
   pares = calloc(MAX_PARES, TAM_PAR);
   ASSEGURA( pares, "Buffer nao foi alocado" );

      /* Tenta criar o arquivo que conter as sries */
   streamS = AbreArquivo(nomeS, "wb");

      /* Garante que a leitura comea no incio do arquivo */
   rewind(streamE);

      /* L o arquivo a ser ordenado e cria as sries */
   while (1) {
         /* L os pares da srie corrente */
      nPares = fread( pares, TAM_PAR, MAX_PARES, streamE );

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA( !ferror(streamE), "Erro de leitura em arquivo" );

         /* Encerra o lao se nenhum par foi lido */
      if (!nPares)
         break;

         /* Ordena os pares lidos */
      qsort(pares, nPares, TAM_PAR, ComparaPares);

         /* Escreve a srie corrente no arquivo */
      fwrite(pares, TAM_PAR, nPares, streamS);

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA( !ferror(streamS), "Erro de escrita de serie" );

      ++nSeries; /* Atualiza o nmero de sries */
   }

      /* Fecha o arquivo que armazena as sries */
   FechaArquivo(streamS, nomeS);

   free(pares); /* Libera o espao ocupado pelo array */

   return nSeries;
}

/****
 *
 * IntercalaMultiArq2(): Intercala sries usando intercalao multplice
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que contm as sries
 *      nomeS (entrada) - nome do arquivo que conter o resultado
 *                        da intercalao
 *      nSeries (entrada) - o nmero de sries que sero intercaladas
 *      tamSerie (entrada) - tamanho mximo de uma srie
 *
 * Retorno: Stream associado ao arquivo resultante da ordenao
 *
 ****/
static FILE *IntercalaMultiArq2( const char *nomeE, const char *nomeS,
                              int nSeries, int tamSerie )
{
   tSerie       *series; /* Ponteiro para um array contendo */
                         /* informaes sobre as sries     */
   int           tamBuffer; /* Tamanho de um buffer */
                            /* de entrada ou sada  */
   FILE         *streamS, /* Stream associado ao arquivo que */
                          /* conter os pares ordenados      */
                *streamE; /* Stream associado ao arquivo */
                          /* que contm as sries        */
   tHeapIM_Arq2         heap; /* Heap usado na ordenao dos elementos */
                       /* dos buffers de entrada                */
   int           i,
                 ibs = 0, /* ndice no qual o prximo par do */
                          /* buffer de sada ser armazenado */
                 iHeap = 0; /* ndice de um elemento do heap */
   tNoHeapIM_Arq2       raiz; /* Uma raiz do heap */
   tChaveIndice *bufferS; /* Apontar para o buffer de sada */

      /* Calcula o tamanho de cada buffer */
   tamBuffer = MAX_PARES/(nSeries + 1);

      /* O tamanho de um buffer no pode ser zero */
   ASSEGURA(tamBuffer, "Tamanho de buffer e' zero");

      /* Aloca espao para o buffer de sada */
   bufferS = calloc(tamBuffer, TAM_PAR);
   ASSEGURA(bufferS, "Impossivel alocar buffer de saida");

      /* Inicia o heap */
   IniciaHeapIM_Arq2(&heap, nSeries);

      /* Tenta abrir o arquivo contendo   */
      /* as sries que sero intercaladas */
   streamE = AbreArquivo(nomeE, "rb");

      /* Aloca um array que contm informaes sobre as sries */
   series = calloc(nSeries, sizeof(tSerie));
   ASSEGURA(series, "Impossivel alocar array de series");

      /* Aloca espao para cada buffer de entrada,  */
      /* preenche cada um com o contedo inicial de */
      /* sua respectiva srie e associa o primeiro  */
      /* par de cada buffer a um n do heap         */
   for (i = 0; i < nSeries; ++i) {
         /* Aloca espao para o buffer de entrada corrente */
      series[i].buffer = calloc(tamBuffer, TAM_PAR);
      ASSEGURA( series[i].buffer,
                "Impossivel alocar um buffer de entrada" );

         /* Obtm a posio do primeiro byte da srie */
      series[i].pos = i*tamSerie*TAM_PAR;

         /* Preenche o contedo do buffer de entrada corrente */
      series[i].nPares = CarregaBufferMS2(streamE, series[i].buffer,
                                       tamBuffer, &series[i].pos);

         /* Obtm o nmero de pares restantes na srie */
      series[i].resto = tamSerie - series[i].nPares;

         /* Associa dados do primeiro par do */
         /* buffer corrente a um n do heap  */
      heap.itens[iHeap].elemento = series[i].buffer;/* Endereo */
                                                    /* do par   */
      heap.itens[iHeap].ir = 0; /* Seu ndice no buffer */
      heap.itens[iHeap].is = i; /* ndice de sua srie */

      ++iHeap; /* Passa para o prximo elemento do heap */
   }

      /* Ordena o heap */
   for (i = (nSeries - 1)/2; i >= 0; --i)
      OrdenaHeapIM_Arq2(&heap, i, ComparaPares);

      /* Abre o arquivo que conter o resultado da ordenao */
   streamS = AbreArquivo(nomeS, "w+b");

      /* Obtm o menor elemento do heap e o substitui */
      /* pelo prximo elemento da srie que o contm  */
      /* at que todas as sries estejam esgotadas    */
   while (!SeriesEstaoVaziasMS2(series, nSeries, &heap)) {
         /* Obtm o menor valor armazenado no heap */
      raiz = ObtemMinimoIM_Arq2(&heap);

         /* Verifica se o buffer de sada est repleto */
      if (ibs > tamBuffer - 1) {
            /* O buffer de sada est repleto e  necessrio  */
            /* descarreg-lo no stream que contm o resultado */
         fwrite( bufferS, TAM_PAR, tamBuffer, streamS );
         ASSEGURA(!ferror(streamS), "Erro de escrita 1");

         ibs = 0; /* Reinicia o buffer de sada */
      }

         /* Armazena o menor valor do heap no buffer */
         /* que conter o resultado da intercalao  */
      bufferS[ibs++] = *raiz.elemento;

         /* O elemento que substituir a raiz pertence */
         /*  mesma srie  qual pertencia a raiz      */
      if (++raiz.ir < series[raiz.is].nPares) {
            /* Obtm o prximo elemento da   */
            /* srie  qual a raiz pertencia */
         raiz.elemento = &series[raiz.is].buffer[raiz.ir];
      } else { /* Esse buffer est esgotado */
            /* Tenta renovar o buffer de entrada  */
         series[raiz.is].nPares = CarregaBufferMS2( streamE,
                                           series[raiz.is].buffer,
                                           MIN(tamBuffer,
                                             series[raiz.is].resto),
                                           &series[raiz.is].pos );

            /* Calcula o restante da srie em arquivo */
         series[raiz.is].resto = series[raiz.is].nPares
                               ? series[raiz.is].resto -
                                 series[raiz.is].nPares
                               : 0;

            /* Se o buffer foi renovado, obtm */
            /* os dados do seu primeiro par    */
         if (series[raiz.is].nPares > 0) {
            raiz.elemento = series[raiz.is].buffer;
            raiz.ir = 0;
         } else {
            /* O buffer no foi renovado porque todos os */
            /* pares de sua srie j foram consumidos. A */
            /* nova raiz ser o ltimo elemento do heap, */
            /* que passar a ter um elemento a menos.    */

               /* Teste de consistncia */
            ASSEGURA(!series[raiz.is].resto,
                     "Nao deveria haver mais pares");

               /* O heap ter um elemento a menos */
            raiz = ReduzHeapIM_Arq2(&heap);
         }
      }

         /* Substitui a raiz do heap pelo novo */
         /* elemento e reordena o heap         */
      SubstituiMinimoIM_Arq2(&heap, &raiz, ComparaPares);
   }

      /* Verifica se restam pares no buffer de sada */
   if (ibs > 0) {
         /* Escreve os pares restantes no */
         /* stream que contm o resultado */
      fwrite( bufferS, TAM_PAR, ibs, streamS );
      ASSEGURA(!ferror(streamS), "Erro de escrita 2");
   }

      /* Libera os buffers de entrada */
   for (i = 0; i < nSeries; ++i)
      free(series[i].buffer);

   free(series); /* Libera array com informaes sobre as sries */

   free(bufferS); /* Libera o buffer de sada */

   FechaArquivo(streamE, nomeE); /* Fecha o arquivo de entrada */

   return streamS;
}

/****************** Funes de Bulkloading ****************/

/****
 *
 * InsereAcimaBM(): Insere uma chave num n interno de uma
 *                  rvore B+ armazenada em arquivo
 *
 * Parmetros:
 *    *nosAtivos (entrada/sada) - lista de ns internos ativos
 *                                 (i.e., armazenados em memria
 *                                 principal)
 *     chave (entrada) - chave a ser inserida num n do ltimo
 *                       nvel de ns internos
 *     pFilhoDireito (entrada) - posio no arquivo do n que ser
 *                               filho direito da chave a ser
 *                               inserida
 *     streamArvore (entrada) - stream associado ao arquivo que contm a rvore
 *
 * Retorno: Nada
 *
 ****/
static void InsereAcimaBM( tListaNosAtivos *nosAtivos, tChave chave,
                           int pFilhoDireito, FILE *streamArvore )
{
   int    iChave, /* ndice da chave no n no qual ela ser inserida */
          i;
   tChave chaveQSobe; /* Chave a ser inserida um nvel acima */
   tNoBM  noDireito, /* N direito que conter as chaves */
                     /* superiores de um n dividido     */
          antigaRaiz, /* Antiga raiz, se foi criada uma nova */
          novaRaiz;   /* Uma nova raiz */

      /* A lista no pode estar vazia */
   ASSEGURA(nosAtivos->nNos > 0, "InsereAcimaBM: lista de nos esta vazia");

   for (i = 0; i < nosAtivos->nNos; ++i) {
         /* O n corrente deve ser um n interno */
      ASSEGURA( nosAtivos->nosPos[i].no.tipoDoNo == INTERNO,
                "InsereAcimaBM: No' deveria ser interno" );

         /* A chave a ser inserida ser a ltima chave do n */
      iChave = nosAtivos->nosPos[i].no.conteudo.noInterno.nFilhos - 1;

         /* O ndice de insero da chave no pode ser negativo */
      ASSEGURA(iChave >= 0, "InsereAcimaBM: Indice de insercao e' negativo");

         /* Se o n corrente no estiver repleto, insere nele */
      if (iChave < G - 1) {
            /* Efetua a insero */
         nosAtivos->nosPos[i].no.conteudo.noInterno.chaves[iChave] = chave;
         nosAtivos->nosPos[i].no.conteudo.noInterno.filhos[iChave+1]=pFilhoDireito;

            /* O nmero de filhos do n aumentou */
         nosAtivos->nosPos[i].no.conteudo.noInterno.nFilhos++;

            /* Atualiza o n no arquivo */
         EscreveNoBM( streamArvore, nosAtivos->nosPos[i].pos,
                      &nosAtivos->nosPos[i].no );

         return; /* Servio completo */
      }

          /* O n corrente  completo e precisa ser dividido */
      DivideNoInternoBM( &nosAtivos->nosPos[i].no, iChave, chave,
                         pFilhoDireito, &noDireito, &chaveQSobe );

         /* Atualiza o n dividido no arquivo */
      EscreveNoBM(streamArvore, nosAtivos->nosPos[i].pos, &nosAtivos->nosPos[i].no);

         /* O filho direito da chave que subir para o nvel superior ser o */
         /* novo n criado no nvel corrente. Armazena este n no arquivo e  */
         /* guarda sua posio neste arquivo.                                */
      pFilhoDireito = AcrescentaNoBM(streamArvore, &noDireito);

         /* Substitui na lista de ns ativos o n */
         /* que foi dividido pelo novo n direito */
      nosAtivos->nosPos[i].no = noDireito;
      nosAtivos->nosPos[i].pos = pFilhoDireito;
   }

   /****                                                                  ****/
   /* Neste ponto, sabe-se que o ltimo n divido no lao for era uma raiz   */
   /* completa. Agora cria-se uma nova raiz contendo a chave que subiu tendo */
   /* como filho esquerdo o n dividido e como filho direito o novo n       */
   /****                                                                  ****/

   IniciaNoBM(INTERNO, &novaRaiz); /* Inicia a nova raiz */

   novaRaiz.conteudo.noInterno.chaves[0] = chaveQSobe; /* Copia a chave */

      /* L a antiga raiz em arquivo, pois ela  */
      /* foi substituda na lista de ns ativos */
   LeNoBM(streamArvore, 0, &antigaRaiz);

      /* O filho da esquerda da nova raiz  a antiga raiz que */
      /* precisa ser armazenada em outra posio no arquivo   */
      /* porque a raiz deve sempre ocupar a primeira posio  */
   novaRaiz.conteudo.noInterno.filhos[0] =
        AcrescentaNoBM(streamArvore, &antigaRaiz);

      /* O filho direito da raiz j foi inserido antes no arquivo e sua */
      /*  posio neste arquivo est guardada em 'pFilhoDireito'        */
   novaRaiz.conteudo.noInterno.filhos[1] = pFilhoDireito;

   novaRaiz.conteudo.noInterno.nFilhos = 2; /* A nova raiz tem dois filhos */

      /* Armazena a raiz da rvore a posio 0 do arquivo */
   EscreveNoBM(streamArvore, 0, &novaRaiz);

      /* Verifica se lista de ns ativos ainda tem espao */
   if (nosAtivos->nNos == nosAtivos->tamArray) {
         /* Array est repleto. Dobra sua capacidade */
      nosAtivos->tamArray *= 2;
      nosAtivos->nosPos = realloc( nosAtivos->nosPos, nosAtivos->
                                      tamArray* sizeof(tNoPos) );
      ASSEGURA( nosAtivos->nosPos, "InsereAcimaBM: "
                "Impossivel redimensionar lista de nos ativos" );
   }

      /* Acrescenta a nova raiz  lista de ns ativos */
   nosAtivos->nosPos[nosAtivos->nNos].no = novaRaiz;
   nosAtivos->nosPos[nosAtivos->nNos].pos = 0;
   nosAtivos->nNos++;
}

/****
 *
 * CriaArvoreBM(): Insere um conjunto de pares chave/ndice
 *                 ordenados numa rvore B+ usando bulkloading
 *
 * Parmetros:
 *     streamOrd (entrada) - stream associado ao arquivo contendo os pares ordenados
 *     streamArv (entrada) - stream associado ao arquivo que conter a rvore
 *     raiz (sada) - ponteiro para a raiz da rvore em memria principal
 *
 * Retorno: Nada
 *
 * Descrio do algoritmo: v. texto
 *
 * Observaes:
 *     1. Esta funo no checa se h chaves repetidas
 *     2. O arquivo que conter a rvore dever estar aberto
 *        num modo que permita escrita
 *     3. Esse arquivo deve estar inicialmente vazio
 *
 ****/
static void CriaArvoreBM( FILE *streamOrd, FILE *streamArv, tNoBM *raiz )
{
   tChave          chaveQSobe; /* Chave a ser inserida num n interno da rvore */
   tNoBM           folhaEsq, /* Folha esquerda ativa */
                   folhaDir; /* Folha direita ativa  */
   int             posFolhaEsq, /* Posio no arquivo de 'folhaEsq' */
                   posFolhaDir, /* Posio no arquivo de 'folhaDir' */
                   i;
   tListaNosAtivos nosAtivos; /* Lista de ns ativos */

      /* Garante que a leitura de pares comea no incio do arquivo que os contm */
   MoveApontador(streamOrd, 0L, SEEK_SET);

   IniciaNoBM(FOLHA, &folhaEsq); /* Inicia a primeira folha da rvore */

      /* Copia pares para a primeira folha */
   folhaEsq.conteudo.noFolha.nChaves =  fread(folhaEsq.conteudo.noFolha.chaves,
                                              sizeof(tChaveIndice), F, streamOrd);

    ASSEGURA(!ferror(streamOrd), "CriaArvoreBM: erro de leitura");

      /* Se o final do arquivo foi atingido, a rvore ter */
      /* apenas um n, que  a nica folha preenchida      */
   if (feof(streamOrd)) {
         /* Torna a folha a raiz da rvore */
      *raiz = folhaEsq;

      EscreveNoBM(streamArv, 0, raiz); /* Escreve a folha (raiz) no arquivo */

      return; /* Servio concludo */
   }

   IniciaNoBM(INTERNO, raiz); /* Inicia a raiz da rvore */

      /* A raiz da rvore ser o primeiro n do arquivo */
   EscreveNoBM(streamArv, 0, raiz);

      /* O primeiro filho da raiz apontar para a primeira folha */
   raiz->conteudo.noInterno.filhos[0] = posFolhaEsq =
         AcrescentaNoBM(streamArv, &folhaEsq);
   raiz->conteudo.noInterno.nFilhos = 1;

      /* Aloca um array para armazenar os ns internos ativos.  */
      /* O nmero mgico 5  uma boa estimativa inicial, j que */
      /* no se espera que a rvore tenha altura maior do que 4 */
   nosAtivos.tamArray = 5;
   nosAtivos.nosPos = calloc(nosAtivos.tamArray, sizeof(tNoPos));
   ASSEGURA(nosAtivos.nosPos, "CriaArvoreBM(): Imposssivel alocar lista de nos");

      /* Por enquanto, a raiz corrente  o nico n ativo */
   nosAtivos.nosPos[0].no = *raiz;
   nosAtivos.nosPos[0].pos = 0;
   nosAtivos.nNos = 1;

   IniciaNoBM(FOLHA, &folhaDir); /* Inicia a segunda folha da rvore */

      /* L os demais pares e acrescenta-os  rvore */
   while (1) {
         /* Copia pares para a folha direita */
      folhaDir.conteudo.noFolha.nChaves = fread(folhaDir.conteudo.noFolha.chaves,
                                                sizeof(tChaveIndice), F, streamOrd);
       ASSEGURA(!ferror(streamOrd), "Erro de leitura em CriaArvoreBM()");

         /* Se no foi lido nenhum par, encerra o lao */
      if (!folhaDir.conteudo.noFolha.nChaves)
         break;

         /* Acrescenta a folha direita ao arquivo */
         /* e obtm sua posio nesse arquivo     */
      posFolhaDir = AcrescentaNoBM(streamArv, &folhaDir);

         /* Efetua o encadeamento das folhas */
      folhaEsq.conteudo.noFolha.proximaFolha = posFolhaDir;
      folhaDir.conteudo.noFolha.proximaFolha = POSICAO_NULA;

         /* Atualiza a folha esquerda no arquivo */
      EscreveNoBM(streamArv, posFolhaEsq, &folhaEsq);

         /* A primeira chave do n da direita ser inserida num n interno */
      chaveQSobe = folhaDir.conteudo.noFolha.chaves[0].chave;

         /* Insere a chave em algum n interno da rvore */
      InsereAcimaBM(&nosAtivos, chaveQSobe, posFolhaDir, streamArv);

         /* A folha direita passar a ser a folha esquerda */
      folhaEsq = folhaDir;
      posFolhaEsq = posFolhaDir;
   }

      /* A raiz  o ltimo n da lista de ns ativos */
   *raiz = nosAtivos.nosPos[nosAtivos.nNos - 1].no;

      /* Atualiza no arquivo todos os ns ativos */
   for (i = 0; i < nosAtivos.nNos; ++i)
      EscreveNoBM(streamArv, nosAtivos.nosPos[i].pos, &nosAtivos.nosPos[i].no);

   free(nosAtivos.nosPos); /* A lista de ns ativos no  mais necessria */
}

/********************* Funes Globais ********************/

/****
 *
 * LeNoBM(): L em arquivo um n de uma rvore B+
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo
 *                        a rvore na qual ser feita a leitura
 *     pos (entrada) - posio no arquivo na qual ser feita a
 *                     leitura
 *    *no (sada) - ponteiro para o n que conter o resultado da leitura
 *
 * Retorno: Nada
 *
 ****/
void LeNoBM(FILE *stream, int pos, tNoBM *no)
{
      /* Verifica a validade do stream */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, LeNoBM));

      /* Tenta mover o apontador de arquivo para o local */
      /* de leitura; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tNoBM)*pos, SEEK_SET);

   fread(no, sizeof(tNoBM), 1, stream); /* Efetua a leitura */

      /* Certifica-se que no houve erro de leitura */
   ASSEGURA(!ferror(stream), ERRO_FREAD(LeNoBM));
}

/****
 *
 * BuscaBM(): Faz buscas numa rvore B+
 *
 * Parmetros:
 *     chave (entrada) - a chave de busca
 *     raiz (entrada) - ponteiro para a raiz da rvore na qual
 *                      ser feita a busca
 *     stream (entrada) - stream representando o arquivo no qual
 *                        a rvore reside
 *
 * Retorno: ndice do registro no arquivo de registros se a
 *          chave for encontrada; POSICAO_NULA, se a chave
 *          no for encontrada.
 *
 ****/
int BuscaBM(FILE *stream, const tNoBM *raiz, tChave chave)
{
   int   i,
         encontrado = 0, /* Informa se a chave foi encontrada */
         posNo = 0; /* Posio no n no arquivo */
   tNoBM no; /* Armazenar cada n visitado */

      /* Verifica se o stream  vlido */
   ASSEGURA( stream, ERRO_OPEN(BuscaBM) );

   no = *raiz;

      /* Desce na rvore at encontrar uma folha */
   while (no.tipoDoNo == INTERNO) {
      i = BuscaEmNoBM(chave, &no, &encontrado);

         /* Desce at o prximo n */
      posNo = encontrado ? no.conteudo.noInterno.filhos[i + 1]
                         : no.conteudo.noInterno.filhos[i];
      LeNoBM(stream, posNo, &no);
   }

      /* Chegou-se a uma folha. Se a chave se encontrar */
      /* na rvore, ela deve estar nessa folha          */
   i = BuscaEmNoBM(chave, &no, &encontrado);

   return encontrado ? no.conteudo.noFolha.chaves[i].indice
                     : POSICAO_NULA;
}

/****
 *
 * InsereBM(): Faz insero numa rvore B+ armazenada em arquivo
 *
 * Parmetros:
 *    *raiz (entrada/sada) - ponteiro para a raiz da rvore
 *                            (armazenada em memria)
 *     streamArvore (entrada) - stream associado ao arquivo que
 *                              contm a rvore
 *    *chaveIndice (entrada/sada) - par chave/ndice contendo
 *                                   a chave a ser inserida
 *
 * Retorno: 1, se houver insero; 0, em caso contrrio.
 *
 * Descrio do algoritmo: v. texto
 *
 * Observaes: A chave  considerada primria; i.e., chaves
 *              repetidas no so permitidas na rvore.
 *
 ****/
int InsereBM( tNoBM *raiz, tChaveIndice *chaveIndice,
              FILE *streamArvore )
{
   int         encontrado, /* A chave foi encontrada? */
               indiceDaChave; /* ndice da chave no n no */
                              /* qual ela , ou deveria   */
                              /* ser, encontrada          */
   tChave      chaveAInserir, /* A chave a ser inserida */
                              /* no nvel corrente      */
               chaveQueSobe;  /* A chave a ser inserida */
                              /* um nvel acima         */
   tNoBM       pNoDireito; /* N direito que conter as chaves */
                           /* superiores de um n dividido     */
   int         pFilhoDireito; /* Posio no arquivo do n que */
                              /* ser filho direito de uma    */
                              /* chave inserida               */
   tNoCaminho  noCaminho; /* Um elemento da pilha que    */
                          /* representa o caminho de ns */
   tPilha      pilha; /* Pilha que armazenar o   */
                      /* caminho de ns visitados */

      /* Verifica a validade do stream */
   ASSEGURA( streamArvore,
             ERRO_STREAM_NULL(streamArvore, InsereBM) );

      /* Verifica se o ponteiro para a raiz  vlido */
   ASSEGURA(raiz, "ERRO: A raiz  NULL em InsereBM\n");

      /* Verifica se a rvore est vazia */
   if (raiz->tipoDoNo == NO_VAZIO) { /* A rvore ainda  virgem */
      IniciaNoBM(FOLHA, raiz); /* Inicia a raiz da rvore */
      raiz->conteudo.noFolha.chaves[0] = *chaveIndice;
      raiz->conteudo.noFolha.nChaves = 1;

         /* Armazena a raiz da rvore no arquivo, */
         /* no qual ela ocupa a posio 0         */
      EscreveNoBM(streamArvore, 0, raiz);

      return 1;
   }

      /* Encontra o caminho que conduz at o local  */
      /* de insero ou onde a chave j se encontra */
   EncontraCaminhoBM( streamArvore, chaveIndice->chave, raiz,
                      &indiceDaChave, &encontrado, &pilha );

      /* Verifica se a chave foi encontrada */
   if (encontrado) {
         /* A chave foi encontrada e no precisa inseri-la. */
         /* O contedo da pilha no  mais necessrio       */
      EsvaziaPilha(&pilha);

      return 0; /* A chave  primria */
   }

      /* A chave no foi encontrada. Insere a chave */
      /* na folha que deve estar no topo da pilha   */

   noCaminho = Desempilha(&pilha);

   ASSEGURA( noCaminho.no.tipoDoNo == FOLHA,
             "Erro em InsereBM(): No' deveria ser folha" );

      /* Caso 1: H espao para insero na folha */
   if (noCaminho.no.conteudo.noFolha.nChaves < F) {
      InsereEmFolhaBM(&noCaminho.no, indiceDaChave, *chaveIndice);

         /* Atualiza o n no arquivo */
      EscreveNoBM(streamArvore, noCaminho.endereco, &noCaminho.no);

         /* Se a raiz da rvore foi atualizada no */
         /* arquivo sua cpia mantida em memria  */
         /* tambm precisa ser atualizada         */
      if (!noCaminho.endereco)
            /* A posio da raiz no arquivo  zero */
         *raiz = noCaminho.no;

         /* O contedo da pilha no  mais necessrio */
      EsvaziaPilha(&pilha);

      return 1;
   }

      /* Caso 2: No h espao para insero na folha */

   DivideFolhaBM( &noCaminho.no, indiceDaChave, *chaveIndice,
                  &pNoDireito, &chaveQueSobe );

      /* Faz o encadeamento de 'pNoDireito' com a folha  */
      /* para qual aponta correntemente a folha dividida */
   pNoDireito.conteudo.noFolha.proximaFolha =
        noCaminho.no.conteudo.noFolha.proximaFolha;

   /****                                                     ****/
   /* Armazena pNoDireito no arquivo. A funo AcrescentaNoBM() */
   /* retorna a posio na qual o n foi armazenado no arquivo. */
   /* 'pFilhoDireito' armazena essa posio.                    */
   /****                                                     ****/

      /* O filho direito da chave que subir para o */
      /* nvel superior ser a nova folha criada    */
   pFilhoDireito = AcrescentaNoBM(streamArvore, &pNoDireito);

      /* Faz o encadeamento da folha   */
      /* dividida com sua prxima irm */
   noCaminho.no.conteudo.noFolha.proximaFolha = pFilhoDireito;

      /* Atualiza a folha dividida no arquivo */
   EscreveNoBM(streamArvore, noCaminho.endereco, &noCaminho.no);

      /* Se a raiz da rvore foi atualizada no arquivo sua   */
      /* cpia mantida em memria tambm deve ser atualizada */
   if (!noCaminho.endereco)
      *raiz = noCaminho.no;

      /* A seguir, percorre-se o caminho inverso at */
      /* encontrar a raiz ou um n incompleto.       */

   indiceDaChave = noCaminho.pos;
   chaveAInserir = chaveQueSobe;

   while (!PilhaVazia(pilha)) {
      noCaminho = Desempilha(&pilha);

      ASSEGURA( noCaminho.no.tipoDoNo == INTERNO,
                "Erro em InsereBM(): No' deveria ser interno" );

         /* Verifica se o n est completo */
      if (noCaminho.no.conteudo.noInterno.nFilhos < G) {
            /* D para inserir neste n */
         InsereEmNoInterno( &noCaminho.no, indiceDaChave,
                            chaveAInserir, pFilhoDireito );

            /* Atualiza o n no arquivo */
         EscreveNoBM( streamArvore, noCaminho.endereco,
                      &noCaminho.no );

            /* Se a raiz da rvore foi atualizada no */
            /* arquivo sua cpia mantida em memria  */
            /* tambm deve ser atualizada            */
         if (!noCaminho.endereco)
               /* A raiz ocupa a posio zero no arquivo */
            *raiz = noCaminho.no;

            /* A pilha no  mais necessria */
         EsvaziaPilha(&pilha);

         return 1;
      }

          /* O n corrente  completo e precisa ser dividido */
      DivideNoInternoBM( &noCaminho.no, indiceDaChave,
                         chaveAInserir, pFilhoDireito,
                         &pNoDireito, &chaveQueSobe );

         /* Atualiza o n dividido no arquivo */
      EscreveNoBM(streamArvore, noCaminho.endereco, &noCaminho.no);

         /* Se a raiz da rvore foi atualizada no */
         /* arquivo sua cpia mantida em memria  */
         /* tambm deve ser atualizada            */
      if (!noCaminho.endereco)
         *raiz = noCaminho.no;

         /* O filho direito da chave que subir para o     */
         /* nvel superior ser o novo n criado no nvel  */
         /* corrente. Armazena este n no arquivo e guarda */
         /* sua posio neste arquivo.                     */
      pFilhoDireito = AcrescentaNoBM(streamArvore, &pNoDireito);

      indiceDaChave = noCaminho.pos;
      chaveAInserir = chaveQueSobe;
   }

   /****                                                   ****/
   /* Neste ponto, sabe-se que o ltimo n desempilhado  uma */
   /* raiz completa, que j foi dividida no lao while ou     */
   /* antes se a raiz era tambm uma folha. Resta criar uma   */
   /* nova raiz contendo a chave que subiu at ela tendo como */
   /* filho esquerdo o n dividido e  direita o novo n      */
   /****                                                   ****/

   IniciaNoBM(INTERNO, raiz); /* Inicia a nova raiz */

      /* Copia a chave */
   raiz->conteudo.noInterno.chaves[0] = chaveQueSobe;

      /* O filho da esquerda da nova raiz  a antiga raiz que */
      /* precisa ser armazenada em outra posio no arquivo   */
      /* porque a raiz deve sempre ocupar a primeira posio  */
   raiz->conteudo.noInterno.filhos[0] =
        AcrescentaNoBM(streamArvore, &noCaminho.no);

      /* O filho direito da raiz j foi inserido antes */
      /* no arquivo e sua posio neste arquivo est   */
      /* guardada em 'pFilhoDireito'                   */
   raiz->conteudo.noInterno.filhos[1] = pFilhoDireito;

      /* A nova raiz tem dois filhos */
   raiz->conteudo.noInterno.nFilhos = 2;

      /* Armazena a raiz da rvore no arquivo, */
      /* no qual ela ocupa a posio 0         */
   EscreveNoBM(streamArvore, 0, raiz);

   /****                                   ****/
   /* Neste ponto, a cpia da raiz armazenada */
   /* em memria j est atualizada           */
   /****                                   ****/

   return 1;
}

/****
 *
 * RemoveBM(): Remove a chave especificada de uma rvore
 *             B+ armazenada em arquivo
 *
 * Parmetros:
 *       streamArvore (entrada) - stream associado ao arquivo
 *                                que armazena a rvore
 *      *raiz (entrada/sada) - raiz da rvore
 *       chave (entrada) - a chave que ser removida
 *
 * Retorno: 1, se a remoo for bem sucedida; 0, em caso contrrio
 *
 * Descrio do algoritmo: v. texto
 *
 ****/
int RemoveBM(FILE *streamArvore, tNoBM *raiz, tChave chave)
{
   int        i,
              encontrado, /* A chave foi encontrada? */
              indiceDaChave, /* ndice da chave na follha na  */
                             /* qual ela est ou deveria estar */
              nChaves, /* Nmero de chaves de uma folha */
              posFolha, /* Posio da folha entre suas irms */
              endFolha, /* Posio da folha no arquivo */
              endIrma; /* Posio de uma irm da folha no arquivo */
   tPilha     pilha; /* Pilha que armazenar o caminho */
                     /* de ns visitados               */
   tNoCaminho caminho; /* Um elemento da pilha que    */
                       /* representa o caminho de ns */
   tChave     substituta, /* Chave que substituir a  */
                          /* chave removida no ndice */
              movida, /* Chave movida de uma folha para outra */
              aRemover; /* Chave que ser removida no n-pai */
   tNoBM      irma, /* Irm da folha na qual ocorreu a remoo */
              folha; /* Folha na qual ocorre a remoo */

      /* Se a rvore for vazia, no h chave para remover */
   if (!raiz)
      return 0; /* No h chave para remover */

      /* Encontra o caminho da raiz at o n */
      /* que contm a chave a ser removida   */
   EncontraCaminhoBM( streamArvore, chave, raiz, &indiceDaChave,
                      &encontrado, &pilha );

      /* Se a chave no for encontrada, no h mais nada a  */
      /* fazer. Mas, antes de retornar, esvazia-se a pilha. */
   if (!encontrado) {
      EsvaziaPilha(&pilha);
      return 0; /* No h chave para remover */
   }

   /***                                           ***/
   /** A chave foi encontrada e informaes sobre  **/
   /** a folha que a contm esto no topo da pilha **/
   /***                                           ***/

      /* Neste ponto, a pilha no pode estar vazia */
   ASSEGURA(!PilhaVazia(pilha),
            "A pilha nao deveria estar vazia em RemoveBM()");

      /* No topo da pilha, encontra-se a folha */
      /* na qual a remoo ser efetuada       */
   caminho = Desempilha(&pilha);

      /* Garante que o elemento no topo da pilha era um folha */
   ASSEGURA( caminho.no.tipoDoNo == FOLHA,
             "ERRO: No' deveria ser folha em RemoveBM()" );

      /* Facilita a legibilidade */
   nChaves = caminho.no.conteudo.noFolha.nChaves;

      /* Move as chaves uma posio para trs */
   for ( i = indiceDaChave; i < nChaves - 1; ++i )
      caminho.no.conteudo.noFolha.chaves[i] =
         caminho.no.conteudo.noFolha.chaves[i + 1];

      /* O nmero de chaves da folha foi reduzido */
   nChaves = --caminho.no.conteudo.noFolha.nChaves;

      /* Atualiza a folha no arquivo */
   EscreveNoBM(streamArvore, caminho.endereco, &caminho.no);

      /* Verifica se a folha  a raiz da rvore */
   if (PilhaVazia(pilha)) {
      *raiz = caminho.no; /* Atualiza a raiz que se encontra */
                          /* em memria principal            */
      return 1; /* Servio completo */
   }

      /* Se a chave removida era a menor chave da folha e o nmero */
      /* de chaves na folha continuar acima do mnimo requerido,   */
      /* resta apenas substituir a chave removida no ndice        */
   if (!indiceDaChave && nChaves >= TF) {
      substituta = caminho.no.conteudo.noFolha.chaves[0].chave;
      AtualizaIndice(streamArvore, &pilha, chave, &substituta, NULL);
      return 1;
   } else if (nChaves >= TF) {
         /* A chave removida no era a primeira da folha e o */
         /* nmero de chaves nela permanece dentro do limite */
      EsvaziaPilha(&pilha);
      return 1; /* No resta mais nada a fazer */
   }

      /***                                                  ***/
      /* Neste ponto, sabe-se que o nmero de chaves na folha */
      /* ficou abaixo do mnimo requerido. Agora existem duas */
      /* opes: (1) redistribuio de chaves ou (2) fuso de */
      /* folhas.                                              */
      /***                                                  ***/

      /* Guarda a folha na qual ocorreu a remoo */
   folha = caminho.no;

      /* Guarda a posio da folha no arquivo */
   endFolha = caminho.endereco;

      /* Guarda a posio da folha entre suas irms */
   posFolha = caminho.pos;

      /* Obtm o n interno que se encontra */
      /* no topo da pilha sem desempilh-lo */
   caminho = ElementoTopo(&pilha);

      /* Garante que o n no topo da pilha  interno */
   ASSEGURA( caminho.no.tipoDoNo == INTERNO,
             "ERRO: No' deveria ser interno em RemoveBM()" );

      /*************************************/
      /* Opo 1: redistribuio de chaves */
      /* =======                           */
      /*************************************/

      /* Verifica se a folha possui irm  esquerda */
   if (posFolha) {
         /* Obtm a posio da irm esquerda no arquivo */
      endIrma = caminho.no.conteudo.noInterno.filhos[posFolha - 1];

         /* L a irm esquerda da folha */
      LeNoBM( streamArvore, endIrma, &irma );

         /* O n lido deve ser uma folha */
      ASSEGURA( irma.tipoDoNo == FOLHA,
                "ERRO: No' deveria ser folha em RemoveBM()" );

         /* Verifica se a irm pode ceder uma chave */
      if (irma.conteudo.noInterno.nFilhos > TF) {
            /*                                           */
            /* Move a ltima chave da irm esquerda para */
            /* o incio da folha que ficou deficiente    */
            /*                                           */

            /* Primeiro, abre espao para a nova chave */
         for ( i = folha.conteudo.noFolha.nChaves - 1; i >= 0;
               i-- )
            folha.conteudo.noFolha.chaves[i + 1] =
                            folha.conteudo.noFolha.chaves[i];

            /* Obtm a chave que ser substituta */
         substituta = irma.conteudo.noFolha.chaves[
                      irma.conteudo.noFolha.nChaves - 1 ].chave;

            /* Copia a chave */
         folha.conteudo.noFolha.chaves[0] =
            irma.conteudo.noFolha.chaves
            [irma.conteudo.noFolha.nChaves - 1];

            /* O nmero de chaves da folha aumentou e  */
            /* o nmero de chaves de sua irm diminuiu */
         folha.conteudo.noFolha.nChaves++;
         irma.conteudo.noFolha.nChaves--;

            /* Atualiza a folha e sua irm no arquivo */
         EscreveNoBM(streamArvore, endFolha, &folha);
         EscreveNoBM(streamArvore, endIrma, &irma);

            /* Atualiza os ns internos */
         AtualizaIndice( streamArvore, &pilha, chave, &substituta,
                         NULL );

         return 1;
      }
   } else {  /* A folha deve possuir irm  direita */
      ASSEGURA(posFolha < caminho.no.conteudo.noInterno.nFilhos - 1,
               "ERRO: A folha deveria possuir irma direita");

         /* Obtm a posio da folha no arquivo */
      endIrma = caminho.no.conteudo.noInterno.filhos[posFolha + 1];

         /* L a irm direita da folha */
      LeNoBM( streamArvore, endIrma, &irma );

         /* O n lido deve ser uma folha */
      ASSEGURA( irma.tipoDoNo == FOLHA,
                "ERRO: No' deveria ser folha em RemoveBM()" );

         /* Verifica se a irm pode ceder uma chave */
      if (irma.conteudo.noInterno.nFilhos > TF) {
            /* Obtm a chave a ser movida */
         movida = irma.conteudo.noFolha.chaves[0].chave;

            /* Move a primeira chave da irm direita para */
            /* o final da folha que ficou deficiente      */
         folha.conteudo.noFolha.chaves
              [folha.conteudo.noFolha.nChaves] =
              irma.conteudo.noFolha.chaves[0];

            /* Remove a chave movida da irm */
         for (i = 0; i > irma.conteudo.noFolha.nChaves - 1; ++i)
            irma.conteudo.noFolha.chaves[i] =
                    irma.conteudo.noFolha.chaves[i + 1];

            /* A irm  uma folha direita. Portanto, */
            /* sua primeira chave aparece no n pai, */
            /* que precisar ser atualizado          */
         for ( i = 0; i < caminho.no.conteudo.noInterno.nFilhos - 1;
               ++i ) {
            if (caminho.no.conteudo.noInterno.chaves[i] == movida) {
               caminho.no.conteudo.noInterno.chaves[i] =
                 irma.conteudo.noFolha.chaves[0].chave;
            }
         }

            /* A chave movida deve ter sido */
            /* encontrada no n interno     */
         ASSEGURA( i < caminho.no.conteudo.noInterno.nFilhos,
                   "Chave a ser substituida nao foi encontrada");

            /* Atualiza o n interno no arquivo */
         EscreveNoBM(streamArvore, caminho.endereco, &caminho.no);

            /* O nmero de chaves da folha aumentou e  */
            /* o nmero de chaves de sua irm diminuiu */
         folha.conteudo.noFolha.nChaves++;
         irma.conteudo.noFolha.nChaves--;

            /* Atualiza a folha e sua irm no arquivo */
         EscreveNoBM(streamArvore, endFolha, &folha);
         EscreveNoBM(streamArvore, endIrma, &irma);

            /* Verifica se a chave removida aparece no ndice */
         if (!indiceDaChave) {
               /* A chave substituta  a primeira chave da folha */
            substituta = folha.conteudo.noFolha.chaves[0].chave;

               /* O elemento que se encontra no topo da pilha */
               /* foi alterado e  preciso desempilh-lo e    */
               /* empilhar o elemento alterado                */
            ASSEGURA( !PilhaVazia(pilha),
                      "ERRO: Pilha vazia em RemoveBM()" );
            (void) Desempilha(&pilha);
            Empilha(caminho, &pilha);

               /* Atualiza os ns internos */
            AtualizaIndice( streamArvore, &pilha, chave,
                            &substituta, NULL );
         } else { /* A chave no aparece no ndice */
            EsvaziaPilha(&pilha);
            return 1; /* No resta mais nada a fazer */
         }

         return 1;
      }
   }

      /****************************/
      /* Opo 2: fuso de folhas */
      /* =======                  */
      /****************************/

      /* Verifica se a folha possui irm  esquerda */
   if (posFolha) {
         /* Obtm a posio da folha no arquivo */
      endIrma = caminho.no.conteudo.noInterno.filhos[posFolha - 1];

         /* L a irm esquerda da folha */
      LeNoBM( streamArvore, endIrma, &irma );

         /* O n lido deve ser uma folha */
      ASSEGURA( irma.tipoDoNo == FOLHA,
                "ERRO: No' deveria ser folha em RemoveBM()" );

         /* A irm deve estar no limite mnimo de chave */
      ASSEGURA( irma.conteudo.noFolha.nChaves == TF,
                "Folha nao tem numero minimo de chaves");
            /*                                         */
            /* Move as chaves da folha na qual ocorreu */
            /* a remoo para sua irm esquerda        */
            /*                                         */

         /* Se a chave removida no era a primeira da folha, */
         /* a chave a ser removida no n pai ser a primeira */
         /* chave da folha na qual ocorreu a remoo         */
      if (indiceDaChave) {
         aRemover = folha.conteudo.noFolha.chaves[0].chave;
      } else {
         aRemover = chave;
      }

         /* Copia as chaves */
      for (i = 0; i < folha.conteudo.noFolha.nChaves; ++i) {
         irma.conteudo.noFolha.chaves[i + TF] =
              folha.conteudo.noFolha.chaves[i];
      }

         /* O nmero de chaves da folha ficou nulo  */
         /* o nmero de chaves de sua irm aumentou */
      folha.tipoDoNo = NO_VAZIO;
      irma.conteudo.noFolha.nChaves +=
          folha.conteudo.noFolha.nChaves;

         /* Preserva o encadeamento das folhas */
      irma.conteudo.noFolha.proximaFolha =
          folha.conteudo.noFolha.proximaFolha;

         /* Atualiza a folha e sua irm no arquivo */
      EscreveNoBM(streamArvore, endFolha, &folha);
      EscreveNoBM(streamArvore, endIrma, &irma);

         /* Chama AtualizaIndice() para completar o servio */
      AtualizaIndice(streamArvore, &pilha, chave, NULL, &aRemover);
   } else {  /* A folha deve possuir irm  direita */
      ASSEGURA(posFolha < caminho.no.conteudo.noInterno.nFilhos - 1,
               "ERRO: A folha deveria possuir irma direita");

         /* Obtm a posio da irm no arquivo */
      endIrma = caminho.no.conteudo.noInterno.filhos[posFolha + 1];

         /* L a irm direita da folha */
      LeNoBM( streamArvore, endIrma, &irma );

         /* O n lido deve ser uma folha */
      ASSEGURA( irma.tipoDoNo == FOLHA,
                "ERRO: No' deveria ser folha em RemoveBM()" );

         /* A irm deve estar no limite mnimo de chave */
      ASSEGURA( irma.conteudo.noFolha.nChaves == TF,
                "Folha nao tem numero minimo de chaves");
            /*                                       */
            /* Move as chaves da irm direita para a */
            /* folha na qual ocorreu a remoo       */
            /*                                       */

         /* Se a chave removida era a primeira da folha, */
         /* ela dever ser substituda no ndice         */
      if (!indiceDaChave) {
         substituta = folha.conteudo.noFolha.chaves[0].chave;
      }

         /* Chave que ser removida no n-pai */
      aRemover = irma.conteudo.noFolha.chaves[0].chave;

         /* Copia as chaves */
      for (i = 0; i < TF; ++i) {
         folha.conteudo.noFolha.chaves[ i +
              folha.conteudo.noFolha.nChaves ] =
              irma.conteudo.noFolha.chaves[i];
      }

         /* O nmero de chaves da irm ficou nulo  */
         /* e o nmero de chaves da folha aumentou */
      irma.tipoDoNo = NO_VAZIO;
      folha.conteudo.noFolha.nChaves += TF;

         /* Preserva o encadeamento das folhas */
      folha.conteudo.noFolha.proximaFolha =
           irma.conteudo.noFolha.proximaFolha;

         /* Atualiza a folha e sua irm no arquivo */
      EscreveNoBM(streamArvore, endFolha, &folha);
      EscreveNoBM(streamArvore, endIrma, &irma);

         /* Chama AtualizaIndice() para completar o servio */
      if (indiceDaChave) {
         AtualizaIndice( streamArvore, &pilha, chave, &substituta,
                         &aRemover );
      } else {
         AtualizaIndice( streamArvore, &pilha, chave, NULL,
                         &aRemover );
      }
   }

   return 1; /* Remoo bem sucedida */
}

/****
 *
 * MenorChaveBM(): Encontra a menor chave de uma rvore B+
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que
 *                        contm a rvore
 *
 * Retorno: A menor chave ou POSICAO_NULA se a rvore estiver vazia
 *
 ****/
int MenorChaveBM(FILE *stream)
{
   tNoBM umNo;

   ASSEGURA( stream,
             ERRO_STREAM_NULL(stream, MenorChaveBM) );

   rewind(stream);

      /* O primeiro n  a raiz */
   fread(&umNo, sizeof(umNo), 1, stream);

   ASSEGURA( !feof(stream) & !ferror(stream),
             ERRO_FREAD(MenorChaveBM) );

   if (umNo.tipoDoNo == NO_VAZIO)
      return POSICAO_NULA;

      /* Desce at encontrar a folha mais  esquerda */
   while (umNo.tipoDoNo == INTERNO)
         /* Desce pelo filho mais  esquerda */
      LeNoBM(stream, umNo.conteudo.noInterno.filhos[0], &umNo);

   ASSEGURA( umNo.tipoDoNo == FOLHA,
             ERRO_INCONSISTENCIA(MenorChaveBM) );

      /* A menor chave da rvore  a primeira chave */
      /* do n mais  esquerda da rvore.           */
   return umNo.conteudo.noFolha.chaves[0].chave;
}

/****
 *
 * MaiorChaveBM(): Encontra a maior chave de uma rvore B+
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que
 *                        contm a rvore
 *
 * Retorno: A maior chave ou POSICAO_NULA se a rvore estiver vazia
 *
 ****/
int MaiorChaveBM(FILE *stream)
{
   tNoBM umNo;

   ASSEGURA( stream,
             ERRO_STREAM_NULL(stream, MaiorChaveBM) );

   rewind(stream);

      /* O primeiro n  a raiz */
   fread(&umNo, sizeof(umNo), 1, stream);

   ASSEGURA( !feof(stream) & !ferror(stream),
             ERRO_FREAD(MaiorChaveBM) );

   if (umNo.tipoDoNo == NO_VAZIO)
      return POSICAO_NULA;

      /* Desce at encontrar a folha mais  direita */
   while (umNo.tipoDoNo == INTERNO)
         /* Desce pelo filho mais  direita */
      LeNoBM( stream, umNo.conteudo.noInterno.filhos
                   [umNo.conteudo.noInterno.nFilhos - 1], &umNo );

      /* O ltimo n lido deve ser uma folha */
   ASSEGURA( umNo.tipoDoNo == FOLHA,
             ERRO_INCONSISTENCIA(MaiorChaveBM) );

      /* A maior chave da rvore  a ltima chave */
      /* da folha mais  direita da rvore        */
   return umNo.conteudo.noFolha.chaves
          [umNo.conteudo.noFolha.nChaves-1].chave;
}

/****
 *
 * EscreveChavesOrdenadas(): Visita todas as folhas de uma
 *                           rvore B+ imprimindo todas as
 *                           chaves da rvore em ordem crescente
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *    *streamArv (entrada) - stream associado ao arquivo que
 *                           contm a rvore
 *     streamChaves (entrada) - stream associado ao arquivo no
 *                              qual sero armazenadas as chaves
 *                              visitadas
 *
 * Retorno: 1, se tudo ocorrer bem; 0, em caso contrrio.
 *
 ****/
void EscreveChavesOrdenadas( const tNoBM* arvore, FILE *streamArv,
                             FILE *streamChaves )
{
   int   i;
   tNoBM umNo;

      /*  */
   ASSEGURA( streamArv,
             ERRO_STREAM_NULL(streamArv, EscreveChavesOrdenadas) );

      /*  */
   ASSEGURA( streamChaves,
             ERRO_STREAM_NULL( streamChaves,
                              EscreveChavesOrdenadas) );

   ASSEGURA( arvore,
             "Erro: A arvore NULL em EscreveChavesOrdenadas");

   ASSEGURA( arvore->tipoDoNo != NO_VAZIO,
             "Erro: A arvore vazia em EscreveChavesOrdenadas");

   umNo = *arvore;

      /* Desce sempre pelo filho mais  esquerda */
      /* at encontrar a primeira folha          */
   while (umNo.tipoDoNo == INTERNO)
      LeNoBM( streamArv, umNo.conteudo.noInterno.filhos[0],
            &umNo );

   ASSEGURA( umNo.tipoDoNo == FOLHA,
             ERRO_INCONSISTENCIA(EscreveChavesOrdenadas) );

      /* Chegou-se  folha mais  esquerda. Como as folhas   */
      /* formam uma lista encadeada e essa folha encabea    */
      /* a lista, deve-se percorr-la imprimindo suas chaves */
   do {
         /* Escreve cada chave da folha */
      for (i = 0; i < umNo.conteudo.noFolha.nChaves; ++i) {
            /* Escreve a chave corrente */
         fprintf( streamChaves, "%ld",
                  umNo.conteudo.noFolha.chaves[i].chave );

         fputc('\n', streamChaves); /* Salta linha */
      }

         /* Passa para a prxima folha */
      LeNoBM( streamArv, umNo.conteudo.noFolha.proximaFolha,
             &umNo );
   } while (umNo.conteudo.noFolha.proximaFolha != POSICAO_NULA);
}

/****
 *
 * BuscaIntervaloBM(): Coleta num arquivo todas as chaves que
 *                     esto entre duas chaves dadas de uma
 *                     rvore B+
 *
 * Parmetros:
 *     chave1 (entrada) - primeira chave que define o intervalo
 *     chave2 (entrada) - segunda chave que define o intervalo
 *     arvore (entrada) - ponteiro para a raiz da rvore na qual
 *                        ser feita a busca
 *    *streamArv (entrada) - stream associado ao arquivo que
 *                           contm a rvore
 *     streamChaves (entrada) - stream associado ao arquivo que
 *                              conter as chaves resultantes
 *                              da busca
 *
 * Retorno: 1, se a coleta de chaves for bem sucedida;
 *          0, em caso contrrio
 *
 * Observaes:
 *    1. No faz diferena qual das duas chaves  menor do
 *       que a outra, mas as chaves devem ser diferentes
 *       para que haja um intervalo.
 *    2. Numa aplicao prtica (e no meramente didtica),
 *       faria mais sentido coletar os registros (e no apenas
 *       as chaves). Mas, implementar uma extenso desta funo
 *       para que ela seja mais pragmtica  considerado trivial
 *
 ****/
int BuscaIntervaloBM( tChave chave1, tChave chave2,
                      const tNoBM* arvore,
                      FILE *streamArv, FILE *streamChaves )
{
   tChave menorChave, /* Menor chave */
          maiorChave; /* Maior chave */
   tNoBM  no; /* Armazenar cada n visitado */
   int    i,
          encontrado = 0, /* Usada para chamar BuscaEmNoBM() */
          posNo = 0; /* Posio do n no arquivo */

      /* Verifica se streamArv  vlido */
   ASSEGURA( streamArv,
             ERRO_STREAM_NULL(streamArv, BuscaIntervaloBM) );

      /* Verifica se streamChaves  vlido */
   ASSEGURA( streamChaves,
             ERRO_STREAM_NULL(streamChaves, BuscaIntervaloBM) );

      /* Verifica se o ponteiro para a raiz  vlido */
   ASSEGURA( arvore,
             "Erro: A arvore NULL em BuscaIntervaloBM");

      /* Verifica se a rvore  vazia */
   if (arvore->tipoDoNo == NO_VAZIO)
      return 0; /* A rvore  vazia */

      /* Se as chaves forem iguais, no h intervalo */
   if (chave1 == chave2)
      return 0; /* No h intervalo */

   /****                                               ****/
   /* O trecho a seguir permite que as chaves que definem */
   /* o intervalo sejam introduzidas em qualquer ordem    */
   /****                                               ****/

   if (chave1 < chave2) { /* chave1 < chave2 */
      menorChave = chave1;
      maiorChave = chave2;
   } else { /* chave1 >= chave2 */
      menorChave = chave2;
      maiorChave = chave1;
   }

      /* A busca comea na raiz da rvore */
   no = *arvore;

      /* Desce na rvore at encontrar uma folha */
   while (no.tipoDoNo == INTERNO) {
      i = BuscaEmNoBM(menorChave, &no, &encontrado);

         /* Desce at o prximo n */
      posNo = encontrado ? no.conteudo.noInterno.filhos[i + 1]
                         : no.conteudo.noInterno.filhos[i];

      LeNoBM(streamArv, posNo, &no);
   }

      /* Chegou-se a uma folha. Encontra, nessa */
      /* folha, a posio da primeira chave que */
      /*  maior do que ou igual a menor chave  */
   i = BuscaEmNoBM(menorChave, &no, &encontrado);

      /* Armazena em arquivo as chaves encontradas a  */
      /* partir desta posio na folha encontrada at */
      /* encontrar a ltima chave da ltima folha ou  */
      /* uma chave maior do que 'maiorChave'          */

      /* A coleta de chaves na primeira folha encontrada    */
      /*  diferente daquela nas demais folhas da sequncia */
      /* porque comea com a chave na posio i             */
   for ( ; i < no.conteudo.noFolha.nChaves &&
         no.conteudo.noFolha.chaves[i].chave <= maiorChave; ++i ){
         /* Escreve a chave corrente */
      fprintf( streamChaves, "%ld\n",
               no.conteudo.noFolha.chaves[i].chave );
   }

      /* Se a ltima chave escrita foi a ltima chave da */
      /* folha,  preciso ajustar o ndice dessa chave   */
   if (i == no.conteudo.noFolha.nChaves)
      --i;

   if (no.conteudo.noFolha.chaves[i].chave > maiorChave)
         /* Foi encontrada uma chave maior do que a maior */
         /* chave do intervalo e nada mais resta a fazer  */
      return 1;

      /* Escreve cada chave de cada folha seguinte que    */
      /* no seja maior do que a maior chave do intervalo */
   do {
         /* Passa para a prxima folha */
      posNo = no.conteudo.noFolha.proximaFolha;

         /* L a folha  */
      LeNoBM(streamArv, posNo, &no);

         /* Escreve cada chave que no  maior */
         /* do que a maior chave do intervalo  */
      for (i = 0; i < no.conteudo.noFolha.nChaves &&
               no.conteudo.noFolha.chaves[i].chave <= maiorChave;
                  ++i) {
            /* Escreve a chave corrente */
         fprintf( streamChaves, "%ld\n",
                  no.conteudo.noFolha.chaves[i].chave );
      }

         /* Checa se  preciso ajustar o ndice da ltima chave */
      if (i == no.conteudo.noFolha.nChaves)
         --i;
   } while ( posNo != POSICAO_NULA &&
             no.conteudo.noFolha.chaves[i].chave <= maiorChave );

   return 1;
}

/****
 *
 * ProfundidadeBM(): Calcula a profundidade de uma rvore B+
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         a rvore
 *
 * Retorno: A referida profundidade
 *
 ****/
int ProfundidadeBM(FILE *stream)
{
   tNoBM umNo;
   int   pos, nNiveis = 0;

   rewind(stream);

      /* O primeiro n  a raiz */
   fread(&umNo, sizeof(umNo), 1, stream);

   ASSEGURA( !feof(stream) & !ferror(stream),
             ERRO_FREAD(ProfundidadeBM) );

      /* Se a rvore estiver vazia, sua profundidade  zero */
   if (umNo.tipoDoNo == NO_VAZIO)
      return 0;

   ++nNiveis; /* O primeiro nvel j foi encontrado */

      /* Desce at encontrar a folha mais  esquerda */
   while (umNo.tipoDoNo == INTERNO) {
         /* Desce pelo filho mais  esquerda */
      pos = umNo.conteudo.noInterno.filhos[0]*sizeof(umNo);
      if (fseek(stream, pos, SEEK_SET)) {
         return 0;
      }

      fread(&umNo, sizeof(umNo), 1, stream);

      if (feof(stream) || ferror(stream))
         return 0;

      ++nNiveis; /* Conta-se um nvel adiantado */
   }

      /* Deve-se ter chegado a uma folha */
   ASSEGURA( umNo.tipoDoNo == FOLHA,
             ERRO_INCONSISTENCIA(ProfundidadeBM) );

   return nNiveis;
}

/****
 *
 * NumeroDeChavesBM(): Calcula o nmero de chaves armazenadas nas
 *                     folhas de uma rvore B+
 *
 * Parmetros:
 *    *streamArv (entrada) - stream associado ao arquivo que
 *                           contm a rvore
 *
 * Retorno: O nmero de chaves
 *
 ****/
int NumeroDeChavesBM(FILE *streamArv)
{
   int   numChaves = 0;
   tNoBM umNo;

      /* Verifica se streamArv  vlido */
   ASSEGURA( streamArv,
             ERRO_STREAM_NULL(streamArv, NumeroDeChavesBM) );

      /* Tenta mover o apontador de posio */
      /* do arquivo para o seu incio       */
   MoveApontador(streamArv, 0, SEEK_SET);

      /* Acessa sequencialmente cada n. Se ele */
      /* for folha, conta seu nmero de chaves  */
   while (1) {
         /* Tenta ler um n */
      fread(&umNo, sizeof(&umNo), 1, streamArv);

         /* Verifica se ocorreu erro ou o */
         /* final de arquivo foi atingido */
      if (feof(streamArv) || ferror(streamArv))
         break;

         /* Se o n for folha, acrescenta seu */
         /* nmero de chaves  contagem       */
      if (umNo.tipoDoNo == FOLHA)
         numChaves += umNo.conteudo.noFolha.nChaves;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(streamArv), ERRO_FREAD(NumeroDeChavesBM));

   return numChaves;
}

/****
 *
 * NumeroDeFolhasBM(): calcula o nmero de folhas de uma rvore B+
 *
 * Parmetros:
 *    streamArv (entrada) - stream associado ao arquivo que
 *                          contm a rvore
 *
 * Retorno: O nmero de folhas da rvore B+
 *
 * Observao: Esta  uma operao bastante custosa, pois envolve
 *             inmeros acessos ao meio de armazenamento externo
 *
 ****/
int NumeroDeFolhasBM(FILE *streamArv)
{
   int   numFolhas = 0;
   tNoBM umNo;

      /* Verifica se streamArv  vlido */
   ASSEGURA( streamArv, ERRO_STREAM_NULL(streamArv, NumeroDeFolhasBM) );

      /* Tenta mover o apontador de posio */
      /* do arquivo para o seu incio       */
   MoveApontador(streamArv, 0, SEEK_SET);

      /* Acessa sequencialmente cada n. Se */
      /* ele for folha, leva-o em conta     */
   while (1) {
         /* Tenta ler um n */
      fread(&umNo, sizeof(&umNo), 1, streamArv);

         /* Verifica se ocorreu erro ou o */
         /* final de arquivo foi atingido */
      if (feof(streamArv) || ferror(streamArv))
         break;

      if (umNo.tipoDoNo == FOLHA)
         ++numFolhas; /* Mais uma folha encontrada */
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(streamArv), ERRO_FREAD(NumeroDeFolhasBM));

   return numFolhas;
}

/****
 *
 * NumeroDeNosInternosBM(): Calcula o nmero de ns internos de
 *                          uma rvore B+
 *
 * Parmetros:
 *    *streamArv (entrada) - stream associado ao arquivo que
 *                           contm a rvore
 *
 * Retorno: Nmero de ns internos da rvore
 *
 ****/
int NumeroDeNosInternosBM(FILE *streamArv)
{
   int   numNos = 0;
   tNoBM umNo;

      /* Verifica se streamArv  vlido */
   ASSEGURA( streamArv,
             ERRO_STREAM_NULL(streamArv, NumeroDeNosInternosBM) );

      /* Tenta mover o apontador de posio */
      /* do arquivo para o seu incio       */
   MoveApontador(streamArv, 0, SEEK_SET);

      /* Acessa sequencialmente cada n. Se  */
      /* ele for n interno, leva-o em conta */
   while (1) {
         /* Tenta ler um n */
      fread(&umNo, sizeof(&umNo), 1, streamArv);

         /* Verifica se ocorreu erro ou o */
         /* final de arquivo foi atingido */
      if (feof(streamArv) || ferror(streamArv))
         break;

      if (umNo.tipoDoNo == INTERNO)
         ++numNos; /* Mais um n interno encontrado */
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(streamArv), ERRO_FREAD(NumeroDeNosInternosBM));

   return numNos;
}

/****
 *
 * NumeroDeNosVaziosBM(): Calcula o nmero de ns vazios de uma
 *                        rvore B+
 *
 * Parmetros:
 *    *streamArv (entrada) - stream associado ao arquivo que contm a rvore
 *
 * Retorno: Nmero de ns vazios da rvore
 *
 ****/
int NumeroDeNosVaziosBM(FILE *streamArv)
{
   int   numNos = 0;
   tNoBM umNo;

      /* Verifica se streamArv  vlido */
   ASSEGURA( streamArv,
             ERRO_STREAM_NULL(streamArv, NumeroDeNosInternosBM) );

      /* Tenta mover o apontador de posio */
      /* do arquivo para o seu incio       */
   MoveApontador(streamArv, 0, SEEK_SET);

      /* Acessa sequencialmente cada n. Se  */
      /* ele for n interno, leva-o em conta */
   while (1) {
         /* Tenta ler um n */
      fread(&umNo, sizeof(&umNo), 1, streamArv);

         /* Verifica se ocorreu erro ou o */
         /* final de arquivo foi atingido */
      if (feof(streamArv) || ferror(streamArv))
         break;

      if (umNo.tipoDoNo == NO_VAZIO)
         ++numNos; /* Mais um n vazio encontrado */
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(streamArv), ERRO_FREAD(NumeroDeNosInternosBM));

   return numNos;
}

/****
 *
 * NumeroDeNosBM(): Calcula o nmero total de ns (internos e
 *                  folhas) de uma rvore B+
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *                         contendo a rvore
 *
 * Retorno: O nmero de ns da rvore
 *
 ****/
int NumeroDeNosBM(FILE *stream)
{
   return NumeroDeNosInternosBM(stream) + NumeroDeFolhasBM(stream);
}

/****
 *
 * BulkloadingBM(): L arquivo de dados (binrio) contendo pares
 *                chave/ndice ordenados e constri uma rvore B+
 *                usando bulkloading
 *
 * Parmetros:
 *     streamReg (entrada) - stream associado ao arquivo de registros
 *     streamArv (entrada) - stream associado ao arquivo que conter a rvore
 *    *raiz (sada) - a raiz da rvore criada
 *
 * Retorno: Nada
 *
 ****/
void BulkloadingBM(FILE *streamReg, FILE *streamArv, tNoBM *raiz)
{
   FILE        *streamTmp, /* Associado ao arquivo contendo pares chave/ndice */
                           /* obtidos do arquivo de registros                  */
               *streamOrd; /* Associado ao arquivo ordenado de pares chave/ndice */
   tRegistroMEC registro; /* Um registro lido em arquivo */
   int          pos = 0, /* ndice de um registro lido em arquivo */
                nPares = 0, /* Nmero de pares lidos no arquivo */
                nSeries; /* Nmero de sries */
   tChaveIndice chaveIndice; /* Um par chave/ndice */

/* >>>>>>>>>>>>> Fase 1: Coleta de Pares <<<<<<<<<<<<< */

   MedidaDeTempo("Criando pares chave/indice...");

   streamTmp = tmpfile(); /* Tenta criar um arquivo temporrio */

      /* Verifica se o arquivo temporrio foi aberto */
   ASSEGURA(streamTmp, "Impossivel criar arquivo temporario");

      /* Os registros sero acessados sequencialmente.  Portanto, deve-se */
      /* assegurar que o processamento comea no incio do arquivo.       */
   MoveApontador(streamReg, 0, SEEK_SET);

   while (1) {
      fread(&registro, sizeof(registro), 1, streamReg); /* L um registro */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(streamReg) || ferror(streamReg))
         break;

         /* Cria o contedo do par chave/ndice */
      chaveIndice.chave = registro.chave;
      chaveIndice.indice = pos++; /* Cuidado com o incremento */

         /* Escreve no arquivo temporrio o par obtido */
      fwrite(&chaveIndice, sizeof(chaveIndice), 1, streamTmp);

         /* Verifica se houve erro de escrita */
      if (ferror(streamTmp))
         break;

      ++nPares;
   }

      /*Verifica se ocorreu algum erro de leitura/escrita */
   ASSEGURA( !ferror(streamReg) && !ferror(streamTmp),
             "Erro de leitura ou escrita em BulkloadingBM()" );

   printf("\n...%d pares criados com sucesso\n", nPares);
   MedidaDeTempo(NULL);

/* >>>>>>>>>>>>>>>> Fase 2: Ordenao <<<<<<<<<<<<<<<< */

   MedidaDeTempo("Criando series ordenadas...");

      /* L o arquivo original de pares e cria as sries */
   nSeries = CriaSeriesMS2(streamTmp, NOME_ARQ_SERIES);

      /* O arquivo temporrio no  mais necessrio */
   ASSEGURA(!fclose(streamTmp), "Arquivo temporario nao foi fechado");

   printf("\n...%d series criadas com sucesso\n", nSeries);
   MedidaDeTempo(NULL);

   MedidaDeTempo("Intercalando series...");

   streamOrd = IntercalaMultiArq2( NOME_ARQ_SERIES, NOME_ARQ_ORD,
                                   nSeries, MAX_PARES );

   printf("\n...series intercaladas com sucesso\n");
   printf( "\nO arquivo ordenado possui %d pares\n",
           NumeroDeRegistros(streamOrd, sizeof(tChaveIndice)) );
   MedidaDeTempo(NULL);

/* >>>>>>>>>>>>> Fase 3: Criao da rvore <<<<<<<<<<< */

   MedidaDeTempo("Criando arvore...");

      /* Constri a rvore usando bulkloading */
   CriaArvoreBM(streamOrd, streamArv, raiz);

   MedidaDeTempo("Arvore criada com sucesso");
}
