/****
 *
 * Arquivo Pilha.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2008
 * ltima modificao: 12/05/2010
 *
 * Descrio: Implementao do mdulo Pilha usado pelo
 *            programa de busca usando rvore B
 *
 ****/

#include <stdio.h>     /* Funes de entrada e sada     */
#include <stdlib.h>    /* Funes de alocao de memria */
#include "Pilha.h"     /* Interface deste mdulo         */
#include "ASSEGURA.h"  /* Macro ASSEGURA e outras macros */
                       /* de tratamento de exceo       */

/***************** Definies de Funes ******************/

/****
 *
 * CriaPilha(): Inicia uma pilha
 *
 * Parmetros: p (sada) - ponteiro para a pilha que ser iniciada
 *
 * Retorno: Nada
 *
 * Observao: O programa ser abortado se o parmetro for NULL
 *
 ****/
void  CriaPilha(tPilha *p)
{
   ASSEGURA(p, "Erro: ponteiro nulo recebido por CriaPilha()");

   *p = NULL;
}

/****
 *
 * PilhaVazia(): Verifica se uma pilha est vazia
 *
 * Parmetros: p (entrada) - a pilha que ser verificada
 *
 * Retorno: 1, se a pilha recebida como parmetro estiver vazia;
 *          0, caso contrrio.
 *
 ****/
int  PilhaVazia(tPilha p)
{
   return !p;
}

/****
 *
 * ElementoTopo(); Retorna o elemento do topo de uma pilha
 *                 sem desempilh-lo
 *
 * Parmetros: p (entrada) - a pilha
 *
 * Retorno: O elemento que est no topo da pilha
 *
 * Observao: Se a pilha estiver vazia, causa o
 *             aborto do programa.
 *
 ****/
tElemento ElementoTopo(tPilha *p)
{
   ASSEGURA(!PilhaVazia(*p),
            "Erro: Pilha vazia recebida por ElementoTopo()");

   return (*p)->dado;
}

/****
 *
 * Empilha(): Empilha um item numa pilha
 *
 * Parmetros:
 *      p (entrada/sada) - ponteiro para a pilha
 *      item (entrada) - o item que ser empilhado
 *
 * Retorno: Nada
 *
 * Observao: Se o parmetro for NULL ou no for possvel
 *             alocar um n para empilhar, o programa ser
 *             abortado
 *
 ****/
void Empilha( tElemento item, tPilha *p)
{
   tPilha  aux;

   ASSEGURA(p, "Erro: ponteiro nulo recebido por Empilha()");

   aux = malloc(sizeof(tNoPilha));

   ASSEGURA(aux, "Erro de alocaca o de um elemento da pilha");

   aux->dado = item;
   aux->proximo = *p;
   *p = aux;
}

/****
 *
 * Desempilha(): Desempilha o elemento do topo de uma pilha
 *
 * Parmetros: p (entrada/sada) - ponteiro para a pilha
 *
 * Retorno: O elemento desempilhado
 *
 * Observao: Se o parmetro for NULL ou a pilha estiver
 *             vazia, o programa ser abortado.
 *
 ****/
tElemento Desempilha(tPilha  *p)
{
   tElemento elementoNoTopo;
   tPilha    aux;

   ASSEGURA(p, "Erro: ponteiro nulo recebido por Desempilha()");

   ASSEGURA(!PilhaVazia(*p), "Erro: Pilha vazia em Desempilha()");

   elementoNoTopo = (*p)->dado;
   aux = *p;
   *p = (*p)->proximo;
   free(aux);

   return elementoNoTopo;
}

/****
 *
 * EsvaziaPilha(): Desempilha e descarta todos os elementos
 *                 da pilha
 *
 * Parmetros: p (sada) - ponteiro para a pilha
 *
 * Retorno: Nada
 *
 * Observao: Se o parmetro for NULL, o programa ser
 *             abortado
 *
 ****/
void EsvaziaPilha(tPilha *p)
{
   ASSEGURA(p,
            "Erro: ponteiro nulo recebido por EsvaziaPilha()");

   while (*p)
      (void) Desempilha(p);
}

/****
 *
 * CopiaPilha(): Faz uma cpia de uma pilha
 *
 * Parmetros:
 *    copia (sada) - ponteiro para a pilha resultante da cpia
 *    pilha (entrada) - ponteiro para a pilha que ser copiada
 *
 * Retorno: Nada
 *
 * Observao: Se o segundo parmetro for NULL, o programa ser
 *             abortado
 *
 ****/
void CopiaPilha(tPilha *copia, const tPilha *pilha)
{
   tPilha p, q = NULL, lista = *pilha;

   ASSEGURA(pilha, "Erro: ponteiro nulo recebido por CopiaPilha()");

   CriaPilha(copia);

   if (PilhaVazia(lista))
      return; /* No h o que copiar */

      /* Aloca o primeiro n da cpia */
   *copia = p = malloc(sizeof(tNoPilha));
   ASSEGURA(p, "Erro de alocaca o de um elemento da pilha");

      /* Copia o primeiro n */
   p->dado = lista->dado;

      /* Passa para o prximo n */
   lista = lista->proximo;

   while (lista) {
         /* Aloca o prximo n */
      q = malloc(sizeof(tNoPilha));
      ASSEGURA(q, "Erro de alocaca o de um elemento da pilha");

         /* Copia o contedo do n */
      q->dado = lista->dado;

      p->proximo = q;

      p = q;
   }

   p->proximo = NULL;
}

